! VNY 25.01.2002; The subroutine dlocate can crash if the
! input xx-array consists of only one element. 
! Protection against this crash has been added
! VNY 29.01.03 Subroutine Marcquardts has been modified (adjusted
! in accordance with the Numerical Recipes book
! VNY 26.01.2004 Changes during the visit to Vilspa 26-28.01.2004
! VNY 02.02.2004 Removing flux calculation for the wavelengths outside
! the range defined by CCF's, that is, 1800-3600 for Grism1 and
! 3000-6000 Angstrom for Grism2
! VNY 06.02.2004 Computing the src/bkg ratio
! VNY 21.04.2004 added subroutine for the re-computation of 
!                zero-order coordinates
! VNY 04.05.2004 Temporarily restoring the wrong algorithm of computing
! the binning factors (to make it compatible with the imaging tasks
! VNY 14.07.2004 Improving the optimal extraction algorithm: subtracting the smoothed
!                background
! VNY 07.09.2004 The wavelength interpolation at the end of the calibration array is improved
!  
! VNY 27.01.2005 Correcting the limits for the background interpolation across the spectrum
!                (SPR-3433)
! VNY 22.02.2005 Correcting the slight bias in the estimation of the spectrum flux
!                when it is comparable to the background   
! VNY 27.02.2006 The count scaling factor is computed from the wavelength-CCF and not 
!                hardwared anymore    
! VNY 29.03.2007 The code is cleaned up, the possible division by zero operation is eliminated
!                (SSC-SPR-3658)       
!    
module omgextraction_mod

  use dal
  use cal
  use sasdal
  use types
  use errorhandling 
  use sort_mod
  use OmParameters_mod     ! For computing source moments
  use omgraph_mod ! For plotting a graph
  use marq_mod

  implicit none

  public

contains
!***********************************************************************
! ExtractSpectrum 
! For extraction of one spectrum from grismSetH using the x and y 
! coordinates of the zero-order image and the extraction region parameters.
!***********************************************************************
! Input parameters:
! -----------------
! grismSetH - handle to the input image, which is rotated so that
!             the dispersion direction is aligned with respect to
!             the y-direction
! xZero - x-coordinate of the image of the zero-order spectrum
! yZero - y-coordinate of the image of the zero-order spectrum 
! spectrumHalfWidth - half-width of the spectrum extraction region  
! bkgOffsetLeft - offset of the left background extraction region from
!             the left edge of the spectrum extraction region 
! bkgWidthLeft - width of the left background extraction region 
! bkgOffsetRight - offset of the  right background extraction region
!             from the right edge of the spectrum extraction region
! bkgWidthRight - width of the right background extraction region
! sizeY -      length of the extraction regions (must be the size
!             in the vertical direction of the input image
! extractionMode - integer, to change the algorithm for spectra
!                  extraction: 
!                  0 - simple extraction,
!                  1 - optimal extraction 
!                  2 - spline smoothing of each cross-section
!                  3 - Gaussian approximation of the cross-sections
! verbFlag - integer flag for test messages (produces messages if 
!            verbFlag=1; for test purporses)
! Output parameters:
! ------------------ 
! wavelengthArray - wavelengths corresponding to the extracted spectrum
! spectrumArray - extracted spectrum (in couts per second)
! spectrumErrArray - corresponding counts errors 
! fluxArray - spectrum converted into energy flux by using CAL_OM
!             calibration file
! fluxErrArray - corresponding flux errors
! bkgArray - extracted background (average from the left and right
!            background strips
!******************************************************************
! 14.05.2007 in the case of very close near-by spectra the background 
!            should be extracted only from one side of the spectrum
!
  subroutine ExtractSpectrum(grismSetH, xZero, yZero, xFirst, yFirst, spectrumHalfWidth, &
       bkgOffsetLeft, bkgWidthLeft, bkgOffsetRight, bkgWidthRight, &
       wavelengthArray, spectrumArray, spectrumErrArray, bkgArray, &
       bkgErrArray, fluxArray, fluxErrArray, sizeY, extractionMode, &
       subtractBkg, verbFlag, s2n, s2n0, crossShape, leftBorder, &
       rightBorder, proximity, nSp)
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xZero, yZero, xFirst, yFirst
    real(kind=double), intent(in) :: spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft
    real(kind=double), intent(in) :: bkgOffsetRight, bkgWidthRight 
    integer, intent(in) :: sizeY, extractionMode
    logical, intent(in) :: subtractBkg
    integer, intent(in) :: verbFlag
    integer :: verbFlag2
    real(kind=double), intent(out), dimension(0:sizeY-1) :: wavelengthArray
    real(kind=double), dimension(0:sizeY-1) :: wavelengthArray2
    real(kind=double), intent(out), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: spectrumErrArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: bkgArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: bkgErrArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: fluxArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: fluxErrArray
    !-----
    integer :: i, j, ny, nx
    real(kind=double) :: deltaY
    real(kind=double), dimension(:), pointer :: countsScaleFactor
    real(kind=double), dimension(:), pointer :: countsScaleFactor2
    real(kind=double), dimension(:), pointer :: bkgVectR, bkgVectL, specVect
    real(kind=double), dimension(:), pointer :: bkgVect2, bkgVect3
    real(kind=double), dimension(:), pointer :: smoothBkgVectL, smoothBkgVectR
    real(kind=double), dimension(:), pointer :: crossSection
    !-----
    real(kind=double), dimension(:), pointer :: lambda
    real(kind=double), dimension(:), pointer :: isfLambda
    real(kind=double), dimension(:), pointer :: isf
    real(kind=double), dimension(:), pointer :: isfError
    real(kind=single), intent(inout), dimension(1:2) :: leftBorder, rightBorder
    integer(int32) :: leftLimit, rightLimit, leftBlimit, rightBlimit
    real(kind=single), intent(in) :: proximity
    integer(kind=int32), intent(in) :: nSp

    integer(kind=int32) :: filterIdInt !+
    integer(kind=int32) ccdMode !+  
    !----
    !real(kind=double), dimension(:), pointer :: zeroOrderVect
    !real(kind=double), dimension(:), pointer :: zeroOrderBkg

    character(len=9) :: filter !+                         ! filter keyword in the input file
    integer(kind=int32) :: binX                                        ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    logical :: binBpe !+
    integer :: jbkg1, jbkg2, nPixBkgR, nPixBkgL, jcross
    integer :: jyspc1, jyspc2
    real(kind=double) :: averBkgR, averRight,numRight, numBkgR
    real(kind=double) :: averBkgL, averLeft, numLeft, numBkgL
    type(ArrayT) :: inArrayH   
    real(kind=single) :: exposureTime         ! The exposure time for an input image
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    integer :: xBkgL, xBkgR
    integer :: jspc1, jspc2, nPixSpec, nPts, iR, iL
    real(kind=double) :: source2spectrum, spectrumLength
    real(kind=double) :: averStrip, numStrip, averSpec, numSpec
    real(kind=double) :: averSpec2
    real(kind=double) :: levelLR, nLevel, a0, a1, wSpec, dSpec, levelLR2
    real(kind=double) :: sum1, sum2, sumTest, sumGauss, sumWeight
    real(kind=double), dimension (:), pointer :: xinput
    real(kind=double), dimension (:), pointer :: yinput
    real(kind=double), dimension (:), pointer :: weights

    integer :: ios, isfSize, lambdaSize, jZero
    real(kind=double) :: modZero
    real(kind=double) :: avStrip, avStripN
    integer :: jstart, jend
    !--- VNY 08.01.2004
    !real(kind=double), dimension(9) :: param9
    real(kind=double), dimension(5) :: param5
    integer(kind=int32) :: numparam !, numpoints
    !---
    real(kind=double)  :: xtest , ytest ! this must be uncomm. for opt. extr. 
    !real(kind=double), dimension(9) ::dydptest9 ! this must be uncomm. for opt. extr.
    real(kind=double), dimension(5) ::dydptest5
    !
    real(kind=double), dimension (:), pointer :: yp
    real(kind=double), dimension (:), pointer :: ysp
    real(kind=double), dimension (:), pointer :: temp
    real(kind=double), dimension (:), pointer :: dispersion
    real(kind=double) :: slp1, slpn, tension
    real(kind=double) :: eps,smooth, delta, averSpecLowess
!, averSpecGauss, averSpecOpt,
    !real(kind=double) :: averSpecSimple
    integer(kind=int32) :: ierr, islpsw, numsteps
    !real(kind=double) :: sbRatio
    real(kind=double) :: optimal1, gaussum, aux1, aux2, znam, simplesum, max1, min1, min2
    integer :: imax1, imin1, imin2, currentExtractionMode
    real(kind=double) :: optimal2, aux3, sumWeights
    integer :: nUsedPoints
    real(kind=double) :: razn
    ! VNY 14.07.2004
    real(kind=double) :: p1sum, p2sum, p3sum, p4sum, p5sum, nPsum ! average Gaussian parameters
    real(kind=double), dimension(:), pointer :: p1vect, p2vect, p3vect, p4vect, p5vect, pAux 
    real(kind=double) :: yFirst2, xZero2

    real(kind=double), dimension(:), pointer :: across0
    real(kind=double) :: auxAver, auxBkgAver
! Variables for plotting the graph
!-----------------------------------------------------
    !real(kind=double), dimension (:), pointer :: xinput
    !real(kind=double), dimension (:), pointer :: yinput
    !real(kind=double), dimension (:), pointer :: weights
    !
    real, dimension (:), pointer :: xdata
    real, dimension (:), pointer :: ydata
    character(len=20) :: graphTitle
    character(len=10) :: xlabel
    character(len=10) :: ylabel    
    real :: x1,x2, y1,y2
    integer :: ndata, pgbeg
    real(kind=double) :: s2nRatio
    real(kind=double), intent(out) :: s2n, s2n0
    real(kind=single) :: crossLateral, crossCentral
    real(kind=double), intent(out) :: crossShape
    integer :: nLat, nCentr
    integer :: i01,i02,j01,j02,ibkgL01,ibkgL02, ibkgR01,ibkgR02, ibkgAux01, ibkgaux02
    real :: averSign0, averBkg0, averBkgL0, averBkgR0, averBkgAux
    integer :: flagAux1
    integer :: jj
    logical :: computeBkgL, computeBkgR
    real(kind=single) :: borderShift
    real(kind=single) :: nsign0, nbkgL, nbkgR
    real(kind=double) :: bkgL, bkgR
!------------------------------------------------------
    s2n=0.0
    s2n0=0.0
    computeBkgL=.true.
    computeBkgR=.true.

    yFirst2= yFirst
    xZero2= xZero
    !
    !numparam=9
    numparam=5
    currentExtractionMode=extractionMode
    !   real :: sigm ! this must be uncomm. for opt. extr. 
    !---
    !   real :: sum1, sum2, sumTest, sumGauss, sumWeight
    !---
    if (.not. subtractBkg) then
       call warning('OMGRISM','In this version the background rates are subtracted from the spectrum')
    endif
    !---VNY 28.01.2004
    if (extractionMode .gt. 3) then
       call warning('OMGRISM','This version works with extractionModes <= 3')
       currentExtractionMode=3
    endif

    !if (verbFlag .eq. 1) then
    !   write(*,*) "OMGEXCTRACTION L234 extractionMode=", extractionMode
    !   write(*,*) "L235 nSp=", nSp, " peacks=", peacks, " proximity=", proximity
    !endif

    if (nSp .eq. 2) then 
       ! disable computing one of the backgrounds (left or right) if there is
       ! a molesting spectrum near-by
       if (proximity .lt. 0) then ! disable the left background
          computeBkgL=.false.
       else ! disable the right backround
          computeBkgR=.false.
       endif

    endif
    if (verbFlag .eq. 1) then
       write(*,*) "L256 computeBkgL=", computeBkgL, " computeBkgR=", computeBkgR
    endif
    ! the image contains the information on the science window
    inArrayH = array(grismSetH, "PRIMARY", read)
    exposureTime = real32Attribute(grismSetH, "EXPOSURE") ! Image exposure time
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  
    !ny=sizeY
    !nx=sizeX
    allocate(countsScaleFactor(0:ny-1))
    allocate(countsScaleFactor2(0:ny-1))
    if (hasAttribute(grismSetH,"FILTER")) then
       filter = stringAttribute(grismSetH, "FILTER")
    else
       call warning('OMGRISM','Keyword FILTER missing in the input file ')
    endif

    !write(*,*) "*******************"
    !write(*,*) " nx=", nx, " ny=", ny
    !write(*,*) "*******************"
    where(oswImage .lt. 1.0) oswImage=-1.0

    spectrumArray=0.0
    spectrumErrArray=0.0
    bkgArray=0.0
    bkgErrArray=0.0
    fluxArray=0.0
    fluxErrArray=0.0

    ccdMode = IMAGE

    !**************************************
    ! Get the filter Id integer value
    !**************************************
    filterIdInt = filterIdToInt(filter)
    !*******************
    ! Set the CAL state
    !********************
    call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)
    !
    allocate(lambda(0:ny-1))
    allocate(isfLambda(0:ny-1))
    allocate(isf(0:ny-1))
    allocate(isfError(0:ny-1))
    lambda=0.0
    ! VNY 21.11.03
    ios=1
    !  open (1,file="isfsas_uv.dat", iostat=ios)
    !  write(*,*) "OMGEXTRACTION L140 iostat=", ios
    !  close(1)

    call CAL_omGetGrismWavelength(lambda)
    call CAL_omGetGrismFlux(isfLambda, isf, isfError)
    isfSize=size(isfLambda)
    lambdaSize=size(lambda)
    !write(*,*) "L237 isfSize=", isfSize, " lambdaSize=", lambdaSize, "ny-1=", ny-1
    !write(*,*) "L238  isfL(0)=", isfLambda(0), " isfL(isfSize-1)=", isfLambda(isfSize-1)
    !write(*,*) "L239  isf(0)=", isf(0), " isf(isfSize-1)=", isf(isfSize-1)

    ! Calculate image binning.  
    binBpe= .false.
    if (hasAttribute(grismSetH,"BINBPE")) then       
       binBpe= booleanAttribute(grismSetH, "BINBPE")
    else
       call warning('OMGRISM','Keyword BINBPE missing in the input file ')
    endif
    ! VNY 04.05.2004 Temporarily restoring the old (wrong) algorithm 
    ! of computing the binning factors (to make it compatible with 
    ! the imaging tasks 
    !if (hasAttribute(grismSetH,"BINAX1")) then       
    !   binAx1 = int32Attribute(grismSetH, "BINAX1")
    !   binX = binAx1
    !   if (binX .lt. 1.0) binX = 1.0
    !else
    !   call warning('OMGRISM','Keyword BINAX1 missing in the input file ')
    !   binX=1.0
    !endif
    if (hasAttribute(grismSetH,"BINAX1")) then       
       binAx1 = int32Attribute(grismSetH, "BINAX1")
       binX = 2 ** binAx1
    else
       call warning('OMGRISM','Keyword BINAX1 missing in the input file ')
       binX=1.0
    endif

    if (hasAttribute(grismSetH,"BINAX2")) then       
       binAx2 = int32Attribute(grismSetH, "BINAX2")
       binY = 2 ** binAx2
    else
       call warning('OMGRISM','Keyword BINAX2 missing in the input file ')
       binY=1.0
    endif

    if (binBPE) then
       binX = 2.0
       binY = 2.0
    end if
    !if (verbFlag .eq. 1) then 
    !   write (*,*) "OMGRISMEXTRACTION L201"
    !   write(*,*) "OMGRISMEXTRACTION L235: xZero=", xZero, " yZero=", yZero, " size=", sizeY
    !   write(*,*) "                  L240 xFirst=", xFirst, " yFirst=", yFirst
    !   write(*,*) "OMGRISMEXTRACTION L236: binBpe=", binBpe, " binAx1=", binAx1, " binAx2=", binAx2
    !   write(*,*) " filter=", filter, " binX=", binX, " binY=", binY
    !endif

    ! compute the zero-to-first-order separation (depending on the filter)
    if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
       source2spectrum= 1100.0 
       spectrumLength= 1000.0
    else
       source2spectrum = 800.0
       spectrumLength = 400.0
    endif
    if (binY .gt. 1.0) then 
       source2spectrum=source2spectrum/binY
       spectrumLength=spectrumLength/binY
    endif

    !if (verbFlag .eq. 1) then 
    !   write(*,*) "L338 source2spectrum=", source2spectrum, " spectrumLength=", spectrumLength
    !endif

    wavelengthArray=0.0

    ! Compute wavelengths
    countsScaleFactor=1.0
    countsScaleFactor2=1.0
    !write(*,*) "++++++++++++++++++++++++++ binY=", binY
    do j=0,ny-1
       deltaY=real(j)+0.5-yZero ! yZero here is the Y-coordinate of the source' zero order image
       jj=deltaY*binY
       ! Y-binning is taken into account for calculation of the wavelengths 
       if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
          countsScaleFactor(j)=0.240774*binY+2.0*0.000047387*deltaY*binY*binY ! this is per unbinned pixelY
          wavelengthArray(j)=0.240774*deltaY*binY+0.000047387*deltaY*deltaY*binY*binY+ &
               77.3484 ! UV grism
          wavelengthArray2(j)=0.240774*deltaY*binY+0.000047387*deltaY*deltaY*binY*binY+ &
               77.3484 ! UV grism
          wavelengthArray2(j)=wavelengthArray2(j)*10.0 ! wavelengths in Angstroms
          countsScaleFactor(j)=countsScaleFactor(j)*10.0
       else
          countsScaleFactor(j)=0.5626*binY
          wavelengthArray2(j)=0.5626*deltaY*binY+20.0898 
          ! for the visual grism (filter="Grism2 "
          wavelengthArray2(j)=wavelengthArray2(j)*10.0
          countsScaleFactor(j)=countsScaleFactor(j)*10.0
       endif
       !if (jj .gt. 0 .and. jj .lt. 350) then
       !   write(*,'(a,i5,a,i5,a,f7.1,a,f7.1,a,f7.1,a,f7.1,a,f7.1)') "L372 j=", j, &
       !        " jj=", jj, " wl2=", wavelengthArray2(j), " l(jj)=", lambda(jj), &
       !        " f=", countsScaleFactor(j), " l(jj+1)-l(jj)=", &
       !        (lambda(jj+1)-lambda(jj))*binY, "  w(jj+1)-w(jj)=", wavelengthArray2(j)-wavelengthArray2(j-1)
       !endif
    enddo ! j=0,ny-1

    ! VNY 26.11.2003 Computing the wavelengths based on CCF
    wavelengthArray=0.0
    jZero=int(yZero)
    ! the wavelength value of each pixcel should be interpolated
    modZero=1.0+yZero-real(jZero)

    do j=jZero, ny-1
       i=(j-jZero)*binY
       if (i .eq. 0) then
          wavelengthArray(j)=lambda(i)-(lambda(i+1)-lambda(i))+(lambda(i+1)-lambda(i))*modZero
       else
          if (i .lt. lambdaSize) then
             wavelengthArray(j)=lambda(i-1)+(lambda(i)-lambda(i-1))*modZero
          endif
       endif
    enddo ! j=jZero+1,ny-1

    ! VNY 27.02.2006 recalculation of the countsScaleFactor
    do j=jZero, ny-1
       if (j .ge. ny-1) then
          countsScaleFactor2(j)=wavelengthArray(j)-wavelengthArray(j-1)
       else
          countsScaleFactor2(j)=wavelengthArray(j+1)-wavelengthArray(j)
       endif
       !write(*,*) "L409 j=", j, "f1=", countsScaleFactor(j), " f2=", countsScaleFactor2(j)
       countsScaleFactor(j)=countsScaleFactor2(j)
    enddo ! j=jZero+1,ny-1

    !write(*,*) "+++++++++++++++++++++++++++++++"
    !do j=0,50
    !   write(*,*) "j=",j, " ccfLambda=", lambda(j), " w2=", wavelengthArray2(j), " w=", wavelengthArray(j) 
    !enddo
    !write(*,*) "+++++++++++++++++++++++++++++++"


    !  write(*,*) " ************************ "
    !  do j=0, 10
    !     write(*,*) " j=", j, " isfLambda=", isfLambda(j), " isf=", isf(j), " isfError=", isfError(j)
    !  enddo
    !  do j=isfSize-10, isfSize-1
    !     write(*,*) " j=", j, " isfLambda=", isfLambda(j), " isf=", isf(j), " isfError", isfError(j)
    !  enddo

    ! Compute the average zero-order spectrum within the extraction region
    i01=int(xZero-spectrumHalfWidth-0.5) ! left border
    if (i01 .lt. 0) i01=0
    i02=i01+int(2*spectrumHalfWidth) ! left border
    if (i02 .gt. nx-1) i02=nx-1

    j01=int(yZero-spectrumHalfWidth-0.5) ! lower border
    if (j01 .lt. 0) j01=0
    j02=j01+int(2*spectrumHalfWidth) ! upper border
    if (j02 .gt. ny-1) j02=ny-1

    ibkgL01=xZero-spectrumHalfWidth-bkgOffsetRight
    if (ibkgL01 .lt. 0) ibkgL01=0
    ibkgL02=ibkgL01+2*spectrumHalfWidth
    if (ibkgL02 .gt. nx-1) ibkgL02=nx-1

    ! left bkg
    averBkgL0=0.0
    nbkgL=0.0
    do j=j01,j02
       do jcross=ibkgL01,ibkgL02
          averBkgL0=averBkgL0+oswImage(jcross,j)
          nbkgL=nbkgL+1.0
       enddo
    enddo

    ibkgR01=xZero+spectrumHalfWidth+bkgOffsetRight
    if (ibkgR01 .gt. nx-2) ibkgR01=nx-2
    ibkgR02=ibkgR01+2*spectrumHalfWidth
    if (ibkgR02 .gt. nx-1) ibkgR02=nx-1

    if (verbFlag .eq. 1) then
       write(*,*) "L490 i01=", i01, " i02=", i02, " j01=",j01, " j02=", j02
       write(*,*) "L491 ibkgL01=", ibkgL01, " ibkgL02=", ibkgL02, &
            " ibkgR01=", ibkgR01, " ibkgR02=", ibkgR02
       write(*,*) "L493 xZero=", xZero
    endif

    auxBkgAver=-1.0
    ! Analyse the cross-secton of the zero-order 
    if (ibkgR02-ibkgL01 .gt. 4) then
       allocate(across0(ibkgL01:ibkgR02))
       do jcross=ibkgL01,ibkgR02
          auxAver=0.0
          do j=j01,j02
             auxAver=auxAver+oswImage(jcross,j)
          enddo
          across0(jcross)=auxAver
       enddo
       if (verbFlag .eq. 1) then
          write(*,*) "L509-------------------------"
          do jcross=ibkgL01, ibkgR02
             write(*,*) " jcross=", jcross, " across=", across0(jcross)
          enddo
          write(*,*) "L513--------------------------"
       endif
       call EvaluateZeroOrderProfile(across0,ibkgL01,ibkgR02,xZero, &
            leftLimit, rightLimit, leftBlimit, rightBlimit, bkgL, bkgR, verbFlag)
       deallocate(across0)
       bkgR=bkgR*real(ibkgR02-ibkgR01+1)
       bkgL=bkgL*real(ibkgL02-ibkgL01+1)
       if (verbFlag .eq. 1) then
          write(*,*) "L523 bkgL =", bkgL, " bkgR = ", bkgR
       endif
       if (bkgR .gt. 0.0 .or. bkgL .gt.0.0) then
          aux2=0.0
          auxBkgAver=0.0
          if (bkgR .gt. 0.0) then
             auxBkgAver=auxBkgAver+bkgR
             aux2=aux2+1.0
          endif
          if(bkgL .gt. 0.0) then
             auxBkgAver=auxBkgAver+bkgL
             aux2=aux2+1.0
          endif
          if (aux2 .gt. 1.0) auxBkgAver=auxBkgAver/aux2
       endif
    endif
    ! right bkg
    averBkgR0=0.0
    nbkgR=0.0
    do j=j01,j02
       do jcross=ibkgR01,ibkgR02
          averBkgR0=averBkgR0+oswImage(jcross,j)
          nbkgR=nbkgR+1
       enddo
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) "L550 averBkgL0=", averBkgL0, " avrBkgR0=", averBkgR0
       write(*,*) "L551 nbkgL=",nbkgL, " nbkgR=",nbkgR
       write(*,*) "L552 auxBkgAver=", auxBkgAver
    endif
    averBkg0=(averBkgL0+averBkgR0)/2.0
    ! 20.10.2007: if the neighbouring background was successfully calculated
    !             then using it and not averBkg0, which can be affected
    !             by the spectra
    if (auxBkgAver .gt. 0.0) averBkg0=auxBkgAver
    
    ! signal
    averSign0=0.0
    nsign0=0.0
    do j=j01,j02
       do jcross=i01,i02
          averSign0=averSign0+oswImage(jcross,j)
          nsign0=nsign0+1.0
       enddo
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) "L566 averSign0=", averSign0
    endif
    !averSign0=averSign0/real(i02-i01+1)
    !averBkg0=averBkg0/real(iBkgR02-iBkgR01+1)
    s2n0=averSign0-averBkg0
    if (verbFlag .eq. 1) then
       write(*,*) "L572 nsign0=", nsign0
       write(*,*) "L573 s2n0=", s2n0, " averBkg0=", averBkg0, " sqrt=", sqrt(averBkg0)
    endif
    if (averBkg0 .gt. 1.0) s2n0=s2n0/sqrt(averBkg0)

    if (verbFlag .eq. 1) then
       write(*,*) "L578 i01=", i01, " i02=", i02, " j01=",j01, " j02=", j02, &
            " averSign=", averSign0, " averBkg0=", averBkg0, " s2n0=", s2n0
       write(*,*) "L580 averBkgL0=", averBkgL0, " averBkgR0=", averBkgR0 
    endif

    ! Check if the background is higher than the signal
    if (auxBkgAver .lt. 0.0) then
       ! To the left of the source
       if (averBkgL0 .gt. averSign0) then ! analyse a closer bkg
          if (verbFlag .eq. 1) then
             write(*,*) "......................................"
             write(*,*) "L593 analysing a closer region of the left background"
          endif
          ibkgAux02=xZero-spectrumHalfWidth
          if (ibkgAux02 .lt. 1) ibkgAux02=1
          ibkgAux01=ibkgAux02-2*spectrumHalfWidth
          if (ibkgAux01 .lt. 0) ibkgAux01=0
          averBkgAux=0.0
          do j=j01,j02
             do jcross=ibkgAux01,ibkgAux02
                averBkgAux=averBkgAux+oswImage(jcross,j)
             enddo
          enddo
          if (verbFlag .eq. 1) then
             write(*,*) "L606 new LEFT background averBkgAux=", averBkgAux
          endif
          if (averBkgAux .lt. averBkgL0) averBkgL0=averBkgAux
       endif

       ! To the right of the source
       if (averBkgR0 .gt. averSign0) then ! analyse a closer right bkg region
          if (verbFlag .eq. 1) then
             write(*,*) "......................................"
             write(*,*) "L615 analysing a closer region of background"
          endif
          ibkgAux01=xZero+spectrumHalfWidth
          if (ibkgAux01 .gt. nx-2) ibkgAux01=nx-2
          ibkgAux02=ibkgAux01+2*spectrumHalfWidth
          if (ibkgAux02 .gt. nx-1) ibkgAux02=nx-1
          averBkgAux=0.0
          do j=j01,j02
             do jcross=ibkgAux01,ibkgAux02
                averBkgAux=averBkgAux+oswImage(jcross,j)
             enddo
          enddo
          if (verbFlag .eq. 1) then
             write(*,*) "L628 new RIGHT background averBkgAux=", averBkgAux
          endif
          if (averBkgAux .lt. averBkgR0) averBkgR0=averBkgAux
       endif
       averBkg0=(averBkgL0+averBkgR0)/2.0
       s2n0=averSign0-averBkg0
       if (averBkg0 .gt. 1.0) s2n0=s2n0/sqrt(averBkg0)
       if (verbFlag .eq. 1) then
          write(*,*) "L636 Correction of the s2n0=", s2n0
       endif
    endif

    !********************************************************************
    ! Compute the background around the spectrum (left and right strips)
    ! Doing the right strip of the background
    !********************************************************************
    !jbkg1=xFirst+spectrumHalfWidth+bkgOffsetRight
    !jbkg2=jbkg1+bkgWidthRight

    ! VNY 17.05.2007
    jbkg1=xFirst+spectrumHalfWidth+bkgOffsetRight
    jbkg2=jbkg1+bkgWidthRight-1


    if (verbFlag .eq. 1) then 
       write(*,*) " L653 jbkg1=", jbkg1, " jbkg2=", jbkg2, " xZero=", xZero, " yZero=", yZero
       write(*,*) " xFirst=", xFirst, " spectrumHalfWidth=", spectrumHalfWidth, &
            " bkgOffsetRight=", bkgOffsetRight
    endif


    iR=(jbkg1+jbkg2)/2 ! iR and iL are for the Gaussian fit
    !write(*,*) " L362 iR=", iR
    nPixBkgR=jbkg2-jbkg1+1

    jyspc1=int(yZero+source2spectrum-spectrumLength/2.0)
    jyspc2=int(yZero+source2spectrum+spectrumLength/2.0)
    if (verbFlag .eq. 1) then
       write(*,*) "L666 jyspc1=", jyspc1, " jyspc2=", jyspc2, " <----"
    endif

    allocate(bkgVectR(0:ny-1))
    allocate(bkgVectL(0:ny-1))
    allocate(specVect(0:ny-1))
    allocate(bkgVect2(0:ny-1))
    allocate(bkgVect3(0:ny-1))
    allocate(smoothBkgVectL(0:ny-1))
    allocate(smoothBkgVectR(0:ny-1))

    ! VNY 14.07.2004
    allocate(p1vect(0:ny-1))
    allocate(p2vect(0:ny-1))
    allocate(p3vect(0:ny-1))
    allocate(p4vect(0:ny-1))
    allocate(p5vect(0:ny-1))
    allocate(pAux(0:ny-1))

    bkgVectR=0.0
    averBkgR=0.0
    numBkgR=0.0
    if (verbFlag .eq. 1) then
       write(*,*) "L689 jbkg1=", jbkg1, " jbkg2=", jbkg2
    endif
    if (jbkg1 .lt. nx-1 .and. jbkg2 .le. nx-1) then
       do j=0, ny-1
          averRight=0.0
          numRight=0.0
          flagAux1=0
          do jcross=jbkg1,jbkg2
             if (jcross .le. nx-1) then
                !if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
                if (oswImage(jcross,j) .ge. 0.0) then ! -0.9) then
                   if (jcross .lt. nx-1) then 
                      averRight=averRight+oswImage(jcross,j)
                      numRight=numRight+1.0
                   endif
                else
                   flagAux1=-1
                   averRight=-abs(averRight)
                endif
             endif
          enddo
          if (numRight .gt. 1.0 .and. flagAux1 .eq. 0) then
             averRight=averRight/numRight
             bkgVectR(j)=averRight
             if (j .gt. jyspc1 .and. j .lt. jyspc2 .and. bkgVectR(j) .gt. 0.0) then
                averBkgR=averBkgR+bkgVectR(j)
                numBkgR=numBkgR+1.0
             endif
          else
             bkgVectR(j)=-1.0
          endif
          !if (verbFlag .eq. 1 .and. j .eq. 540) then
          !   write(*,*) " L367 numRight=", numRight, " averRight=", averRight, &
          !        " numBkgR=", numBkgR, " aver/num=", averRight/numBkgR
          !endif
          !if (verbFlag .eq. 1) then
          !   write(*,*) "L583 j=", j, " bkgVectR=", bkgVectR(j)
          !endif
       enddo
    endif
! 27.01.2005 Correcting the interpolation limits again
!---
    xBkgR=iR
    if (numBkgR .gt. 1.0) then 
       averBkgR=averBkgR/numBkgR
!       xBkgR=jbkg1
!---
    else
       xBkgR=0
    endif
    if (verbFlag .eq. 1 .and. j .eq. 540) then 
       write(*,*) " L740 averBkgR=", averBkgR, " numBkgR=", numBkgR
    endif
    averBkgR=smoothBkgExcludeZeros(bkgVectR,jyspc1,jyspc2,ny-1,verbFlag)

    !if (verbFlag .eq. 1) then 
    !   write(*,*) "L606 averBkgR=", averBkgR
    !   write(*,*) "--------------------------"
    !   do j=jyspc1,jyspc2
    !      write(*,*) "L609 j=",j, " bkgVectR=", bkgVectR(j)
    !   enddo
    !   write(*,*) "--------------------------"
    !endif

    ! Doing the left strip of the background
    jbkg1=xFirst-spectrumHalfWidth-bkgOffsetLeft
    jbkg2=jbkg1-bkgWidthLeft

    ! VNY 17.05.2007
    jbkg1=xFirst-spectrumHalfWidth-bkgOffsetLeft
    jbkg2=jbkg1-bkgWidthLeft+1

    if (verbFlag .eq. 1) then
       write(*,*) "L762 Left background limits: jbkg1=", jbkg1, " jbkg2=", jbkg2
    endif
    iL=(jbkg2+jbkg1)/2
    nPixBkgL=jbkg1-jbkg2+1
    if (verbFlag .eq. 1) then
       write(*,*) "L767 nPixBkgL=", nPixBkgL
    endif    
    bkgVectL=0.0
    averBkgL=0.0
    numBkgL=0.0
    if (jbkg1 .gt. 0) then
       do j=0, ny-1
          averLeft=0.0
          numLeft=0.0
          flagAux1=0
          do jcross=jbkg1,jbkg2,-1
             if (jcross .ge. 0 .and. jcross .lt. nx-1) then
                !if (oswImage(jcross,j) .gt. 0.0) then !-0.9) then
                if (oswImage(jcross,j) .ge. 0.0) then !-0.9) then
                   if (jcross .lt. nx-1) then
                      averLeft=averLeft+oswImage(jcross,j)
                      numLeft=numLeft+1.0
                   endif
                else
                   flagAux1=-1
                   averLeft=-abs(averLeft)
                endif
             endif
          enddo
          if (numLeft .gt. 1.0 .and. flagAux1 .eq. 0) then
             averLeft=averLeft/numLeft
             bkgVectL(j)=averLeft ! bkgVectL value is per pixel
             if (j .gt. jyspc1 .and. j .lt. jyspc2 .and. bkgVectL(j) .gt. 0.0) then
                averBkgL=averBkgL+bkgVectL(j)
                numBkgL=numBkgL+1.0
             endif
          else
             bkgVectL(j)=-1.0
          endif
          !if (verbFlag .eq. 1) then
          !   write(*,*) " L414 numLeft=", numLeft, " averLeft=", averLeft
          !endif
       enddo
    endif
! 27.01.2005 correcting the left and right interpolation limits 
!---- 
    xBkgL=iL
    if (numBkgL .gt. 1.0) then
       averBkgL=averBkgL/numBkgL ! averBkgL is per pixel
!       xBkgL=jbkg1
!----
    else
       xBkgL=0
    endif
    averBkgL=smoothBkgExcludeZeros(bkgVectL,jyspc1,jyspc2,ny-1,verbFlag)

    !if (verbFlag .eq. 1) then 
    !   write(*,*) " L509 averBkgL=", averBkgL, " numBkgL=", numBkgL
    !   write(*,*) "--------------------------"
    !   do j=jyspc1,jyspc2
    !      write(*,*) " j=",j, " bkgVectL=", bkgVectL(j)
    !   enddo
    !   write(*,*) "--------------------------"
    !   averBkgL=smoothBkgExcludeZeros(bkgVectL,jyspc1,jyspc2,ny-1,verbFlag)
    !   write(*,*) "L516 new averBkgL=",averBkgL, " <-----"
    !endif

    
    ! Compute the average spectrum within the extraction region
    !jspc1=int(xFirst-spectrumHalfWidth-0.5) ! left border 
    !jspc2=int(xFirst+spectrumHalfWidth+0.5) ! right border

    ! VNY 17.05.2007
    jspc1=int(xFirst-spectrumHalfWidth+0.5) ! left border 
    jspc2=int(xFirst+spectrumHalfWidth-0.5) ! right border

    if (verbFlag .eq. 1) then
       write(*,*) "L839 jspc1=", jspc1, " jspc2=", jspc2
    endif
    ! In the case of a close double spectrum the limits should be 
    ! slightly squeezed on the side between the two spectra

    !write(*,*) "L762 nSp=", nSp, " leftBorder=", leftBorder, &
    !     " rightBorder=", rightBorder
    if (nSp .eq. 2) then
       if (computeBkgR .and. .not. computeBkgL) then ! adjust the left border
          if (verbFlag .eq. 1) then
             write(*,*) "L849 adjusting the LEFT border"
             write(*,*) "L850 jspc1=", jspc1, " jspc2=", jspc2, &
                  " leftBorder(2)=", leftBorder(2), " rightBorder(2)=", &
                  rightBorder(2)
          endif
          if (jspc1 .lt. leftBorder(2)) then
             borderShift=leftBorder(2)-jspc1
             jspc1=leftBorder(2) ! shifting the left limit to the right
             ! Now, in order to compemsate and keep the width of the spectrum extraction
             ! region the same we have to shift also the right border
             jspc2=jspc2+borderShift
             if (jspc2 .gt. nx-1) jspc2=nx-1
             rightBorder(2)=jspc2
          else 
             rightBorder(2)=jspc2
          endif
       endif
       if (computeBkgL .and. .not. computeBkgR) then ! adjust the right border
          if (verbFlag .eq. 1) then
             write(*,*) "L868 adjusting the RIGHT border"
             write(*,*) "L869 jspc1=", jspc1, " jspc2=", jspc2, &
                  " leftBorder(1)=", leftBorder(1), " rightBorder(1)=", &
                  rightBorder(1)
          endif
!*********************** !!!!!!!!!!!!!!!!!   VNY 16.10.2007 NB ************
!          if (jspc2 .gt. rightBorder(1)) then
          if (jspc2 .gt. rightBorder(1) &
               .and. rightBorder(1) .gt. leftBorder(1) .and. rightBorder(1) .gt. 0) then
!**************************************************************************
             borderShift=jspc2-rightBorder(1)
             jspc2=rightBorder(1) ! right limit is shifted to the left
             ! respectively, shifting the left limit to the left
             jspc1=jspc1-borderShift
             if (jspc1 .lt. 0) jspc1=0.0
             leftBorder(1)=jspc1
             if (verbFlag .eq. 1) then
                write(*,*) "L885 after adjusting the RIGHT border"
                write(*,*) "L886 ispc1=", jspc1, " jspc2=", jspc2, &
                     " leftBorder(1)=", leftBorder(1), " rightBorder(1)=", &
                     rightBorder(1)
             endif
          else
             leftBorder(1)=jspc1
             if (verbFlag .eq. 1) then
                write(*,*) "L893 NOT adjusting the RIGHT border"
                write(*,*) "L894 ispc1=", jspc1, " jspc2=", jspc2, &
                     " leftBorder(1)=", leftBorder(1), " rightBorder(1)=", &
                     rightBorder(1)
             endif
          endif
       endif
    endif
    if (jspc2-jspc1 .lt. 2) then
       ! For very narrow strip enlarging it 
       jspc1=jspc1-1
       if (jspc1 .lt. 0.0) jspc1=0.0
       jspc2=jspc2+1
       if (jspc2 .gt. nx-1) jspc2=nx-1
    endif
    allocate(crosssection(jspc1:jspc2))
    nPixSpec=jspc2-jspc1+1
    !jyspc1=int(yZero+source2spectrum-spectrumLength/2.0)
    !jyspc2=int(yZero+source2spectrum+spectrumLength/2.0)
    if (verbFlag .eq. 1) then
       write(*,*) "L913 jspc1=", jspc1, " jspc2=", jspc2, " jy1=", jyspc1, " jy2=", jyspc2
       write(*,*) "-------------------------------------"
    endif
    averStrip=0.0
    numStrip=0.0
    crossSection=0.0
    do j=jyspc1,jyspc2
       if (j .ge. 0 .and. j .le. ny-1) then
          averSpec=0.0
          numSpec=0.0
          flagAux1=0
          do jcross=jspc1,jspc2
             if (jcross .ge. 0 .and. jcross .le. nx-1) then
                if (oswImage(jcross,j) .ge. 1.0) then ! -0.9) then
                   averSpec=averSpec+oswImage(jcross,j)
                   crossSection(jcross)=crossSection(jcross)+oswImage(jcross,j)
                   numSpec=numSpec+1.0
                else
                   flagAux1=-1
                   averSpec=-abs(averSpec)
                endif
             endif
          enddo
          if (numSpec .gt. 1.0 .and. flagAux1 .eq. 0) then
             averSpec=averSpec/numSpec ! per pixel
             specVect(j)=averSpec
             if (j .gt. jyspc1 .and. j .lt. jyspc2) then
                averStrip=averStrip+averSpec
                numStrip=numStrip+1.0
             endif
          else
             specVect(j)=-1.0
          endif
       endif
    enddo

    if (verbFlag .eq. 1) then
       !write(*,*) "L728 =========== nPixSpec=",nPixSpec
       !do j=jyspc1,jyspc2
       !   write(*,*) "L736 j=", j, " specVect=", specVect(j)
       !enddo
       do jcross=jspc1,jspc2
          write(*,*) "L955 jcross=", jcross, " crossSection=", crossSection(jcross)
       enddo
       write(*,*) "==============="
    endif

!    if (nPixSpec .ge. 3) then
!       nLat=nPixSpec/3
!    else
       nLat=1
!    endif
    nCentr=nPixSpec-2*nLat
    if (nCentr .le. 0) nCentr=1
    ! compute the lateral level (supposed to be the background)
    crossLateral=0.0

    if (verbFlag .eq. 1) then
       write(*,*) "L971 jspc1=", jspc1, " jspc2=", jspc1+nLat-1
    endif
    do jcross=jspc1,jspc1+nLat-1
       if (verbFlag .eq. 1) then
          write(*,*) "L975 jcross=", jcross, " crossSection=", crossSection(jcross), " ..."
       endif
       crossLateral=crossLateral+crossSection(jcross)
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) "L980 jspc1=", jspc2-nLat+1, " jspc2=", jspc2
    endif

    do jcross=jspc2-nlat+1,jspc2
       if (verbFlag .eq. 1) then
          write(*,*) "L985 jcross=", jcross, " crossSection=", crossSection(jcross), " ..."
       endif
       crossLateral=crossLateral+crossSection(jcross)
    enddo
    crossLateral=crossLateral/real(2*nLat)
    ! Compute the central part of the cross-section (signal)
    crossCentral=0.0

    if (verbFlag .eq. 1) then
       write(*,*) "L994 jspc1=", jspc1+nLat, " jspc2=", jspc1+nLat+nCentr-1
    endif
    do jcross=jspc1+nLat, jspc1+nLat+nCentr-1
       if (verbFlag .eq. 1) then
          write(*,*) "L998 jcross=", jcross, " crossSection=", crossSection(jcross), " ..."
       endif
       if (crossSection(jcross) .gt. crossCentral) then
          !crossCentral=crossCentral+crossSection(jcross)
          crossCentral=crossSection(jcross)
       endif
    enddo
    !crossCentral=crossCentral/real(nCentr)
    crossShape=crosscentral-crossLateral
    !if (crossLateral .gt. 1.0) crossShape=crossShape/crossLateral
    if (crossLateral .gt. 1.0) crossShape=crossShape/crossCentral
    if (verbFlag .eq. 1) then
       write(*,*) "L1010 crossLateral=", crossLateral, " crossCentral=", crossCentral, &
            " crossShape=", crossShape
    endif
!-----------
    if (numStrip .gt. 1.0) averStrip=averStrip/numStrip ! per pixel
    averStrip=smoothBkgExcludeZeros(specVect,jyspc1,jyspc2,ny-1,verbFlag)
    s2nRatio=averStrip-(averBkgL+averBkgR)/2.0
    if ((averBkgL+averBkgR)/2.0 .gt. 1.0) then
       s2nRatio=s2nRatio/(sqrt((averBkgL+averBkgR)/2.0))
    endif
    s2n=s2nRatio
    if (verbFlag .eq. 1) then
       write(*,*) " L1022 averStrip=", averStrip, " numStrip=", numStrip, &
            " above the bkg=", averStrip-(averBkgL+averBkgR)/2.0, " s/n=", &
            s2nRatio
       write(*,*) "L1025 averBkgL=", averBkgL, " averBkgR=", averBkgR
    endif
    
    if (s2nRatio .lt. 0.0 .and. crossShape .gt. 0.0) then 
       ! something is wrong with the background
       if (verbFlag .eq. 1) write(*,*) "L850 something is wrong with the background"
       ! 15.05.2007 it is likely that one of the background regions coincides with a bright source
       if (averStrip .gt. averBkgL .and. averStrip .lt. averBkgR) then
          computeBkgR=.false.
          if (verbFlag .eq. 1) then
             write(*,*) "L1035 ignoring the RIGHT background strip"
          endif
       endif
       if (averStrip .lt. averBkgL .and. averStrip .gt. averBkgR) then
          computeBkgL=.false.
          if (verbFlag .eq. 1) then
             write(*,*) "L1041 ignoring the LEFT background strip"
          endif
       endif
       ! re-computing the s/n ratio
       if (.not. computeBkgL) then ! using only the right background
          s2nRatio=averStrip-averBkgR
          if (averBkgR .gt. 1.0) then
             s2nRatio=s2nRatio/sqrt(averBkgR)
             s2n=s2nRatio
          endif
       endif
       if (.not. computeBkgR) then ! using only the left background
          s2nRatio=averStrip-averBkgL
          if (averBkgL .gt. 1.0) then
             s2nRatio=s2nRatio/sqrt(averBkgL)
             s2n=s2nRatio
          endif
       endif
       if (verbFlag .eq. 1) then
          write(*,*) "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
          write(*,*) "L1060 New s/n ratio=", s2nRatio
          write(*,*) "++++++++++++++++++++++++++++++++++++++++++++++++++++++"
       endif
    endif
    !*******************************************************************************
    !smoothing the background 
    !
    do j=0,ny-1
       if (bkgVectL(j) .gt. 0.0) then 
          smoothBkgVectL(j)=smoothMedian(bkgVectL,j,ny-1)
       else
          smoothBkgVectL(j)=0.0
       endif
       if (bkgVectR(j) .gt. 0.0) then
          smoothBkgVectR(j)=smoothMedian(bkgVectR,j,ny-1)
       else
          smoothBkgVectR(j)=0.0
       endif
    enddo

    do j=0,ny-1
       if (smoothBkgVectL(j) .gt. 0.0) then
          bkgVectL(j)=smoothAver(smoothBkgVectL,j,ny-1)
       else
          bkgVectL(j)=0.0
       endif
       if (smoothBkgVectR(j) .gt. 0.0) then
          bkgVectR(j)=smoothAver(smoothBkgVectR,j,ny-1)
       else
          bkgVectR(j)=0.0
       endif
    enddo

    avStrip=0.0
    avStripN=0.0
    ! VNY 14.07,.2004
    p1sum=0.0
    p2sum=0.0
    p3sum=0.0
    p4sum=0.0
    p5sum=0.0
    nPsum=0.0
    !
    p1vect=0.0
    p2vect=0.0
    p3vect=0.0
    p4vect=0.0
    p5vect=0.0
    pAux=0.0

    ! Subtract the background
    do j=0,ny-1
       ! Calculation of average value for the strip
       a0=0.0 ! average value
       a1=0.0 ! number of averaging pixels in the strip     
       ! Compute the mean background value 
       levelLR=0.0
       nLevel=0.0
       ! across the left bkg-strip
       if (xBkgL .gt. 0.0 .and. computeBkgL) then
          levelLR=levelLR+bkgVectL(j)
          nLevel=nLevel+1.0
       endif
       ! across the right bkg-strip
       if (xBkgR .gt. 0.0 .and. computeBkgR) then
          levelLR=levelLR+bkgVectR(j)
          nLevel=nLevel+1.0
       endif
       ! if there is only one background strip then 
       ! levelLR would correspond to this one strip
       if (nLevel .gt. 1.0) levelLR=levelLR/nLevel 

       averSpec=0.0 ! average value for the spectrum cross-section with
       averSpec2=0.0
       ! interpolation of the background
!!!!!!!

!!!!!!
       !---11.01.05
       if (verbFlag .eq. 1 .and. j .eq. 540) then
          write(*,*) "...........>"
          write(*,*) "L1140 bkgVectL(j)=", bkgVectL(j), " bkgVectR(j)=", &
               bkgVectR(j), " xBkgL=", xBkgL, " xBkgR=", xBkgR
       endif
       numSpec=0.0
       if (jspc1 .gt. 0 .and. jspc2 .lt. nx-1 .and. jspc2 .gt. jspc1) then
          ! if both edges are within the OSW then interpolate
          ! supposing that the background levels at the left and right
          ! edges of the strip are bkgVectL(j) and bkgVectR(j), respectively

!--- 11.01.05 correction of the left anf right interpolation points
!          wSpec=real(jspc2-jspc1+2)
!          if (verbFlag .eq. 1 .and. j .eq. 650) then
!             write(*,*) " L584 wSpec=", wSpec, " jspc1=", jspc1, " jspc2=", jspc2 
!          endif
!---

! 27.01.2005 over again, correcting the left and right interpolation points
!------
!          wSpec=real(xBkgR-xBkgL+1)
          wSpec=real(xBkgR-xBkgL)
!------
          if (verbFlag .eq. 1 .and. j .eq. 540) then
             write(*,*) " L1162 wSpec=", wSpec 
          endif
!---
!--
          if (bkgVectL(j) .eq. 0.0 .or. .not. computeBkgL) bkgVectL(j)=bkgVectR(j)
          if (bkgVectR(j) .eq. 0.0 .or. .not. computeBkgR) bkgVectR(j)=bkgVectL(j)
          dSpec=bkgVectR(j)-bkgVectL(j)
          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
          if (verbFlag .eq. 1 .and. j .eq. 540) then
             write(*,*) "L1171 jspc1=", jspc1, " jspc2=", jspc2
          endif
          do jcross=jspc1,jspc2
!--- 11.01.05 correction
!             wSpec=real(jcross-jspc1+1)
             wSpec=real(jcross-xBkgL)
!--
             levelLR=bkgVectL(j)+dSpec*wSpec
             if (jcross .lt. nx .and. jcross .gt. 0) then
                averSpec2=averSpec2+oswImage(jcross,j)-levelLR
                if (oswImage(jcross,j) -levelLR .gt. 0.0) then
                   averSpec=averSpec+oswImage(jcross,j)-levelLR
                endif
!------------------------ 11.01.05
                !if (verbFlag .eq. 1 .and. j .eq. 540) then
                !   write(*,*) " jcross=",jcross, " levelLR=", &
                !        levelLR, " dif=", oswImage(jcross,j)-levelLR
                !endif ! verbFlag .eq. 1
!------------------------ 11.01.05
             endif
          enddo
          !if (verbFlag .eq. 1 .and. j .eq. 900) then
          !   write(*,*) " L662 averSpec=", averSpec, " averSpec2=", averSpec2
          !   write(*,*) " j=",j, " jspc1=", jspc1, " jspc2=", jspc2
          !   write(*,*) " ------------------"
          !   do jcross=xBkgL,xBkgR
          !      write(*,*) " jcross=", jcross, " oswImage=", oswImage(jcross,j)
          !   enddo
          !   write(*,*) " ------------------"
          !   !----
          !   ndata=xBkgR-xBkgL+1
          !   allocate(xdata(1:ndata))
          !   allocate(ydata(1:ndata))
          !   !----
          !   ! plotting the graph
          !   if (pgbeg(0,'plot250405/PS',1,1)/=1) then
          !      write(*,*) " OMGEXTRACTION: L1889 cannot opent plot device"
          !   else
          !      write(*,*) " L678 The plot device is open"
          !      xlabel="pixels"
          !      ylabel="counts"
          !      graphTitle="plot"
          !      do jcross=xBkgL,xBkgR
          !         ydata(jcross-xBkgL+1)=oswImage(jcross,j)
          !         xdata(jcross-xBkgL+1)=jcross
          !      enddo                
          !      x1=0.0
          !      x2=1.0
          !      y1=0.0
          !      y2=1.0
          !      write(*,*) " x1=", x1, " x2=", x2, " y1=", y1, " y2=", y2
          !      call DrawGraph(ndata, xdata, ydata, &
          !           graphTitle, xlabel, ylabel, x1, x2, y1, y2)
          !      call pgend
          !      write(*,*) " L1895 The plot device is closed"
          !   endif
          !   !---
          !   deallocate(xdata)
          !   deallocate(ydata)
          !   !---!
!
!          endif ! verbFlag .eq. 1

          !otherwise compute just the average value    
       else
          do jcross=jspc1,jspc2
             if (jcross .gt. 0 .and. jcross .lt. nx-1) then
                averSpec2=averSpec2+oswImage(jcross,j)-levelLR               
                if (oswImage(jcross,j)-levelLR .gt. 0.0) then
                   averSpec=averSpec+oswImage(jcross,j)-levelLR
                endif
             endif
          enddo
       endif
! VNY 22.02.2005
       if (verbFlag .eq. 1 .and. j .eq. 540) then
          write(*,*) "L1247 averSpec2=", averSpec2, " averSpec=", averSpec
       endif
       averSpec=averSpec2
!
       ! VNY test 08.01.2004
       jstart=jspc1-10
       jend=jspc2+10
       if (jstart .lt. 0) jstart=0
       if (jend .gt. nx-1) jend=nx-1
       !if (verbFlag .eq. 1) then
       !   if (j .eq. 1041) then ! jyspc1+10) then
       !      write(*,*) " j=", j, ":"
       !      write(*,*) "------------"
       !      do jcross=jstart, jend
       !         write (*,*) " jcross=", jcross, " i=", oswImage(jcross,j)
       !      enddo
       !      write(*,*) "------------"
       !      !write(*,*) " iR=", iR, " iL=", iL, " iR-il-9=", iR-iL-9
       !   endif
       !endif
       !---
       !------
       !     if (j .ge. 600 .and. j .lt. 603) then
       !        write(*,*) " j=", j, " averSpec=", averSpec, " levelLR=", levelLR
       !     endif
       !---

       !nPts=iR-iL-9
       ! VNY 29.03.2004
       !nPts=jspc2-jspc1+1
       nPts=jend-jstart+1
       !if (verbFlag .eq. 1 .and. j .eq. 540) then
       !   write (*,*) "L663: j=", j, " nPts=", nPts, " jstart=", &
       !     jstart, " jend=", jend, " averSpec=", averSpec
       !endif

       allocate(xinput(1:nPts))
       allocate(yinput(1:nPts))
       allocate(weights(1:nPts))
       !---VNY 15.01.2004
       allocate(yp(1:nPts))
       !allocate(temp(1:nPts*9))
       allocate(temp(1:nPts))
       allocate(ysp(1:nPts))
       allocate(dispersion(1:nPts))

       ! VNY 14.01.2004
       !numparam=4
       !numparam=9
       numparam=5
       ! Different  extraction modes

!---------------------------------------------------- 14.07.2004
       do i=jstart,jend
          xinput(i-jstart+1)=real(i)
          if (i .gt. 0 .and. i .lt. nx)  then
             yinput(i-jstart+1)=oswImage(i,j)
          endif       
       enddo
!----------------------------------------------------
! VNY 14.07.2004 Subtracting the interpolated background from the image 
! VNY 15.07.2004 (for the case of optimal or Gaussian extraction)
       !**********************************************************************
       ! Extraction modes with smoothing
       !**********************************************************************
       if (currentExtractionMode .eq. 1 &
            .or. currentExtractionMode .eq. 2 &
            .or. currentExtractionMode .eq. 3) then
          dSpec=bkgVectR(j)-bkgVectL(j)
          wSpec=real(nPts)
          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
          !if (verbFlag .eq. 1 .and. j .eq. 540) write(*,*) "aaaaaaaaaaaaaaaaaaaaaaaaaaa"
          do i=jstart,jend
             wSpec=real(i-jstart+1)
             levelLR=bkgVectL(j)+dSpec*wSpec
             xinput(i-jstart+1)=real(i)
             if (i .gt. 0 .and. i .lt. nx)  then
                yinput(i-jstart+1)=oswImage(i,j)-levelLR
             endif
             ! 25.04.2005
             if (verbFlag .eq. 1 .and. j .eq. 540) then
                write(*,*) "L1328 i=",i, " yinput=", yinput(i-jstart+1)
             endif
             !
          enddo
          if (verbFlag .eq. 1 .and. j .eq. 540) then
             write(*,*) "L1333 *************  currentExtractionMode=", currentExtractionMode
          endif
          ! 25.04.2005
          !if (verbFlag .eq. 1 .and. j .eq. 900) then
          !   !----
          !   ndata=jend-jstart+1
          !   allocate(xdata(1:ndata))
          !   allocate(ydata(1:ndata))
          !   !----
          !   ! plotting the graph
          !   if (pgbeg(0,'plot250405/PS',1,1)/=1) then
          !      write(*,*) " OMGEXTRACTION: L1889 cannot opent plot device"
          !   else
          !      write(*,*) " L809 The plot device is open"
          !      xlabel="pixels"
          !      ylabel="counts"
          !      graphTitle="plot"
          !      do jcross=jstart,jend
          !         ydata(jcross-jstart+1)=oswImage(jcross,j)
          !         xdata(jcross-jstart+1)=jcross
          !      enddo
          !      x1=0.0
          !      x2=1.0
          !      y1=0.0
          !      y2=1.0
          !      write(*,*) " x1=", x1, " x2=", x2, " y1=", y1, " y2=", y2
          !      call DrawGraph(ndata, xdata, ydata, &
          !           graphTitle, xlabel, ylabel, x1, x2, y1, y2)
          !      call pgend
          !      write(*,*) " L825 The plot device is closed"
          !   endif
          !   !---
          !   deallocate(xdata)
          !   deallocate(ydata)
          !   !---
          !endif

          !if (verbFlag .eq. 1 .and. j .eq. 650) write(*,*) "aaaaaaaaaaaaaaaaaaaaaaaaaaa"
       endif ! currentExtractionMode .eq. 1
!-----------------------------------------------

       !  Gaussian fit of the strip
       if (currentExtractionMode .eq. 1 .or. currentExtractionMode .eq. 3) then
          if (verbFlag .eq. 1 .and. j .eq. 900) then
             write(*,*) "%%%%%%%%%%% computing the gaussian approximation for j=", j 
          endif
          ! call gaussiana(xinput,yinput, nPts, param)
          ! VNY 27.01.2004
          verbFlag2=0
          if (verbFlag .eq. 1 .and. j .eq. 900) then
             verbFlag2=1 
          endif
          !if (verbFlag .eq. 1) then
          !   if (j .eq. 650) then
          !      write(*,*) " L714 calling gaussianal9 j=",j
          !      verbFlag2=1
          !   endif
          !endif

!          call gaussianae(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)

          !call gaussianal(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)
          call gaussianalSim(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)

          !call gaussianal9(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)

          ! VNY Test 14.07.2004
          !param5(4)=2.5
          ! ------- 14.07.2004

          if (verbFlag .eq. 1 .and. j .eq. 540) then
             write(*,*) " L1404 after calling gaussianal9: param5=", param5, " nUsedPoints=", nUsedPoints
          endif

          ! VNY 14.07.2004
          p1sum=p1sum+param5(1)
          p2sum=p2sum+param5(2)
          p3sum=p3sum+param5(3)
          p4sum=p4sum+param5(4)
          p5sum=p5sum+param5(5)
          nPsum=nPsum+1.0
          !
          p1vect(j)=param5(1)
          p2vect(j)=param5(2)
          p3vect(j)=param5(3)
          p4vect(j)=param5(4)
          p5vect(j)=param5(5)

          !numparam=9
          numparam=5
          slp1=0.0
          slpn=0.0
          tension=0.9
          islpsw=1 ! means that dispersion is scalar in CURVS
          ierr=0
          dispersion=10.0 ! dispersion of the data points
          smooth=real(2*nPts) ! smoothing factor
          eps= 0.8 ! sqrt(2.0/real(nPts))

          !!           call curvs(nPts, xinput, yinput, dispersion, islpsw, smooth, &
          !!                eps, yp, ysp, tension, temp, ierr)
          !tension=0.001
          !islpsw=3 
          !call curv1(nPts, xinput, yinput, slp1, slpn, islpsw, yp, temp, tension, ierr)
          !!           write (*,*) " L523 ierr=", ierr
          !        call gaussiana2(xinput,yinput, nPts, param, numparam)
          !write(*,*) " L565: computed parameters: param5=", param5
          !sum1=gFunc(param, numparam, xinput, yinput, nPts)
          !write(*,*) " L517; residual=", sum1
          !        write(*,*) " RESULT AFTER CALLING GAUSSIANA:"
          !           write(*,*) " ysp=", ysp
          !           write(*,*) " tension=", tension
          !write(*,*) " --------------------------------"


          optimal1=0.0
          optimal2=0.0
          gaussum=0.0
          simplesum=0.0
          znam=0.0
          sumWeights=0.0
          if (nUsedPoints .gt. 9) then

             if (verbFlag .eq. 1 .and. j .eq. 900) then
                !----
                ndata=nPts
                allocate(xdata(1:ndata))
                allocate(ydata(1:ndata))
             endif

             do i=1, nPts
                xtest=xinput(i)
                !   ! VNY 14.01.2004
                !!           !call fgauss(xtest,param,ytest,dydptest,numparam)
                !!           call fgaussup(xtest,param,ytest,dydptest,numparam)
                ! VNY 15.01.2003!
                !              ytest=curv2(xtest, nPts, xinput, yinput, ysp, tension)
                ! VNY 27.01.2004
                numparam=5
                call fgaussl(xtest,param5,ytest,dydptest5,numparam)
                aux1=param5(1)+param5(5)*xtest
                aux2=1.0
                razn=0.0
! VNY 22.02.2005
                if (yinput(i) .gt. 0.0) then
                   aux2=yinput(i)
!                   razn=yinput(i)-aux1
                endif
                razn=yinput(i)-aux1
!
                weights(i)=ytest-aux1 !aux2
                sumWeights=sumWeights+ytest-aux1 !aux2
                aux3=(ytest-aux1)/aux2
                aux2=(ytest-aux1)*(ytest-aux1)/aux2 ! weight
                optimal1=optimal1+razn*aux3 ! /(ytest-aux1) ! it is divided to make not squared
                znam=znam+aux2 ! znameatel is squared
                gaussum=gaussum+ytest-aux1 ! This will give an estimation for the pure
                ! gaussian
                simplesum=simplesum+razn ! This is a simple sum of the spectrum
                ! minus background
                if (verbFlag .eq. 1 .and. j .eq. 900) then
                   write(*,'(a,i3,a,f7.1,a,f7.1,a,f7.1,a, f7.1, a, f7.1,a,f8.1, a, f7.1)') &
                        "i=", i, &
                        " x(i)=", xtest, " y(i)=",yinput(i), &
                        " ytest=", ytest, " razn=", razn, &
                        " aux2=", aux2, " opt1=", optimal1, " aux1=", aux1
                   !!           !write(*,*) " dydp=", dydptest
                   xdata(i)=i
                   ydata(i)=aux2 ! ytest
                endif
             enddo
             ! 25.04.2005
             if (verbFlag .eq. 1 .and. j .eq. 900) then
                !----
                !ndata=nPts
                !allocate(xdata(1:ndata))
                !allocate(ydata(1:ndata))
                !----
                ! plotting the graph
                if (pgbeg(0,'plot250405/PS',1,1)/=1) then
                   write(*,*) " OMGEXTRACTION: L1513 cannot opent plot device"
                else
                   write(*,*) " L959 The plot device is open"
                   xlabel="pixels"
                   ylabel="counts"
                   graphTitle="plot"
                   !do jcross=1,nPts
                   !   ydata(jcross+1)=ytest(jcross)
                   !   xdata(jcross+1)=jcross
                   !enddo
                   x1=0.0
                   x2=1.0
                   y1=0.0
                   y2=1.0
                   write(*,*) " x1=", x1, " x2=", x2, " y1=", y1, " y2=", y2
                   call DrawGraph(ndata, xdata, ydata, &
                        graphTitle, xlabel, ylabel, x1, x2, y1, y2)
                   call pgend
                   write(*,*) " L1531 The plot device is closed"
                endif
                !---
                deallocate(xdata)
                deallocate(ydata)
                !---
             endif

          else
             simplesum=0.0
             znam=1.0
             do i=1,nPts
                aux1=param5(1)
                razn=0.0
! VNY 22.02.2005
!                if (yinput(i) .gt. 0.0) then
                   razn=yinput(i)-aux1
!                endif
!
                simplesum=simplesum+razn
             enddo
             gaussum=simplesum
             optimal1=simplesum
          endif ! nUsedPoints .gt. 9
          if (verbFlag .eq. 1 .and. j .eq. 540) then
             write(*,*) "%%%%%%%%%%%"
             write(*,*) " L1557 optimal1=", optimal1, " znam=", znam !, " optimal2=", optimal2
             write(*,*) " gaussum=", gaussum, " simplesum=", simplesum
          endif
          if (znam .gt. 0.0) optimal1=optimal1/znam
          !
          if (verbFlag .eq. 1 .and. j .eq. 540) then
             write(*,*) " L1563 optimal1/znam=", optimal1
             write(*,*) "%%%%%%%%%%%"
          endif
          !
          if (nUsedPoints .gt. 9) then 
             aux2=0.0
             optimal1=0.0
             ! normalise the weights to unity
             max1=0.0
             if (sumWeights .gt. 0.0) then
                do i=1, nPts
                   weights(i)=weights(i)/sumWeights
                enddo
             endif
             if (verbFlag .eq. 1 .and. j .eq. 540) then
                !write(*,*) " Maxumal value of weights=", max1
                write(*,*) " L1579 sumWeight=", sumWeights
             endif
             znam=0.0
             do i=1, nPts
                xtest=xinput(i)
                call fgaussl(xtest,param5,ytest,dydptest5,numparam)
                aux1=param5(1)+param5(5)*xtest
                aux2=1.0
                razn=0.0
! VNY 22.02.2005
!                if (yinput(i) .gt. 0.0) then
                   aux2=yinput(i)
                   razn=yinput(i)-aux1
!                endif
!
                aux3=weights(i)/aux2
                aux2=weights(i)*weights(i)/aux2
                optimal1=optimal1+razn*aux3
                znam=znam+aux2
                !if (verbFlag .eq. 1 .and. j .eq. 540) then
                !   write(*,*) "L810 i=", i, " weight(i)=", weights(i), " w/sum=", weights(i)/znam
                !endif
             enddo
             if (znam .gt. 0.0) optimal1=optimal1/znam
             if (verbFlag .eq. 1 .and. j .eq. 540) then
                write(*,*) "%%%%%%%%%%%"
                write(*,*) " L1605 znam=", znam, " optimal1/aux2=", optimal1 
             endif
             !
          else
             simplesum=0.0
             znam=1.0
             do i=1,nPts
                aux1=param5(1)
                razn=0.0
                if (yinput(i) .gt. 0.0) then
                   razn=yinput(i)-aux1
                endif
                simplesum=simplesum+razn
             enddo
             gaussum=simplesum
             optimal1=0.0 !simplesum
          endif ! nUsedPoints .lt. 9
          !if (verbFlag .eq. 1 .and. j .eq. 650) then
          !   write(*,*) " j=", j, " optimal1=", optimal1, " simpleSum=", simpleSum, " nUsedPoints=", nUsedPoints
          !endif
          if (optimal1 .lt. 0.0) optimal1=0.0
       endif ! extractionMode .eq. 1 .or. extractionMode .eq. 3
       ! VNY 22.01.2004
       !--- 

       averSpecLowess=0.0
       if (currentExtractionMode .eq. 2) then
          !if (verbFlag .eq. 1 .and. j .eq. 866) then !!!!!!
          !   write(*,*) "L730 %%%%%%%%%%% computing the lowess approximation for j=", j
          !endif
!          smooth=0.3 ! smoothing factor for the lowess approximation algorithm
          smooth=0.2 ! smoothing factor for the lowess approximation algorithm
          numsteps=0
          delta=0.0
          call lowess(xinput, yinput, nPts, smooth, numsteps, delta, &
               ysp, yp, dispersion)


          ! if (verbFlag .eq. 1) then !  .and. j .eq.  757) then !!!!!!!!
             ! Find maximum and two minima
             max1=0.0
             imax1=-1
             do i=1, nPts
                if (ysp(i) .gt. max1) then
                   max1=ysp(i)
                   imax1=i
                endif
             enddo
             !if (verbFlag .eq. 1 .and. j .eq.  550) then !!!!!!!!
             !   write(*,*) " L751 found a maximum at i=", imax1, " value=", max1
             !endif
             imin1=1
             imin2=nPts
             if (imax1 .gt. 1 .and. imax1 .lt. nPts) then
                ! Find two minima
                imin1=1
                min1=ysp(1)
                ! first find a positive number
                do i=1,imax1-1
                   imin1=i
                   min1=ysp(i)
                   if (yinput(i) .gt. 0.0) then 
                      exit
                   endif
                enddo
                !if (verbFlag .eq. 1 .and. j .eq.  650) then !!!!!!!!
                !   write(*,*) " L809 the first starting point is  at i=", imin1, " value=", min1
                !endif
                do i=imin1+1, imax1-1
!                   if (ysp(i) .lt. min1) then
                   if (ysp(i) .lt. min1 .and. ysp(i) .gt. 0.0) then
                      min1=ysp(i)
                      imin1=i
                   endif
                enddo
                imin2=nPts
                min2=ysp(nPts)
                do i=imax1+1, nPts-1
!                   if (ysp(i) .lt. min2) then
                   if (ysp(i) .lt. min2 .and. ysp(i) .gt. 0.0) then
                      min2=ysp(i)
                      imin2=i
                   endif
                enddo
             endif
             !if (verbFlag .eq. 1 .and. j .eq.  860) then !!!!!!!!
             !   write(*,*) " L758 two minima: i1=", imin1, " value1=", min1, &
             !        " i2=", imin2, " value2=", min2
             !   write(*,*) " --------------------------------"!
             !endif
             ! Find maximum and two minima
             max1=0.0
             imax1=-1
             !if (verbFlag .eq. 1 .and. j .eq.  860) then !!!!!!!!
             !   do i=1, nPts
             !      write(*,*) "i=", i, " x(i)=", xinput(i), " y(i)=",yinput(i), " ysp(i)=", ysp(i)
             !   enddo
             !endif
          !endif (verbFlag .eq. 1
          ! Compute averSpecLowess
          averSpecLowess=0.0
          wSpec=real(imin2-imin1)
          dSpec=ysp(imin2)-ysp(imin1)
          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
          do jcross=imin1,imin2
             wSpec=real(jcross-imin1)
             levelLR2=ysp(imin1)+dSpec*wSpec
             if (yinput(jcross) .gt. 0.0 .and. ysp(jcross) .gt. 0.0) then
                if (ysp(jcross)-levelLR2 .gt. 0.0) then
                   averSpecLowess=averSpecLowess+ysp(jcross)-levelLR2
                endif
             endif
             !if (verbFlag .eq. 1 .and. j .eq. 860) then
             !   write(*,*) " L742 j=", jcross, " level=", levelLR, " diff=", &
             !        ysp(jcross)-levelLR
             !endif
          enddo

          ! Compute averSpecSimple
!!          averSpecSimple=0.0
!!          wSpec=real(imin2-imin1)
!!          dSpec=yinput(imin2)-yinput(imin1)
!!          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
!!          do jcross=imin1,imin2
!!             wSpec=real(jcross-imin1)
!!             levelLR=yinput(imin1)+dSpec*wSpec
!!             if (yinput(jcross) .gt. 0.0) then
!!                if (yinput(jcross)-levelLR .gt. 0.0) then
!!                   averSpecSimple=averSpecSimple+yinput(jcross)-levelLR
!!                endif
!!             endif
!!          enddo
          
          !if (verbFlag .eq. 1 .and. j .eq. 860) then
          !   write(*,*) "-----------------------------------_"
          !   write(*,*) " L608: averSpec=", averSpec, " averSpecSimple=", &
          !        averSpecSimple," averSpecLowess=", averSpecLowess
          !   write(*,*) "%%%%%%%%%%%"
          !endif
       endif ! extractionMode .eq. 2
       !--- 

       ! Calculate the sums
       sum1=0.0
       sum2=0.0
       sumTest=0.0
       sumGauss=0.0
       !if (verbFlag .eq. 1) then
!      !    if (j .eq. int(yFirst)) then
       !   if (j .eq. 757) then
       !      write(*,*) " L659 nPts=", nPts, " j=", j
       !      write(*,*) " ------------------"
       !      do i=1,nPts
       !         write(*,*) " i=", i, " x=", xinput(i), " y=", yinput(i)
       !      enddo ! i=1,nPts
       !      write(*,*) " ------------------"
       !   endif ! j .eq. int(yFirst)
       !endif ! verbFlag .eq. 1
       !..........................................
       !
       ! The following peice of code should be uncommented for the optimal extraction algorithm
       !                     ! Optimal extraction
       !                     do i=1, nPts
       !                        xtest=xinput(i)
       !                        !numparam=4
       !                        call fgauss(xtest,param,ytest,dydptest,numparam)
       !                        ytest=ytest-param(1)
       !                        sigm=sqrt(yinput(i))
       !                        !sigm=sqrt(abs(yinput(i)-param(1)))
       !                        !sigm=1.0
       !                        if (sigm .eq. 0.0) sigm=1.0
       !                        ! VNY 6.02.03 Using the smoothed background
       !!                        sum1=sum1+(yinput(i)-param(1))*ytest*ytest/(sigm*sigm) 
       !                        sum1=sum1+(yinput(i)-bkgVect3(j))*ytest*ytest/(sigm*sigm) 
       !                        sum2=sum2+ytest*ytest/(sigm*sigm)
       !!                        sumTest=sumTest+(yinput(i)-param(1))
       !                        sumTest=sumTest+(yinput(i)-bkgVect3(j))
       !                        sumGauss=sumGauss+ytest ! sumGauss is the integral over the gaussian calculated 
       !!                                                 with the use of parameters determined by gaussiana()
       !                     enddo
       !...........................................
       sumWeight=sum1 ! sumWeight is the result of optimal extraction
       ! 29.03.2007 since sum2 is not changed, eliminating the following three lines
       !if (sum2 .gt. 0.0) then 
       !   sumWeight=sumWeight/sum2
       !endif
       !
       !if (j .gt. y1 .and. j .le. y1+10) then
       !!   write(*,*) "L891; j=", j, "a0=", a0, " a1=", a1, &
       !!        " iL+4=", iL+4, " nPts=", nPts
       !   write(*,*) " param=", param
       !   write(*,*) " sum1=", sum1, " sum2=", sum2, " s1/s2=", sumWeight, &
       !        "sumTest=", sumTest, " sumGauss=", sumGauss
       !   write(*,*) "........"
       !endif

       !if (j .gt. y1+9 .and. j .le. y1+10) then
       !   write(*,*) "%%%%%%%%%%% j=", j
       !   do i=1, nPts
       !      xtest=xinput(i)
       !      call fgauss(xtest,param,ytest,dydptest,numparam)
       !      write(*,*) "i=", i, " y(i)=",yinput(i), " ytest=", ytest
       !   enddo
       !   write(*,*) "%%%%%%%%%%%"
       !endif
       deallocate(xinput)
       deallocate(yinput)
       deallocate(weights)
       deallocate(yp)
       deallocate(temp)
       deallocate(ysp)
       deallocate(dispersion)
       !endif !(j .ge. y1 .and. j .le. y2)
       !----------------------------------------------------
       !
       !**********************************
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! The following line should be uncommented for the optimal extraction algorithm
       !                     rateSourceOutArray(j)=sumWeight  ! Optimal extraction value 
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! rateSourceOutArray(j)=sumGauss ! approximation of data with a Gaussian
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! The following line should be uncommented for the normal extraction algorithm
       !rateSourceOutArray(j)=a0 !oswImage(j1,i) ! this is just the average across the spectrum
       !                 !    !if (a1 .gt. 0.0) rateSourceOutArray(j)=rateSourceOutArray(j)/a1
       !rateSourceErrOutArray(j)=0.0
       !if (rateSourceOutArray(j) .gt. 1.0) rateSourceErrOutArray(j)=sqrt(rateSourceOutArray(j))
       !---VNY 30.08.03
       !!spectrumArray(j)=a0 ! this is just the average across the spectrum
       ! VNY 13.11.2003
       ! VNY 27.01.2004
       ! For the extractionMode=0 the output spectrum array takes the average across 
       spectrumArray(j)=averSpec ! this is the average with interpolation of the backgroung
       !spectrumArray(j)=averSpecSimple 
       !if (verbFlag .eq. 1 .and. j .eq. 650) then
       !   write(*,*) "%%%%%%%%%%%%"
       !   write(*,*) " L836 j= 757, averSpec=", averSpec, " simplesum=", simplesum
       !   write(*,*) "%%%%%%%%%%%%"
       !endif
       ! If extractionMode=2 then the output is the integral of the spline-smoothed (lowess) cross-section
       if (currentExtractionMode .eq. 2) spectrumArray(j)=averSpecLowess
       if (currentExtractionMode .eq. 1) spectrumArray(j)=optimal1
       if (currentExtractionMode .eq. 3) spectrumArray(j)=gaussum
       !if (verbFlag .eq. 1) then
       !   write(*,'(a,i5,a,f5.1,a,f5.1,a,f5.1,a,f5.1)') " L968 j=", j, &
       !        " simple=", averSpec, " optimal=", optimal1, &
       !        " gauss=", gaussum, " spline=", averSpecLowess
       !endif

       !------
       !     if (j .ge. 600 .and. j .lt. 603) then
       !        write(*,*) " j=", j, " averSpec=", averSpec, " levelLR=", levelLR
       !     endif
       !---
       !---
       avStrip=avStrip+averSpec
       avStripN=avStripN+1.0
       !---
       ! across the spectrum
       spectrumErrArray(j)=0.0
       if (spectrumArray(j) .gt. 1.0) spectrumErrArray(j)= &
            sqrt(spectrumArray(j))
       !---
       ! VNY 17.10.2003 here the ISF calibration must go
       ! VNY 18.11.2003 the ISF calibration can be implemented outside this subroutine
       !     fluxOutArray0(j)=nullValue
       !     fluxErrOutArray0(j)=nullValue
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! The following line should be  uncommented  for normal extraction
       !if (a1 .ge. 2.0) rateSourceErrOutArray(j)=rateSourceErrOutArray(j)/sqrt(a1-1.0)
       !---
       if (a1 .ge. 2.0) spectrumErrArray(j)=spectrumErrArray(j)/sqrt(a1-1.0)
       !---
       ! The following line should be uncommented for the optimal extraction algorithm
       !                  if (sumWeight .ge. 2.0) rateSourceErrOutArray(j)=rateSourceErrOutArray(j)/sqrt(sumWeight-1.0)
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       !                  rateBackOutArray(j)=levelLR
       ! The following line should be uncommeted for the optimal extraction algorithm
       !                 rateBackOutArray(j)=param(1)
       !.............The three following lines should be uncommented for the normal extraction
       !rateBackOutArray(j)=bkgVect2(j)
       !rateBackErrOutArray(j)=0.0
       !if (rateBackOutArray(j) .gt. 0.0) rateBackErrOutArray(j)=sqrt(rateBackOutArray(j)) 
       !--VNY 30.08.03
       bkgArray(j)=bkgVect2(j)
       ! VNY 13.11.2003 
       bkgArray(j)=levelLR !(bkgVectR(j)+bkgVectL(j))/2

       bkgErrArray(j)=0.0
       if (bkgArray(j) .gt. 0.0) bkgErrArray(j)=sqrt(bkgArray(j)) 
       !---

       !!                  if (levelLR .gt. 0.0) rateBackErrOutArray(j)=sqrt(levelLR)
       ! The following line should be uncommented for the optimal extraction algorithm 
       !                  if (param(1) .gt. 0.0) rateBackErrOutArray(j)=sqrt(param(1))
       !^^^^^^^^^^^^^^^^^^^^^^^
       !  Scaling counts
       if (exposureTime .gt. 0.0) then
          spectrumArray(j)=spectrumArray(j)/exposureTime
          spectrumErrArray(j)=spectrumErrArray(j)/exposureTime
          bkgArray(j)=bkgArray(j)/exposureTime
          bkgErrArray(j)=bkgErrArray(j)/exposuretime
       endif
       if (countsScaleFactor(j) .gt. 0.0) then
          spectrumArray(j)=spectrumArray(j)/countsScaleFactor(j)
          spectrumErrArray(j)=spectrumErrArray(j)/countsScaleFactor(j)
          bkgArray(j)=bkgArray(j)/countsScaleFactor(j)
          bkgErrArray(j)=bkgErrArray(j)/countsScaleFactor(j)
       endif
       ! Compute the flux and its error
       !     call CountsToFlux(wavelengthArray(j), spectrumArray(j), spectrumErrArray(j), &
       !          fluxArray(j), fluxErrArray(j), isfLambda, isf, isfError, isfSize)
    enddo ! j=0,ny-1
    ! VNY 14.07.2004
    !write(*,*) "--------- current extr mode=", currentExtractionMode
    !write(*,*) " L1150 sum: p1=", p1sum, " p2=", p2sum, " p3=", p3sum, " p4=", p4sum, &
    !     " p5=", p5sum, " nPsum=", nPsum
    if (nPsum .gt. 0.0) then
       p1sum=p1sum/nPsum
       p2sum=p2sum/nPsum
       p3sum=p3sum/nPsum
       p4sum=p4sum/nPsum
       p5sum=p5sum/nPsum
    endif
    !if (verbFlag .eq. 1) then
    !   write(*,*) " L1159 p1sum=", p1sum, " p2sum=", p2sum, " p3sum=", p3sum, &
    !        " p4sum=", p4sum, " p5sum=", p5sum
    !   write(*,*) "*************************"
    !   do j=600,700
    !      write(*,'(a,i6,a,f6.2,a,f6.2,a,f6.2,a,f6.2,a,f6.2)') " j=", j, " p1=", p1vect(j), " p2=", p2vect(j), &
    !           " p3=", p3vect(j), " p4=", p4vect(j), " p5=", p5vect(j)
    !   enddo
    !endif
    ! smooth the Gaussian parameters
    ! median p1
    if (currentExtractionMode .eq. 1) then
       ! smooth all the Gaussian parameters
       do j=0,ny-1
          if (p1vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p1vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p1
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p1vect(j)=smoothAver(pAux,j,ny-1)
          else
             p1vect(j)=0.0
          endif
       enddo
       !
       ! median p2
       do j=0,ny-1
          if (p2vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p2vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p2
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p2vect(j)=smoothAver(pAux,j,ny-1)
          else
             p2vect(j)=0.0
          endif
       enddo
       !
       ! median p3
       do j=0,ny-1
          if (p3vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p3vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p3
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p3vect(j)=smoothAver(pAux,j,ny-1)
          else
             p3vect(j)=0.0
          endif
       enddo
       !
       ! median p4
       do j=0,ny-1
          if (p4vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p4vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p4
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p4vect(j)=smoothAver(pAux,j,ny-1)
          else
             p4vect(j)=0.0
          endif
       enddo
       !
       ! median p5
       do j=0,ny-1
          if (p5vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p5vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p5
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p5vect(j)=smoothAver(pAux,j,ny-1)
          else
             p5vect(j)=0.0
          endif
       enddo
    endif ! currentExtractionMode .eq. 1
    
    if (currentExtractionMode .eq. 3) then
       ! smooth the gaussian parameters except for the amplitude (p2) 
       do j=0,ny-1
          if (p1vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p1vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p1
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p1vect(j)=smoothAver(pAux,j,ny-1)
          else
             p1vect(j)=0.0
          endif
       enddo
       !
       ! median p2
       do j=0,ny-1
          if (p2vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p2vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p2
       !do j=0,ny-1
       !   if (pAux(j) .gt. 0.0) then
       !      p2vect(j)=smoothAver(pAux,j,ny-1)
       !   else
       !      p2vect(j)=0.0
       !   endif
       !enddo
       !
       ! median p3
       do j=0,ny-1
          if (p3vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p3vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p3
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p3vect(j)=smoothAver(pAux,j,ny-1)
          else
             p3vect(j)=0.0
          endif
       enddo
       !
       ! median p4
       do j=0,ny-1
          if (p4vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p4vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p4
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p4vect(j)=smoothAver(pAux,j,ny-1)
          else
             p4vect(j)=0.0
          endif
       enddo
       !
       ! median p5
       do j=0,ny-1
          if (p5vect(j) .gt. 0.0) then
             pAux(j)=smoothMedian(p5vect,j,ny-1)
          else
             pAux(j)=0.0
          endif
       enddo
       ! averaging p5
       do j=0,ny-1
          if (pAux(j) .gt. 0.0) then
             p5vect(j)=smoothAver(pAux,j,ny-1)
          else
             p5vect(j)=0.0
          endif
       enddo
    endif ! currentExtractionMode .eq. 3
       !
       !if (verbFlag .eq. 1) then
       !   write(*,*) "************************* after smoothing"
       !   do j=600,700
       !      write(*,'(a,i6,a,f6.2,a,f6.2,a,f6.2,a,f6.2,a,f6.2)') " j=", j, " p1=", p1vect(j), " p2=", p2vect(j), &
       !           " p3=", p3vect(j), " p4=", p4vect(j), " p5=", p5vect(j)
       !   enddo
       !endif
       !
       !88888888888888888888888888888888888888888888888888888888

    if (currentExtractionMode .eq. 1 .or. currentExtractionMode .eq. 3) then
       ! Recomputing optimal extraction 
       ! Subtract the background
       do j=0,ny-1

          ! VNY test 08.01.2004
          jstart=jspc1-10
          jend=jspc2+10
          if (jstart .lt. 0) jstart=0
          if (jend .gt. nx-1) jend=nx-1
          nPts=jend-jstart+1

          allocate(xinput(1:nPts))
          allocate(yinput(1:nPts))
          allocate(weights(1:nPts))
          allocate(yp(1:nPts))
          allocate(temp(1:nPts))
          allocate(ysp(1:nPts))
          allocate(dispersion(1:nPts))
          numparam=5

          ! VNY 14.07.2004 Subtracting the interpolated background from the image 
          dSpec=bkgVectR(j)-bkgVectL(j)
          wSpec=real(nPts)
          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
          do i=jstart,jend
             wSpec=real(i-jstart+1)
             levelLR=bkgVectL(j)+dSpec*wSpec
             xinput(i-jstart+1)=real(i)
             if (i .gt. 0 .and. i .lt. nx)  then
                yinput(i-jstart+1)=oswImage(i,j)-levelLR
             endif
             !if (verbFlag .eq. 1 .and. j .eq. 650) then
             !   write(*,*) " i=",i, " yinput=", yinput(i-jstart+1)
             !endif
          enddo
          !if (currentExtractionMode .eq. 1 .or. currentExtractionMode .eq. 3) then
 
          ! VNY 14.07.2004 using smoothed values for the Gaussian
          param5(1)=p1vect(j)
          param5(2)=p2vect(j)
          param5(3)=p3vect(j)
          param5(4)=p4vect(j)
          param5(5)=p5vect(j)

          ! VNY Test 14.07.2004 
          !          param5(4)=1.5


          numparam=5
          optimal1=0.0
          optimal2=0.0
          gaussum=0.0
          simplesum=0.0
          znam=0.0
          sumWeights=0.0
          if (nPts .gt. 9) then
             do i=1, nPts
                xtest=xinput(i)
                numparam=5
                call fgaussl(xtest,param5,ytest,dydptest5,numparam)
                aux1=param5(1)+param5(5)*xtest
                aux2=1.0
                razn=0.0
                if (yinput(i) .gt. 0.0) then
                   aux2=yinput(i)
                   !razn=yinput(i)-aux1
                   razn=yinput(i)
                endif
                weights(i)=ytest-aux1
                sumWeights=sumWeights+ytest-aux1
                aux3=(ytest-aux1)/aux2
                aux2=(ytest-aux1)*(ytest-aux1)/aux2 ! weight
                optimal1=optimal1+razn*aux3 ! /(ytest-aux1) ! it is divided to make not squared
                znam=znam+aux2 ! znameatel is squared
                gaussum=gaussum+ytest-aux1 ! This will give an estimation for the pure gaussian
                simplesum=simplesum+razn ! This is a simple sum of the spectrum
             enddo
          else
             simplesum=0.0
             znam=1.0
             do i=1,nPts
                aux1=param5(1)
                razn=0.0
                if (yinput(i) .gt. 0.0) then
                   !razn=yinput(i)-aux1
                   razn=yinput(i)
                endif
                simplesum=simplesum+razn
             enddo
             gaussum=simplesum
             optimal1=simplesum
          endif ! nPts .gt. 9
          if (znam .gt. 0.0) optimal1=optimal1/znam
          ! VNY 14.07.2004 NB: is this optimal1 used?

          if (nPts .gt. 9) then 
             aux2=0.0
             optimal1=0.0
             ! normalise the weights to unity
             max1=0.0
             if (sumWeights .gt. 0.0) then
                do i=1, nPts
                   weights(i)=weights(i)/sumWeights
                enddo
             endif

             znam=0.0
             do i=1, nPts
                xtest=xinput(i)
                call fgaussl(xtest,param5,ytest,dydptest5,numparam)
                aux1=param5(1)+param5(5)*xtest
                aux2=1.0
                razn=0.0
                if (yinput(i) .gt. 0.0) then
                   aux2=yinput(i)
                   !razn=yinput(i)-aux1
                   razn=yinput(i)
                endif
                aux3=weights(i)/aux2
                aux2=weights(i)*weights(i)/aux2
                optimal1=optimal1+razn*aux3
                znam=znam+aux2
             enddo
             if (znam .gt. 0.0) optimal1=optimal1/znam
             !
          else
             simplesum=0.0
             znam=1.0
             do i=1,nPts
                aux1=param5(1)
                razn=0.0
                if (yinput(i) .gt. 0.0) then
                   !razn=yinput(i)-aux1
                   razn=yinput(i)
                endif
                simplesum=simplesum+razn
             enddo
             gaussum=simplesum
             optimal1=0.0 !simplesum
          endif ! nUsedPoints .lt. 9
          if (optimal1 .lt. 0.0) optimal1=0.0
          !endif ! extractionMode .eq. 1 .or. extractionMode .eq. 3
          ! VNY 22.01.2004
          !--- 
          
          deallocate(xinput)
          deallocate(yinput)
          deallocate(weights)
          deallocate(yp)
          deallocate(temp)
          deallocate(ysp)
          deallocate(dispersion)

          ! If extractionMode=2 then the output is the integral of the spline-smoothed (lowess) cross-section
          if (currentExtractionMode .eq. 1) spectrumArray(j)=optimal1
          if (currentExtractionMode .eq. 3) spectrumArray(j)=gaussum
          

          !---
          ! across the spectrum
          spectrumErrArray(j)=0.0
          if (spectrumArray(j) .gt. 1.0) spectrumErrArray(j)= &
               sqrt(spectrumArray(j))

          !  Scaling counts
          if (exposureTime .gt. 0.0) then
             spectrumArray(j)=spectrumArray(j)/exposureTime
             spectrumErrArray(j)=spectrumErrArray(j)/exposureTime
          endif


          if (countsScaleFactor(j) .gt. 0.0) then
             spectrumArray(j)=spectrumArray(j)/countsScaleFactor(j)
             spectrumErrArray(j)=spectrumErrArray(j)/countsScaleFactor(j)
          endif

          ! Compute the flux and its error
          !     call CountsToFlux(wavelengthArray(j), spectrumArray(j), spectrumErrArray(j), &
          !          fluxArray(j), fluxErrArray(j), isfLambda, isf, isfError, isfSize)
       enddo ! j=0,ny-1

       !88888888888888888888888888888888888888888888888888888888
    endif ! currentExtractionMode .eq. 1

    ! VNY 05.12.2003
    if (avStripN .gt. 1.0) then
       avStrip=avStrip/avStripN
    endif

    do j=jZero, ny-1
       call CountsToFlux(wavelengthArray(j), spectrumArray(j), spectrumErrArray(j), &
            fluxArray(j), fluxErrArray(j), isfLambda, isf, isfError)
       !if (wavelengthArray(j) .gt. 5900 .and. wavelengthArray(j) .lt. 6500) then
       !   write(*,*) " L1183 j=",j, " wl=",wavelengthArray(j), "spectrumArray=", spectrumArray(j), &
       !        " fluxArray=", fluxArray(j)
       !endif
    enddo
    !---
    deallocate(bkgVectR)
    deallocate(bkgVectL)
    deallocate(specVect)
    deallocate(bkgVect2)
    deallocate(bkgVect3)
    deallocate(smoothBkgVectL)
    deallocate(smoothBkgVectR)
    !deallocate(lambda)
    !deallocate(isf)
    !deallocate(isfError)
    deallocate(countsScaleFactor)
    deallocate(countsScaleFactor2)

    !VNY 14.07.2004
    deallocate(p1vect)
    deallocate(p2vect)
    deallocate(p3vect)
    deallocate(p4vect)
    deallocate(p5vect)
    deallocate(pAux)

  end subroutine ExtractSpectrum


subroutine TestFit()

  real(kind=double), dimension (1:28) :: xinput
  real(kind=double), dimension (1:28) :: yinput
!  real(kind=double), dimension (1:28) :: weights
  integer(kind=int32) :: nPts, verbFlag2, nUsedPoints
  real(kind=double), dimension(5) :: param5

  nPts=28
  nUsedPoints=28
  verbFlag2=1


  call gaussianae(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)


end subroutine TestFit



!**********************************************************************
! Subroutine EvaluateSpectrum for evaluation of the signal-to-noise
! ratio within a given limits of the extraction strip
!******************************************************************
  subroutine EvaluateSpectrum(spectrumArray, bkgArray, sizeY, blueEdge, &
       redEdge, bkgWidth, spectrumWidth, sbRatio, maskArray, verbFlag)
    integer, intent(in) :: blueEdge, redEdge 
    integer, intent(in) :: sizeY
    integer, intent(in) :: verbFlag
    real(kind=double), intent(in) :: bkgWidth, spectrumWidth
    real(kind=double), intent(in), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(in), dimension(0:sizeY-1) :: bkgArray
    real(kind=double), intent(out) :: sbRatio
    integer(kind=int32), intent(in), dimension(0:sizeY-1) :: maskArray
    !integer :: i, correctedRedEdge, redMinusBlue !, j
    integer :: correctedRedEdge, redMinusBlue
    real(kind=double) :: averSpec, averBkg
    integer, dimension(0:1024) :: iBlue, iRed
    integer :: nBlue, nRed, switch !, flag !, nBlueRed
    real(kind=double) :: nAver
    !real(kind=single) :: averSpec2 , averBkg2
    iBlue=0
    iRed=0
    nBlue=0
    nRed=0
    averSpec=0.0
    nAver=0.0
    averBkg=0.0
    sbRatio=0.0
    switch=0
    correctedRedEdge=redEdge
    if (redEdge .ge. sizeY) correctedRedEdge=sizeY-1
    redMinusBlue=correctedRedEdge-blueEdge
    redMinusBlue=int(0.8*real(redMinusBlue))
!    do i=blueEdge,correctedRedEdge
! VNY 14.11.2005 improvements
!    do i=blueEdge,sizeY-1
!       if (maskArray(i) .ne. 0) then 
!          averSpec=averSpec+spectrumArray(i)
!          nAver=nAver+1.0
!          averBkg=averBkg+bkgArray(i)
!          !if (verbFlag .eq. 1) then
!          !   write(*,*) "L827 i=", i, " spec=", spectrumArray(i), " bkg=", bkgArray(i)
!          !endif
!       endif
!    enddo
    averSpec=smoothBkgExcludeZeros2(spectrumArray,maskArray,blueEdge,correctedRedEdge,sizeY-1,verbFlag)
    averBkg=smoothBkgExcludeZeros2(bkgArray,maskArray,blueEdge,correctedRedEdge,sizeY-1,verbFlag) 
    if (verbFlag .eq. 1) then
       write(*,*) "L1921 averSpec2=", averSpec, " averBkg=", averBkg, " <------"
       !do i=blueEdge,correctedRedEdge
       !   write(*,*) "        L1923 i=",i, " spec=", spectrumArray(i), " bkg=", bkgArray(i) 
       !enddo
    endif
    ! Normalise the background and spectrum to one-pixel width
    if (bkgWidth .gt. 0.0) averBkg=averBkg/bkgWidth
    if (spectrumWidth .gt. 0.0) averSpec=averSpec/spectrumWidth
    if (averBkg .gt. 0.0) then
       !sbRatio=(averSpec+averBkg)/averBkg
       !sbRatio=(averSpec)/sqrt(averSpec+averBkg)
       sbRatio=(averSpec)/(averBkg)
    else
       sbRatio=0.0
    endif
    if (verbFlag .eq. 1) then
       write(*,*) " L834 averSpec=", averSpec, " averBkg=", averBkg, " sbRatio=", sbRatio
    endif
! VNY 12.02.2004 Check if the sbRatio>2 was caused by a zero-order
! within the spectrum extraction area
    if (nAver .gt. 1.0) then
       averSpec=averSpec/nAver
       averBkg=averBkg/nAver
    endif
! VNY 18.03.2004 commented sice the zero-orders are already taken into account
!    if (sbRatio .gt. 2.0) then
!       ! Search for the maximum
!       flag=0
!       do i=blueEdge, sizeY-1
!          flag=0
!          if (switch .eq. 0) then
!             if (spectrumArray(i) .gt. averSpec) then
!                iBlue(nBlue)=i
!                nBlue=nBlue+1
!                switch=1
!                flag=1
!             endif! spectrumArray(i) .gt. averSpec
!          endif ! switch .eq. 0
!          if (switch .eq. 1 .and. flag .eq. 0) then
!             if (spectrumArray(i) .lt. averSpec) then
!                iRed(nRed)=i
!                nRed=nRed+1
!                switch=0
!             endif ! spectrumArray(i) .lt. averSpec
!          endif ! switch .eq. 1
 !      enddo ! i=blueEdge, correctedRedEdge
!       nBlueRed=min(nBlue,nRed)
!
!       flag=1
!       do i=0,nBlueRed-1
!          !if (verbFlag .eq. 1) then
!          !   write(*,*) " i=", i, " iBlue=", iBlue(i), " iRed=", iRed(i), " delta=", &
!          !        iRed(i)-iBlue(i), " redMinusBlue=", redMinusBlue
!          !endif
!          if (iRed(i)-iBlue(i) .ge. redMinusBlue/4) then
!             flag=0 ! do not recalculate the sbRatio
!          endif
!       enddo ! i=0,nBlueRed
!       if (flag .eq. 1) then ! recalculate sbRatio
!          if (verbFlag .eq. 1) write(*,*) " L875 recalculating sbRatio"
!          do i=blueEdge,correctedRedEdge
!             ! Check if i lies within one of the exlusion regions
!             switch=1
!             do j=0,nBlueRed-1
!                if (i .ge. iBlue(j)-20 .and. i .le. iRed(j)+20) then
!                   switch=0
!                endif
!             enddo ! j=0,nBluered-1
!             if (switch .eq. 1) then
!                if (maskArray(i) .ne. 0) then
 !                  averSpec=averSpec+spectrumArray(i)
!                   nAver=nAver+1.0
!                   averBkg=averBkg+bkgArray(i)
!                endif
!             endif ! switch .eq. 1
!          enddo
!          if (verbFlag .eq. 1) then
!             write(*,*) " L893 averSpec=", averSpec, " averBkg=", averBkg
!          endif
!          if (averBkg .ne. 0.0) then
!             sbRatio=(averSpec+averBkg)/averBkg
!          endif
!       endif ! flag .eq. 1
!    endif ! sbRatio .gt. 2.0
  end subroutine EvaluateSpectrum


!**********************************************************************
! Subroutine EvaluateProfile for evaluation of the minima and maxima
! in the cross-section profile of a spectrum 
! 13.05.2007
!******************************************************************
  subroutine EvaluateProfile(sProfile, dprofile, spiMin, spiMax, &
       nSp, peacks, leftBorder, rightBorder, verbFlag)

    integer(int32), intent(in) :: spiMin, spiMax
    real(kind=single), intent(in), dimension(spiMin:spiMax) :: sProfile ! profile
    real(kind=single), intent(in), dimension(spiMin:spiMax) :: dProfile ! its derivative
    integer, intent(in) :: verbFlag
    integer :: i

    real(kind=single) :: mx1,mx2,px1,px2
    real(kind=single) :: dmx1, dmn1, dpx1, dpn1
    real(kind=single) :: aux1, aux2
    integer(int32) :: leftSlope, leftSlope2, rightSlope, rightSlope2
    integer(int32) :: rightSlopeLimit
    integer(int32), intent(out) :: nSp ! number of detected maxima 
    real(kind=single), intent(out), dimension(1:2) :: peacks ! the maxima
    real(kind=single), intent(out), dimension(1:2) :: leftBorder, rightBorder

    if (verbFlag .eq. 1) then
       write(*,*) "L2451 subroutine EvaluateProfile"
       write(*,*) "L2452 spiMin=", spiMin, " spiMax=", spiMax
       write(*,*) "---------------------------------"
       do i=spiMin, spiMax
          write(*,*) "i=", i, " profile=", sProfile(i), " d=", dProfile(i)
       enddo
    endif
    ! search for the maxima and minima

    mx1=0.0
    px1=(real(spiMin+spiMax))/2
    mx2=0.0
    px2=px1

    dmx1=0.0
    dmn1=999999.0
    dpx1=px1
    dpn1=px1
    aux1=0.0
    aux2=0.0

    ! compute the average value of dProfile
    ! and the maximum and minimum of sProfile
    do i=spiMin, spiMax
       if (i .ne. spiMin) then
          aux1=aux1+ abs(dProfile(i))
          aux2=aux2+1.0
       endif

    enddo
    if (aux2 .gt. 1.0) then
       aux1=aux1/aux2
    endif
    ! Check the left and right limits of the spectrum
    leftSlope=0
    rightSlope=0
    !aux1=aux1/2.0
    aux1=aux1/1.5
    do i=spiMin,spiMax
       if (abs(dProfile(i)) .gt. aux1) then
          leftSlope=i
          exit
       endif
    enddo
    do i=spiMax,spiMin+1,-1
       if (abs(dProfile(i)) .gt. aux1) then
          rightSlope=i
          exit
       endif
    enddo

   if (verbFlag .eq. 1) then
       write(*,*) "L2503 leftSlope=", leftSlope, " rightSlope=", &
            rightSlope
    endif
    rightSlopeLimit=rightSlope
    if (rightSlopeLimit .ge. spiMax) rightSlopeLimit=spiMax-1
   if (verbFlag .eq. 1) then
       write(*,*) "L2509 leftSlope=", leftSlope, " rightSlope=", &
            rightSlope
    endif
    ! check for the intermediate minimum (in case of a double spectrum)
    leftSlope2=0
    rightSlope2=0
    do i=leftSlope,rightSlopeLimit
       if (abs(dProfile(i)) .gt. aux1 .and. dProfile(i) .lt. 0.0 &
            .and. abs(dProfile(i+1)) .lt. aux1) then
          leftSlope2=i
          exit
       endif
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) "L2498 leftSlope2=", leftSlope2
    endif

    do i=rightSlope-1,leftSlope,-1
       if (abs(dProfile(i)) .gt. aux1 .and. dProfile(i) .gt. 0.0 &
            .and. abs(dProfile(i-1)) .lt. aux1) then
          rightSlope2=i
          exit
       endif
    enddo
!    endif
   
!    if (rightSlope2 .lt. leftSlope2) rightSlope2=leftSlope2

    if (verbFlag .eq. 1) then
       write(*,*) "--------------------"
       write(*,*) " L2440 aux1=", aux1, " aux2=", aux2, &
            " leftSlope=", leftSlope, " rightSlope=", rightSlope
       write(*,*) "L2442 leftSlope2=", leftSlope2, " rightSlope2=", rightSlope2
    endif

    peacks=-1.0
    leftBorder=-1.0
    rightBorder=-1.0
    nSp=1
!----------------- NB 16.10.2007 -------------
!    if (rightSlope2-leftSlope2 .gt. 0) then
    if (rightSlope2-leftSlope2 .gt. 0 &
         .and. leftSlope2 .gt. 0 .and. rightSlope2 .gt. 0) then
!-------------------------------------------
       !write(*,*) "L2449 twospectra are found"
       nSp=2
    endif

    ! in case of a single spectrum find one maximum
    if (nSp .eq. 1) then
       do i=leftSlope, rightSlope
          if (sProfile(i) .gt. mx1) then
             mx1=sProfile(i)
             px1=i
          endif
       enddo
       peacks(1)=px1
       leftBorder(1)=leftSlope
       rightBorder(1)=rightSlope
    else ! find two maxima
       do i=leftSlope,leftSlope2
          if (sProfile(i) .gt. mx1) then
             mx1=sProfile(i)
             px1=i
          endif
       enddo
       peacks(1)=px1
       leftBorder(1)=leftSlope
       rightBorder(1)=leftSlope2

       do i=rightSlope2,rightSlope
          if (sProfile(i) .gt. mx2) then
             mx2=sProfile(i)
             px2=i
          endif
       enddo
       peacks(2)=px2
       leftBorder(2)=rightSlope2
       rightBorder(2)=rightSlope
    endif
    if (verbFlag .eq. 1) then
       write(*,*) "L2587 nSp=", nSp, " px1=", px1, " mx1=", mx1, &
            " px2=", px2, " mx2=", mx2
       write(*,*) "&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&"
    endif
  end subroutine EvaluateProfile



!**********************************************************************
! Subroutine EvaluateZeroOrderProfile is for checking whether the 
! background extraction regions near the zero-order ar OK (they might
! fall onto the neighbouring spectra) 
! 20.10.2007
!******************************************************************
  subroutine EvaluateZeroOrderProfile(profile, iMin, iMax, &
       xZero, leftLimit, rightLimit, leftBlimit, rightBlimit, &
       bkgL, bkgR, verbFlag)

    integer(int32), intent(in) :: iMin, iMax
    real(kind=double), intent(in), dimension(iMin:iMax) :: profile ! profile
    real(kind=double), intent(in) :: xZero ! x-coordinate of the zero-order
    real(kind=double), dimension(iMin:iMax) :: dProfile ! derivative of the profile
    integer, intent(in) :: verbFlag
    integer :: i, i0

    real(kind=double) :: mx1,mx2,px1,px2
    real(kind=double) :: dmx1, dmn1, dpx1, dpn1
    real(kind=double) :: aux1, aux2
    integer(int32) :: leftSlope, leftSlope2, rightSlope, rightSlope2
    integer(int32) :: rightSlopeLimit
    integer(int32), intent(out) :: leftLimit, rightLimit, leftBlimit, rightBlimit
    real(kind=double), intent(out) :: bkgL, bkgR

    bkgL=-1.0
    bkgR=-1.0
    do i=iMin,iMax
       mx1=0.0
       if (i-iMin .gt. 0) then
          mx1=profile(i)-profile(i-1)
       endif
       dProfile(i)=mx1
    enddo

    if (verbFlag .eq. 1) then
       write(*,*) "L2678 subroutine EvaluateZeroOrderProfile"
       write(*,*) "L2679 iMin=", iMin, " iMax=", iMax
       write(*,*) "---------------------------------"
       do i=iMin, iMax
          write(*,*) "i=", i, " profile=", profile(i), " dProfile=", dProfile(i)
       enddo
    endif

    ! search for the maxima and minima

    mx1=0.0
    px1=xZero
    i0=int(xZero)
    mx2=0.0
    px2=px1

    dmx1=0.0
    dmn1=999999.0
    dpx1=px1
    dpn1=px1
    aux1=0.0
    aux2=0.0

    ! compute the average value of dProfile
    ! and the maximum and minimum of sProfile
    do i=iMin, iMax
       if (i .ne. iMin) then
          aux1=aux1+ abs(dProfile(i))
          aux2=aux2+1.0
       endif

    enddo
    if (aux2 .gt. 1.0) then
       aux1=aux1/aux2
    endif
    ! Check the left and right limits of the spectrum
    leftSlope=0
    rightSlope=0
    !aux1=aux1/2.0
    aux1=aux1/1.5
    do i=i0,iMin+1,-1
       if (abs(dProfile(i)) .gt. aux1 .and. &
            abs(dProfile(i-1)) .le. aux1) then
          leftSlope=i
          exit
       endif
    enddo
    do i=i0,iMax-1
       if (abs(dProfile(i)) .gt. aux1 .and. &
            abs(dProfile(i+1)) .le. aux1) then
          rightSlope=i
          exit
       endif
    enddo

    if (verbFlag .eq. 1) then
       write(*,*) "L2734 leftSlope=", leftSlope, " rightSlope=", &
            rightSlope
    endif
    rightLimit=iMax
    if (rightSlope .ne. 0) rightLimit=rightSlope
    leftLimit=iMin
    if (leftSlope .ne. 0) leftLimit=leftSlope

    ! find left and right background limits
    leftSlope=0
    rightSlope=0
    do i=leftLimit,iMin+1,-1
       if (abs(dProfile(i)) .le. aux1 .and. &
            abs(dProfile(i-1)) .gt. aux1) then
          leftSlope=i
          exit
       endif
    enddo
    do i=rightLimit,iMax-1
       if (abs(dProfile(i)) .le. aux1 .and. &
            abs(dProfile(i+1)) .gt. aux1) then
          rightSlope=i
          exit
       endif
    enddo
    rightBlimit=iMax
    if (rightSlope .ne. 0) rightBlimit=rightSlope
    leftBlimit=iMin
    if (leftSlope .ne. 0) leftBlimit=leftSlope    
    if(verbFlag .eq. 1) then
       write(*,*) "L2765 leftBlimit=", leftBlimit, " rightBlimit=", rightBlimit
    endif

    if (leftlimit-leftBlimit .ge. 1) then
       ! compute left background
       aux1=0.0
       aux2=0.0

       do i=leftBlimit, leftLimit
          aux1=aux1+ profile(i)
          aux2=aux2+1.0
       enddo
       if (aux2 .gt. 1.0) then
          aux1=aux1/aux2
       endif
       bkgL=aux1
    endif

    if (rightBlimit-rightLimit .ge. 1) then
       ! compute right background
       aux1=0.0
       aux2=0.0

       do i=rightLimit, rightBlimit
          aux1=aux1+ profile(i)
          aux2=aux2+1.0
       enddo
       if (aux2 .gt. 1.0) then
          aux1=aux1/aux2
       endif
       bkgR=aux1
    endif

  end subroutine EvaluateZeroOrderProfile

!******************************************************************
! Subroutine for finding a zero-order position based on the 
! coordinates of the first-order.
! yZero is the output coordinate of the zero-order
! but at the beginning it must contain the expected position
! of the zero-order
!******************************************************************
  subroutine FindZeroOrder(grismSetH, xFirst, yFirst, &
       spectrumHalfLength, yZero, &
       maxValue, spectrumHalfWidth, shapeCriterium1, verbFlag, &
       maxValue2, meanValue2, yZero2, spectrumHalfHeight)  
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xFirst, yFirst
    real(kind=double), intent(in) :: spectrumHalfLength
    real(kind=double), intent(in) :: spectrumHalfWidth
    real(kind=double), intent(in) :: shapeCriterium1
    real(kind=double), intent(in) :: spectrumHalfHeight
    integer, intent(in) :: verbFlag ! auxiliary flag for test output
    real(kind=double), intent(out) :: yZero2 ! 
    real(kind=double), intent(inout) :: yZero
    real(kind=double), intent(out) :: maxValue
    real(kind=double), intent(out) :: maxValue2
    real(kind=double), intent(out) :: meanValue2
    real(kind=double), dimension(:), pointer :: verticalStrip
    integer :: i, j, ny, nx
    integer :: iLeft, iRight, jBlue, jRed, jMaxValue
    real(kind=double) :: stripWidth
    type(ArrayT) :: inArrayH   
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    real(kind=double) :: maxValueFirst
    real(kind=double) :: convValue, nMean
    integer :: jMaxValue1, jMaxValue2
    !
    ! the input image image 
    inArrayH = array(grismSetH, "PRIMARY", read)
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  
    allocate(verticalStrip(0:ny-1))
    iLeft=int(xFirst-spectrumHalfWidth)
    if (iLeft .lt. 0) iLeft=0
    iRight=int(xFirst+spectrumHalfWidth)
    if (iRight .gt. nx-1) iRight=nx-1
    stripWidth=real(iRight-iLeft+1)

    ! Fill the strip in with the values from the image
    do j=0,ny-1
       verticalStrip(j)=0.0
       do i=iLeft,iRight
          verticalStrip(j)=verticalStrip(j)+oswImage(i,j)
       enddo ! i=iLeft,iRight
       if (stripWidth .gt. 1.0) verticalStrip(j)=verticalStrip(j)/stripWidth
    enddo ! j=0,ny-1

    ! Compute maximum of the spectrum itself
    jBlue=int(yFirst-spectrumHalfLength)
    jRed=int(yFirst+spectrumHalfLength)
    if (verbFlag .eq.1 ) then
       write(*,*) "L2232 jBlue=", jBlue, " jRed=", jRed, " yZero=", yZero
    endif
    if (jBlue .lt. 0) jBlue=0
    if (jRed .gt. ny-1) jRed=ny-1
    maxValueFirst=0.0
    jMaxValue1=0
    do j=jBlue,jRed
       if (verticalStrip(j) .gt. maxValueFirst) then
          maxValueFirst=verticalStrip(j)
          jMaxValue1=j
       endif
    enddo ! j=jBlue,jRed
    if (verbFlag .eq. 1) then
       write(*,*) "L2241 maxValueFirst=", maxValueFirst, " jMaxValue1=", jMaxValue1
    endif
    ! Set the interval for searching the maximum near the 
    ! expected position of the zero-order
!    jBlue=int(yZero)-400
!    jRed=jBlue+450
    jBlue=int(yZero)-ny/4.0
    jRed=int(yZero)+ny/4.0
    if (jBlue .lt. 0) jBlue=0
    if (jRed .gt. ny-1) jRed=ny-1
    if (verbFlag .eq. 1 ) then
       write(*,*) " L2256 spectrumHalfLength=", spectrumHalfLength
       write(*,*) "OMGEXTRACTION L2257 jBlue=", jBlue, " jRed=", jRed
    endif
    ! Search for the maximum in between jBlue and jRed
    maxValue=0.0
    jMaxValue=-1
    do j=jBlue,jRed
       if (verticalStrip(j) .gt. maxValue) then
         !write(*,*) " j=",j, " v=", verticalStrip(j), " ********", &
         !     " current MAX=", maxValue
          maxValue=verticalStrip(j)
          jMaxValue=j
       else
          !write(*,*) " j=",j, " v=", verticalStrip(j)  
       endif
    enddo ! j=jBlue,jRed
    if (verbFlag .eq. 1) then
       write(*,*) " L957 jMaxValue=", jMaxValue, " maxValue=", maxValue, &
            " maxValueFirst=", maxValueFirst
    endif
    yZero=-1
    if (jMaxValue .gt. -1 .and. maxValue .gt. maxValueFirst*2.0 &
         .and. abs(jMaxValue-jMaxValue1) .gt. spectrumHalfHeight) yZero=jMaxValue
    if (verbFlag .eq. 1) then
       write(*,*) "L2279 yZero=", yZero
    endif
    ! VNY 03.03.2004 Test of the ExpConv
    jMaxValue2=-1
    maxValue2=0.0
    meanValue2=0.0
    nMean=0.0
    yZero2=-1
    do j=jBlue,jRed
       convValue=ConvoluteExp(verticalStrip, ny, j, shapeCriterium1)
       meanValue2=meanValue2+convValue
       nMean=nMean+1
       if (convValue .gt. maxValue2) then
          maxValue2=convValue
          jMaxValue2=j
       endif
       !if (verbFlag .eq. 1) then
       !   write(*,*) "L2292  j=", j, " conv=", convValue
       !end if ! verbFlag .eq. 1
    enddo ! j=jBlue, jRed
    if (nMean .gt. 1.0 .and. abs(jMaxValue2-jMaxValue1) .gt. spectrumHalfHeight) then
       meanValue2=meanValue2/nMean
       yZero2=real(jMaxValue2)
    endif
    deallocate(verticalStrip)
  end subroutine FindZeroOrder

!**********************************************************************
! Subroutine RemoveZeroOrders for removing the zero orders overlapping 
! the spectrum 
!******************************************************************
  subroutine RemoveZeroOrders(spectrumArray, bkgArray, fluxArray, &
       maskArray, sizeY, blueEdge, redEdge, verbFlag)
    integer, intent(in) :: sizeY
    integer, intent(in) :: blueEdge, redEdge
    real(kind=double), intent(inout), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(inout), dimension(0:sizeY-1) :: bkgArray
    real(kind=double), intent(inout), dimension(0:sizeY-1) :: fluxArray
    integer(kind=int32), intent(in), dimension(0:sizeY-1) :: maskArray
    integer, intent(in) :: verbFlag 
    integer :: i, j, limit1, limit2, redEdgeModified
    integer :: regionCount, switch1 !, switch2
    integer, dimension(1:10) :: firstEdge, secondEdge
    real(kind=double), dimension(1:10) :: featureMax
    real(kind=double) :: meanSpectrumValue, nMean
    integer :: maxRegionCount, regionCount2
    maxRegionCount=10

    !if (verbFlag .eq. 1) then
    !   write(*,*) " RemoveZeroOrders L1868, blueEdge=", blueEdge, " redEdge=", redEdge
    !endif
    if (redEdge .gt. sizeY) then
       redEdgeModified=sizeY
    else
       redEdgeModified=redEdge
    endif

    ! Check the number of masked regions
    regionCount=0
    switch1=0
    firstEdge=0
    secondEdge=0
    featureMax=0.0
    meanSpectrumValue=0.0
    do i=blueEdge, redEdge-1
       if (maskArray(i) .eq. 0 .and. switch1 .eq. 0) then
          regionCount=regionCount+1
          switch1=1
          if (regionCount .le. 10) then
             firstEdge(regionCount)=i
          endif
       endif
       if (switch1 .eq. 1) then
          if (maskArray(i) .eq. 1) then
             switch1=0
             if (regionCount .le. 10) then
                secondEdge(regionCount)=i-1
             endif
          endif
       endif
       if (maskArray(i) .eq. 1) then
          meanSpectrumValue=meanSpectrumValue+spectrumArray(i)
          nMean=nMean+1.0
       endif
       !if (verbFlag .eq. 1) write(*,*) " i=", i, " sp=", &
       !     spectrumArray(i), " mask=", maskArray(i), &
       !     " switch1=", switch1, " count=", regionCount
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) " meanSpectrumValue=", meanSpectrumValue, " nMean=", nMean
    endif
       if (nMean .gt. 1.0) meanSpectrumValue=meanSpectrumValue/nMean
    !if (verbFlag .eq. 1) then
    !   write(*,*) " averaged meanSpectrumValue=", meanSpectrumValue
    !   write (*,*) "L1060 regionCount=", regionCount 
    !endif
    !find mean values for the features
    featureMax=0.0
    regionCount2=regionCount
    if (regionCount2 .gt. maxRegionCount) regionCount2=maxRegionCount
    do i=1, regionCount2
       limit1=firstEdge(i)
       limit2=secondEdge(i)
       nMean=0.0
       do j=limit1,limit2
          if (spectrumArray(j) .gt. featureMAx(i)) then
             featureMax(i)=spectrumArray(j)
          endif
          !featureMax(i)=featureMax(i)+spectrumArray(j)
          !nMean=nMean+1.0
       enddo
       !if (nMean .gt. 1.0) featureMax(i)=featureMax(i)/nMean
    enddo
    !if (verbFlag .eq. 1) then
    !   write(*,*) "----------------------------"
    !   do i=1, regionCount2
    !      write(*,*) " L1937 reg=", i, " firstEdge=", firstEdge(i), " secondEdge=", secondEdge(i), &
    !           " featureMax=", featureMax(i), " mean=", meanSpectrumValue
    !   enddo
    !   write(*,*) "----------------------------"
    !endif

!    if (verbFlag .eq. 1) write(*,*) " --------------------------"
!    do i=blueEdge, redEdge
!       if (verbFlag .eq. 1) write(*,*) " i=", i, " sp=", spectrumArray(i), ", mask=", maskArray(i)
!    enddo
    
    ! eliminate features
    do i=1, regionCount2
       if (featureMax(i) .gt. meanSpectrumValue*6.0) then ! eliminate
          limit1=firstEdge(i)
          limit2=secondEdge(i)
          do j=limit1, limit2
             spectrumArray(j)=0.0
             bkgArray(j)=0.0
             fluxArray(j)=0.0
          enddo
       endif
    enddo
    !do i=0,sizeY-1
       !if (verbFlag .eq. 1) then
       !   write(*,*) " i=", i, " sp=", spectrumArray(i)
       !endif
       !if (maskArray(i) .eq. 0) then
       !   spectrumArray(i)=0.0
       !   bkgArray(i)=0.0
       !   fluxArray(i)=0.0
       !endif
    !enddo
  end subroutine RemoveZeroOrders


!******************************************************************
! Subroutine for finding possible features (zero-orders)
! sitting on the spectrum 
!******************************************************************
  subroutine FindPossibleFeatures2 (spectrumArray, windowHalfWidth, &
       maxValue, maxValue2, meanValue2, &
       maskArray, ny, blueEdge, redEdge, verbFlag, nFeatures) 
    integer, intent(in) :: ny
    integer, intent(in) :: blueEdge, redEdge
    integer, intent(in) :: windowHalfWidth
    real(kind=double), intent(inout), dimension(0:ny-1) :: spectrumArray
    integer, intent(in) :: verbFlag ! auxiliary flag for test output
    real(kind=double), intent(out) :: maxValue
    real(kind=double), intent(out) :: maxValue2
    real(kind=double), intent(out) :: meanValue2
    real(kind=double), dimension(:), pointer :: verticalStrip
    real(kind=double), dimension(:), pointer :: verticalStrip2
    real(kind=double), dimension(:), pointer :: convStrip
    real(kind=double), dimension(:), pointer :: convStrip2
    integer, intent(out) :: nFeatures
    integer(kind=int32), intent(out), dimension(0:ny-1) :: maskArray
    integer(kind=int32), dimension(0:ny-1) :: maskArray2
    integer :: j
    integer :: jBlue, jRed, jMaxValue
!    real(kind=double) :: stripWidth
!    type(ArrayT) :: inArrayH   
    !real(kind=single), pointer :: oswImage(:, :)
!    integer(kind=int32) :: dim(2)
    real(kind=double) :: maxValueFirst
    real(kind=double) :: nMean
!    integer :: jMaxValue2, nmax, flagMax, flagMax2, kMax
!    real(kind=double) :: dRight, dLeft1, dRight1
!    real(kind=double) :: dLeftPrev, dRightPrev
    integer :: leftOkay, rightOkay

    allocate(verticalStrip(0:ny-1))
    allocate(verticalStrip2(0:ny-1))
    allocate(convStrip(0:ny-1))
    allocate(convStrip2(0:ny-1))
    maskArray=1
    maskArray2=1
    verticalStrip=spectrumArray
    verticalStrip2=spectrumArray

    ! smooth the strip
    !-----------------------------------------------------
!    windowHalfWidth=50
    do j=0, ny-1
       convStrip(j)=smoothMedianWindow(verticalStrip, j, ny-1, windowHalfWidth)
    enddo
    do j=0, ny-1
       verticalStrip(j)=smoothAver(convStrip, j, ny-1)
    enddo
    !-----------------------------------------------------
    ! Compute maximum of the spectrum itself
    jBlue=blueEdge
    jRed=redEdge
    if (verbFlag .eq. 1 ) then
       write(*,*) "FindPossibleFeatures2 L2286  jBlue=", jBlue, " jRed=", jRed
    endif
    if (jBlue .lt. 0) jBlue=0
    if (jRed .gt. ny-1) jRed=ny-1
    maxValueFirst=0.0
    do j=jBlue,jRed
       if (verticalStrip(j) .gt. maxValueFirst) maxValueFirst=verticalStrip(j)
    enddo ! j=jBlue,jRed
    if (verbFlag .eq. 1 ) then
       write(*,*) " L2296 maxValueFirst=", maxValueFirst
    endif
    ! Search for the maximum in between jBlue and jRed
    maxValue=0.0
    jMaxValue=-1
    do j=jBlue,jRed
       if (verticalStrip(j) .gt. maxValue) then
          maxValue=verticalStrip(j)
          jMaxValue=j
       endif
    enddo ! j=jBlue,jRed
!-----------------------------
    verticalStrip2=verticalStrip2-verticalStrip
    where (verticalStrip2 .lt. 0.0) verticalStrip2=0.0
!-----------------------------

!    if (verbFlag .eq. 1 ) then
!       write(*,*) " FindFeature2 L2325 jMaxValue=", jMaxValue, " maxValue=", maxValue, &
!            " maxValueFirst=", maxValueFirst, " windowHalfWidth=", windowHalfWidth
!    endif

    ! Initialise the mask
    do j=jBlue, jRed
       maskArray(j)=1
       maskArray2(j)=1
       if (verticalStrip2(j) .gt. verticalStrip(j)) then
          maskArray(j)=0
          maskArray2(j)=0
       endif
!       if (verbFlag .eq. 1) &
!            write(*,*) " j=",j, " v=", verticalStrip(j), &
!            " v2=", verticalStrip2(j), " mask=", maskArray(j)
    enddo

    ! Filter the mask 
    do j=jBlue, jRed
       ! check if there are more clean mask pixels
       if (maskArray(j) .eq. 0) then
          if (.not. maskArrayOkay(maskArray, ny, j, windowHalfWidth/10, leftOkay, rightOkay)) maskArray(j)=1  
       endif
    enddo

    if (verbFlag .eq. 1) then
       do j=jBlue, jRed
          write(*,*) "L2313  j=",j, " v=", verticalStrip(j), &
               " v2=", verticalStrip2(j), " mask=", maskArray(j)
       enddo
    endif
    nFeatures=0
    do j=jBlue+1,jRed
       if (maskArray(j) .eq. 0 .and. maskArray(j-1) .eq. 1) nFeatures=NFeatures+1
    enddo

    maxValue2=0.0
    meanValue2=0.0
    nMean=0.0
    do j=jBlue+1,jRed
       if (maskArray(j) .eq. 1) then
          meanValue2=meanValue2+verticalStrip(j)
          nMean=nMean+1.0
       endif
    enddo
    if (nMean .gt. 1.0) meanValue2=meanValue2/nMean
    if (verbFlag .eq. 1) then
       write(*,*) "L2328 meanValue2=", meanValue2, " nMean=", nMean
    endif
    deallocate(verticalStrip)
    deallocate(verticalStrip2)
    deallocate(convStrip)
    deallocate(convStrip2)
  end subroutine FindPossibleFeatures2

!********************************************************************
! Checking if the zeroed pixels in the mask array are surrounded by
! at least "nClearPixels" 
!********************************************************************
logical function MaskArrayOkay(maskArray, ny, j, nClearPixels, leftOkay, rightOkay)
  integer, intent(in) :: ny
  integer(int32), intent(in), dimension(0:ny-1)  :: maskArray
  integer, intent(in) :: j
  integer, intent(in) :: nClearPixels
  integer :: i
  integer, intent(out)  :: leftOkay, rightOkay
  MaskArrayOkay = .true.
  leftOkay=0
  rightOkay=0
  !if (j .eq. 585) then
  !   write(*,*) " L2348 nClearPixels=", nClearPixels
  !endif
  if (maskArray(j) .eq. 1) return
  do i=1,nClearPixels
     !if (j .eq. 585) then
     !   write(*,*) " L2353, i=", i, " maskArray(j+i)=", maskArray(j+i)
     !endif
     if (j+i .ge. ny-1) exit
     if (maskArray(j+i) .eq. 1) then
        exit
     else
        rightOkay=rightOkay+1
     endif
  enddo
  !if (j .eq. 585) then
  !   write(*,*) " L2363 j=", j, " rightOkay=", rightOkay 
  !endif
  do i=1,nClearPixels
     if (j-i .lt. 0) exit
     if (maskArray(j-i) .eq. 1) then
        exit
     else
        leftOkay=leftOkay+1
     endif
  enddo
  if (leftOkay+rightOkay .lt. nClearPixels) MaskArrayOkay=.false.
end function MaskArrayOkay

!******************************************************************
! Re-compute zero-order coordinates
!******************************************************************
  subroutine RecomputeZeroOrder(grismSetH, xZeroIn, yZeroIn, &
       spectrumHalfWidth, spectrumHalfHeight, computingMode, &
       verbFlag, xZeroOut, yZeroOut)
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xZeroIn, yZeroIn
    integer, intent(in) :: computingMode
    real(kind=double), intent(in) :: spectrumHalfWidth
    real(kind=double), intent(in) :: spectrumHalfHeight
    integer, intent(in) :: verbFlag
    real(kind=double), intent(out) :: xZeroOut, yZeroOut
    type(ArrayT) :: inArrayH   
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    integer :: i, j, ny, nx
    character(len=9) :: filter ! filter keyword in the input file
    integer(kind=int32) ccdMode !+  
    integer(kind=int32) :: binX                                        ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    logical :: binBpe !+
    real(kind=double), dimension(:), pointer :: specVect, smoothSpecVect
    integer :: ixLeft, ixRight, jcross
    real(kind=double) :: averSpec,numSpec
    integer :: jstart, jend, nPts
    real(kind=double), dimension (:), pointer :: xinput
    real(kind=double), dimension (:), pointer :: yinput
    real(kind=double), dimension (:), pointer :: weights
    !
    real, dimension (:), pointer :: xdata
    real, dimension (:), pointer :: ydata
!    character(len=20) :: graphTitle
!    character(len=10) :: xlabel
!    character(len=10) :: ylabel    
!    real :: x1,x2, y1,y2
    integer :: ndata!, pgbeg
!    real(kind=double), dimension(5) ::dydptest 
    !
    real(kind=double), dimension (:), pointer :: yp
    real(kind=double), dimension (:), pointer :: ysp
    real(kind=double), dimension (:), pointer :: temp
    real(kind=double), dimension (:), pointer :: dispersion
!    real(kind=double) :: slp1, slpn, tension
    integer :: verbFlag2
    real(kind=double), dimension(5) :: param5
    integer :: nUsedPoints
    integer(kind=int32) :: numparam!, numpoints
!    real(kind=double) :: optimal1
!    real(kind=double) :: optimal2 
!    real(kind=double) :: gaussum
!    real(kind=double) :: simplesum
!    real(kind=double) :: znam, sumWeights
!    real(kind=double)  :: xtest , ytest, aux1, aux2, aux3, razn
!    real(kind=double), dimension(5) ::dydptest5
    real(kind=double) :: smooth, delta
    integer(kind=int32) :: numsteps
    real(kind=double) :: max1, min1, min2
    integer :: imax1, imin1, imin2
!    real(kind=double) :: averSpecLowess
    real(kind=double) :: wSpec, dSpec
    real(kind=double) :: levelLR
    real(kind=double) :: averSpecSimple
    real(kind=double) :: averLeftSimple, averRightSimple!, averLeftLowess, averRightLowess
    real(kind=double) :: ratio, ratio2, averLeft0, averRight0, averLeft1, averRight1
    integer :: jEquilibrium
    real(kind=double) :: averLeftEq, averRightEq, l1,r1,l0,r0, ratio3
    ! --- for computing moments:
    integer(kind=int16) :: sourceNumber
    integer(kind=int32) :: iMin, iMax, jMin, jMax
    real(kind=single) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 
!    real(kind=double), dimension(6) :: param6
!    real(kind=double) :: xD, yD, fxy
    integer :: jstartM
    integer :: jendM
    integer :: icentre, jcentre, jdistanceMin, jdistanceMax
    real(kind=double) gaussianMaximum, gaussianAmplitude, gaussianDisp

    !write(*,*) "RecomputeZeroOrder L1765:  verbosity flag=", verbFlag 


    ! information on the science window
    inArrayH = array(grismSetH, "PRIMARY", read)
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  

    if (hasAttribute(grismSetH,"FILTER")) then
       filter = stringAttribute(grismSetH, "FILTER")
    else
       call warning('OMGRISM','Keyword FILTER missing in the input file ')
    endif

    ccdMode = IMAGE

    ! Calculate image binning.  
    binBpe= .false.
    if (hasAttribute(grismSetH,"BINBPE")) then       
       binBpe= booleanAttribute(grismSetH, "BINBPE")
    else
       call warning('OMGEXTRACTION','Keyword BINBPE missing in the input file ')
    endif
    ! VNY 04.05.2004 Temporarily returning to the old (wrong)
    ! computation of the binning factors
    !if (hasAttribute(grismSetH,"BINAX1")) then       
    !   binAx1 = int32Attribute(grismSetH, "BINAX1")
    !   binX = binAx1
    !   if (binX .lt. 1.0) binX = 1.0
    !else
    !   call warning('OMGEXTRACTION','Keyword BINAX1 missing in the input file ')
    !   binX=1.0
    !endif

    !if (hasAttribute(grismSetH,"BINAX2")) then       
    !   binAx2 = int32Attribute(grismSetH, "BINAX2")
    !   binY = binAx2
    !   if (binY .lt. 1.0) binY = 1.0
    !else
    !   call warning('OMGEXTRACTION','Keyword BINAX2 missing in the input file ')
    !   binY=1.0
    !endif
    if (hasAttribute(grismSetH,"BINAX1")) then       
       binAx1 = int32Attribute(grismSetH, "BINAX1")
       binX = 2 ** binAx1
    else
       call warning('OMGEXTRACTION','Keyword BINAX1 missing in the input file ')
       binX=1.0
    endif

    if (hasAttribute(grismSetH,"BINAX2")) then       
       binAx2 = int32Attribute(grismSetH, "BINAX2")
       binY = 2 ** binAx2
    else
       call warning('OMGEXTRACTION','Keyword BINAX2 missing in the input file ')
       binY=1.0
    endif


    if (binBPE) then
       binX = 2.0
       binY = 2.0
    end if
    !if (verbFlag .eq. 1) then 
    !   write (*,*) "ComputZeroOrder L1820"
    !   write(*,*) "ComputeZeroOrder L1821: xZeroIn=", xZeroIn, " yZeroIn=", yZeroIn
    !   write(*,*) "OMGRISMEXTRACTION L236: binBpe=", binBpe, " binAx1=", binAx1, " binAx2=", binAx2
    !   write(*,*) " filter=", filter, " binX=", binX, " binY=", binY, " nx=", nx, " ny=", ny
    !endif



    allocate(specVect(0:ny-1))
    allocate(smoothSpecVect(0:ny-1))
    specVect=0.0
    smoothSpecVect=0.0
    ixRight=nx-1
    ixLeft=xZeroIn-spectrumHalfWidth
    if (ixLeft .lt. 0) ixLeft=0
    ixRight=xZeroIn+spectrumHalfWidth
    if (ixRight .gt. nx-1) ixRight=nx-1
    if (ixRight .le. ixLeft) ixRight=ixLeft+1
    !write(*,*) " RecomputeZeroOrder L1839 ixLeft=", ixLeft, " ixRight=", ixRight
    ! VY 23.04.2004 
    iMin=xZeroIn-spectrumHalfWidth
    iMax=xZeroIn+spectrumHalfWidth
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yZeroIn-spectrumHalfHeight
    jMax=yZeroIn+spectrumHalfHeight
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L1811 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    do j=0, ny-1
       averSpec=0.0
       numSpec=0.0
       do jcross=ixLeft,ixRight
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(j)=averSpec
       !write(*,*) " j=",j, " spec=", averSpec
    enddo

    !*******************************************************************************
    !smoothing the background 
    !
    if (computingMode .ge. 10) then ! smoothing
       ! smoothing median
       do j=0,ny-1
          smoothSpecVect(j)=smoothMedian(specVect,j,ny-1)
       enddo
       ! smoothing average
       do j=0,ny-1
          specVect(j)=smoothAver(smoothSpecVect,j,ny-1)
       enddo
    endif

    ! VNY test 08.01.2004
    jstart=int(yZeroIn)-int(spectrumHalfHeight)/3
    jend=int(yZeroIn)+int(spectrumHalfHeight)/3

    if (jstart .lt. 0) jstart=0
    if (jend .gt. ny-1) jend=ny-1
    ! Check for non-zero values and choose the proper length for xinput and yinput
    ndata=0
    jstartM=-1
    jendM=-1
    do j=jstart,jend
       if (j .gt. 0 .and. j .lt. ny)  then
          if (specVect(j) .gt. 0.0) then
             if (jstartM .eq. -1) jstartM=j
             ndata=ndata+1
          else
             if (jstartM .gt. 0 .and. jendM .eq. -1) jendM=j
          endif
       endif
    enddo
    if (jendM .eq. -1) jendM=jend
    if (jendM .ge. ny) jendM=ny-1

    nPts=jend-jstart+1
    !write(*,*) " L1878 ndata=", ndata, " nPts=", nPts
    !write(*,*) " L1879 jstart=", jstart, " jstartM=", jstartM, " jend=", jend, " jendM=", jendM

    nPts=jendM-jstartM+1

    allocate(xinput(1:nPts))
    allocate(yinput(1:nPts))
!----
    allocate(xdata(1:nPts))
    allocate(ydata(1:nPts))
!----
    allocate(weights(1:nPts))

    allocate(yp(1:nPts))
    !allocate(temp(1:nPts*9))
    allocate(temp(1:nPts))
    allocate(ysp(1:nPts))
    allocate(dispersion(1:nPts))

    ! VNY 14.01.2004
    !numparam=4
    !numparam=9
    numparam=5!
! VNY 26.04.04
!    if (verbFlag .eq. 1) then
!       write(*,*) "********************"
!       do j=jstart,jend
!          write(*,*) " j=", j, " sp=", specVect(j)
!       enddo
!       write(*,*) "********************"
!    endif ! verbFlag .eq. 1
!
    do j=jstartM,jendM
       if (j .gt. 0 .and. j .lt. ny)  then
          xinput(j-jstartM+1)=real(j+1)
          yinput(j-jstartM+1)=specVect(j)
       endif
    enddo

    ! 25.04.2005
    !if (verbFlag .eq. 1) then
    !   ! plotting the graph
    !   if (pgbeg(0,'plot250405/PS',1,1)/=1) then
    !      write(*,*) " OMGEXTRACTION: L1889 cannot opent plot device"
    !   else
    !      write(*,*) " L1891 The plot device is open"
    !      xlabel="pixels"
    !      ylabel="counts"
    !      graphTitle="plot"
    !      xdata=xinput
    !      ydata=yinput
    !      ndata=nPts
    !      x1=0.0
    !      x2=1.0
    !      y1=0.0
    !      y2=1.0
    !      write(*,*) " x1=", x1, " x2=", x2, " y1=", y1, " y2=", y2
    !      call DrawGraph(ndata, xdata, ydata, &
    !           graphTitle, xlabel, ylabel, x1, x2, y1, y2)
    !      call pgend
    !      write(*,*) " L1895 The plot device is closed"
    !   endif
    !endif ! verbFlag .eq. 1
!
    !write(*,*) "%%%%%%%%%%% ComputeZeroOrder L1912: computing the gaussian approximation" 
    verbFlag2=0
!    call gaussianal(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)
    call gaussianalSim(xinput,yinput, nPts, param5, verbFlag2, nUsedPoints)
    !if (verbFlag .eq. 1) then
    !   write(*,*) " L1974 GAUSSIAN APPROXIMATION: P2=", param5(2), &
    !        " P3=", param5(3), " P4=", param5(4)
    !endif
    gaussianMaximum=param5(3) ! The Gaussian maximum will be used to define the centre of the box
    ! for the method of momenta
    gaussianAmplitude=param5(2)
    gaussianDisp=param5(4)
    numparam=5
    !if (verbFlag .eq. 1) then
    !   ! plotting the obtained gaussiana
    !   do j=1, nPts
    !      xtest=xinput(j)
    !      call fgaussl(xtest,param5,ytest,dydptest,numparam)
    !         write(*,*) "     j=", j, " x(j)=", xtest, " yinput(j)=",yinput(j), " youtput=", ytest
    !         xdata(j)=xtest
    !         ydata(j)=ytest
    !   enddo
    !   ! plotting the graph
    !   if (pgbeg(0,'plot260404_gauss.ps/PS',1,1)/=1) then
    !      write(*,*) " OMGEXTRACTION: L1889 cannot opent plot device"
    !   else
    !      write(*,*) " L1938 The plot device is open"
    !      xlabel="pixels"
    !      ylabel="counts"
    !      graphTitle="plot"
    !      !xdata=xinput
    !      !ydata=yinput
    !      ndata=nPts
    !      x1=0.0
    !      x2=1.0
    !      y1=0.0
    !      y2=1.0
    !      write(*,*) " x1=", x1, " x2=", x2, " y1=", y1, " y2=", y2
    !      call DrawGraph(ndata, xdata, ydata, &
    !           graphTitle, xlabel, ylabel, x1, x2, y1, y2)
    !      call pgend
    !      write(*,*) " L1953 The plot device is closed"
    !   endif
    !endif

    !write(*,*) "L2045 %%%%%%%%%%% RecomputeZeroOrder computing the lowess approximation"
    !endif
    smooth=0.2 ! smoothing factor for the lowess approximation algorithm
    numsteps=0
    delta=0.0
    !if (verbFlag .eq. 1) write(*,*) " Computing Lowess ..."
    call lowess(xinput, yinput, nPts, smooth, numsteps, delta, &
         ysp, yp, dispersion)
    ! Find maximum and two minima
    max1=0.0
    imax1=-1
    do i=1, nPts
       if (ysp(i) .gt. max1) then
          max1=ysp(i)
          imax1=i
       endif
    enddo
    !if (verbFlag .eq. 1 .and. j .eq.  866) then !!!!!!!!
    !if (verbFlag .eq. 1) then !!!!!!!!
    !   write(*,*) " L1996 found a Lowess maximum at i=", imax1, "(", imax1+jstart-1, ") value=", max1, &
    !        " jstart=", jstart
    !endif
    imin1=1
    imin2=nPts
    if (imax1 .gt. 1 .and. imax1 .lt. nPts) then
       ! Find two minima
       !write(*,*) " L2003 ..."
       imin1=1
       min1=ysp(1)
       ! first find a positive number
       do i=1,imax1-1
          imin1=i
          min1=ysp(i)
          if (yinput(i) .gt. 0.0) then 
             exit
          endif
       enddo
       !if (verbFlag .eq. 1) then !!!!!!!!
       !   write(*,*) " L2081 the first starting point is  at i=", imin1, " value=", min1
       !endif
       !write(*,*) " L2017 ... imin1+1=", imin1+1, " imax1-1=", imax1-1
       do i=imin1+1, imax1-1
          if (ysp(i) .lt. min1 .and. ysp(i) .gt. 0.0) then
             min1=ysp(i)
             imin1=i
          endif
       enddo
       imin2=nPts
       min2=ysp(nPts)
       !write(*,*) " L2026 ... imax1+1=", imax1+1
       do i=imax1+1, nPts-1
          if (ysp(i) .lt. min2 .and. ysp(i) .gt. 0.0) then
             min2=ysp(i)
             imin2=i
          endif
       enddo
    else
       imax1=nPts/2
    endif ! imax1 .gt. 1

    ! Find the point of the equal areas for the spline
    averSpecSimple=0.0
    averLeftSimple=0.0
    averRightSimple=0.0
    !if (verbFlag .eq. 1) then
    !   write(*,*) " L2077 imin1=", imin1, " imin2=",imin2
    !endif
    wSpec=real(imin2-imin1)
    dSpec=ysp(imin2)-ysp(imin1)
    if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
    averLeft0=0.0
    averright0=99999.0
    averLeft1=0.0
    averRight1=0.0
    jEquilibrium=0.0
    averLeftEq=0.0
    averRightEq=0.0
    !write(*,*) " L2048 ... imin1=", imin1, " imin2=", imin2
    do jcross=imin1,imin2
       !compute left and right areas
       averLeft1=0.0
       averRight1=0.0
       do j=imin1,imin2
          wSpec=real(j-imin1)
          levelLR=ysp(imin1)+dSpec*wSpec
          if (ysp(j) .gt. 0.0) then
             if (ysp(j)-levelLR .gt. 0.0) then
                if (j .lt. jcross) averLeft1=averLeft1+ysp(j)-levelLR
                if (j .gt. jcross) averRight1=averRight1+ysp(j)-levelLR
                if (j .eq. jcross) then 
                   averLeft1=averLeft1+(ysp(j)-levelLR)/2.0
                   averRight1=averRight1+(ysp(j)-levelLR)/2.0
                endif
             endif
          endif
       enddo
       if (averLeft1 .gt. averRight1 .and. averLeft0 .lt. averRight0) then
          ! equilibrium pixel is found)
          jEquilibrium=jcross-1
          averLeftEq=averLeft1
          averRightEq=averRight1
          l0=averLeft0
          r0=averRight0
          l1=averLeft1
          r1=averRight1
       endif
       !write(*,*) " L2077 jcross=", jcross+jstart-1, " averLeft1=", averLeft1, &
       !     " averRight1=", averRight1
       averLeft0=averLeft1
       averRight0=averRight1
    enddo
    !write(*,*) "-------------"
    !write(*,*) " jEquilibrium=", jEquilibrium+jstart-1, " l0=", l0, " l1=", l1, &
    !     " r0=", r0, " r1=", r1
    ratio=averLeftEq/(averLeftEq+averRightEq)
    ratio2=(averRightEq-averLeftEq)/(averRightEq+averLeftEq)
    ratio3=(r0-l0)/(l1-l0+r0-r1)
    !write(*,*) " ratio=", ratio, " ratio2=", ratio2, " ratio3=", ratio3
    if (verbFlag .eq. 1) write(*,*) " L2045 EQUILIBRIUM OF SPLINE=", jEquilibrium+jstart-1+ratio3

    ! Find the point of the equal areas for the non-smoothed curve
    averSpecSimple=0.0
    averLeftSimple=0.0
    averRightSimple=0.0
    wSpec=real(imin2-imin1)
    dSpec=yinput(imin2)-yinput(imin1)
    if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
    averLeft0=0.0
    averright0=99999.0
    averLeft1=0.0
    averRight1=0.0
    jEquilibrium=0.0
    averLeftEq=0.0
    averRightEq=0.0
    do jcross=imin1,imin2
       !compute left and right areas
       averLeft1=0.0
       averRight1=0.0
       do j=imin1,imin2
          wSpec=real(j-imin1)
          levelLR=yinput(imin1)+dSpec*wSpec
          if (yinput(j) .gt. 0.0) then
             if (yinput(j)-levelLR .gt. 0.0) then
                if (j .lt. jcross) averLeft1=averLeft1+yinput(j)-levelLR
                if (j .gt. jcross) averRight1=averRight1+yinput(j)-levelLR
                if (j .eq. jcross) then 
                   averLeft1=averLeft1+(yinput(j)-levelLR)/2.0
                   averRight1=averRight1+(yinput(j)-levelLR)/2.0
                endif
             endif
          endif
       enddo
       if (averLeft1 .gt. averRight1 .and. averLeft0 .lt. averRight0) then
          ! equilibrium pixel is found)
          jEquilibrium=jcross-1
          averLeftEq=averLeft1
          averRightEq=averRight1
          l0=averLeft0
          r0=averRight0
          l1=averLeft1
          r1=averRight1
       endif
       !write(*,*) " jcross=", jcross+jstart-1, " averLeft1=", averLeft1, &
       !     " averRight1=", averRight1
       averLeft0=averLeft1
       averRight0=averRight1
    enddo
    !write(*,*) "-------------"
    !write(*,*) " jEquilibrium=", jEquilibrium+jstart-1, " l0=", l0, " l1=", l1, &
    !     " r0=", r0, " r1=", r1
    ratio=averLeftEq/(averLeftEq+averRightEq)
    ratio2=(averRightEq-averLeftEq)/(averRightEq+averLeftEq)
    ratio3=(r0-l0)/(l1-l0+r0-r1)
    !write(*,*) " ratio=", ratio, " ratio2=", ratio2, " ratio3=", ratio3
    if (verbFlag .eq. 1) write(*,*) " L2101 EQUILIBRIUM NON_SMOOTHED=", jEquilibrium+jstart-1+ratio3

    deallocate(xinput)
    deallocate(yinput)
!---
    deallocate(xdata)
    deallocate(ydata)
!---
    deallocate(weights)
    deallocate(yp)
    deallocate(temp)
    deallocate(ysp)
    deallocate(dispersion)

    !---
    deallocate(specVect)
    deallocate(smoothSpecVect)
    ! VNY 23.04.2004 Compute using the method of moments
    sourceNumber=1
    !if (verbFlag .eq. 1) then 
    !   write(*,*) " L2199 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax
    !   write(*,*) " gMax=", gaussianMaximum, " gAmpl=", gaussianAmplitude, " gDisp=", gaussianDisp
    !   write(*,*) " xZeroIn=", xZeroIn, " yZeroIn=", yZeroIn
    !endif
    ! Find limits in the vertical direction
    icentre=int(xZeroIn+0.5)
    jcentre=int(gaussianMaximum+0.5)
    jMin=jcentre-1
    jMax=jcentre+1
    !if (verbFlag .eq. 1) then 
    !   write(*,*) "icentre=", icentre, " jcentre=",jcentre
    !   write(*,*) " detecting maximum"
    !   write(*,*) "-------------------"
    !endif
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    do j=jcentre, jcentre+int(spectrumHalfHeight)
       if (j .ge. ny-1) exit
       !if (verbFlag .eq. 1) write(*,*) " j=", j, " osw=", oswImage(icentre,j)
       if (oswImage(icentre,j) .lt. gaussianAmplitude/3.0) exit
    enddo
    jMax=j
    !if (verbFlag .eq. 1) then
    !   write(*,*) " detecting minimum"
    !   write(*,*) "------------------"
    !endif
    do j=jcentre,jcentre-int(spectrumHalfHeight),-1
       if (j .le. 0) exit
       !if (verbFlag .eq. 1) write(*,*) " j=", j, " osw=",oswImage(icentre,j)
       if (oswImage(icentre,j) .lt. gaussianAmplitude/3.0) exit
    enddo
    jMin=j
    !if (verbFlag .eq. 1) write(*,*) " L2230 jMin=", jMin, " jMax=", jMax  
    jdistanceMin=jcentre-jMin
    jdistanceMax=jMax-jcentre
    !if (verbFlag .eq. 1) write(*,*) " L2240 jdistanceMin=", jdistanceMin, " jdistanceMax=", jdistanceMax
    jdistanceMin=min(jdistanceMin, jdistanceMax)
    !if (verbFlag .eq. 1) write(*,*) " jdistanceMin=", jdistanceMin
    jMin=jcentre-jdistanceMin
    jMax=jcentre+jdistanceMin

    call ComputeMoments(sourceNumber, oswImage, iMin, iMax, &
         jMin, jMax, xPos, yPos, xPosErr, yPosErr, aValue, bValue, &
         theta)
    !if (verbFlag .eq. 1) then
    !   write(*,*) " L2134 COMPUTE MOMENTS: xPos=", xPos, " yPos=", yPos, &
    !        " xPosErr=", xPosErr, " yPosErr=", yPosErr, " aValue=", aValue, &
    !        " bValue=", bValue, " theta=", theta
    !endif
    xZeroOut=xPos
    yZeroOut=yPos

    ! Test of the function fgaussxy
!    param6(1)=bValue
!    param6(2)=aValue
!    param6(3)=80.0
!    param6(4)=0.5
!    param6(5)=344.0
!    param6(6)=0.5
!    numparam=6
!    iMin=75
!    iMax=85
!    jMin=339
!    jMax=349
!    do j=jMin, jMax
!       yD=real(j)
!       do i=iMin,iMax
!          xD=real(i)
!          !fxy=fgaussxy(xD,yD, param6, numparam)
!          write(*,'(f6.2,f6.2,a,f6.2)') xD,yD, "     f=", &
!               fgaussxy(xD,yD, param6, numparam)
!       enddo
!    enddo
!    write(*,*) "----------"
!    xD=80.0
!    yD=344.0
!    write(*,*) " x=", xD, " y=", yD, " f=", fgaussxy(xD,yD,param6,numparam)

  end subroutine RecomputeZeroOrder

!**********************************************************************




!******************************************************************
! Convolutes an exponent with the spectra at a given point
!******************************************************************
real function ConvoluteExp(spectrumArray, sizeY, currentY, expWidth)
    integer, intent(in) :: sizeY, currentY
    real(kind=double), intent(in), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(in) :: expWidth
    integer :: k, l, nSample
    real(kind=double) :: a1, a2, sum1, sum2
    real(kind=double), dimension(:), pointer :: verticalStrip
    real(kind=double) :: kSlope, aux
    nSample=2*int(expWidth)
    allocate(verticalStrip(0:nSample-1))
    ! fill in the sample
    verticalStrip=0.0
    do k=0, nSample-1
       l=currentY-nSample/2+k
       if (l .ge. 0 .and. l .lt. sizeY) then
          verticalStrip(k)=spectrumArray(l)
       endif
    enddo
    ! remove the slope
    kSlope=(verticalStrip(nSample-1)-verticalStrip(0))/nSample
    aux=verticalStrip(0)
    do k=0, nSample-1
       verticalStrip(k)=verticalStrip(k)-aux-real(k)*kSlope
    enddo
    sum1=0.0
    sum2=0.0
    do k=0,nSample-1
       a1=abs(k-nSample/2)
       a1=4.0*real(a1)/expWidth
       a2=exp(-a1)
       sum1=sum1+verticalStrip(k)*a2
       sum2=sum2+a2
    enddo
    if (sum2 .gt. 0.0) sum1=sum1/sum2
    ConvoluteExp=sum1
    deallocate(verticalStrip)
  end function ConvoluteExp

!**********************************************
! Test 05.05.04
real(kind=double) function sExcess(ratio)
  real(kind=double), intent(in) :: ratio ! R/a ratio
  real(kind=double) :: e, r, arg1, arg2, arg3
  r=ratio/sqrt(1-ratio*ratio)
  arg1=atan(r)+atan(ratio/2.0)
  arg2=atan(r)+2.0*atan(ratio/2.0)
  arg3=sqrt(tan(arg1)*r)*tan(arg2)
  e=4.0*atan(arg3)
  sExcess=e
  end function sExcess
!************************************************

!*******************************************************************
! Subroutine for conversion of count rates to fluxes
!*******************************************************************
subroutine CountsToFlux(wavelength, countRate, countRateErr, flux, fluxErr, &
     calWavelength, calFlux, calFluxErr)
real(kind=double), intent(in) :: wavelength
real(kind=double), intent(in) :: countRate
real(kind=double), intent(in) :: countRateErr
real(kind=double), intent(in), dimension(:) :: calWavelength
real(kind=double), intent(in), dimension(:) :: calFlux
real(kind=double), intent(in), dimension(:) :: calFluxErr
real(kind=double), intent(out) :: flux
real(kind=double), intent(out) :: fluxErr
!---
integer :: cSize
integer :: j, indexW
real(kind=double) :: fluxIncrement, wIncrement, extrapolatedFlux, normCountRateErr 
real(kind=double) :: limitWavelength

flux=0.0
fluxErr=0.0
cSize= size(calWavelength)
limitWavelength=calWavelength(csize)

!if (wavelength .gt. calWavelength(cSize)-50.0 .and. &
!     wavelength .lt. calWavelength(cSize)) then
!   write(*,*) " L2843  omgextrcation: wavelength=", wavelength, &
!     " calWavelength(cSize)=", calWavelength(cSize)
!endif

! return if the wavelength is out of range of CAL
if (wavelength .lt. calWavelength(1) .or. wavelength .gt. calWavelength(cSize)) then
   flux=0.0
   fluxErr=0.0
   return
endif

indexW=-1
do j=1,cSize
   if (calWavelength(j) .gt. wavelength) then
      indexW=j
      exit 
   endif
enddo


!if (indexW .gt. -1 .and. indexW .ne. cSize .and. &
!     wavelength .lt. limitWavelength) then 
if (indexW .gt. -1 .and. indexW .le. cSize .and. &
     wavelength .le. limitWavelength) then 

   if (indexW .gt. 0 .and. indexW .lt. cSize) then
      fluxIncrement=calFlux(indexW+1)-calFlux(indexW)
      wIncrement=calWavelength(indexW+1)-calWavelength(indexW)
      fluxIncrement=fluxIncrement/wIncrement
      wIncrement=wavelength-calWavelength(indexW)
      extrapolatedFlux=calFlux(indexW)+fluxIncrement*wIncrement
      flux=countRate*extrapolatedFlux
      if (abs(countRate) .gt. 1.0e-6) then
         normCountRateErr=countRateErr/abs(countRate)
      else
         normCountRateErr=0.0
      endif
      fluxErr=calFluxErr(indexW)*calFluxErr(indexW)/10000.0+ &
           normCountRateErr*normCountRateErr
      fluxErr=sqrt(fluxErr)
      fluxErr=countRate*fluxErr*extrapolatedFlux
   endif
   if (indexW .eq. cSize) then
      fluxIncrement=calFlux(indexW)-calFlux(indexW-1)
      wIncrement=calWavelength(indexW)-calWavelength(indexW-1)
      fluxIncrement=fluxIncrement/wIncrement
      wIncrement=wavelength-calWavelength(indexW-1)
      extrapolatedFlux=calFlux(indexW-1)+fluxIncrement*wIncrement
      if (abs(countRate) .gt. 1.0e-6) then
         normCountRateErr=countRateErr/abs(countRate)
      else
         normCountRateErr=0.0
      endif
      flux=countRate*extrapolatedFlux
      fluxErr=calFluxErr(indexW)*calFLuxErr(indexW)/10000.0+ &
           normCountRateErr*normCountRateErr
      fluxErr=sqrt(fluxErr)
!      fluxErr=calFluxErr(indexW)/100.0
      fluxErr=countRate*fluxErr*extrapolatedFlux
   endif
endif

! Find the wavelenght index in cal-array

end subroutine CountsToFlux

!----
! Median smoothing function 11 points
!real function smoothMedian(vect, j, nmax)
!  integer, intent(in) :: nmax
!  real, intent(in) :: vect(0:nmax)
!  integer, intent(in) :: j
!  real :: a(11) ! smoothing for +-5 elements
!  integer :: k,l
!  real :: dum
!  ! Copy the current 11 elements of the vector (indexed +-5)
!  do k=1,11 !j-5,j+5
!     if (j+k-6 .lt. 0 .or. j+k-6 .gt. nmax) then
!        a(k) = 9.E+20
!     else
!        a(k)=vect(j+k-6)
!     endif
!  enddo
!  ! ordering the vector a
!  do k=1,10
!     do l=k+1, 11
!        if (a(l) .lt. a(k)) then
!           dum=a(k)
!           a(k)=a(l)
!           a(l)=dum
!        endif
!     enddo
!  enddo
!  smoothMedian=a(6)
!end function smoothMedian

!----
! Median smoothing function 41 points
real function smoothMedian(vect, j, nmax)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer, intent(in) :: j
  real(kind=double) :: a(41) ! smoothing for +-5 elements
  integer :: k,l
  real(kind=double) :: dum
  ! Copy the current 11 elements of the vector (indexed +-5)
  do k=1,41 !j-5,j+5
     if (j+k-21 .lt. 0 .or. j+k-21 .gt. nmax) then
        a(k) = 9.E+20
     else
        a(k)=vect(j+k-21)
     endif
  enddo
  ! ordering the vector a
  do k=1,40
     do l=k+1, 41
        if (a(l) .lt. a(k)) then
           dum=a(k)
           a(k)=a(l)
           a(l)=dum
        endif
     enddo
  enddo
  smoothMedian=a(21)
end function smoothMedian


!----
! smoothing background excluding possible zero-orders
real function smoothBkgExcludeZeros(vect, jfirst, jlast, nmax, verbFlag)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer, intent(in) :: jfirst,jlast
  real(kind=double), dimension(:), pointer :: vectAux
  integer :: nAux, i, k
  real(kind=double) :: dum, sum
  integer, intent(in) :: verbFlag

  ! Fill in the auxiliar vector
  nAux=jlast-jfirst
  if (nAux .gt. nmax) nAux=nmax
  allocate(vectAux(0:nAux))
  do i=0,nAux
     if (i+jfirst .le. nmax) then
        if (vect(i+jfirst) .gt. 0.0) then
           vectAux(i)=vect(i+jfirst)
        else
           vectAux(i)=999999.0
        endif
     else
        vectAux(i)=999999.0
     endif
  enddo

  if (verbFlag .eq. 1) then
     write(*,*) "L3272 jfirst=", jfirst, " jlast=", jlast, " nAux=", nAux
  endif
  ! sort the auxiliar vector
  do i=0,nAux-1
     do k=i+1,nAux
        if (vectAux(i) .gt. vectAux(k)) then
           dum=vectAux(i)
           vectAux(i)=vectAux(k)
           vectAux(k)=dum
        endif
     enddo
  enddo

  ! compute the average of the first half of the auxiliar vector
  dum=0.0
  sum=0.0
  do i=0,nAux/2
     if (vectAux(i) .lt. 99999.0) then
        sum=sum+vectAux(i)
        dum=dum+1.0
     endif
  enddo
  if (dum .gt. 1.0) sum=sum/dum
  smoothBkgExcludeZeros=sum
  deallocate(vectAux)
end function smoothBkgExcludeZeros

real function smoothBkgExcludeZeros2(vect, maskArray, jfirst, jlast, nmax, verbFlag)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer(kind=int32), intent(in) :: maskArray(0:nmax)
  integer, intent(in) :: jfirst,jlast
  real(kind=double), dimension(:), pointer :: vectAux
  integer :: nAux, i, k
  real(kind=double) :: dum, sum
  integer, intent(in) :: verbFlag

  ! Fill in the auxiliar vector
  nAux=jlast-jfirst
  if (nAux .gt. nmax) nAux=nmax
  allocate(vectAux(0:nAux))
  do i=0,nAux
     if (i+jfirst .le. nmax .and. maskArray(i) .ne. 0) then
        if (vect(i+jfirst) .gt. 0.0) then
           vectAux(i)=vect(i+jfirst)
        else
           vectAux(i)=999999.0
        endif
     else
        vectAux(i)=999999.0
     endif
  enddo

  if (verbFlag .eq. 1) then
     write(*,*) "L3322 jfirst=", jfirst, " jlast=", jlast, " nAux=", nAux
  endif
  ! sort the auxiliar vector
  do i=0,nAux-1
     do k=i+1,nAux
        if (vectAux(i) .gt. vectAux(k)) then
           dum=vectAux(i)
           vectAux(i)=vectAux(k)
           vectAux(k)=dum
        endif
     enddo
  enddo

  ! compute the average of the first half of the auxiliar vector
  dum=0.0
  sum=0.0
  do i=0,nAux/2
     if (vectAux(i) .lt. 99999.0) then
        sum=sum+vectAux(i)
        dum=dum+1.0
     endif
  enddo
  if (dum .gt. 1.0) sum=sum/dum
  smoothBkgExcludeZeros2=sum
  deallocate(vectAux)
end function smoothBkgExcludeZeros2


! Median smoothing function within a given window
real function smoothMedianWindow(vect, j, nmax, windowHalfWidth)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer, intent(in) :: j
  integer, intent(in) :: windowHalfWidth
  real(kind=double) :: a(1:2*windowHalfWidth+1) ! smoothing for +-5 elements
  integer :: k,l
  real(kind=double) :: dum
  ! Copy the current 11 elements of the vector (indexed +-5)
  do k=1,2*windowHalfWidth+1 !j-5,j+5
     if (j+k-windowHalfWidth-1 .lt. 0 .or. j+k-windowHalfWidth-1 .gt. nmax) then
        a(k) = 9.E+20
     else
        a(k)=vect(j+k-windowHalfWidth-1)
     endif
  enddo
  ! ordering the vector a
  do k=1,2*windowHalfWidth+1
     do l=k+1, 2*windowHalfWidth+1
        if (a(l) .lt. a(k)) then
           dum=a(k)
           a(k)=a(l)
           a(l)=dum
        endif
     enddo
  enddo
  smoothMedianWindow=a(windowHalfWidth+1)
end function smoothMedianWindow


!---smoothing by averaging
real function smoothAver(vect, j, nmax)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer, intent(in) :: j
  integer :: k
  real(kind=double) :: sum, nsum
  sum=0.0
  nsum=0.0
  do k=1,41
     if (j+k-21 .gt. 0 .and. j+k-21 .le. nmax) then
        sum=sum+vect(j+k-21)
        nsum=nsum+1.0
     endif
  enddo
  if (nsum .gt. 0.0) sum=sum/nsum
  smoothAver=sum  
end function smoothAver


!******************************************************************
! Check zero-order coordinates
!******************************************************************
  subroutine CheckZeroOrder(grismSetH, xZeroIn, yZeroIn, &
       spectrumHalfWidth, spectrumHalfHeight, computingMode, &
       verbFlag, xZeroOut, yZeroOut, averCountIn, averCountOut)
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xZeroIn, yZeroIn
    integer, intent(in) :: computingMode
    integer :: computingMode2
    real(kind=double), intent(in) :: spectrumHalfWidth
    real(kind=double), intent(in) :: spectrumHalfHeight
    integer, intent(in) :: verbFlag
    real(kind=double), intent(out) :: xZeroOut, yZeroOut
    real(kind=double), intent(out) :: averCountIn, averCountOut
    type(ArrayT) :: inArrayH   
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    integer :: j, ny, nx
    character(len=9) :: filter ! filter keyword in the input file
    integer(kind=int32) ccdMode !+  
    integer(kind=int32) :: binX                                        ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    logical :: binBpe !+
    real(kind=double), dimension(:), pointer :: specVect!, smoothSpecVect
    integer :: ixLeft, ixRight, jcross
    real(kind=double) :: averSpec,numSpec
!    integer :: jstart
!    integer :: jend, nPts
!    real(kind=double), dimension (:), pointer :: xinput
!    real(kind=double), dimension (:), pointer :: yinput
!    real(kind=double), dimension (:), pointer :: weights
    !
!    real, dimension (:), pointer :: xdata
!    real, dimension (:), pointer :: ydata
!    character(len=20) :: graphTitle
!    character(len=10) :: xlabel
!    character(len=10) :: ylabel    
!    real :: x1,x2, y1,y2
!    integer :: ndata
!    integer :: pgbeg
!    real(kind=double), dimension(5) ::dydptest 
    !
!    real(kind=double), dimension (:), pointer :: yp
!    real(kind=double), dimension (:), pointer :: ysp
!    real(kind=double), dimension (:), pointer :: temp
!    real(kind=double), dimension (:), pointer :: dispersion
!    real(kind=double) :: slp1, slpn, tension
    integer :: verbFlag2
!    real(kind=double), dimension(5) :: param5
!    integer :: nUsedPoints
!    integer(kind=int32) :: numparam, numpoints
!    real(kind=double) :: optimal1, optimal2
!    real(kind=double) :: gaussum, simplesum, znam, sumWeights
!    real(kind=double)  :: xtest , ytest, aux1, aux2, aux3, razn
!    real(kind=double), dimension(5) ::dydptest5
!    real(kind=double) :: smooth, delta
!    integer(kind=int32) :: numsteps
!    real(kind=double) :: max1, min1, min2
!    integer :: imax1, imin1, imin2
!    real(kind=double) :: averSpecLowess
!    real(kind=double) :: wSpec, dSpec
!    real(kind=double) :: levelLR
!    real(kind=double) :: averSpecSimple
!    real(kind=double) :: averLeftSimple, averRightSimple, averLeftLowess, averRightLowess
!    real(kind=double) :: ratio, ratio2, averLeft0, averRight0, averLeft1, averRight1
!    integer :: jEquilibrium
!    real(kind=double) :: averLeftEq, averRightEq, l1,r1,l0,r0, ratio3
    ! --- for computing moments:
    integer(kind=int16) :: sourceNumber
    integer(kind=int32) :: iMin, iMax, jMin, jMax
    real(kind=single) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 
!    real(kind=double), dimension(6) :: param6
!    real(kind=double) :: xD, yD, fxy
!    integer :: jstartM, jendM
    integer ::  icentre, jcentre, jdistanceMin, jdistanceMax
!    real(kind=double) gaussianMaximum
    real(kind=double) :: gaussianAmplitude
!    real(kind=double) :: gaussianDisp
    real(kind=double) :: maxValue, xZero2, yZero2, xZero3, yZero3
    integer :: imaxValue
    real(kind=double) :: averArea, numArea

    computingMode2=computingMode
    verbFlag2=verbFlag

    !write(*,*) "CheckZeroOrder L2837:  verbosity flag=", verbFlag 


    ! information on the science window
    inArrayH = array(grismSetH, "PRIMARY", read)
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  

    if (hasAttribute(grismSetH,"FILTER")) then
       filter = stringAttribute(grismSetH, "FILTER")
    else
       call warning('OMEXTRACTION','Keyword FILTER missing in the input file ')
    endif

    ccdMode = IMAGE

    ! Calculate image binning.  
    binBpe= .false.
    if (hasAttribute(grismSetH,"BINBPE")) then       
       binBpe= booleanAttribute(grismSetH, "BINBPE")
    else
       call warning('OMGEXTRACTION','Keyword BINBPE missing in the input file ')
    endif

    if (hasAttribute(grismSetH,"BINAX1")) then       
       binAx1 = int32Attribute(grismSetH, "BINAX1")
       binX = 2 ** binAx1
    else
       call warning('OMGEXTRACTION','Keyword BINAX1 missing in the input file ')
       binX=1.0
    endif

    if (hasAttribute(grismSetH,"BINAX2")) then       
       binAx2 = int32Attribute(grismSetH, "BINAX2")
       binY = 2 ** binAx2
    else
       call warning('OMGEXTRACTION','Keyword BINAX2 missing in the input file ')
       binY=1.0
    endif


    if (binBPE) then
       binX = 2.0
       binY = 2.0
    end if
    !if (verbFlag .eq. 1) then 
    !   write (*,*) "ComputZeroOrder L2885"
    !   write(*,*) "ComputeZeroOrder L2886: xZeroIn=", xZeroIn, " yZeroIn=", yZeroIn
    !   write(*,*) "OMGRISMEXTRACTION L2887: binBpe=", binBpe, " binAx1=", binAx1, " binAx2=", binAx2
    !   write(*,*) " filter=", filter, " binX=", binX, " binY=", binY, " nx=", nx, " ny=", ny
    !endif


    allocate(specVect(0:ny-1))
    !allocate(smoothSpecVect(0:ny-1))
    specVect=0.0
    !smoothSpecVect=0.0
    ixRight=nx-1
    ixLeft=xZeroIn-2.0*spectrumHalfWidth
    if (ixLeft .lt. 0) ixLeft=0
    ixRight=xZeroIn+2.0*spectrumHalfWidth
    if (ixRight .gt. nx-1) ixRight=nx-1
    if (ixRight .le. ixLeft) ixRight=ixLeft+1
    !write(*,*) " CheckZeroOrder L2902 ixLeft=", ixLeft, " ixRight=", ixRight
   
    iMin=xZeroIn-2.0*spectrumHalfWidth
    iMax=xZeroIn+2.0*spectrumHalfWidth
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yZeroIn-2.0*spectrumHalfWidth!spectrumHalfHeight
    jMax=yZeroIn+2.0*spectrumHalfWidth!spectrumHalfHeight
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L2912 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    !write(*,*) " ---------------------------------------"
    !write(*,*) " In horizontal direction:"
    !write(*,*) "----------------------------------------"
    do jcross=iMin,iMax ! in horizontal direction
       averSpec=0.0
       numSpec=0.0
       do j=jMin,jMax ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(jcross)=averSpec
       !write(*,*) " jcross=",jcross, " spec=", averSpec
    enddo

    maxValue=0.0
    imaxValue=xZeroIn
    do jcross=iMin,iMax
       if (specVect(jcross) .gt. maxValue) then
          maxValue=specVect(jcross)
          imaxValue=jcross
       endif
    enddo
    !write(*,*) " L2954 maxValue=", maxValue, " imaxValue=", imaxValue, " <--- horizontal"
    xZero2=imaxValue

    !write(*,*) " -----------------------------------------"
    !write(*,*) " In the vertical direction:"
    !write(*,*) "------------------------------------------"
    ! Find maximum in the y-direction
    specVect=0.0
    do j=jMin,jMax ! in vertical direction
       averSpec=0.0
       numSpec=0.0
       do jcross=iMin,iMax ! in horizontal direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(j)=averSpec
       !write(*,*) " j=",j, " spec=", averSpec
    enddo
    maxValue=0.0
    imaxValue=yZeroIn
    do j=jMin,jMax
       if (specVect(j) .gt. maxValue) then
          maxValue=specVect(j)
          imaxValue=j
       endif
    enddo
    !write(*,*) " L2978 maxValue=", maxValue, " imaxValue=", imaxValue, " <- vertical"
    !write(*,*) "-------------------------------"
    yZero2=imaxValue

!------------Check again
    iMin=xZero2-spectrumHalfWidth
    iMax=xZero2+spectrumHalfWidth
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yZero2-spectrumHalfWidth!spectrumHalfHeight
    jMax=yZero2+spectrumHalfWidth!spectrumHalfHeight
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L2982 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    !write(*,*) " ---------------------------------------"
    !write(*,*) " In horizontal direction (2):"
    !write(*,*) "----------------------------------------"
    do jcross=iMin,iMax ! in horizontal direction
       averSpec=0.0
       numSpec=0.0
       do j=jMin,jMax ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(jcross)=averSpec
       !write(*,*) " jcross=",jcross, " spec=", averSpec
    enddo
    maxValue=0.0
    imaxValue=xZeroIn
    do jcross=iMin,iMax
       if (specVect(jcross) .gt. maxValue) then
          maxValue=specVect(jcross)
          imaxValue=jcross
       endif
    enddo
    !write(*,*) " L3008 maxValue=", maxValue, " imaxValue=", imaxValue, " <--- horizontal"
    xZero3=imaxValue

    !write(*,*) " -----------------------------------------"
    !write(*,*) " In the vertical direction (2):"
    !write(*,*) "------------------------------------------"
    ! Find maximum in the y-direction
    specVect=0.0
    do j=jMin,jMax ! in vertical direction
       averSpec=0.0
       numSpec=0.0
       do jcross=iMin,iMax ! in horizontal direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(j)=averSpec
       !write(*,*) " j=",j, " spec=", averSpec
    enddo
    maxValue=0.0
    imaxValue=yZeroIn
    do j=jMin,jMax
       if (specVect(j) .gt. maxValue) then
          maxValue=specVect(j)
          imaxValue=j
       endif
    enddo
    !write(*,*) " L3037 maxValue=", maxValue, " imaxValue=", imaxValue, " <- vertical"
    !write(*,*) "-------------------------------"
    yZero3=imaxValue

    !---
    deallocate(specVect)

    ! VNY 23.04.2004 Compute using the method of moments
    sourceNumber=1
    !if (verbFlag .eq. 1) then 
    !   write(*,*) " L2199 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax
    !   write(*,*) " gMax=", gaussianMaximum, " gAmpl=", gaussianAmplitude, " gDisp=", gaussianDisp
    !   write(*,*) " xZeroIn=", xZeroIn, " yZeroIn=", yZeroIn
    !endif
    ! Find limits in the vertical direction
    icentre=int(xZero3+0.5)
    jcentre=int(yZero3+0.5)

    !--- VNY 09.06.2004
    gaussianAmplitude= oswImage(icentre,jcentre)
    !-------------------------------------------- NB

    jMin=jcentre-1
    jMax=jcentre+1
    !if (verbFlag .eq. 1) then 
       !write(*,*) "icentre=", icentre, " jcentre=",jcentre
       !write(*,*) " detecting maximum"
       !write(*,*) "-------------------"
    !endif
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    do j=jcentre, jcentre+int(spectrumHalfHeight)
       if (j .ge. ny-1) exit
       !if (verbFlag .eq. 1) write(*,*) " j=", j, " osw=", oswImage(icentre,j)
       if (oswImage(icentre,j) .lt. gaussianAmplitude/3.0) exit
    enddo
    jMax=j
    !if (verbFlag .eq. 1) then
       !write(*,*) " detecting minimum"
       !write(*,*) "------------------"
    !endif
    do j=jcentre,jcentre-int(spectrumHalfHeight),-1
       if (j .le. 0) exit
       !if (verbFlag .eq. 1) write(*,*) " j=", j, " osw=",oswImage(icentre,j)
       if (oswImage(icentre,j) .lt. gaussianAmplitude/3.0) exit
    enddo
    jMin=j
    !if (verbFlag .eq. 1) write(*,*) " L3079 jMin=", jMin, " jMax=", jMax  
    jdistanceMin=jcentre-jMin
    jdistanceMax=jMax-jcentre
    !if (verbFlag .eq. 1) write(*,*) " L3082 jdistanceMin=", jdistanceMin, " jdistanceMax=", jdistanceMax
    jdistanceMin=min(jdistanceMin, jdistanceMax)
    !if (verbFlag .eq. 1) write(*,*) " jdistanceMin=", jdistanceMin
    jMin=jcentre-jdistanceMin
    jMax=jcentre+jdistanceMin

    call ComputeMoments(sourceNumber, oswImage, iMin, iMax, &
         jMin, jMax, xPos, yPos, xPosErr, yPosErr, aValue, bValue, &
         theta)
    !if (verbFlag .eq. 1) then
       !write(*,*) " L3092 COMPUTE MOMENTS: xPos=", xPos, " yPos=", yPos, &
       !     " xPosErr=", xPosErr, " yPosErr=", yPosErr, " aValue=", aValue, &
       !     " bValue=", bValue, " theta=", theta
    !endif
    !xZeroOut=xPos
    !yZeroOut=yPos

    xZeroOut=xZero3
    yZeroOut=yZero3
!----------------------------------------
    iMin=xZeroIn-spectrumHalfWidth/2.0
    iMax=xZeroIn+spectrumHalfWidth/2.0
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yZeroIn-spectrumHalfWidth/2.0
    jMax=yZeroIn+spectrumHalfWidth/2.0
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L3110 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    averArea=0.0
    numArea=0.0
    do jcross=iMin,iMax ! in horizontal direction
       do j=jMin,jMax ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averArea=averArea+oswImage(jcross,j)
             numArea=numArea+1.0
          endif
       enddo
    enddo
    if (numArea .gt. 1.0) averCountIn=averArea/numArea
    !write(*,*) " averCountIn=", averCountIn, " (in)"
!-----------------------------------------
    iMin=xZeroOut-spectrumHalfWidth/2.0
    iMax=xZeroOut+spectrumHalfWidth/2.0
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yZeroOut-spectrumHalfWidth/2.0
    jMax=yZeroOut+spectrumHalfWidth/2.0
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L3133 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    averArea=0.0
    numArea=0.0
    do jcross=iMin,iMax ! in horizontal direction
       do j=jMin,jMax ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averArea=averArea+oswImage(jcross,j)
             numArea=numArea+1.0
          endif
       enddo
    enddo
    if (numArea .gt. 1.0) averCountOut=averArea/numArea
    !write(*,*) " averCountOut=", averCountOut, " (out)"

  end subroutine CheckZeroOrder

!**********************************************************************
!******************************************************************
! Check first-order coordinates
!******************************************************************
  subroutine CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
       spectrumHalfWidth, spectrumHalfHeight, computingMode, &
       verbFlag, xFirstOut, yFirstOut, averCountIn, averCountOut)
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xFirstIn, yFirstIn
    integer, intent(in) :: computingMode
    integer :: computingMode2
    real(kind=double), intent(in) :: spectrumHalfWidth
    real(kind=double), intent(in) :: spectrumHalfHeight
    integer, intent(in) :: verbFlag
    real(kind=double), intent(out) :: xFirstOut, yFirstOut
    real(kind=double), intent(out) :: averCountIn, averCountOut
    type(ArrayT) :: inArrayH   
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    integer :: j, ny, nx
    character(len=9) :: filter ! filter keyword in the input file
    integer(kind=int32) ccdMode !+  
    integer(kind=int32) :: binX                                        ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    logical :: binBpe !+
    real(kind=double), dimension(:), pointer :: specVect!, smoothSpecVect
    integer :: ixLeft, ixRight, jcross
    real(kind=double) :: averSpec,numSpec

!    integer :: jstart
!    integer :: jend, nPts
!    real(kind=double), dimension (:), pointer :: xinput
!    real(kind=double), dimension (:), pointer :: yinput
!    real(kind=double), dimension (:), pointer :: weights
    !
!    real, dimension (:), pointer :: xdata
!    real, dimension (:), pointer :: ydata
!    character(len=20) :: graphTitle
!    character(len=10) :: xlabel
!    character(len=10) :: ylabel    
!    real :: x1,x2, y1,y2
!    integer :: ndata, pgbeg
!    real(kind=double), dimension(5) ::dydptest 
    !
!    real(kind=double), dimension (:), pointer :: yp
!    real(kind=double), dimension (:), pointer :: ysp
!    real(kind=double), dimension (:), pointer :: temp
!    real(kind=double), dimension (:), pointer :: dispersion
!    real(kind=double) :: slp1, slpn, tension
    integer :: verbFlag2
!    real(kind=double), dimension(5) :: param5
!    integer :: nUsedPoints
!    integer(kind=int32) :: numparam, numpoints
!    real(kind=double) :: optimal1
!    real(kind=double) :: optimal2
!    real(kind=double) :: gaussum, simplesum, znam, sumWeights
!    real(kind=double)  :: xtest , ytest, aux1, aux2, aux3, razn
!    real(kind=double), dimension(5) ::dydptest5
!    real(kind=double) :: smooth, delta
!    integer(kind=int32) :: numsteps
!    real(kind=double) :: max1, min1, min2
!    integer :: imax1, imin1, imin2
!    real(kind=double) :: averSpecLowess, wSpec, dSpec
!    real(kind=double) :: levelLR
!    real(kind=double) :: averSpecSimple
!    real(kind=double) :: averLeftSimple, averRightSimple, averLeftLowess, averRightLowess
!    real(kind=double) :: ratio, ratio2, averLeft0, averRight0, averLeft1, averRight1
!    integer :: jEquilibrium
!    real(kind=double) :: averLeftEq, averRightEq, l1,r1,l0,r0, ratio3
    ! --- for computing moments:
    integer(kind=int16) :: sourceNumber
    integer(kind=int32) :: iMin, iMax, jMin, jMax
    real(kind=single) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 
!    real(kind=double), dimension(6) :: param6
!    real(kind=double) :: xD, yD, fxy
!    integer :: jstartM, jendM
    integer :: icentre, jcentre, jdistanceMin, jdistanceMax
!    real(kind=double) gaussianMaximum
    real(kind=double) :: gaussianAmplitude
!    real(kind=double) :: gaussianDisp
    real(kind=double) :: maxValue, xFirst2, yFirst2, xFirst3, yFirst3
    integer :: imaxValue
    real(kind=single) :: rmaxValue
    real(kind=double) :: averArea, numArea
    integer:: nCentreArray, jCentreBin, jAlong
    real(kind=double), dimension(:), pointer :: centreArray
    real(kind=double), dimension(:), pointer :: centreShapeArray
    real(kind=double) :: edgeValue
    real(kind=double), dimension(:), pointer :: crossSectionArray
    real(kind=double), dimension(:), pointer :: crossSectionAux
    real(kind=double) :: averCross

    !write(*,*) "CheckFirstOrder L3698:  verbosity flag=", verbFlag 
    computingMode2=computingMode
    verbFlag2=verbFlag

    ! information on the science window
    inArrayH = array(grismSetH, "PRIMARY", read)
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  

    !write(*,*) "L3745 nx=", nx, " ny=", ny
    if (hasAttribute(grismSetH,"FILTER")) then
       filter = stringAttribute(grismSetH, "FILTER")
    else
       call warning('OMEXTRACTION','Keyword FILTER missing in the input file ')
    endif

    ccdMode = IMAGE

    ! Calculate image binning.  
    binBpe= .false.
    if (hasAttribute(grismSetH,"BINBPE")) then       
       binBpe= booleanAttribute(grismSetH, "BINBPE")
    else
       call warning('OMGEXTRACTION','Keyword BINBPE missing in the input file ')
    endif

    if (hasAttribute(grismSetH,"BINAX1")) then       
       binAx1 = int32Attribute(grismSetH, "BINAX1")
       binX = 2 ** binAx1
    else
       call warning('OMGEXTRACTION','Keyword BINAX1 missing in the input file ')
       binX=1.0
    endif

    if (hasAttribute(grismSetH,"BINAX2")) then       
       binAx2 = int32Attribute(grismSetH, "BINAX2")
       binY = 2 ** binAx2
    else
       call warning('OMGEXTRACTION','Keyword BINAX2 missing in the input file ')
       binY=1.0
    endif


    if (binBPE) then
       binX = 2.0
       binY = 2.0
    end if
    if (verbFlag .eq. 1) then 
       write (*,*) "CheckFirstOrder L4198....................."
       write(*,*) "CheckFirstOrder L4199: xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn
       write(*,*) "OMGRISMEXTRACTION L4200: binBpe=", binBpe, " binAx1=", binAx1, " binAx2=", binAx2
       write(*,*) " filter=", filter, " binX=", binX, " binY=", binY, " nx=", nx, " ny=", ny
    endif


    allocate(specVect(0:ny-1))
    !allocate(smoothSpecVect(0:ny-1))
    specVect=0.0
    !smoothSpecVect=0.0
    ixRight=nx-1
    ixLeft=xFirstIn-2.0*spectrumHalfWidth
    if (ixLeft .lt. 0) ixLeft=0
    ixRight=xFirstIn+2.0*spectrumHalfWidth
    if (ixRight .gt. nx-1) ixRight=nx-1
    if (ixRight .le. ixLeft) ixRight=ixLeft+1
    if (verbFlag .eq. 1) then
       write(*,*) " CheckFirstOrder L4216 ixLeft=", ixLeft, " ixRight=", ixRight
    endif
    iMin=xFirstIn-2.0*spectrumHalfWidth
    iMax=xFirstIn+2.0*spectrumHalfWidth
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yFirstIn-2.0*spectrumHalfHeight
    jMax=yFirstIn+2.0*spectrumHalfHeight
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    if (jMin .gt. jMax) jMin=jMax-1
    if (verbFlag .eq. 1) then
       write(*,*) " L3773 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax
       write(*,*) " ---------------------------------------"
       write(*,*) " In horizontal direction:"
       write(*,*) "----------------------------------------"
    endif ! verbFlag
    !-------------------------------------------------------------------------------
    ! VNY 07.12.2005 trying to check the horizontal profile in differnt points along 
    ! the spectrum
    !-------------------------------------------------------------------------------
    jCentreBin=30/binY
    nCentreArray=(jMax-jMin)/jCentreBin
    if (verbFlag .eq. 1) then
       write(*,*) "L4140 nCentreArray=", nCentreArray, " jCentreBin=", jCentreBin
    endif
    allocate(centreArray(1:nCentreArray))
    allocate(centreShapeArray(1:nCentreArray))
    do jAlong=1,nCentreArray
       maxValue=0.0
       imaxValue=xFirstIn
       do jcross=iMin,iMax ! in horizontal direction
          averSpec=0.0
          numSpec=0.0
          do j=jMin+(jAlong-1)*jCentreBin,jMin+jAlong*jCentreBin-1 ! in vertical direction
             if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
                averSpec=averSpec+oswImage(jcross,j)
                numSpec=numSpec+1.0
             endif
          enddo
          if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
          if (averSpec .gt. maxValue) then
             maxValue=averSpec
             iMaxValue=jcross
          endif
       enddo ! jcross
       centreArray(jAlong)=iMaxValue
       edgeValue=0.0
       numSpec=0.0
       do j=jMin+(jAlong-1)*jCentreBin,jMin+jAlong*jCentreBin-1 ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             edgeValue=edgeValue+(oswImage(iMin,j)+oswImage(iMax,j))/2.0
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) edgeValue=edgeValue/numSpec         
       centreShapeArray(jAlong)=maxValue-edgeValue
       if (maxValue .gt. 0.0) centreShapeArray(jAlong)=centreShapeArray(jAlong)/maxValue
       if (verbFlag .eq. 1) then
          write(*,*) "L4255 jAlong=", jAlong, " centreArray=", centreArray(jAlong), &
               " centreshapeArray=", centreShapeArray(jAlong)
       endif
    enddo ! jAlong
    ! compute the average of the cross-section shape 
    !averCross=0.0
    !numSpec=0.0
    !do jAlong=1,nCentreArray
    !   if (centreShapeArray(jAlong) .gt. 0.0) then
    !      averCross=averCross+centreShapeArray(jAlong)
    !      numSpec=numSpec+1.0
    !   endif
    !enddo
    !if (numSpec .gt. 1.0) averCross=averCross/numSpec
    !maxValue=computeVectorMedian(centreArray,nCentreArray,nCentreArray)
    !write(*,*) "L4273 median averaged valuer=", maxValue, " average shape=", averSpec
    averCross=ComputeVectorMedian(centreShapeArray,nCentreArray,nCentreArray)

    maxValue=ComputeSpecialAverage(centreArray,nCentreArray,centreShapeArray)
    !*******!
    if (verbFlag .eq. 1) then
       write(*,*) "L4285 special average=", maxValue
    endif
    deallocate(centreArray)
    deallocate(centreShapeArray)

    ! VNY 07.12.2005
    !do jcross=iMin,iMax ! in horizontal direction
    !   averSpec=0.0
    !   numSpec=0.0
    !   do j=jMin,jMax ! in vertical direction
    !      if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
    !         averSpec=averSpec+oswImage(jcross,j)
    !         numSpec=numSpec+1.0
    !      endif
    !   enddo
    !   if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
    !   specVect(jcross)=averSpec
    !   !write(*,*) " jcross=",jcross, " spec=", averSpec
    !enddo

    !maxValue=0.0
    !imaxValue=xFirstIn
    !do jcross=iMin,iMax
    !   if (specVect(jcross) .gt. maxValue) then
    !      maxValue=specVect(jcross)
    !      imaxValue=jcross
    !   endif
    !enddo
    !if (verbFlag .eq. 1) then
    !   write(*,*) " L3800 maxValue=", maxValue, " imaxValue=", imaxValue, " <--- horizontal (1)"
    !endif
    !xFirst2=imaxValue

    xFirst2=int(maxValue+0.5) ! 0.5 is for rounding the value
    if (verbFlag .eq. 1) then
       write(*,*) " L4327 xFirst2=", xFirst2
       write(*,*) " -----------------------------------------"
       write(*,*) " In the vertical direction:"
       write(*,*) "------------------------------------------"
    endif
    ! Find maximum in the y-direction
    specVect=0.0
    do j=jMin,jMax ! in vertical direction
       averSpec=0.0
       numSpec=0.0
       do jcross=iMin,iMax ! in horizontal direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(j)=averSpec
       !write(*,*) " j=",j, " spec=", averSpec
    enddo
    maxValue=0.0
    imaxValue=yFirstIn
    do j=jMin,jMax
       if (specVect(j) .gt. maxValue) then
          maxValue=specVect(j)
          imaxValue=j
       endif
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) " L3829 maxValue=", maxValue, " imaxValue=", imaxValue, " <- vertical (1)"
       write(*,*) "-------------------------------"
    endif
    yFirst2=imaxValue

!------------Check again
    iMin=xFirst2-spectrumHalfWidth
    iMax=xFirst2+spectrumHalfWidth
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    !jMin=yFirst2-spectrumHalfHeight
    !jMax=yFirst2+spectrumHalfHeight
    !if (jMin .lt. 0) jMin=0
    !if (jMax .gt. ny-1) jMax=ny-1
    !if (jMin .gt. jMax) jMin=jMax-1 
    if (verbFlag .eq. 1) then
       write(*,*) " L3842 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax!!

       write(*,*) " ---------------------------------------"
       write(*,*) " In horizontal direction (2):"
       write(*,*) "----------------------------------------"
    endif
!==========================================
    !-------------------------------------------------------------------------------
    ! VNY 07.12.2005 trying to check the horizontal profile in differnt points along 
    ! the spectrum
    !-------------------------------------------------------------------------------
    jCentreBin=30/binY
    nCentreArray=(jMax-jMin)/jCentreBin
    if (verbFlag .eq. 1) then
       write(*,*) "L4386 nCentreArray=", nCentreArray, " jCentreBin=", jCentreBin
    endif
    allocate(centreArray(1:nCentreArray))
    allocate(centreShapeArray(1:nCentreArray))
    allocate(crossSectionArray(iMin:iMax))
    allocate(crossSectionAux(iMin:iMax))
    crossSectionArray=0.0
    do jAlong=1,nCentreArray
       maxValue=0.0
       imaxValue=xFirstIn
       crossSectionAux=0.0
       do jcross=iMin,iMax ! in horizontal direction
          averSpec=0.0
          numSpec=0.0
          do j=jMin+(jAlong-1)*jCentreBin,jMin+jAlong*jCentreBin-1 ! in vertical direction
             if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
                averSpec=averSpec+oswImage(jcross,j)
                numSpec=numSpec+1.0
             endif
          enddo
          if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
          if (averSpec .gt. maxValue) then
             maxValue=averSpec
             iMaxValue=jcross
          endif
          crossSectionAux(jcross)=averSpec
       enddo ! jcross
       centreArray(jAlong)=iMaxValue
       edgeValue=0.0
       numSpec=0.0
       do j=jMin+(jAlong-1)*jCentreBin,jMin+jAlong*jCentreBin-1 ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             edgeValue=edgeValue+(oswImage(iMin,j)+oswImage(iMax,j))/2.0
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) edgeValue=edgeValue/numSpec         
       centreShapeArray(jAlong)=maxValue-edgeValue
       if (maxValue .gt. 0.0) centreShapeArray(jAlong)=centreShapeArray(jAlong)/maxValue
       if (verbFlag .eq. 1) then
          write(*,*) "L4421 jAlong=", jAlong, " centreArray=", centreArray(jAlong), &
               " centreshapeArray=", centreShapeArray(jAlong)
       endif
       if (abs(centreShapeArray(jAlong)-averCross) .lt. averCross/1.5 &
            .and. centreShapeArray(jAlong) .gt. 0.1) then 
          ! use this cross-section for averaging
          crossSectionArray=crossSectionArray+crossSectionAux
       endif
    enddo ! jAlong

    maxValue=ComputeSpecialAverage(centreArray,nCentreArray,centreShapeArray)
    !******* this is the x-position of the maximum of the spectrum
    if (verbFlag .eq. 1) then
       write(*,*) "L4443 new special average=", maxValue
       write(*,*) "--------------------------"
       do jcross=iMin,iMax
          write(*,*) "L4446 jcross=", jcross, " crossSectionArray=", crossSectionArray(jcross)
       enddo
    endif
    deallocate(centreArray)
    deallocate(centreShapeArray)

!=========================================
    do jcross=iMin,iMax ! in horizontal direction
!       averSpec=0.0
!       numSpec=0.0
!       do j=jMin,jMax ! in vertical direction
!          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
!             averSpec=averSpec+oswImage(jcross,j)
!             numSpec=numSpec+1.0
!          endif
!       enddo
!       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
!       specVect(jcross)=averSpec
       specVect(jcross)=crossSectionArray(jcross)
       !write(*,*) "L4474 jcross=",jcross, " specVect=", specVect(jcross)
    enddo
    maxValue=0.0
    imaxValue=xFirstIn
    do jcross=iMin,iMax
       if (specVect(jcross) .gt. maxValue) then
          maxValue=specVect(jcross)
       !if (crossSectionArray(jcross) .gt. maxValue) then
       !   maxValue=crossSectionArray(jcross)
          imaxValue=jcross
       endif
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) "L4485 maxValue=", maxValue, " imaxValue=", iMaxValue
    endif
    ! Check the maximal neighbouring pixel
    jcross=imaxValue-1
    if (jcross .lt. 0) jcross=0
    j=imaxValue+1
    if (j .gt. nx-1) j=nx-1
    if (specVect(j) .gt. specVect(jcross)) then
    !if (crossSectionArray(j) .gt. crossSectionArray(jcross)) then
       if (verbFlag .eq. 1) write(*,*) "L4497 j=", j, " specVect(j)=", &
            specVect(j), " jcross=", jcross, " specVect(jcross)=", specVect(jcross)
       jcross=j
       j=1
    else
       if (verbFlag .eq. 1) write(*,*) "L4501 j=", j, " specVect(j)=", &
            specVect(j), " jcross=", jcross, " specVect(jcross)=", specVect(jcross)
       j=-1
    endif
    if (verbFlag .eq. 1) then
       write(*,*) "L4502 j=", j, " jcross=", jcross, " imaxValue=", iMaxValue
       write(*,*) " specVect(jcross=", specVect(jcross)
       !write(*,*) " crossSectionArray(jcross)=", crossSectionArray(jcross)
    endif
    if (specVect(imaxValue) .gt. 0.0) then
       xFirst3=real(j)*specVect(jcross)/specVect(imaxValue)*0.5
    else
       xFirst3=specVect(jcross)
    endif
    !if (crossSectionArray(imaxValue) .gt. 0.0) then
    !   xFirst3=real(j)*crossSectionArray(jcross)/crossSectionArray(imaxValue)*0.5
    !else
    !   xFirst3=crossSectionArray(jcross)
    !endif
    if (verbFlag .eq. 1) then
       write(*,*) " L4517 xFirst3=", xFirst3
       write(*,*) " L4518 maxValue=", maxValue, " imaxValue=", imaxValue, " <--- horizontal (2)"
    endif
    rmaxValue=imaxValue+xFirst3
    if (verbFlag .eq. 1) then
       write(*,*) " L4522 corrected rmaxValue=", rmaxValue
    endif
    xFirst3=rmaxValue
    if (verbFlag .eq. 1) write(*,*) "L4530 xFirst3=", xFirst3
    if (verbFlag .eq. 1) then
       write(*,*) " -----------------------------------------"
       write(*,*) " L4527 In the vertical direction (2):"
       write(*,*) "------------------------------------------"
    endif
    ! Find maximum in the y-direction
    specVect=0.0
    do j=jMin,jMax ! in vertical direction
       averSpec=0.0
       numSpec=0.0
       do jcross=iMin,iMax ! in horizontal direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averSpec=averSpec+oswImage(jcross,j)
             numSpec=numSpec+1.0
          endif
       enddo
       if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
       specVect(j)=averSpec
       !write(*,*) "L3887  j=",j, " spec=", averSpec
    enddo
    maxValue=0.0
    imaxValue=yFirstIn
    do j=jMin,jMax
       if (specVect(j) .gt. maxValue) then
          maxValue=specVect(j)
          imaxValue=j
       endif
    enddo
    if (verbFlag .eq. 1) then
       write(*,*) " L3897 maxValue=", maxValue, " imaxValue=", imaxValue, " <- vertical (2)"
       write(*,*) "-------------------------------"
    endif
    yFirst3=imaxValue
    if (yFirst3 .gt. jMax) then 
       yFirst3=jMax-1
       !yFirstIn=jMax-1
    endif
    !---
    deallocate(specVect)
    deallocate(crossSectionArray)
    deallocate(crossSectionAux)

    ! VNY 23.04.2004 Compute using the method of moments
    sourceNumber=1
    if (verbFlag .eq. 1) then 
       write(*,*) " L4576 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax
       !write(*,*) "L3988  gMax=", gaussianMaximum, " gAmpl=", gaussianAmplitude, &
       !     " gDisp=", gaussianDisp
       write(*,*) "L4579 xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn
    endif
    ! Find limits in the vertical direction
    icentre=int(xFirst3+0.5)
    jcentre=int(yFirst3+0.5)
    ! NB ------ VNY 09.06.2005 gaussianAmplitude was never set
    gaussianAmplitude=oswImage(icentre,jcentre)
    !------------------------------------------ NB

    jMin=jcentre-1
    jMax=jcentre+1
    
    if (verbFlag .eq. 1) then 
       write(*,*) "L4592 icentre=", icentre, " jcentre=",jcentre
       write(*,*) " detecting maximum"
       write(*,*) "-------------------"
    endif
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    do j=jcentre, jcentre+int(spectrumHalfHeight)
       if (j .ge. ny-1) exit
       !if (verbFlag .eq. 1) write(*,*) " j=", j, " osw=", oswImage(icentre,j)
       if (oswImage(icentre,j) .lt. gaussianAmplitude/3.0) exit
    enddo
    jMax=j
    !if (verbFlag .eq. 1) then
       !write(*,*) " detecting minimum"
       !write(*,*) "------------------"
    !endif
    do j=jcentre,jcentre-int(spectrumHalfHeight),-1
       if (j .le. 0) exit
       !if (verbFlag .eq. 1) write(*,*) " j=", j, " osw=",oswImage(icentre,j)
       if (oswImage(icentre,j) .lt. gaussianAmplitude/3.0) exit
    enddo
    jMin=j
    !if (verbFlag .eq. 1) write(*,*) " L3079 jMin=", jMin, " jMax=", jMax  
    jdistanceMin=jcentre-jMin
    jdistanceMax=jMax-jcentre
    !if (verbFlag .eq. 1) write(*,*) " L3082 jdistanceMin=", jdistanceMin, " jdistanceMax=", jdistanceMax
    jdistanceMin=min(jdistanceMin, jdistanceMax)
    !if (verbFlag .eq. 1) write(*,*) " jdistanceMin=", jdistanceMin
    jMin=jcentre-jdistanceMin
    jMax=jcentre+jdistanceMin

    call ComputeMoments(sourceNumber, oswImage, iMin, iMax, &
         jMin, jMax, xPos, yPos, xPosErr, yPosErr, aValue, bValue, &
         theta)
    !if (verbFlag .eq. 1) then
       !write(*,*) " L3952 COMPUTE MOMENTS: xPos=", xPos, " yPos=", yPos, &
       !     " xPosErr=", xPosErr, " yPosErr=", yPosErr, " aValue=", aValue, &
       !     " bValue=", bValue, " theta=", theta
    !endif
    !xFirstOut=xPos
    !yFirstOut=yPos

    xFirstOut=xFirst3
    yFirstOut=yFirst3
!----------------------------------------
    iMin=xFirstIn-spectrumHalfWidth/2.0
    iMax=xFirstIn+spectrumHalfWidth/2.0
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yFirstIn-spectrumHalfWidth/2.0
    jMax=yFirstIn+spectrumHalfWidth/2.0
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L3110 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    averArea=0.0
    numArea=0.0
    do jcross=iMin,iMax ! in horizontal direction
       do j=jMin,jMax ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averArea=averArea+oswImage(jcross,j)
             numArea=numArea+1.0
          endif
       enddo
    enddo
    if (numArea .gt. 1.0) averCountIn=averArea/numArea
    !write(*,*) "L3983  averCountIn=", averCountIn, " (in)"
!-----------------------------------------
    iMin=xFirstOut-spectrumHalfWidth/2.0
    iMax=xFirstOut+spectrumHalfWidth/2.0
    if (iMin .lt. 0) iMin=0
    if (iMax .gt. nx-1) iMax=nx-1
    jMin=yFirstOut-spectrumHalfWidth/2.0
    jMax=yFirstOut+spectrumHalfWidth/2.0
    if (jMin .lt. 0) jMin=0
    if (jMax .gt. ny-1) jMax=ny-1
    !write(*,*) " L3993 iMin=", iMin, " iMax=", iMax, " jMin=", jMin, " jMax=", jMax

    averArea=0.0
    numArea=0.0
    do jcross=iMin,iMax ! in horizontal direction
       do j=jMin,jMax ! in vertical direction
          if (oswImage(jcross,j) .gt. 0.0) then ! -0.9) then
             averArea=averArea+oswImage(jcross,j)
             numArea=numArea+1.0
          endif
       enddo
    enddo
    if (numArea .gt. 1.0) averCountOut=averArea/numArea
    !write(*,*) " L4006 averCountOut=", averCountOut, " (out)"

  end subroutine CheckFirstOrder

!**********************************************************************


!---
!---test of Gaussian function 28.01.03
real(double) function gFunc(param, numparam, xinput, yinput, numpoints)
   integer, intent(in)      :: numparam, numpoints
   real(kind=double), intent(in)   :: param(numparam)
   real(kind=double), intent(in)   :: xinput(numpoints), yinput(numpoints)
   real(kind=double)    :: residual
   integer    :: i
! The function is: y=param(1)+param(2)*exp(-((x-param(3))/param(4))**2)
! The result is the sum of residuals squared
   gFunc=0.0
   do i=1,numpoints
      if (param(4) .gt. 0.0) then
         residual=param(1)+param(2)*exp(-((xinput(i)-param(3))/param(4))**2)
      else
         residual=param(1)+param(2)*exp(-((xinput(i)-param(3)))**2)
      endif
      residual=(yinput(i)-residual)
      gFunc=gFunc+residual*residual
   enddo
end function gFunc

! Test 2 of the Gaussian function
subroutine fgauss(x, p, y, dydp, np)
! The function is: y=p(1)+p(2)*exp(-((x-p(3))/p(4))**2)
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac

  y=0.0
  dydp=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  y=p(1)+p(2)*ex
  dydp(1)=1.0
  dydp(2)=ex
  dydp(3)=fac/p(4)
  dydp(4)=fac*arg/p(4)
end subroutine fgauss

! Superposition of the Gaussian and a linear function
subroutine fgaussl(x, p, y, dydp, np)
! The function is: y=p(1) + p(5)*(x-p(3)) + p(2)*exp(-((x-p(3))/p(4))**2)
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac

  y=0.0
  dydp=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  y=p(1)+p(2)*ex +p(5)*x
  dydp(1)=1.0
  dydp(2)=ex

  dydp(3)=fac/p(4)
  dydp(4)=fac*arg/p(4)
  dydp(5)=x-p(3)

end subroutine fgaussl


! Superposition of the Gaussian and a linear function
real function fgausse(x, p, np)
! The function is: y=p(1) + p(5)*(x-p(3)) + p(2)*exp(-((x-p(3))/p(4))**2)
  integer :: np ! number of parameters
  real :: x ! x-point
  real :: p(np) ! parameters
!  real :: y ! value of the function
!  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac

  fgausse=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  fgausse=p(1)+p(2)*ex +p(5)*x
  !dydp(1)=1.0
  !dydp(2)=ex

  !dydp(3)=fac/p(4)
  !dydp(4)=fac*arg/p(4)
  !dydp(5)=x-p(3)

end function fgausse

! Superposition of the Gaussian and a linear function
subroutine d_gausse(x, p, da, np, dydp)
! The function is: y=p(1) + p(5)*(x-p(3)) + p(2)*exp(-((x-p(3))/p(4))**2)
  integer :: np ! number of parameters
  real :: x ! x-point
  real :: p(np) ! parameters
  real :: da, da2 ! ?
  real :: dydp(np) ! derivatives of the function
  real :: arg, ex, fac
  da2=da

  !y=0.0
  dydp=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  !y=p(1)+p(2)*ex +p(5)*x
  dydp(1)=1.0
  dydp(2)=ex

  dydp(3)=fac/p(4)
  dydp(4)=fac*arg/p(4)
  dydp(5)=x-p(3)

end subroutine d_gausse


! Superposition of the Gaussian and a linear function
! for f90
subroutine fgaussl9(x, p, y, dydp)
! The function is: y=p(1) + p(5)*x + p(2)*exp(-((x-p(3))/p(4))**2)
  implicit none
  real(kind=double), dimension(:), intent(in) :: x
  real(kind=double), dimension(:), intent(inout) :: p 
  real(kind=double), dimension(:), intent(out) :: y 
  ! value of the function
  real(kind=double), dimension(:,:), intent(out) :: dydp 
  ! derivatives of the function
!  integer(kind=int32) :: i
  integer(kind=int32) :: np,nx
  !real(kind=double) :: arg, ex, fac
  real(kind=double), dimension(size(x)) :: arg,ex,fac

  !write(*,*) " FGAUSSL9 L3065 size(x)=", size(x), " size(p)=", size(p) 
  nx=assert_eq3(size(x), size(y), size(dydp,1), 'fgaussl9: nx')
  np=assert_eq2(size(p),size(dydp,2), 'fgaussl9: np')
  arg=0.0
  ex=0.0
  fac=0.0

  y(:)=0.0
  dydp(:,:)=0.0
  !write(*,*) " L3076 dydp=", dydp(1,:)
  !do i=1,5
  !   write(*,*) " i=", i, " dydp(i)=", dydp(1,i)
  !enddo
  if (p(4) .eq. 0.0) p(4)=1.0
  arg(:)=(x(:)-p(3))/p(4)
  !write(*,*) " L3080"
  ex(:)=exp(-arg(:)**2)
  !write(*,*) " L3082"
  fac(:)=p(2)*ex(:)*2.0_double*arg(:)
  !write(*,*) " L3084"
  y(:)=p(1)+p(2)*ex(:) +p(5)*x(:)
  !write(*,*) " L3086"
  dydp(:,1)=1.0
  dydp(:,2)=ex(:)
  dydp(:,3)=fac(:)/p(4)
  !write(*,*) " L3090"
  dydp(:,4)=fac(:)*arg/p(4)
  !write(*,*) " L3092"
  dydp(:,5)=x(:)-p(3)
  !write(*,*) " L3094"
end subroutine fgaussl9


!*********************************************************************
! This subroutine computes the two-dimentional Gaussian function
! on a background:
! f=p(1)+p(2)*exp-(((x-p(3))/p(4))**2+((y-p(5))/p(6))**2)
!********************************************************************
real(double) function fgaussxy(x, y, p, np)
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in) :: y ! y-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double) :: arg1, arg2, ex
  fgaussxy=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  if (p(5) .eq. 0.0) p(5)=1.0
  arg1=(x-p(3))/p(4)
  arg2=(y-p(5))/p(6)
  ex=exp(-(arg1*arg1+arg2*arg2))
  fgaussxy=p(1)+p(2)*ex
end function fgaussxy

! Superposition  of the Gaussian function
subroutine fgaussup(x, p, y, dydp, np)
!
! The function computes the superposition of various gaussians:
! y(x;a)=\sum_{k=1}^K B_k exp[-((x-E_k)/G_k)**2],
! which is a sum of K Gaussians (np/3), each having a variable position,
! amplitude and width. The parameters are stored in the 
! order B_1,E_1,G_1, B_2,E_2,G_2, ... , B_K,E_K,G_K:
! p(i)=B_k, p(i+1)=E_k, p(i+2)=G_k, k=1,...,np/3
!
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac
  integer(kind=int32) :: i

  y=0.0
  dydp=0.0
  do i=1,np-1,3
     if (p(i+2) .eq. 0.0) p(i+2)=1.0
     arg=(x-p(i+1))/p(i+2)
     ex=exp(-arg*arg)
     fac=p(i)*ex*2.0*arg
     y=y+p(i)*ex
     dydp(i)=ex
     dydp(i+1)=fac/p(i+2)
     dydp(i+2)=fac*arg/p(i+2)
  enddo
  return
end subroutine fgaussup

!---
subroutine gaussiana(xinput,yinput, nPts, param)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a four-parameter gaussian curve y=p1+p2*exp(-(x-p3**2)/p4**2).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(4)
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(4) :: fitList
  real(kind=double), dimension(4,4) :: covMatrix
  real(kind=double), dimension(4,4) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(4) :: atry, beta, da
  real(kind=double), dimension(4,4) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1
  ! 08.01.2004
  real(kind=double)  :: xtest , ytest ! this must be uncomm. for opt. extr. 
  real(kind=double), dimension(4) ::dydptest ! this must be uncomm. for opt. extr. 
  integer :: verbFlag
  verbFlag=0
  numparam=4
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  do i1=1,nPts
     meany=meany+yinput(i1)
  enddo
  if (nPts .gt. 0) then
     meany=meany/real(nPts)
  else
     write(*,*) "Error (no input points for Gaussian approximation"
  endif
!  write(*,*) " GAUSSIANA L927: meany=", meany

  !Initial values for the input parameters
!  param(1)=0.0 ! 2.0
!  param(2)=meany ! 6.0
!  param(3)=nPts/2.0 ! 10.0
!  param(4)=param(3)/3.0 ! 5.0
  param(1)=meany ! 10.0 ! meany/10.0
  param(2)=10.0 !meany !*1.2
  param(3)=nPts/2.0 ! 10.0
  param(4)=1.0 ! param(3)/3.0 ! 5.0
  !write(*,*) "******************000000000000000000"
  !write(*,*) " L943 param(1)=", param(1), " param(2)=", param(2), &
  !     " param(3)=", param(3), " param(4)=", param(4)
  siginput=6.0
  aLambda=-10

  
  call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
       covMatrix, alphaMatrix, chiSqr, fgauss, aLambda, atry, beta, da, &
       oneda, ochisqr, verbFlag)
  chi0=chiSqr
  !write(*,*) "  First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, " aLambda1=", aLambda 
  !write(*,*) "  First   p1=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgauss(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "*********************1111111111111111"

  !write(*,*) " Consecutive steps :"
  istep=0
  fchi: do 
     istep=istep+1
     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, chiSqr, fgauss, aLambda, atry, beta, da, &
          oneda, ochiSqr, verbFlag)
     if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
     chi0=chiSqr
  enddo fchi
  !write (*,*) " Intermediate chiSqr=", chiSqr, " istep=", istep
  !write (*,*) " Intermediate parameters=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgauss(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************2222222222222222222"
  finchi0=chiSqr
  !write(*,*) " steps :"
  jstep=0
  fchi0 : do
     jstep=jstep+1
     istep=0
     aLambda=-10
     fchi1: do 
        istep=istep+1
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, &
             numparam, covMatrix, alphaMatrix, chiSqr, fgauss, aLambda, &
             atry, beta, da, oneda, ochiSqr, verbFlag)
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
        chi0=chiSqr
     enddo fchi1
     if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
     finchi0=chiSqr
  enddo fchi0
  !write (*,*) " Final chiSqr=", chiSqr, " jstep=", jstep
  !write (*,*) " Final parameters2=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgauss(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************3333333333333333333"
end subroutine gaussiana

!---

! Approximation of the input data with a Gaussian plus linear functions
!---
subroutine gaussianal(xinput,yinput, nPts, param, verbFlag, nUsedPoints)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a five-parameter gaussian curve superimposed with a linear function 
  ! y=p1+p2*exp(-(x-p3**2)/p4**2) + p(5)*(x-p(3)).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(5)
  integer :: verbFlag
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(5) :: fitList
  real(kind=double), dimension(5,5) :: covMatrix
  real(kind=double), dimension(5,5) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(5) :: atry, beta, da
  real(kind=double), dimension(5,5) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1
  real(kind=double)  :: xtest , ytest  
  real(kind=double), dimension(5) ::dydptest 
  real(kind=double) :: maxy, miny, nmeany
  integer(kind=int32) :: imaxy 
  integer :: nUsedPoints
  real(kind=double) :: distributionWidth
  
  numparam=5
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  nmeany=0.0
  maxy=0.0
  miny=99999.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .lt. miny) miny=yinput(i1)
  enddo
  yinput=yinput-miny

  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     if (yinput(i1) .gt. 0.0) then
        meany=meany+yinput(i1)
        nmeany=nmeany+1.0
     endif
  enddo
  if (nmeany .gt. 0) then
     meany=meany/nmeany
  else
     meany=0.0
     param(1)=0.0
     param(2)=0.0
     param(3)=0.0
     param(4)=1.0
     param(5)=0.0
     return
     !write(*,*) "Error (no input points for Gaussian approximation"
  endif
  !if (verbFlag .eq. 1) then
  !   write(*,*) " GAUSSIANAL L4356: meany=", meany, " miny=", miny, " maxy=", maxy, &
  !        " imaxy=", imaxy, " nmeany=", nmeany
  !endif

  ! VNY 15.07.2004 Find the width of the distribution
  distributionWidth=0.0
  if (meany .gt. 0.0) then
     do i1=1,nPts
        if (yinput(i1) .gt. meany) then
           distributionWidth=distributionWidth+1.0
        endif
     enddo
  endif
  if (verbFlag .eq. 1) then
     write(*,*) " L4370 distributionWidth=", distributionWidth, " <----------"
  endif

  !Initial values for the input parameters
  !param(1)=meany
  !param(1)=0.8*miny+0.2*meany
  param(1)=0.0
  !param(2)=maxy-meany ! (meany+maxy)/2.0
  param(2)=maxy-param(1)
  param(3)=xinput(imaxy)
  if (distributionWidth .gt. 0.0) then
     param(4)=distributionWidth/2.8
  else
     param(4)=2.0
  endif
!  param(4)=1.0
  param(5)=0.0
!

  if (verbFlag .eq. 1) then
     write(*,*) "******************STEP:  000000000000000000"
     write(*,*) " L4391 param(1)=", param(1), " param(2)=", param(2), &
          " param(3)=", param(3), " param(4)=", param(4), " param(5)=", param(5)
  endif
!  siginput=6.0
!  siginput=sqrt(yinput)
  siginput=20.0

  aLambda=-10

  nUsedPoints=nmeany
  if (nmeany .gt. 9) then  
     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
          da, oneda, ochisqr, verbFlag)

     chi0=chiSqr

       if (verbFlag .eq. 1) then
          write(*,*) "  L4405 First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, &
               " aLambda1=", aLambda 
          !write(*,*) " L4407 aLambda=", aLambda
          write(*,*) " L4407  First   p1=", param
       endif
     do i1=1, nPts
        xtest=xinput(i1)
        call fgaussl(xtest,param,ytest,dydptest,numparam)
        if (verbFlag .eq. 1) then
           write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), &
                " yt=", ytest, " dy3=", dydptest(3) 
        endif
     enddo
       if (verbFlag .eq. 1) then
          write(*,*) "*********************STEP: 1111111111111111"
          write(*,*) " Consecutive steps :"
       endif

     istep=0
     fchi: do 
        istep=istep+1
             if (verbFlag .eq. 1) then
                write(*,*) " L4431 istep=", istep, " nPts=", nPts
                do i1=1, nPts
                   xtest=xinput(i1)
                   call fgaussl(xtest,param,ytest,dydptest,numparam)
                   write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, &
                        " x(i)=", xtest, " y(i)=",yinput(i1), " yt=", ytest, &
                        " dy4=", dydptest(4)
                enddo
                write(*,*) " numparam=", numparam
                write(*,*) " param=", param
             endif
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
             covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
             da, oneda, ochiSqr, verbFlag)
             if (verbFlag .eq. 1) then
                write(*,*) " L4440 chiSqr=", chiSqr, " chi0=", chi0, &
                     " diff=", abs(chiSqr-chi0)
             endif
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
        chi0=chiSqr
     enddo fchi
       if (verbFlag .eq. 1) then
          write (*,*) " L4447 Intermediate chiSqr=", chiSqr, " istep=", istep
          write (*,*) " Intermediate parameters=", param
       endif
     do i1=1, nPts
        xtest=xinput(i1)
        call fgaussl(xtest,param,ytest,dydptest,numparam)
             if (verbFlag .eq. 1) then
                write(*,'(a, i4, 3(a, f8.2))') "     i=", i1, &
                     " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
             endif
     enddo
       if (verbFlag .eq. 1) then
          write(*,*) "**************2222222222222222222"
       endif
     finchi0=chiSqr
     !write(*,*) " steps :"
     jstep=0
     ! VNY 28.01.2004
     aLambda=-10
     fchi0 : do
        jstep=jstep+1
        istep=0
        ! VNY 28.01.2004
        !     aLambda=-10
        fchi1: do 
           istep=istep+1
           call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, &
                numparam, covMatrix, alphaMatrix, chiSqr, fgaussl, &
                aLambda, atry, beta, da, oneda, ochiSqr, verbFlag)
           !write(*,*) " L4482 chiSqr=", chiSqr, " chi0=", chi0, " ochiSqr=", ochiSqr
           if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
           chi0=chiSqr
        enddo fchi1
        if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
        finchi0=chiSqr
     enddo fchi0
       if (verbFlag .eq. 1) then
          write (*,*) " L4490 Final chiSqr=", chiSqr, " jstep=", jstep
          write (*,*) " L4491 Final parameters2=", param     
          do i1=1, nPts
             xtest=xinput(i1)
             call fgaussl(xtest,param,ytest,dydptest,numparam)
             if (verbFlag .eq. 1) then
                write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, " x(i)=", xtest, &
                     " y(i)=",yinput(i1), &
                     " yt=", ytest, " dy4=", dydptest(4)
             endif
          enddo      
          write(*,*) "**************3333333333333333333"
       endif
  endif ! nmeany .gt. 9 

end subroutine gaussianal


! Approximation of the input data with a Gaussian plus linear functions
!---
subroutine gaussianalSim(xinput,yinput, nPts, param, verbFlag, nUsedPoints)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a five-parameter gaussian curve superimposed with a linear function 
  ! y=p1+p2*exp(-(x-p3**2)/p4**2) + p(5)*(x-p(3)).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam, nca
  real(kind=double) :: param(5)
  integer :: verbFlag
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(5) :: fitList
  real(kind=double), dimension(5,5) :: covMatrix
  real(kind=double), dimension(5,5) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(5) :: atry, beta, da
  real(kind=double), dimension(5,5) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1
  real(kind=double)  :: xtest , ytest  
  real(kind=double), dimension(5) ::dydptest 
  real(kind=double) :: maxy, miny, nmeany
  integer(kind=int32) :: imaxy 
  integer :: nUsedPoints
  real(kind=double) :: distributionWidth
  
  numparam=5
  nca=5
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  nmeany=0.0
  maxy=0.0
  miny=99999.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .lt. miny) miny=yinput(i1)
  enddo
  yinput=yinput-miny

  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     if (yinput(i1) .gt. 0.0) then
        meany=meany+yinput(i1)
        nmeany=nmeany+1.0
     endif
  enddo
  if (nmeany .gt. 0) then
     meany=meany/nmeany
  else
     meany=0.0
     param(1)=0.0
     param(2)=0.0
     param(3)=0.0
     param(4)=1.0
     param(5)=0.0
     return
     !write(*,*) "Error (no input points for Gaussian approximation"
  endif
  if (verbFlag .eq. 1) then
     write(*,*) " GAUSSIANALsim L4735: meany=", meany, " miny=", miny, " maxy=", maxy, &
          " imaxy=", imaxy, " nmeany=", nmeany
  endif

  ! VNY 15.07.2004 Find the width of the distribution
  distributionWidth=0.0
  if (meany .gt. 0.0) then
     do i1=1,nPts
        if (yinput(i1) .gt. meany) then
           distributionWidth=distributionWidth+1.0
        endif
     enddo
  endif
  if (verbFlag .eq. 1) then
     write(*,*) " L4749 distributionWidth=", distributionWidth, " <----------"
  endif

  !Initial values for the input parameters
  !param(1)=meany
  !param(1)=0.8*miny+0.2*meany
  param(1)=0.0
  !param(2)=maxy-meany ! (meany+maxy)/2.0
  param(2)=maxy-param(1)
  param(3)=xinput(imaxy)
  if (distributionWidth .gt. 0.0) then
     param(4)=distributionWidth/2.8
  else
     param(4)=2.0
  endif
!  param(4)=1.0
  param(5)=0.0
!

  if (verbFlag .eq. 1) then
     write(*,*) "******************STEP:  000000000000000000"
     write(*,*) " L4770 param(1)=", param(1), " param(2)=", param(2), &
          " param(3)=", param(3), " param(4)=", param(4), " param(5)=", param(5)
  endif
!  siginput=6.0
!  siginput=sqrt(yinput)
  siginput=20.0

  aLambda=-10

  nUsedPoints=nmeany
  if (nmeany .gt. 9) then  
!     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
!          covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
!          da, oneda, ochisqr, verbFlag)
     call mrqminSim(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, nca, chiSqr, aLambda)


     chi0=chiSqr

       if (verbFlag .eq. 1) then
          write(*,*) "  L4791 First chiSqr1=", chiSqr, &
               " aLambda1=", aLambda 
          !write(*,*) " L4407 aLambda=", aLambda
          write(*,*) " L4794  First   p1=", param
       endif
     do i1=1, nPts
        xtest=xinput(i1)
        call fgausslSim(xtest,numparam,param,ytest,dydptest)
        if (verbFlag .eq. 1) then
           write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), &
                " yt=", ytest, " dy3=", dydptest(3) 
        endif
     enddo
       if (verbFlag .eq. 1) then
          write(*,*) "*********************STEP: 1111111111111111"
          write(*,*) " Consecutive steps :"
       endif

     istep=0
     fchi: do 
        istep=istep+1
             if (verbFlag .eq. 1) then
                write(*,*) " L4813 istep=", istep, " nPts=", nPts
                do i1=1, nPts
                   xtest=xinput(i1)
                   call fgausslSim(xtest,numparam,param,ytest,dydptest)
                   write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, &
                        " x(i)=", xtest, " y(i)=",yinput(i1), " yt=", ytest, &
                        " dy4=", dydptest(4)
                enddo
                write(*,*) " numparam=", numparam
                write(*,*) " param=", param
             endif
!        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
!             covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
!             da, oneda, ochiSqr, verbFlag)
        call mrqminSim(xinput, yinput, siginput, nPts, param, fitList, numparam, &
             covMatrix, alphaMatrix, nca, chiSqr, aLambda)

             if (verbFlag .eq. 1) then
                write(*,*) " L4831 chiSqr=", chiSqr, " chi0=", chi0, &
                     " diff=", abs(chiSqr-chi0)
             endif
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
        chi0=chiSqr
     enddo fchi
       if (verbFlag .eq. 1) then
          write (*,*) " L4838 Intermediate chiSqr=", chiSqr, " istep=", istep
          write (*,*) " Intermediate parameters=", param
       endif
     do i1=1, nPts
        xtest=xinput(i1)
        call fgausslSim(xtest,numparam,param,ytest,dydptest)
             if (verbFlag .eq. 1) then
                write(*,'(a, i4, 3(a, f8.2))') "     i=", i1, &
                     " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
             endif
     enddo
       if (verbFlag .eq. 1) then
          write(*,*) "**************2222222222222222222"
       endif
     finchi0=chiSqr
     !write(*,*) " steps :"
     jstep=0
     ! VNY 28.01.2004
     aLambda=-10
     fchi0 : do
        jstep=jstep+1
        istep=0
        ! VNY 28.01.2004
        !     aLambda=-10
        fchi1: do 
           istep=istep+1
!           call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, &
!                numparam, covMatrix, alphaMatrix, chiSqr, fgaussl, &
!                aLambda, atry, beta, da, oneda, ochiSqr, verbFlag)
           call mrqminSim(xinput, yinput, siginput, nPts, param, fitList, numparam, &
                covMatrix, alphaMatrix, nca, chiSqr, aLambda)
           !write(*,*) " L4482 chiSqr=", chiSqr, " chi0=", chi0, " ochiSqr=", ochiSqr
           if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
           chi0=chiSqr
        enddo fchi1
        if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
        finchi0=chiSqr
     enddo fchi0
       if (verbFlag .eq. 1) then
          write (*,*) " L4877 Final chiSqr=", chiSqr, " jstep=", jstep
          write (*,*) " L4878 Final parameters2=", param     
          do i1=1, nPts
             xtest=xinput(i1)
             call fgausslSim(xtest,numparam,param,ytest,dydptest)
             if (verbFlag .eq. 1) then
                write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, " x(i)=", xtest, &
                     " y(i)=",yinput(i1), &
                     " yt=", ytest, " dy4=", dydptest(4)
             endif
          enddo      
          write(*,*) "**************3333333333333333333"
       endif
  endif ! nmeany .gt. 9 

end subroutine gaussianalSim




! Approximation of the input data with a Gaussian plus linear functions
! using the opt_extr.f90 module
!---
subroutine gaussianae(xinput,yinput, nPts, param, verbFlag, nUsedPoints)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a five-parameter gaussian curve superimposed with a linear function 
  ! y=p1+p2*exp(-(x-p3)**2/p4**2) + p(5)*(x-p(3)).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(5)
  integer :: verbFlag
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(5) :: fitList
!  real(kind=double), dimension(5,5) :: covMatrix
!  real(kind=double), dimension(5,5) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(5) :: atry, beta, da
  real(kind=double), dimension(5,5) :: oneda
  real(kind=double) :: ochiSqr, chi0!, finchi0
!  integer(kind=int32) :: istep
!  integer(kind=int32) :: jstep
  integer(kind=int32) :: i1
  real(kind=double)  :: xtest , ytest  
  real(kind=double), dimension(5) ::dydptest 
  real(kind=double) :: maxy, miny, nmeany
  integer(kind=int32) :: imaxy 
  integer :: nUsedPoints
  real(kind=double) :: distributionWidth
  !
  integer :: n_par, mode_w, n_term, icurvf, nPts2
  real, dimension(1:nPts) :: x1, y1, w1
  real, dimension(1:5) :: a_par, da2
  real, dimension(1:5,1:5) :: covar
  real :: b_chi

!--- NB ! chisqr never set!
  chisqr=0.
!--- VNY 09.06.2005

  numparam=5
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  nmeany=0.0
  maxy=0.0
  miny=99999.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .lt. miny) miny=yinput(i1)
  enddo
  yinput=yinput-miny

  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     if (yinput(i1) .gt. 0.0) then
        meany=meany+yinput(i1)
        nmeany=nmeany+1.0
     endif
  enddo
  if (nmeany .gt. 0) then
     meany=meany/nmeany
  else
     meany=0.0
     param(1)=0.0
     param(2)=0.0
     param(3)=0.0
     param(4)=1.0
     param(5)=0.0
     return
     !write(*,*) "Error (no input points for Gaussian approximation"
  endif
  if (verbFlag .eq. 1) then
     write(*,*) " GAUSSIANAL L4356: meany=", meany, " miny=", miny, " maxy=", maxy, &
          " imaxy=", imaxy, " nmeany=", nmeany
  endif

  ! VNY 15.07.2004 Find the width of the distribution
  distributionWidth=0.0
  if (meany .gt. 0.0) then
     do i1=1,nPts
        if (yinput(i1) .gt. meany) then
           distributionWidth=distributionWidth+1.0
        endif
     enddo
  endif
  if (verbFlag .eq. 1) then
     write(*,*) " L4370 distributionWidth=", distributionWidth, " <----------"
  endif

  !Initial values for the input parameters
  !param(1)=meany
  !param(1)=0.8*miny+0.2*meany
  param(1)=0.0
  !param(2)=maxy-meany ! (meany+maxy)/2.0
  param(2)=maxy-param(1)
  param(3)=xinput(imaxy)
  if (distributionWidth .gt. 0.0) then
     param(4)=distributionWidth/2.8
  else
     param(4)=2.0
  endif
!  param(4)=1.0
  param(5)=0.0
!

  if (verbFlag .eq. 1) then
     write(*,*) "******************STEP:  000000000000000000"
     write(*,*) " L4391 param(1)=", param(1), " param(2)=", param(2), &
          " param(3)=", param(3), " param(4)=", param(4), " param(5)=", param(5)
  endif
!  siginput=6.0
!  siginput=sqrt(yinput)
  siginput=20.0

  aLambda=-10

  nUsedPoints=nmeany
  if (nmeany .gt. 9) then  
!     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
!          covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
!          da, oneda, ochisqr, verbFlag)

     mode_w=-1
     icurvf=1001
     x1=xinput
     y1=yinput
     nPts2=nPts
     a_par=param
     n_par=5
     n_term=5
  
     call curfit(x1, y1, w1, nPts2, mode_w, a_par, da2, n_par, &
          n_term, covar, 0.01, b_chi, -1.0, 20, icurvf, fgausse, d_gausse)

     chi0=chiSqr

       if (verbFlag .eq. 1) then
          write(*,*) "  L4711 First chiSqr1=", b_chi, " old chiSqr=0.01"
          write(*,*) " L4712  First   a_par=", a_par
       endif

       param=a_par

     do i1=1, nPts
        xtest=xinput(i1)
        call fgaussl(xtest,param,ytest,dydptest,numparam)
        if (verbFlag .eq. 1) then
           write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), &
                " yt=", ytest, " dy3=", dydptest(3) 
        endif
     enddo
!       if (verbFlag .eq. 1) then
!          write(*,*) "*********************STEP: 1111111111111111"
!          write(*,*) " Consecutive steps :"
!       endif
!
!     istep=0
!     fchi: do 
!        istep=istep+1
!             if (verbFlag .eq. 1) then
!                write(*,*) " L4431 istep=", istep, " nPts=", nPts
!                do i1=1, nPts
!                   xtest=xinput(i1)
!                   call fgaussl(xtest,param,ytest,dydptest,numparam)
!                   write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, &
!                        " x(i)=", xtest, " y(i)=",yinput(i1), " yt=", ytest, &
!                        " dy4=", dydptest(4)
!                enddo
!                write(*,*) " numparam=", numparam
!                write(*,*) " param=", param
!             endif
!        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
!             covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
!             da, oneda, ochiSqr, verbFlag)
!             if (verbFlag .eq. 1) then
!                write(*,*) " L4440 chiSqr=", chiSqr, " chi0=", chi0, &
!                     " diff=", abs(chiSqr-chi0)
!             endif
!        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
!        chi0=chiSqr
!     enddo fchi
!       if (verbFlag .eq. 1) then
!          write (*,*) " L4447 Intermediate chiSqr=", chiSqr, " istep=", istep
!          write (*,*) " Intermediate parameters=", param
!       endif
!     do i1=1, nPts
!        xtest=xinput(i1)
!        call fgaussl(xtest,param,ytest,dydptest,numparam)
!             if (verbFlag .eq. 1) then
!                write(*,'(a, i4, 3(a, f8.2))') "     i=", i1, &
!                     " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
!             endif
!     enddo
!       if (verbFlag .eq. 1) then
!          write(*,*) "**************2222222222222222222"
!       endif
!     finchi0=chiSqr
!     !write(*,*) " steps :"
!     jstep=0
!     ! VNY 28.01.2004
!     aLambda=-10
!     fchi0 : do
!        jstep=jstep+1
!        istep=0
!        ! VNY 28.01.2004
!        !     aLambda=-10
!        fchi1: do 
!           istep=istep+1
!           call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, &
!                numparam, covMatrix, alphaMatrix, chiSqr, fgaussl, &
!                aLambda, atry, beta, da, oneda, ochiSqr, verbFlag)
!           !write(*,*) " L4482 chiSqr=", chiSqr, " chi0=", chi0, " ochiSqr=", ochiSqr
!           if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
!           chi0=chiSqr
!        enddo fchi1
!        if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
!        finchi0=chiSqr
!     enddo fchi0
!       if (verbFlag .eq. 1) then
!          write (*,*) " L4490 Final chiSqr=", chiSqr, " jstep=", jstep
!          write (*,*) " L4491 Final parameters2=", param     
!          do i1=1, nPts
!             xtest=xinput(i1)
!             call fgaussl(xtest,param,ytest,dydptest,numparam)
!             if (verbFlag .eq. 1) then
!                write(*,'(a, i4, 4(a, f8.2))') "     i=", i1, " x(i)=", xtest, &
!                     " y(i)=",yinput(i1), &
!                     " yt=", ytest, " dy4=", dydptest(4)
!             endif
!          enddo      
!          write(*,*) "**************3333333333333333333"
!       endif
  endif ! nmeany .gt. 9 

end subroutine gaussianae

!---
! Approximation of the input data with a Gaussian plus linear functions
!--- VNY 19.07.2004
subroutine gaussianal9(xinput,yinput, nPts, param, verbFlag, nUsedPoints)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a five-parameter gaussian curve superimposed with a linear function 
  ! y=p1+p5*x+p2*exp(-(x-p3**2)/p4**2).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(5)
  integer :: verbFlag
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(5) :: fitList
  logical, dimension(5) :: maska
  real(kind=double), dimension(5,5) :: covMatrix
  real(kind=double), dimension(5,5) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(5) :: atry, beta, da
  real(kind=double), dimension(5,5) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1
  real(kind=double)  :: xtest , ytest  
  real(kind=double), dimension(5) ::dydptest 
  real(kind=double) :: maxy, miny, nmeany
  integer(kind=int32) :: imaxy 
  integer :: nUsedPoints
  real(kind=double) :: distributionWidth

  
  numparam=5
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1
  maska=.true.

  ! averaging y
  meany=0.0
  nmeany=0.0
  maxy=0.0
  miny=99999.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     if (yinput(i1) .gt. 0.0) then
        if (yinput(i1) .lt. miny) miny=yinput(i1)
        meany=meany+yinput(i1)
        nmeany=nmeany+1.0
     endif
  enddo
  if (nmeany .gt. 0) then
     meany=meany/nmeany
  else
     meany=0.0
     param(1)=0.0
     param(2)=0.0
     param(3)=0.0
     param(4)=1.0
     param(5)=0.0
     return
     !write(*,*) "Error (no input points for Gaussian approximation"
  endif
  if (verbFlag .eq. 1) then
     write(*,*)  " GAUSSIANAL9 L3537: meany=", meany, " miny=", miny, " maxy=", maxy, &
          " imaxy=", imaxy, " nmeany=", nmeany
  endif

  ! VNY 15.07.2004 Find the width of the distribution
  distributionWidth=0.0
  if (meany .gt. 0.0) then
     do i1=1,nPts
        if (yinput(i1) .gt. meany) then
           distributionWidth=distributionWidth+1.0
        endif
     enddo
  endif
  !if (verbFlag .eq. 1) then
     !write(*,*) " L3307 distributionWidth=", distributionWidth, " <----------"
  !endif

  !Initial values for the input parameters
  !param(1)=meany
  !param(1)=0.8*miny+0.2*meany
  param(1)=0.0
  !param(2)=maxy-meany ! (meany+maxy)/2.0
  param(2)=maxy-param(1)
  param(3)=xinput(imaxy)
  if (distributionWidth .gt. 0.0) then
     param(4)=distributionWidth/2.8
  else
     param(4)=2.0
  endif
!  param(4)=1.0
  param(5)=0.0
!

  if (verbFlag .eq. 1) then
     write(*,*) "GL9+++++++++++++++++++++++++++++++++++++++++++++++++++++"
     write(*,*) "GL9******************STEP:  000000000000000000"
     write(*,'(5(a,f6.2))') " ****L1226 param(1)=", param(1), " param(2)=", param(2), &
       " param(3)=", param(3), " param(4)=", param(4), " param(5)=", param(5)
  endif
  siginput=6.0
  aLambda=-10

  nUsedPoints=nmeany
  if (nmeany .gt. 9) then  
     call mrqmin9(xinput, yinput, siginput, param, maska, &
          covMatrix, alphaMatrix, chiSqr, fgaussl9, aLambda, verbFlag)
     chi0=chiSqr

       if (verbFlag .eq. 1) then
          write(*,'(3(a,f6.2))') " **** First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, &
               " aLambda1=", aLambda 
          write(*,*) " ***** First   p1=", param
       endif
     do i1=1, nPts
        xtest=xinput(i1)
        call fgaussl(xtest,param,ytest,dydptest,numparam)
        if (verbFlag .eq. 1) then
           write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), &
                " yt=", ytest, " dy4=", dydptest(4) 
        endif
     enddo
       if (verbFlag .eq. 1) then
          write(*,*) "*********************STEP: 1111111111111111"
          write(*,*) " Consecutive steps :"
       endif

     istep=0
     fchi: do 
        istep=istep+1
             if (verbFlag .eq. 1) then
                write(*,*) " istep=", istep, " nPts=", nPts
                do i1=1, nPts
                   xtest=xinput(i1)
                   call fgaussl(xtest,param,ytest,dydptest,numparam)
                   write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " yt=", ytest, &
                        " dy4=", dydptest(4)
                enddo
                write(*,*) " numparam=", numparam
                write(*,*) " param=", param
             endif
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
             covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, &
             da, oneda, ochiSqr, verbFlag)
             if (verbFlag .eq. 1) then
                write(*,*) " L1962 chiSqr=", chiSqr, " chi0=", chi0, &
                     " diff=", abs(chiSqr-chi0)
             endif
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
        chi0=chiSqr
     enddo fchi
       if (verbFlag .eq. 1) then
          write (*,*) " L1969 Intermediate chiSqr=", chiSqr, " istep=", istep
          write (*,*) " Intermediate parameters=", param
       endif
     do i1=1, nPts
        xtest=xinput(i1)
        call fgaussl(xtest,param,ytest,dydptest,numparam)
             if (verbFlag .eq. 1) then
                write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
             endif
     enddo
       if (verbFlag .eq. 1) then
          write(*,*) "**************2222222222222222222"
       endif
     finchi0=chiSqr
     !write(*,*) " steps :"
     jstep=0
     ! VNY 28.01.2004
     aLambda=-10
     fchi0 : do
        jstep=jstep+1
        istep=0
        ! VNY 28.01.2004
        !     aLambda=-10
        fchi1: do 
           istep=istep+1
           call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, &
                numparam, covMatrix, alphaMatrix, chiSqr, fgaussl, &
                aLambda, atry, beta, da, oneda, ochiSqr, verbFlag)
           if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
           chi0=chiSqr
        enddo fchi1
        if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
        finchi0=chiSqr
     enddo fchi0
       if (verbFlag .eq. 1) then
          write (*,*) " Final chiSqr=", chiSqr, " jstep=", jstep
          write (*,*) " Final parameters2=", param     
          do i1=1, nPts
             xtest=xinput(i1)
             call fgaussl(xtest,param,ytest,dydptest,numparam)
             if (verbFlag .eq. 1) then
                write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), &
                     " yt=", ytest, " dy4=", dydptest(4)
             endif
          enddo      
          write(*,*) "**************3333333333333333333"
       endif
  endif ! nmeany .gt. 9 

end subroutine gaussianal9

!---

!---


subroutine gaussiana2(xinput,yinput, nPts, param, numparam)
! This subroutine approximates a series of input data 
! xinput[1:nPts], yinput[1:nPts]
! with a superposition of various gaussians:
! y(x;a)=\sum_{k=1}^K B_k exp[-((x-E_k)/G_k)**2],
! which is a sum of K Gaussians (np/3), each having a variable position,
! amplitude and width. The parameters are stored in the 
! order B_1,E_1,G_1, B_2,E_2,G_2, ... , B_K,E_K,G_K:
! p(i)=B_k, p(i+1)=E_k, p(i+2)=G_k, k=1,...,numparam/3
! numparam should be ge 3
! 
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(numparam)
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(numparam) :: fitList
  real(kind=double), dimension(numparam,numparam) :: covMatrix
  real(kind=double), dimension(numparam,numparam) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(numparam) :: atry, beta, da
  real(kind=double), dimension(numparam,numparam) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1,i
  ! 08.01.2004
  real(kind=double)  :: xtest , ytest ! this must be uncomm. for opt. extr. 
  real(kind=double), dimension(numparam) ::dydptest ! this must be uncomm. for opt. extr. 
  real(kind=double) :: signature, amplitude, amplitude0, maxy
  integer(kind=int32) :: imaxy
  integer :: verbFlag

  !write(*,*) " L1079: Starting GAUSSIANA2, numparam=", numparam
  !numparam=3
  verbFlag=0
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  maxy=0.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     meany=meany+yinput(i1)
  enddo
  if (nPts .gt. 0) then
     meany=meany/real(nPts)
  else
     write(*,*) "Error (no input points for Gaussian approximation"
  endif
  !write(*,*) " GAUSSIANA L1105: meany=", meany, " maxy=", maxy, " imaxy=", imaxy

  !Initial values for the input parameters
!  param(1)=0.0 ! 2.0
!  param(2)=meany ! 6.0
!  param(3)=nPts/2.0 ! 10.0
!  param(4)=param(3)/3.0 ! 5.0

  signature=1.0
  amplitude=1.0
  amplitude0=0.0
  do i=1,numparam-1,3
     signature=(-1.0)*signature
     if (i .gt. 3) then 
        amplitude=2.0
        amplitude0=1.0
     endif
     param(i)=maxy/(amplitude+amplitude0*5.0) 
     param(i+1)=real(imaxy)+amplitude0*signature*nPts/4.0 
!     param(i+1)=real(nPts)/2+amplitude0*signature*real(nPts)/4.0 
!     param(i+2)=nPts/2.0
     param(i+2)=2.5 ! nPts/amplitude
  enddo
  !write(*,*) "******************000000000000000000"
  !write(*,*) "             Initial parameter values:"
  do i=1, numparam-1,3
     !write(*,*) " L1103 i=", i, " param(i+1)=", param(i), &
     !  " param(i+2)=", param(i+1), " param(i+3)=", param(i+2)
  enddo
!  siginput=16.0
  siginput=sqrt(yinput)
  aLambda=-10
  
  call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
       covMatrix, alphaMatrix, chiSqr, fgaussup, aLambda, atry, beta, &
       da, oneda, ochisqr, verbFlag)
!  VNY 14.01.2004
!  chi0=chiSqr
  chi0=chiSqr*1.2
  ochiSqr=chi0
  !write(*,*) "  First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, " aLambda1=", aLambda 
  !write(*,*) "  First   p1=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgaussup(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i1=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "*********************1111111111111111"

  !write(*,*) " Consecutive steps :"
  istep=0
  fchi: do 
     istep=istep+1
     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, chiSqr, fgaussup, aLambda, atry, beta, &
          da, oneda, ochiSqr, verbFlag)
     !write(*,*) "1) istep=", istep, "    chi0=", chi0, " chiSqr=", chiSqr, " ochiSqr=", ochiSqr
     if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
     chi0=chiSqr
  enddo fchi
  !write (*,*) " Intermediate chiSqr=", chiSqr, " istep=", istep
  !write (*,*) " Intermediate parameters=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgaussup(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i1=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************2222222222222222222"
  finchi0=chiSqr
  !write(*,*) " steps :"
  jstep=0
  fchi0 : do
     jstep=jstep+1
     istep=0
     aLambda=-10
     fchi1: do 
        istep=istep+1
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, &
             numparam, covMatrix, alphaMatrix, chiSqr, fgaussup, &
             aLambda, atry, beta, da, oneda, ochiSqr, verbFlag)
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
        chi0=chiSqr
     enddo fchi1
     if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
     finchi0=chiSqr
  enddo fchi0
  !write (*,*) " Final chiSqr=", chiSqr, " jstep=", jstep
  !write (*,*) " Final parameters2=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgaussup(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************3333333333333333333"
end subroutine gaussiana2


!---exponential integral function 27.02.03
real(double) function eiFunc(x)
! using normal function: seems to be valid for -2<x<0
   real(kind=double), intent(in)   :: x
   real(kind=double)    :: smallgamma,x0, x1, xd, ei, ei0, nfact
   real(kind=double) :: sum,sig
   integer    :: i
   smallgamma=0.5772156649015328606
   x0=x
   ei=0.0
   ei0=0.0
   if (x0 .lt. 0) then
      x1=abs(x0)
      if (x1 .lt. 4.885) then
         ei0=0.5772156649015328606+log(x1)
         nfact=1.0
         xd=1.0
         do i=1,20
            nfact=nfact*real(i)
            xd=x0**real(i)
            if (xd .gt. 1.0E10) exit
            ei=ei0+xd/(real(i)*nfact)
            if (abs(ei-ei0)  .lt. 1.0E-8) exit
            ei0=ei
         enddo
      endif
      if (x1 .ge. 4.885) then
         sum=1.0
         nfact=1.0
         sig=1.0
         xd=1.0
         do i=1,5
            xd=xd*x1
            nfact=nfact*real(i)
            sig=-1.0*sig
            sum=sum+sig*nfact/xd
         enddo
         ei0=exp(-x1)/x1
         ei0=ei0*sum
         ei0=-ei0
      endif
   endif
   eiFunc=ei0
end function eiFunc
!--
!---exponential integral function 27.02.03
real(double) function eiFunc2(x)
! using formula for x>>1
   real(kind=double), intent(in)   :: x
   real(kind=double)    :: smallgamma,x0, x1, xd, ei, ei0, nfact
   real(kind=double) :: sum,sig
   integer    :: i
   smallgamma=0.5772156649015328606
   ei=0.0
   ei0=0.0
   x0=x
   if (x0 .lt. 0) then
      x1=abs(x0)
!      if (x1 .lt. 5.0) then
!         ei0=0.5772156649015328606+log(x1)
!         nfact=1.0
!         xd=1.0
!         do i=1,20
!            nfact=nfact*real(i)
!            xd=x**real(i)
!            if (xd .gt. 1.0E10) exit
!            ei=ei0+xd/(real(i)*nfact)
!            if (abs(ei-ei0)  .lt. 1.0E-8) exit
!            ei0=ei
!         enddo
!!         write (*,*) " eiFunc: i=",i, " ei0=", ei0
!      endif
!      if (x1 .ge. 5.0) then
         sum=1.0
         nfact=1.0
         sig=1.0
         xd=1.0
         do i=1,5
            !xd=x1**real(i)
            xd=xd*x1
            nfact=nfact*real(i)
            sig=-1.0*sig
            sum=sum+sig*nfact/xd
!            write (*,*) " i=", i, " xd=", xd, " nfact=", nfact, &
!                 " t=", sig*nfact/xd, " sum=", sum
         enddo
         ei0=exp(-x1)/x1
         ei0=ei0*sum
!         write (*,*) "x1=", x1, " sum=", sum, " e1func=", ei0
         ei0=-ei0
!      endif
   endif
   eiFunc2=ei0
!   eiFunc2=sum
!   write (*,*) "x1=", x1, " sum=", sum, " e1func=", ei0, " f2=", eiFunc2
end function eiFunc2


subroutine polyfit(x,y,sigmay,npts,nterms,mode,a,chisqr)

   real(kind=double) x(:)
   real(kind=double) y(:)
   real(kind=double) sigmay(:)
   integer(kind=int32) npts
   integer(kind=int32) nterms
   integer(kind=int32) mode
   real(kind=double) a(:)
   real(kind=double) chisqr

   integer(kind=int32) nmax
   integer(kind=int32) n
   integer(kind=int32) j
   integer(kind=int32) i
   integer(kind=int32) k
   integer(kind=int32) l
   real(kind=double) chisq
   real(kind=double) weight
   real(kind=double) xterm
   real(kind=double) yterm
   real(kind=double) delta
   real(kind=double) free
   real(kind=double) xi
   real(kind=double) yi
   real(kind=double), allocatable :: array(:,:) 
   real(kind=double), allocatable :: sumx(:)
   real(kind=double), allocatable :: sumy(:)

   allocate(array(nterms, nterms))
   allocate(sumx(npts))
   allocate(sumy(npts))

   nmax  =  2 * nterms - 1
   sumx  =  0.0d0
   sumy  =  0.0d0
   chisq  =  0.0d0
   do i  =  1, npts
      xi  =  x(i)
      yi  =  y(i)
      if (mode < 0) then
         if (yi < 0) then
            weight  =  -1.0d0 / yi
         else if (yi == 0) then
            weight = 1.d0
         else
            weight = 1.0d0 / yi
         end if
      else if (mode == 0) then
         weight = 1.d0
      else
         weight = 1.0d0 / sigmay(i)**2
      end if

      xterm = weight
      do n = 1, nmax
         sumx(n) = sumx(n) + xterm
         xterm = xterm * xi
      end do

      yterm = weight * yi
      do n  =  1, nterms
         sumy(n) = sumy(n) + yterm
         yterm = yterm * xi
      end do

      chisq = chisq + weight * yi**2
   end do

   do j = 1, nterms
      do k = 1, nterms
         n = j + k - 1
         array(j, k) = sumx(n)
      end do
   end do

   delta = determ(array, nterms)
   if (delta /= 0) then
      do l = 1, nterms
         do j = 1, nterms
            do k = 1, nterms
               n  = j + k - 1
               array(j, k) = sumx(n)
            end do

            array(j, l) = sumy(j)
         end do

         a(l) = determ(array, nterms) / delta
      end do

      do j = 1, nterms
         chisq = chisq - 2.0d0 * a(j) * sumy(j)
         do k = 1, nterms
            n = j + k - 1
            chisq = chisq + a(j) * a(k) * sumx(n)
         end do
      end do

      free = dble(npts - nterms)
      chisqr = chisq / free
   else
      chisqr = 0.0d0
      do j = 1, nterms
         a(j) = 0.0d0
      end do
   end if

   deallocate(array)
   deallocate(sumx)
   deallocate(sumy)

end subroutine polyfit

function determ(array, norder) result(res)

   integer(kind=int32) :: norder
   real(kind=double) array(norder, norder)

   real(kind=double) :: res

   integer(kind=int32) :: i
   integer(kind=int32) :: j
   integer(kind=int32) :: k
   integer(kind=int32) :: k1
   real(kind=double) :: save

   res = 1.d0

   do k = 1, norder
      if (array(k, k) /= 0) goto 100

      do j = k, norder
         if (array(k,j) /= 0) goto 50
      end do

      res = 0.00
      goto 200

50    continue

      do i = k, norder
         save = array(i, j)
         array(i, j) = array(i, k)
         array(i, k) = save
      end do

      res = -res

100   continue

      res = res * array(k, k)

      if (k < norder) then
         k1 = k + 1

         do i = k1, norder
            do j = k1, norder
               array(i, j) = array(i, j) - &
                             array(i, k) * &
			     array(k, j) / &
			     array(k, k)
            end do
         end do
      end if
   end do

200 continue

end function determ
  
subroutine locate10(xx, n, x, j)

   real(kind=double) :: xx(:)
   integer(kind=int32) n
   real(kind=double) :: x
   integer(kind=int32) j

   integer(kind=int32) ju
   integer(kind=int32) jl
   integer(kind=int32) jm
   
   j = 0

   if (x < xx(1)) then
      j = 1
   endif

   if (x > xx(n)) then
      j = n
   endif


   if ( j == 0) then

      if (x .eq. xx(n)) then
         ju = n-1
         jl = n-1
      endif

      if (x .eq. xx(1)) then
         jl = 1
         ju = 1 
      endif

      jl = 0
      ju = n + 1

      do while(ju - jl > 1)
         jm = (ju + jl) / 2
 
         if ((xx(n) > xx(1)) .eqv. (x > xx(jm))) then
            jl = jm
         else
            ju = jm
         end if
      end do

      j = jl
   endif

end subroutine locate10



!*************************************************************
! subroutine dlocate
! VNY 25.01.2002; If the dimension of the array xx is 1 or 0
! this subroutin can crash due to subscript. Protection 
! against this crash has been added
! Modified 12.4.2002 to use arrays starting at subscript 0
!**************************************************************
subroutine dlocate(xx, n, x, j)
   integer (kind=int32), intent(in) :: n
   real(kind=double)   :: xx(0 : n - 1)
   
   real(kind=double)   :: x
   integer(kind=int32) :: j

   integer(kind=int32) :: ju
   integer(kind=int32) :: jl
   integer(kind=int32) :: jm

   jl = 0
   ju = n + 1

! VNY 25.01.2002 Protection against one-element array
   if (n .le. 1) then
      j = 0
      return
   endif

   if (x .ge. xx(n - 1)) then
       j = n - 1
       return
   end if 
      

   if (x .le. xx(0)) then
      j = 0
      return
   endif

   do while(ju - jl > 1)
      jm = (ju + jl) / 2

      if (x >= xx(jm - 1)) then
         jl = jm
      else
         ju = jm
      end if
   end do

   j = jl - 1

end subroutine dlocate



subroutine Mrqmin(x, y, sig, nData, a, ia, ma, covMatrix, &
                    alphaMatrix, chiSqr, funcs, aLambda, atry, &
                    beta, da, oneda, ochiSqr, verbFlag)
! Leveenberg-Marquardt method, attempting to reduce the value 
! of chi**2 of a fit between a set of data points x[1..nData],
! y[1..nData] with individual standard deviations sig[1..nData],
! and a nonlinear function dependent on ma coefficients a[1..ma].
! The input array ia[1..ma] indicates by nonzero entries those components
! of a that should be fitted for, and by zero entries those compoents
! that should be held fixed at their input values. The program 
! returns current best-fit values for the parameters a[1..ma],
! and chi**2=chiSqr. The arrays covar[1..ma][1..ma], alpha[1..ma][1..ma]
! are used as working space during most iterations. Supply a routine
! funcs(x,a,yfit,dyda,ma) that evaluates the fitting function yfit,
! and its derivatives dyda[1..ma] with respect to the fitting parameters
! a at x. On the first call provide an initial guess for the parameters
! a, and set aLambda<0 for initialization (which then sets aLambda=.001).
! If a step succeeds chiSqr becomes smaller and aLambda decreases 
! by a factor of 10. If a step fails aLambda grows by a factor of 10.
! You must call this routine repeatedly until convergence is achieved.
! Then, make one final call with aLambda=0, so that covar[1..ma][1.ma]
! returns the covariance matrix, and alpha the curvature matrix.
! (Parameters held fixed will return zero covariances.)
!  
   integer(kind=int32) :: nData
   real(kind=double)   :: x(nData)
   real(kind=double)   :: y(nData)
   real(kind=double)   :: sig(nData)
   integer(kind=int32) :: ma
   real(kind=double)   :: a(ma)
   integer(kind=int32) :: ia(ma)
!   integer(kind=int32) :: fitList(*)
   integer(kind=int32) :: mFit
!   integer(kind=int32) :: nca
   real(kind=double)   :: covMatrix(ma, ma)
   real(kind=double) :: chiSqr
   real(kind=double) :: aLambda
   integer :: verbFlag
   real(kind=double) :: alphaMatrix(ma, ma)
   integer(kind=int32), parameter :: mmax = 20
   real(kind=double) :: atry(ma)
   real(kind=double) :: beta(ma)
   real(kind=double) :: da(ma)
   real(kind=double) :: oneda(ma,ma)
!   integer(kind=int32) :: ihit
   integer(kind=int32) :: j
   integer(kind=int32) :: k
   integer(kind=int32) :: l
!   integer(kind=int32) :: kk
   real(kind=double) :: ochiSqr

   external funcs

   !write(*,*) " OMGNUMERICAL L1873 Marquardt input aLambda=", aLambda
   
   ! Initialisation
   if (aLambda .lt. 0.0) then
      mFit=0
      do j=1,ma
         if (ia(j) .ne. 0) mFit=mFit+1
      enddo

      aLambda = 0.001
      if (verbFlag .eq. 1) then
         write (*,*) "L5317                      Mrqmin: aLambda=", aLambda
         write(*,*) " a=", a
         write(*,*) " beta=", beta, " mFit=", mFit
      endif
      call MrqCof(x, y, sig, nData, a, ia, ma, alphaMatrix, beta, &
                  chiSqr, funcs, verbFlag)
      if (verbFlag .eq. 1) then
         write(*,*) " Mrqmin: L5324, a=", a, " chiSqr=", chiSqr
         write(*,*) " beta=", beta
      endif
      ochiSqr = chiSqr
      do j=1, ma
         atry(j) = a(j)
      enddo
   end if

   ! Alter linearised fitting matrix, by augmenting
   ! diagonal elements
   !mFit=0
   !do j=1,ma
   !   if (ia(j) .ne. 0) mFit=mFit+1
   !enddo
   !if (verbFlag .eq. 1) then
   !   write (*,*) " Mrqmin: L1896, mFit=", mFit, " at the entrance atry=", atry 
   !   write(*,*) " alphaMatrix="
  ! 
  !    do j=1, mFit
  !       write (*,*) alphaMatrix(j,1:mFit)
  !    enddo
  !    write(*,*) "---"
  !    write(*,*) "L1903  beta=", beta
  !    write(*,*) "---"
  ! endif

   if (verbFlag .eq. 1) then
      write(*,*) "L5359 !!!!!!!!!!!!!!!!!!!!!! Alpha matrix"
   endif
   do j = 1, mFit
      do k = 1, mFit
         CovMatrix(j, k) = alphaMatrix(j, k)
      end do
      CovMatrix(j, j) = alphaMatrix(j, j) * (1.0 + aLambda)
      if (verbFlag .eq. 1) then
         write(*,*) (alphaMatrix(j,k), k=1,mFit)
      endif
      ! VNY 27.04.2005
      oneda(j,1) = beta(j)
      !da(j) = beta(j)
      !
   end do
   if (verbFlag .eq. 1) then
      write(*,*) "!!!!!!!!!!!!!!!!!!!!!!"
   endif

   !Test VNY 30.01.03
   if (verbFlag .eq. 1) then
      write (*,*) " Mqtmin: L5364, da=", da
   endif

   ! Matrix solution
   call gaussj(covMatrix, mFit, da, 1)
   if (verbFlag .eq. 1) then
      write (*,*) " Mqtmin: L5370, da=", da
   endif

   ! VNY uncommented 15.07.2004
   da(1:mFit) = oneda(1:mFit,1)
   if (verbFlag .eq. 1) then
      write (*,*) " Mqtmin: L5384, da=", da
   endif

   if (aLambda .eq. 0.0) then
      !write (*,*) " Mqtmin: L1922, before call CovarianceSort-1"
      ! Once converged, evaluate covariance matrix
      call CovarianceSort(covMatrix, ma, ia, mFit)
      !write (*,*) " Mqtmin: Line 1925, before call CovarianceSort-2"
      ! Spread out alpha to its full size too
      call CovarianceSort(alphaMatrix, ma, ia, mFit)
      !write (*,*) " Mqtmin: Line 1928, after call Covariance Sort"
      return
   end if
   
   ! Did the trial succeed?
   !write(*,*) "((((((((((( before the modification: atry=a+da )))))))))))"
   !write(*,*) " L1934: atry=", atry
   !write(*,*) " L1935: a=", a
   !write(*,*) " L1936: da=", da
   j=0
   if (verbFlag .eq. 1) then
      write(*,*) "L5394 ia=", ia
   endif

   do l = 1, ma
      !j=j+1 ! VNY  27.04.2005
      if (ia(l) .ne. 0) then
         j=j+1 ! VNY 27.04.2005
         ! VNY 13.01.2004
         atry(l) = a(l) + da(j)
         !atry(l) = atry(l) + da(j)
      endif
   end do

   if (verbFlag .eq. 1) then
      write (*,*) " L5408: MODIFIED atry=", atry
      write (*,*) "         da=", da
   endif
   call MrqCof(x, y, sig, nData, atry, ia, ma, covMatrix, da, &
               chiSqr, funcs, verbFlag)
   if (verbFlag .eq. 1) then
      write (*,*) " L5414: atry=", atry
      write (*,*) " chiSqr=", chiSqr, " old chiSqr=", ochiSqr
   endif
   if (chiSqr .lt. ochiSqr) then
      ! Success, accept the new solution
      if (verbFlag .eq. 1) then
         write(*,*) " The trial has succeed: decreasing aLambda"
      endif
      aLambda = 0.1 * aLambda
      ochiSqr = chiSqr

      do j = 1, mFit
         do k = 1, mFit
            alphaMatrix(j, k) = CovMatrix(j,k)
         end do
         beta(j) = da(j)
      end do
      !write (*,*) " Mqtmin: Line 1965, after do j=1,mFit"
      do l=1,ma
         a(l)=atry(l)
      enddo
   else
      ! Failure, increase aLambda and return
      if (verbFlag .eq. 1) then
         write(*,*) " The trial is not successful: chiSqr=", chiSqr, &
              " ochiSqr=", ochiSqr
      endif
      aLambda = 10.0 * aLambda
      chiSqr = ochiSqr
   end if
   !write (*,*) " Mqtmin: Line 1974, before return from Mrqmin"
   return
 end subroutine Mrqmin

subroutine MrqCof(x, y, sig, nData, a, ia, ma, alphaMatrix, beta, &
                 chiSqr, funcs, verbFlag)
! Used by Mrqmin to evaluate the linearized fitting matrix alphaMatrix,
! and vector beta, and calculate chi**2
   integer(kind=int32) nData
   real(kind=double) x(nData)
   real(kind=double) y(nData)
   real(kind=double) sig(nData)
   integer(kind=int32) ma
   real(kind=double) a(ma)
   integer(kind=int32) ia(*)
!   integer(kind=int32) fitList(*)
   integer(kind=int32) mFit
!   integer(kind=int32) nalp
   real(kind=double) alphaMatrix(ma, ma)
   real(kind=double) beta(ma)
   real(kind=double) chiSqr
   real(kind=double) sig2i,dy,ymod,wt
   integer :: verbFlag

   integer(kind=int32) i
   integer(kind=int32) j
   integer(kind=int32) k
   integer(kind=int32) l
   integer(kind=int32) m

   external funcs

   integer(kind=int32), parameter :: mmax = 20
!   real dyda(mmax)
   real(kind=double) dyda(ma)

   mFit=0
   do j=1,ma
      if (ia(j) .ne. 0) mFit=mFit+1
   enddo
!   write(*,*) " MRQCOF L2012 ---"
!   write(*,*) "L2013: mFit=", mFit
   do j = 1, mFit
      ! Initialise (symmetric) alpha, beta
      do k = 1, j
         alphaMatrix(j, k) = 0.0
      end do
      beta(j) = 0.0
   end do
!---VNY 03.02.03, commented 13.01.2004
!   alphaMatrix=0.0
!   beta=0.0
!---
!   write(*,*) "L493  alphaMatrix="
!   do j=1, mFit
!      write (*,*) alphaMatrix(j,1:mFit)
!   enddo
!   write(*,*) "L497: beta=", beta

   chiSqr = 0.0
   !if (verbFlag .eq. 1) then 
   !   write(*,*) "....................................................................."
   !   !write(*,*) " MRQCOF L2033: Computing the function values for the given parameters" 
   !   do i=1, nData
   !      write(*,*) " i=", i, " sig=", sig(i)
   !   enddo
   !endif

   do i = 1, nData
      ! summation loop over all data
      
      !write(*,*) " before ymod=", ymod, " dyda=", dyda
      call funcs(x(i), a, ymod, dyda, ma)
      
      !write(*,*) " i=", i, " x(i)=", x(i), " y(i)=", ymod, " sig(i)=", sig(i)

      sig2i = 1.0 / (sig(i) * sig(i))
      dy = y(i)- ymod
      !if (verbFlag .eq. 1) then
      !   write(*,*) " L5547 i=", i, " y(i)=", y(i), " ymod=", ymod, " dy=", dy
      !endif

      ! VNY 13.01.2004
      !sig2i=sig2i/abs(dy)

      !if (verbFlag .eq. 1) then
      !   write(*,*) "L5541  i=", i, "x(i)=", x(i), " y(i)=", y(i), " ymod=", ymod, " dy=", dy
      !   write(*,*) "L5542  dyda=", dyda, " sig2i=", sig2i
      !endif

      j=0
      do l = 1, ma
         if (ia(l) .ne. 0) then
            j=j+1
! VNY 28.01.2004
            !wt = -0.01*dyda(l) * sig2i
!            wt = dyda(l) * sig2i
            wt = dyda(l) * sig2i*sig2i

!            wt = 0.01*dyda(l) * sig2i
!
            !if (verbFlag .eq. 1) then
            !   write(*,*) " l=",l, " dyda=", dyda(l), " sig2i=", sig2i
            !endif

            k=0
            do m = 1, l
               if (ia(m) .ne. 0) then
                  k=k+1
                  alphaMatrix(j, k) = alphaMatrix(j, k) + wt * dyda(m)
               endif
            enddo
            !if (verbFlag .eq. 1 .and. j .eq. 3) then
            !   write(*,*) " L5581 j=", j, " before: beta(j)=2", beta(j), &
            !        " dy=", dy, " wt=", wt, " dy*wt=", dy*wt
            !endif
               beta(j)=beta(j)+dy*wt
            !if (verbFlag .eq. 1 .and. j .eq. 3) then
            !   write(*,*) " L5586 j=", j, " after: beta(j)=2", beta(j), "............."
            !endif
!

         endif
      end do

      if (verbFlag .eq. 1) then
         wt = dyda(3) * sig2i*sig2i
         write (*,*) "L5594 i=", i, " dy=", dy, " dyda(3)=", dyda(3), " wt=", wt, & 
              " beta(3)=", beta(3)
      endif

      chiSqr = chiSqr + dy * dy * sig2i ! and find chi**2.
   end do ! i=1, nData
!   write(*,*) "....................................................................."

   do j = 2, mFit ! Fill in the synnetric side
      do k = 1, j - 1
         alphaMatrix(k, j) = alphaMatrix(j, k)
      end do
   end do
   return
end subroutine MrqCof

subroutine GaussJ(a, n, b, m)
! Linear equation solution by Gaus-Jordan elimination.
! a[1..n][1..n] is the input matrix. b[1..n][1..m] is
! input containing the m right-hand side vectors. On output,
! a is replaced by its matrix inverse, and b is replaced by the 
! corresponding set of solution vectors
   integer(kind=int32) n
   real(kind=double) :: a(n, n)
   integer(kind=int32) m
   real(kind=double) ::  b(n, m)
   integer i,j,k
   integer, parameter :: nmax=50
   real(kind=double):: big,pivinv,dum !, temp
   integer irow, icol, l, ll
   integer ipiv(nmax),indxr(nmax),indxc(nmax)

   ! The integer arrays ipiv, indxr, and indxc are used 
   ! for bookkeeping on the pivoting.

   do j=1,n
      ipiv(j) = 0
   enddo

   do i = 1, n
      !This is the main loop over the columns to be reduced
      big = 0.0

      do j = 1, n
         if (ipiv(j) .ne. 1) then
            do k = 1, n
               if (ipiv(k) .eq. 0) then
                  if (abs(a(j, k)) .ge. big) then
                     big = abs(a(j, k))
                     irow = j
                     icol = k
                  end if
                  ! VNY 27.04.05
               !else if (ipiv(k) > 1) then
               !   call fatal("singMat", "Singular matrix in Marquardt minimization")
               end if
            end do
         end if
      end do

      ipiv(icol) = ipiv(icol) + 1
      ! We now have the pivot element, so we interchange rows,
      ! if needed, to put the pivot element on the diagonal.
      ! The column are not physically interchanged, only relabeled:
      ! indxc[i], the column of the ith pivot element, is the ith column
      ! that is reduced, while indxr[i] is the row i which that pivot
      ! element was originallylocated. If indxr[i] .ne. indxc[i] there
      ! is an implied column interchange. With this form of bookkeeping,
      ! the solution b's will end up in the correct order, and the inverse
      ! matrix will be scrambled by columns.
      if (irow .ne. icol) then
         do l = 1, n
            dum = a(irow, l)
            a(irow, l) = a(icol, l)
            a(icol, l) = dum
         end do

         do l = 1, m
            dum = b(irow, l)
            b(irow, l) = b(icol, l)
            b(icol, l) = dum
         end do
      end if
      
      ! we are now ready to divide the pivot row by the 
      ! pivot element, located at irow and icol.
      indxr(i) = irow
      indxc(i) = icol

      if (a(icol, icol) .eq.  0.0) then
         call warning("L5634singMat", "Singular matrix in Marquardt minimization")
         a(icol,icol)=1.0E-10
      end if

      pivinv = 1.0 / a(icol, icol)
      a(icol, icol) = 1.0

      do l = 1, n
         a(icol, l) = a(icol, l) * pivinv
      end do

      do l = 1, m
         b(icol, l) = b(icol, l) * pivinv
      end do

      do ll = 1, n
         if (ll .ne. icol) then
            ! Next, we reduce the rows except for the pivot one
            dum = a(ll, icol)
            a(ll, icol) = 0.0

            do l = 1, n
               a(ll, l) = a(ll, l) - a(icol, l) * dum
            end do

            do l = 1, m
               b(ll, l) = b(ll, l) - b(icol, l) * dum
            end do
         end if
      end do
      !This this the end of the main loop over columns of the 
      ! reduction. It only remains to unscramble the solution
      ! in view of the column interchanges. We do this by 
      ! interchanging pairs of columns in the reverse order
      ! that the permutation was build up.
   end do

   do  l = n, 1, -1
      if (indxr(l) .ne. indxc(l)) then
         do k = 1, n
            dum = a(k, indxr(l))
            a(k, indxr(l)) = a(k, indxc(l))
            a(k, indxc(l)) = dum
         end do
      end if
   end do

end subroutine GaussJ

subroutine CovarianceSort(covMatrix, ma, ia, mFit)
! expand in storage the covariance matrix covMatrix, so as to take
! into account parameters that are being held fixed. (For the latter,
! return zero covariances)
  integer :: mFit, i, j, k, ma
  real(kind=double) :: CovMatrix(ma,ma)
  integer :: ia(*)
  real(kind=double) ::  dum
  do i=mFit+1, ma
     do j=1,i
        covMatrix(i,j)=0.
        covMatrix(j,i)=0
     enddo
  enddo
  k=mFit
  do j=ma,1,-1
     if (ia(j) .gt. 0)  then
        do i=1,ma
           dum=covMatrix(i,k)
           covMatrix(i,k)=covMatrix(i,j)
           covMatrix(i,j)=dum
        enddo
        do i=1,ma
           dum=covMatrix(k,i)
           covMatrix(k,i)=covMatrix(j,i)
           covMatrix(j,i)=dum
        enddo
        k=k-1
     endif
  enddo
end subroutine CovarianceSort

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Median smoothing of a vector
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
real function ComputeVectorMedian(vect, vectSize, nmax) result(median)
  integer(int32), intent(in) :: vectSize
  integer(int32), intent(in) :: nmax ! number of significant elements in the vector, including 0-th
  real(kind=double), intent(in) :: vect(1:vectSize)
  real(kind=double) :: a(1:vectSize) ! auxiliar vector
  integer :: k,l
  real(kind=double) :: dum
  if (nmax .gt. 2) then
     ! Copy the vector
     a=vect
     ! ordering the vector a
     do k=1,nmax-1
        do l=k+1, nmax
           if (a(l) .lt. a(k)) then
              dum=a(k)
              a(k)=a(l)
              a(l)=dum
           endif
        enddo
     enddo
     if (mod(nmax,2) .eq. 0) then
        l=nmax/2
        median=(a(l)+a(l+1))/2.0
     else
        l=(nmax-1)/2+1
        median=a(l)
     endif
  else
     median=0.0
     do k=1,nmax
        median=median+vect(k)
        dum=dum+1.0
     enddo
     if (dum .gt. 1.0) median=median/dum
  endif
end function ComputeVectorMedian

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Median smoothing of a vector
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
real function ComputeSpecialAverage(vect, vectSize, specialVect) result(average)
  integer(int32), intent(in) :: vectSize
  real(kind=double), intent(in) :: vect(1:vectSize), specialVect(1:vectSize)
  integer :: k
  real(kind=double) :: averSpec, numSpec !, averCross

  ! compute the average of the special vector
  averSpec=ComputeVectorMedian(specialVect,vectSize,vectSize)
  !averSpec=0.0
  !numSpec=0.0
  !do k=1,vectSize
  !   if (specialVect(k) .gt. 0.0) then
  !      averSpec=averSpec+specialVect(k)
  !      numSpec=numSpec+1.0
  !   endif
  !   if (numSpec .gt. 1.0) averSpec=averSpec/numSpec
  !enddo
  !write(*,*) "L6943 averSpec=", averSpec
  ! compute the average of the main vector using only
  ! those values that fall inside if the interval
  ! (averSpec-0.5*averSpec, averSpec+0.5*averSpec)
  average=0.0
  numSpec=0.0
  do k=1,vectSize
     if (abs(specialVect(k)-averSpec) .lt. averSpec/1.5 .and. specialVect(k) .gt. 0.1) then
        average=average+vect(k)
        numSpec=numSpec+1.0
        !write(*,*) "L6953 k=", k, " vect=", vect(k), " specialVect=", specialVect(k),&
        !     " diff=", abs(specialVect(k)-averSpec), "+++++"
     else
        !write(*,*) "L6956 k=", k, " vect=", vect(k), " specialVect=", specialVect(k),&
        !     " diff=", abs(specialVect(k)-averSpec)
     endif
  enddo
  if (numSpec .gt. 1.0) average=average/numSpec
end function ComputeSpecialAverage

real function computeMedian10(image, nx, ny, i1, i2, j1, j2) result (median)

   ! computes the median of an image over the specific limits

   integer :: nx, ny
   real(kind=double) :: image(0:nx-1, 0:ny-1)
   integer :: i1, i2, j1, j2
   integer :: i, j
   real(kind=single), allocatable :: array(:)
   integer :: npts
   
   npts = (i2-i1+1)*(j2-j1+1)
   allocate(array(npts))

   npts = 0
   do i = i1, i2
      do j = j1, j2
         npts = npts + 1
         array(npts) = image(i,j)
      enddo
   enddo

   call sort(npts, array)

   if (NINT(real(npts)/2.) /= INT(npts/2)) then
      median = array(NINT(real(npts)/2.))
       else
      median = 0.5*(array(npts/2) + array(npts/2+1))
   endif

end function computeMedian10


real function arrayMedian10(array, npts) result (median)

   ! computes the median of an array

   integer :: npts
   real(kind=single) :: array(*)
   
!   integer i

   if (npts .eq. 1) then
      median = array(1)
    else
      call sort(npts, array)
      if (NINT(real(npts)/2.) /= INT(npts/2)) then
        median = array(NINT(real(npts)/2.))
         else
        median = 0.5*(array(npts/2) + array(npts/2+1))
      endif
   endif

!   write (*, *) "npts : ", npts
!   do i = 1, npts
!      write (*, *) i, array(i)
!   end do

end function arrayMedian10

!------ Subroutines related to fitpack
subroutine curv1 (n,x,y,slp1,slpn,islpsw,yp,temp,sigma,ierr)
!
      integer(kind=int32) :: n,islpsw,ierr
      real(kind=double) :: x(n),y(n),slp1,slpn,yp(n),temp(n),sigma
      integer(kind=double) :: nm1, np1
      real(kind=double) :: sigmap, delx1, delx2, slpp1, c1, c2, c3
      real(kind=double) :: delxn, delxnm, slppn, dx1, diag1, sdiag1
      real(kind=double) :: diag, dx2, diag2, sdiag2
      integer(kind=int32) :: i, ibak
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the parameters necessary to
! compute an interpolatory spline under tension through
! a sequence of functional values. the slopes at the two
! ends of the curve may be specified or omitted.  for actual
! computation of points on the curve it is necessary to call
! the function curv2.
!
! on input--
!
!   n is the number of values to be interpolated (n.ge.2).
!
!   x is an array of the n increasing abscissae of the
!   functional values.
!
!   y is an array of the n ordinates of the values, (i. e.
!   y(k) is the functional value corresponding to x(k) ).
!
!   slp1 and slpn contain the desired values for the first
!   derivative of the curve at x(1) and x(n), respectively.
!   the user may omit values for either or both of these
!   parameters and signal this with islpsw.

!   islpsw contains a switch indicating which slope data
!   should be used and which should be estimated by this
!   subroutine,
!          = 0 if slp1 and slpn are to be used,
!          = 1 if slp1 is to be used but not slpn,
!          = 2 if slpn is to be used but not slp1,
!          = 3 if both slp1 and slpn are to be estimated
!              internally.
!
!   yp is an array of length at least n.
!
!   temp is an array of length at least n which is used for
!   scratch storage.
!
! and
!
!   sigma contains the tension factor. this value indicates
!   the curviness desired. if abs(sigma) is nearly zero
!   (e.g. .001) the resulting curve is approximately a
!   cubic spline. if abs(sigma) is large (e.g. 50.) the
!   resulting curve is nearly a polygonal line. if sigma
!   equals zero a cubic spline results.  a standard value
!   for sigma is approximately 1. in absolute value.
!
! on output--
!
!   yp contains the values of the second derivative of the
!   curve at the given nodes.
!
!   ierr contains an error flag,
!        = 0 for normal return,
!        = 1 if n is less than 2,
!        = 2 if x-values are not strictly increasing.
!
! and
!
!   n, x, y, slp1, slpn, islpsw and sigma are unaltered.

! this subroutine references package modules ceez, terms,
! and snhcsh.
!
!-----------------------------------------------------------
!
      nm1 = n-1
      np1 = n+1
      ierr = 0
      if (n .le. 1) then
         !
         ! too few points
         !
         ierr = 1
         return
      endif

      if (x(n) .le. x(1)) then
         !
         ! x-values not strictly increasing
         !
         ierr = 2
         return
      endif

      !
      ! denormalize tension factor
      !
      sigmap = abs(sigma)*real(n-1)/(x(n)-x(1))
      !
      ! approximate end slopes
      !
      if (islpsw .ge. 2) then
         delx1 = x(2)-x(1)
         delx2 = delx1+delx1
         if (n .gt. 2) delx2 = x(3)-x(1)
         if (delx1 .le. 0. .or. delx2 .le. delx1) then ! go to 9
            ierr = 2
            return
         endif
         call ceez (delx1,delx2,sigmap,c1,c2,c3,n)
         slpp1 = c1*y(1)+c2*y(2)
         if (n .gt. 2) slpp1 = slpp1+c3*y(3)
      else
         slpp1 = slp1
      endif
      if (islpsw .eq. 1 .or. islpsw .eq. 3) then
         delxn = x(n)-x(nm1)
         delxnm = delxn+delxn
         if (n .gt. 2) delxnm = x(n)-x(n-2)
         if (delxn .le. 0. .or. delxnm .le. delxn) then !go to 9
            ierr = 2
            return
         endif
         call ceez (-delxn,-delxnm,sigmap,c1,c2,c3,n)
         slppn = c1*y(n)+c2*y(nm1)
         if (n .gt. 2) slppn = slppn+c3*y(n-2)
      else
         slppn = slpn
      endif

      !
      ! set up right hand side and tridiagonal system for yp and
      ! perform forward elimination
      !
      delx1 = x(2)-x(1)
      if (delx1 .le. 0.) then
         ierr = 2
         return
      endif
      dx1 = (y(2)-y(1))/delx1
      call terms (diag1,sdiag1,sigmap,delx1)
      yp(1) = (dx1-slpp1)/diag1
      temp(1) = sdiag1/diag1
      if (n .eq. 2) then
         diag = diag1-sdiag1*temp(nm1)
         yp(n) = (slppn-dx1-sdiag1*yp(nm1))/diag
      else
         do i = 2,nm1
            delx2 = x(i+1)-x(i)
            if (delx2 .le. 0.) then
               ierr = 2
               return
            endif
            dx2 = (y(i+1)-y(i))/delx2
            call terms (diag2,sdiag2,sigmap,delx2)
            diag = diag1+diag2-sdiag1*temp(i-1)
            yp(i) = (dx2-dx1-sdiag1*yp(i-1))/diag
            temp(i) = sdiag2/diag
            dx1 = dx2
            diag1 = diag2
            sdiag1 = sdiag2
         enddo
         diag = diag1-sdiag1*temp(nm1)
         yp(n) = (slppn-dx1-sdiag1*yp(nm1))/diag
      endif
      !
      ! perform back substitution
      !
      do  i = 2,n
         ibak = np1-i
         yp(ibak) = yp(ibak)-temp(ibak)*yp(ibak+1)
      enddo
      return
end subroutine curv1

!****************************************************
!  subroutine CEEZ
!****************************************************
      subroutine ceez (del1,del2,sigma,c1,c2,c3,n)
!
      real(kind=double) :: del1,del2,sigma,c1,c2,c3
      integer(kind=int32) :: n
      real(kind=double) :: del,delp, delm, denom, coshm1, sinhmp
      real(kind=double) :: sinhmm, coshm2, dummy
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the coefficients c1, c2, and c3
! used to determine endpoint slopes. specifically, if
! function values y1, y2, and y3 are given at points x1, x2,
! and x3, respectively, the quantity c1*y1 + c2*y2 + c3*y3
! is the value of the derivative at x1 of a spline under
! tension (with tension factor sigma) passing through the
! three points and having third derivative equal to zero at
! x1. optionally, only two values, c1 and c2 are determined.
!
! on input--
!
!   del1 is x2-x1 (.gt. 0.).
!
!   del2 is x3-x1 (.gt. 0.). if n .eq. 2, this parameter is
!   ignored.
!
!   sigma is the tension factor.
!
! and
!
!   n is a switch indicating the number of coefficients to
!   be returned. if n .eq. 2 only two coefficients are
!   returned. otherwise all three are returned.
!
! on output--
!
!   c1, c2, and c3 contain the coefficients.
!
! none of the input parameters are altered.
!
! this subroutine references package module snhcsh.
!
!-----------------------------------------------------------
!
      if (n .eq. 2) go to 2
      if (sigma .ne. 0.) go to 1
      del = del2-del1
!
! tension .eq. 0.
!
      c1 = -(del1+del2)/(del1*del2)
      c2 = del2/(del1*del)
      c3 = -del1/(del2*del)
      return
!
! tension .ne. 0.
!
    1 call snhcsh (dummy,coshm1,sigma*del1,1)
      call snhcsh (dummy,coshm2,sigma*del2,1)
      delp = sigma*(del2+del1)/2.
      delm = sigma*(del2-del1)/2.
      call snhcsh (sinhmp,dummy,delp,-1)
      call snhcsh (sinhmm,dummy,delm,-1)
      denom = coshm1*(del2-del1)-2.*del1*delp*delm* &
           (1.+sinhmp)*(1.+sinhmm)
      c1 = 2.*delp*delm*(1.+sinhmp)*(1.+sinhmm)/denom
      c2 = -coshm2/denom
      c3 = coshm1/denom
      return
!
! two coefficients
!
    2 c1 = -1./del1
      c2 = -c1
      return
    end subroutine ceez
!*************************************************
! subroutine TERMS
!*************************************************
      subroutine terms (diag,sdiag,sigma,del)
!
      real(kind=double) :: diag,sdiag,sigma,del
      real(kind=double) :: sigdel, denom, sinhm, coshm
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine computes the diagonal and superdiagonal
! terms of the tridiagonal linear system associated with
! spline under tension interpolation.
!
! on input--
!
!   sigma contains the tension factor.
!
! and
!
!   del contains the step size.
!
! on output--
!
!                sigma*del*cosh(sigma*del) - sinh(sigma*del)
!   diag = del*--------------------------------------------.
!                     (sigma*del)**2 * sinh(sigma*del)
!
!                   sinh(sigma*del) - sigma*del
!   sdiag = del*----------------------------------.
!                (sigma*del)**2 * sinh(sigma*del)
!
! and
!
!   sigma and del are unaltered.
!
! this subroutine references package module snhcsh.
!
!-----------------------------------------------------------
!
      if (sigma .ne. 0.) go to 1
      diag = del/3.
      sdiag = del/6.
      return
    1 sigdel = sigma*del
      call snhcsh (sinhm,coshm,sigdel,0)
      denom = sigma*sigdel*(1.+sinhm)
      diag = (coshm-sinhm)/denom
      sdiag = sinhm/denom
      return
    end subroutine terms
!--------------------------

!*************************************************
! subroutine SNHCSH
!*************************************************
      subroutine snhcsh (sinhm,coshm,x,isw)
!
      integer(kind=int32) :: isw
      real(kind=double) :: sinhm,coshm,x
      real(kind=double) :: ax, xs, sp14, sp13, sp12, sp11, sq12, sq11
      real(kind=double) :: sq10, sp25, sp24, sp23, sp22, sp21, sq22
      real(kind=double) :: sq21, sq20, sp35, sp34, sp33, sp32, sp31
      real(kind=double) :: sq32, sq31, sq30, sp45, sp44, sp43, sp42
      real(kind=double) :: sp41, sq42, sq41, sq40, expx, cp5, cp4
      real(kind=double) :: cp3, cp2, cp1, cq2, cq1, cq0, zp4, zp3
      real(kind=double) :: zp2, zp1, zq2, zq1, zq0
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine returns approximations to
!       sinhm(x) = sinh(x)/x-1
!       coshm(x) = cosh(x)-1
! and
!       coshmm(x) = (cosh(x)-1-x*x/2)/(x*x)
! with relative error less than 4.0e-14.
!
! on input--
!
!   x contains the value of the independent variable.
!
!   isw indicates the function desired
!           = -1 if only sinhm is desired,
!           =  0 if both sinhm and coshm are desired,
!           =  1 if only coshm is desired,
!           =  2 if only coshmm is desired,
!           =  3 if both sinhm and coshmm are desired.
!
! on output--
!
!   sinhm contains the value of sinhm(x) if isw .le. 0 or
!   isw .eq. 3 (sinhm is unaltered if isw .eq.1 or isw .eq.
!   2).
!
!   coshm contains the value of coshm(x) if isw .eq. 0 or
!   isw .eq. 1 and contains the value of coshmm(x) if isw
!   .ge. 2 (coshm is unaltered if isw .eq. -1).
!
! and
!
!   x and isw are unaltered.
!
!-----------------------------------------------------------
!
      data sp14/.227581660976348d-7/, &
           sp13/.612189863171694d-5/, &
           sp12/.715314759211209d-3/, &
           sp11/.398088289992973d-1/, &
           sq12/.206382701413725d-3/, &
           sq11/-.611470260009508d-1/, &
           sq10/.599999999999986d+1/ 
      data sp25/.129094158037272d-9/, &
           sp24/.473731823101666d-7/, &
           sp23/.849213455598455d-5/, &
           sp22/.833264803327242d-3/, &
           sp21/.425024142813226d-1/, &
           sq22/.106008515744821d-3/, &
           sq21/-.449855169512505d-1/, &
           sq20/.600000000268619d+1/ 
      data sp35/.155193945864942d-9/, &
           sp34/.511529451668737d-7/, &
           sp33/.884775635776784d-5/, &
           sp32/.850447617691392d-3/, &
           sp31/.428888148791777d-1/, &
           sq32/.933128831061610d-4/, &
           sq31/-.426677570538507d-1/, &
           sq30/.600000145086489d+1/
      data sp45/.188070632058331d-9/, &
           sp44/.545792817714192d-7/, &
           sp43/.920119535795222d-5/, &
           sp42/.866559391672985d-3/, &
           sp41/.432535234960858d-1/, &
           sq42/.824891748820670d-4/, &
           sq41/-.404938841672262d-1/, &
           sq40/.600005006283834d+1/
      data cp5/.552200614584744d-9/, &
           cp4/.181666923620944d-6/, &
           cp3/.270540125846525d-4/, &
           cp2/.206270719503934d-2/, &
           cp1/.744437205569040d-1/, &
           cq2/.514609638642689d-4/, &
           cq1/-.177792255528382d-1/, &
           cq0/.200000000000000d+1/
      data zp4/.664418805876835d-8/, &
           zp3/.218274535686385d-5/, &
           zp2/.324851059327161d-3/, &
           zp1/.244515150174258d-1/, &
           zq2/.616165782306621d-3/, &
           zq1/-.213163639579425d0/, &
           zq0/.240000000000000d+2/
!
      ax = abs(x)
      if (isw .ge. 0) go to 5
!
! sinhm approximation
!
      if (ax .gt. 3.9) go to 2
      xs = ax*ax
      if (ax .gt. 2.2) go to 1
!
! sinhm approximation on (0.,2.2)
!
      sinhm = xs*((((sp14*xs+sp13)*xs+sp12)*xs+sp11)*xs+1.)/ &
                   ((sq12*xs+sq11)*xs+sq10)
      return
!
! sinhm approximation on (2.2,3.9)
!
    1 sinhm = xs*(((((sp25*xs+sp24)*xs+sp23)*xs+sp22)*xs+sp21) &
              *xs+1.)/((sq22*xs+sq21)*xs+sq20)
      return
    2 if (ax .gt. 5.1) go to 3
!
! sinhm approximation on (3.9,5.1)
!
      xs = ax*ax
      sinhm = xs*(((((sp35*xs+sp34)*xs+sp33)*xs+sp32)*xs+sp31) &
              *xs+1.)/((sq32*xs+sq31)*xs+sq30)
      return
    3 if (ax .gt. 6.1) go to 4
!
! sinhm approximation on (5.1,6.1)
!
      xs = ax*ax
      sinhm = xs*(((((sp45*xs+sp44)*xs+sp43)*xs+sp42)*xs+sp41) &
              *xs+1.)/((sq42*xs+sq41)*xs+sq40)
      return
!
! sinhm approximation above 6.1
!
    4 expx = exp(ax)
      sinhm = (expx-1./expx)/(ax+ax)-1.
      return
!
! coshm and (possibly) sinhm approximation
!
    5 if (isw .ge. 2) go to 7
      if (ax .gt. 2.2) go to 6
      xs = ax*ax
      coshm = xs*(((((cp5*xs+cp4)*xs+cp3)*xs+cp2)*xs+cp1) &
              *xs+1.)/((cq2*xs+cq1)*xs+cq0)
      if (isw .eq. 0) sinhm = xs*((((sp14*xs+sp13)*xs+sp12) &
                *xs+sp11)*xs+1.)/((sq12*xs+sq11)*xs+sq10)
      return
    6 expx = exp(ax)
      coshm = (expx+1./expx)/2.-1.
      if (isw .eq. 0) sinhm = (expx-1./expx)/(ax+ax)-1.
      return
!
! coshmm and (possibly) sinhm approximation
!
    7 xs = ax*ax
      if (ax .gt. 2.2) go to 8
      coshm = xs*((((zp4*xs+zp3)*xs+zp2)*xs+zp1)*xs+1.)/ &
                   ((zq2*xs+zq1)*xs+zq0)
      if (isw .eq. 3) sinhm = xs*((((sp14*xs+sp13)*xs+sp12) &
                *xs+sp11)*xs+1.)/((sq12*xs+sq11)*xs+sq10)
      return
    8 expx = exp(ax)
      coshm = ((expx+1./expx-xs)/2.-1.)/xs
      if (isw .eq. 3) sinhm = (expx-1./expx)/(ax+ax)-1.
      return
    end subroutine snhcsh
!----------------

!*********************************************************
! function CURV2
!*********************************************************

real  function curv2 (t,n,x,y,yp,sigma) result (curv2result)
!
      integer(kind=int32) ::  n
      real(kind=double) t,x(n),y(n),yp(n),sigma
      integer(kind=int32) :: im1, i
      real(kind=double) :: sigmap, del1, del2, dels, sum, sigdel, s1, s2, ss
      real(kind=double) :: dummy
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this function interpolates a curve at a given point
! using a spline under tension. the subroutine curv1 should
! be called earlier to determine certain necessary
! parameters.
!
! on input--
!
!   t contains a real value to be mapped onto the interpo-
!   lating curve.
!
!   n contains the number of points which were specified to
!   determine the curve.
!
!   x and y are arrays containing the abscissae and
!   ordinates, respectively, of the specified points.
!
!   yp is an array of second derivative values of the curve
!   at the nodes.
!
! and
!
!   sigma contains the tension factor (its sign is ignored).
!
! the parameters n, x, y, yp, and sigma should be input
! unaltered from the output of curv1.
!
! on output--
!
!   curv2 contains the interpolated value.
!
! none of the input parameters are altered.
!
! this function references package modules intrvl and
! snhcsh.
!
!-----------------------------------------------------------
!
! determine interval
!
      im1 = intrvl(t,x,n)
      i = im1+1
!
! denormalize tension factor
!
      sigmap = abs(sigma)*float(n-1)/(x(n)-x(1))
!
! set up and perform interpolation
!
      del1 = t-x(im1)
      del2 = x(i)-t
      dels = x(i)-x(im1)
      sum = (y(i)*del1+y(im1)*del2)/dels
      if (sigmap .ne. 0.) go to 1
      curv2result=sum-del1*del2*(yp(i)*(del1+dels)+yp(im1)* &
              (del2+dels))/(6.*dels)
      return
    1 sigdel = sigmap*dels
      call snhcsh (ss,dummy,sigdel,-1)
      call snhcsh (s1,dummy,sigmap*del1,-1)
      call snhcsh (s2,dummy,sigmap*del2,-1)
      curv2result = sum+(yp(i)*del1*(s1-ss)+yp(im1)*del2*(s2-ss))/ &
                  (sigdel*sigmap*(1.+ss))
      return
    end function curv2
!-------------------------------------------
integer function intrvl (t,x,n)
!
      integer(kind=int32) ::  n
      real(kind=double) ::  t,x(n)
      real(kind=double) :: tt
      integer(kind=int32) :: i, il, ih
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this function determines the index of the interval
! (determined by a given increasing sequence) in which
! a given value lies.
!
! on input--
!
!   t is the given value.
!
!   x is a vector of strictly increasing values.
!
! and
!
!   n is the length of x (n .ge. 2).
!
! on output--
!
!   intrvl returns an integer i such that
!
!          i =  1       if         e   t .lt. x(2)  ,
!          i =  n-1     if x(n-1) .le. t            ,
!          otherwise       x(i)  .le. t .le. x(i+1),
!
! none of the input parameters are altered.
!
!-----------------------------------------------------------
!
      save i
      data i /1/
!
      tt = t
!
! check for illegal i
!
      if (i .ge. n) i = n/2
!
! check old interval and extremes
!
      if (tt .lt. x(i)) then
        if (tt .le. x(2)) then
          i = 1
          intrvl = 1
          return
        else
          il = 2
          ih = i
        end if
      else if (tt .le. x(i+1)) then
        intrvl = i
        return
      else if (tt .ge. x(n-1)) then
        i = n-1
        intrvl = n-1
        return
      else
        il = i+1
        ih = n-1
      end if
!
! binary search loop
!
    1 i = (il+ih)/2
      if (tt .lt. x(i)) then
         ih = i
      else if (tt .gt. x(i+1)) then
         il = i+1
      else
         intrvl = i
         return
      end if
      go to 1
    end function intrvl

!---------------------

 subroutine curvs (n,x,y,d,isw,s,eps,ys,ysp,sigma,temp,ierr)
!
      integer(kind=int32) ::  n,isw,ierr
      real(kind=double) :: x(n),y(n),d(n),s,eps,ys(n),ysp(n),sigma,temp(n,9)

!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the parameters necessary to
! compute a smoothing spline under tension. for a given
! increasing sequence of abscissae (x(i)), i = 1,..., n and
! associated ordinates (y(i)), i = 1,..., n, the function
! determined minimizes the summation from i = 1 to n-1 of
! the square of the second derivative of f plus sigma
! squared times the difference of the first derivative of f
! and (f(x(i+1))-f(x(i)))/(x(i+1)-x(i)) squared, over all
! functions f with two continuous derivatives such that the
! summation of the square of (f(x(i))-y(i))/d(i) is less
! than or equal to a given constant s, where (d(i)), i = 1,
! ..., n are a given set of observation weights. the
! function determined is a spline under tension with third
! derivative discontinuities at (x(i)), i = 2,..., n-1. for
! actual computation of points on the curve it is necessary
! to call the function curv2. the determination of the curve
! is performed by subroutine curvss, the subroutine curvs
! only decomposes the workspace for curvss.
!
! on input--
!
!   n is the number of values to be smoothed (n.ge.2).
!
!   x is an array of the n increasing abscissae of the
!   values to be smoothed.
!
!   y is an array of the n ordinates of the values to be
!   smoothed, (i. e. y(k) is the functional value
!   corresponding to x(k) ).
!
!   d is a parameter containing the observation weights.
!   this may either be an array of length n or a scalar
!   (interpreted as a constant). the value of d
!   corresponding to the observation (x(k),y(k)) should
!   be an approximation to the standard deviation of error.
!
!   isw contains a switch indicating whether the parameter
!   d is to be considered a vector or a scalar,
!          = 0 if d is an array of length n,
!          = 1 if d is a scalar.
!
!   s contains the value controlling the smoothing. this
!   must be non-negative. for s equal to zero, the
!   subroutine does interpolation, larger values lead to
!   smoother funtions. if parameter d contains standard
!   deviation estimates, a reasonable value for s is
!   float(n).
!
!   eps contains a tolerance on the relative precision to
!   which s is to be interpreted. this must be greater than
!   or equal to zero and less than or equal to one. a
!   reasonable value for eps is sqrt(2./float(n)).
!
!   ys is an array of length at least n.
!
!   ysp is an array of length at least n.
!
!   sigma contains the tension factor. this value indicates
!   the degree to which the first derivative part of the
!   smoothing functional is emphasized. if sigma is nearly
!   zero (e. g. .001) the resulting curve is approximately a
!   cubic spline. if sigma is large (e. g. 50.) the
!   resulting curve is nearly a polygonal line. if sigma
!   equals zero a cubic spline results. a standard value for
!   sigma is approximately 1.
!
! and
!
!   temp is an array of length at least 9*n which is used
!   for scratch storage.
!
! on output--
!
!   ys contains the smoothed ordinate values.
!
!   ysp contains the values of the second derivative of the
!   smoothed curve at the given nodes.
!
!   ierr contains an error flag,
!        = 0 for normal return,
!        = 1 if n is less than 2,
!        = 2 if s is negative,
!        = 3 if eps is negative or greater than one,
!        = 4 if x-values are not strictly increasing,
!        = 5 if a d-value is non-positive.
!
! and
!
!   n, x, y, d, isw, s, eps, and sigma are unaltered.
!
! this subroutine references package modules curvss, terms,
! and snhcsh.
!
!-----------------------------------------------------------
!
! decompose temp into nine arrays and call curvss
!
      call curvss (n,x,y,d,isw,s,eps,ys,ysp,sigma,temp(1,1), &
                   temp(1,2),temp(1,3),temp(1,4),temp(1,5), &
                   temp(1,6),temp(1,7),temp(1,8),temp(1,9), &
                   ierr)
      return
    end subroutine curvs
!--------------------
!**********************************************************
! subroutine CURVSS
!**********************************************************
      subroutine curvss (n,x,y,d,isw,s,eps,ys,ysp,sigma,td, &
                         tsd1,hd,hsd1,hsd2,rd,rsd1,rsd2,v, &
                         ierr)
!
      integer(kind=int32) ::  n,isw,ierr
      real(kind=double) ::  x(n),y(n),d(n),s,eps,ys(n),ysp(n),sigma,td(n) 
      real(kind=double) ::  tsd1(n),hd(n),hsd1(n),hsd2(n),rd(n),rsd1(n)
      real(kind=double) ::  rsd2(n),v(n)
!
      real(kind=double) :: p, rdim1, yspim2, sigmap, delxi1, delyi1
      real(kind=double) :: dim1, delxi, delyi, di, sl, su, betapp, betap
      real(kind=double) :: alphap, alpha, beta, hsd1p, hdim1, hdi, rsd2i
      real(kind=double) :: rsd1i, sum, f, g, wim2, wim1, tui, wi, h, step
      integer(kind=int32) :: nm1, nm3, i, ibak
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the parameters necessary to
! compute a smoothing spline under tension. for a given
! increasing sequence of abscissae (x(i)), i = 1,..., n and
! associated ordinates (y(i)), i = 1,..., n, the function
! determined minimizes the summation from i = 1 to n-1 of
! the square of the second derivative of f plus sigma
! squared times the difference of the first derivative of f
! and (f(x(i+1))-f(x(i)))/(x(i+1)-x(i)) squared, over all
! functions f with two continuous derivatives such that the
! summation of the square of (f(x(i))-y(i))/d(i) is less
! than or equal to a given constant s, where (d(i)), i = 1,
! ..., n are a given set of observation weights. the
! function determined is a spline under tension with third
! derivative discontinuities at (x(i)), i = 2,..., n-1. for
! actual computation of points on the curve it is necessary
! to call the function curv2.
!
! on input--
!
!   n is the number of values to be smoothed (n.ge.2).
!
!   x is an array of the n increasing abscissae of the
!   values to be smoothed.
!
!   y is an array of the n ordinates of the values to be
!   smoothed, (i. e. y(k) is the functional value
!   corresponding to x(k) ).
!
!   d is a parameter containing the observation weights.
!   this may either be an array of length n or a scalar
!   (interpreted as a constant). the value of d
!   corresponding to the observation (x(k),y(k)) should
!   be an approximation to the standard deviation of error.
!
!   isw contains a switch indicating whether the parameter
!   d is to be considered a vector or a scalar,
!          = 0 if d is an array of length n,
!          = 1 if d is a scalar.
!
!   s contains the value controlling the smoothing. this
!   must be non-negative. for s equal to zero, the
!   subroutine does interpolation, larger values lead to
!   smoother funtions. if parameter d contains standard
!   deviation estimates, a reasonable value for s is
!   float(n).
!
!   eps contains a tolerance on the relative precision to
!   which s is to be interpreted. this must be greater than
!   or equal to zero and less than equal or equal to one. a
!   reasonable value for eps is sqrt(2./float(n)).
!
!   ys is an array of length at least n.
!
!   ysp is an array of length at least n.
!
!   sigma contains the tension factor. this value indicates
!   the degree to which the first derivative part of the
!   smoothing functional is emphasized. if sigma is nearly
!   zero (e. g. .001) the resulting curve is approximately a
!   cubic spline. if sigma is large (e. g. 50.) the
!   resulting curve is nearly a polygonal line. if sigma
!   equals zero a cubic spline results. a standard value for
!   sigma is approximately 1.
!
! and
!
!   td, tsd1, hd, hsd1, hsd2, rd, rsd1, rsd2, and v are
!   arrays of length at least n which are used for scratch
!   storage.
!
! on output--
!
!   ys contains the smoothed ordinate values.
!
!   ysp contains the values of the second derivative of the
!   smoothed curve at the given nodes.
!
!   ierr contains an error flag,
!        = 0 for normal return,
!        = 1 if n is less than 2,
!        = 2 if s is negative,
!        = 3 if eps is negative or greater than one,
!        = 4 if x-values are not strictly increasing,
!        = 5 if a d-value is non-positive.
!
! and
!
!   n, x, y, d, isw, s, eps, and sigma are unaltered.
!
! this subroutine references package modules terms and
! snhcsh.
!
!-----------------------------------------------------------
! 
      if (n .lt. 2) go to 16
      if (s .lt. 0.) go to 17
      if (eps .lt. 0. .or. eps .gt. 1.) go to 18
      ierr = 0
      p = 0.
      v(1) = 0.
      v(n) = 0.
      ysp(1) = 0.
      ysp(n) = 0.
      if (n .eq. 2) go to 14
      rsd1(1) = 0.
      rd(1) = 0.
      rsd2(n) = 0.
      rdim1 = 0.
      yspim2 = 0.
!
! denormalize tension factor
!
      sigmap = abs(sigma)*float(n-1)/(x(n)-x(1))
!
! form t matrix and second differences of y into ys
!
      nm1 = n-1
      nm3 = n-3
      delxi1 = 1.
      delyi1 = 0.
      dim1 = 0.
      do  i = 1,nm1
        delxi = x(i+1)-x(i)
        if (delxi .le. 0.) go to 19
        delyi = (y(i+1)-y(i))/delxi
        ys(i) = delyi-delyi1
        call terms (di,tsd1(i+1),sigmap,delxi)
        td(i) = di+dim1
        hd(i) = -(1./delxi+1./delxi1)
        hsd1(i+1) = 1./delxi
        delxi1 = delxi
        delyi1 = delyi
        dim1 = di
     enddo
!
! calculate lower and upper tolerances
!
      sl = s*(1.-eps)
      su = s*(1.+eps)
      if (isw .eq. 1) go to 3
!
! form h matrix - d array
!
      if (d(1) .le. 0. .or. d(2) .le. 0.) go to 20
      betapp = 0.
      betap = 0.
      alphap = 0.
      do  i = 2,nm1
        alpha = hd(i)*d(i)*d(i)
        if (d(i+1) .le. 0.) go to 20
        beta = hsd1(i+1)*d(i+1)*d(i+1)
        hd(i) = (hsd1(i)*d(i-1))**2+alpha*hd(i) &
                                   +beta*hsd1(i+1)
        hsd2(i) = hsd1(i)*betapp
        hsd1(i) = hsd1(i)*(alpha+alphap)
        alphap = alpha
        betapp = betap
        betap = beta
     enddo
      go to 5
!
! form h matrix - d constant
!
    3 if (d(1) .le. 0.) go to 20
      sl = d(1)*d(1)*sl
      su = d(1)*d(1)*su
      hsd1p = 0.
      hdim1 = 0.
      do  i = 2,nm1
        hdi = hd(i)
        hd(i) = hsd1(i)*hsd1(i)+hdi*hdi+hsd1(i+1)*hsd1(i+1)
        hsd2(i) = hsd1(i)*hsd1p
        hsd1p = hsd1(i)
        hsd1(i) = hsd1p*(hdi+hdim1)
        hdim1 = hdi
     enddo
!
! top of iteration
! cholesky factorization of p*t+h into r
!
    5 do i = 2,nm1
        rsd2i = hsd2(i)
        rsd1i = p*tsd1(i)+hsd1(i)-rsd2i*rsd1(i-1)
        rsd2(i) = rsd2i*rdim1
        rdim1 = rd(i-1)
        rsd1(i) = rsd1i*rdim1
        rd(i) = 1./(p*td(i)+hd(i)-rsd1i*rsd1(i) &
                                 -rsd2i*rsd2(i))
        ysp(i) = ys(i)-rsd1(i)*ysp(i-1)-rsd2(i)*yspim2
        yspim2 = ysp(i-1)
     enddo
!
! back solve of r(transpose)* r * ysp = ys
!
      ysp(nm1) = rd(nm1)*ysp(nm1)
      if (n .eq. 3) go to 8
      do  ibak = 1,nm3
        i = nm1-ibak
        ysp(i) = rd(i)*ysp(i)-rsd1(i+1)*ysp(i+1) &
                             -rsd2(i+2)*ysp(i+2)
     enddo
    8 sum = 0.
      delyi1 = 0.
      if (isw .eq. 1) go to 10
!
! calculation of residual norm
!  - d array
!
      do  i = 1,nm1
        delyi = (ysp(i+1)-ysp(i))/(x(i+1)-x(i))
        v(i) = (delyi-delyi1)*d(i)*d(i)
        sum = sum+v(i)*(delyi-delyi1)
        delyi1 = delyi
     enddo
      v(n) = -delyi1*d(n)*d(n)
      go to 12
!
! calculation of residual norm
!  - d constant
!
   10 do  i = 1,nm1
        delyi = (ysp(i+1)-ysp(i))/(x(i+1)-x(i))
        v(i) = delyi-delyi1
        sum = sum+v(i)*(delyi-delyi1)
        delyi1 = delyi
     enddo
      v(n) = -delyi1
   12 sum = sum-v(n)*delyi1
!
! test for convergence
!
      if (sum .le. su) go to 14
!
! calculation of newton correction
!
      f = 0.
      g = 0.
      wim2 = 0.
      wim1 = 0.
      do  i = 2,nm1
        tui = tsd1(i)*ysp(i-1)+td(i)*ysp(i) &
                              +tsd1(i+1)*ysp(i+1)
        wi = tui-rsd1(i)*wim1-rsd2(i)*wim2
        f = f+tui*ysp(i)
        g = g+wi*wi*rd(i)
        wim2 = wim1
        wim1 = wi
     enddo
      h = f-p*g
      if (h .le. 0.) go to 14
!
! update p - newton step
!
      step = (sum-sqrt(sum*sl))/h
      if (sl .ne. 0.) step = step*sqrt(sum/sl)
      p = p+step
      go to 5
!
! store smoothed y-values and second derivatives
!
   14 do  i = 1,n
        ys(i) = y(i)-v(i)
        ysp(i) = p*ysp(i)
     enddo
      return
!
! n less than 2
!
   16 ierr = 1
      return
!
! s negative
!
   17 ierr = 2
      return
!
! eps negative or greater than 1
!
   18 ierr = 3
      return
!
! x-values not strictly increasing
!
   19 ierr = 4
      return
!
! weight non-positive
!
   20 ierr = 5
      return
    end subroutine curvss
!--------------------

! wsc@research.bell-labs.com Mon Dec 30 16:55 EST 1985
! W. S. Cleveland
! Bell Laboratories
! Murray Hill NJ 07974
! 
! outline of this file:
!    lines 1-72   introduction
!        73-177   documentation for lowess
!       178-238   ratfor version of lowess
!       239-301   documentation for lowest
!       302-350   ratfor version of lowest
!       351-end   test driver and fortran version of lowess and lowest
! 
!   a multivariate version is available by "send dloess from a"
! 
!              COMPUTER PROGRAMS FOR LOCALLY WEIGHTED REGRESSION
! 
!             This package consists  of  two  FORTRAN  programs  for
!        smoothing    scatterplots   by   robust   locally   weighted
!        regression, or lowess.   The  principal  routine  is  LOWESS
!        which   computes   the  smoothed  values  using  the  method
!        described in The Elements of Graphing Data, by William S.
!        Cleveland    (Wadsworth,    555 Morego   Street,   Monterey,
!        California 93940).
! 
!             LOWESS calls a support routine, LOWEST, the code for
!        which is included. LOWESS also calls a routine  SORT,  which
!        the user must provide.
! 
!             To reduce the computations, LOWESS  requires  that  the
!        arrays  X  and  Y,  which  are  the  horizontal and vertical
!        coordinates, respectively, of the scatterplot, be such  that
!        X  is  sorted  from  smallest  to  largest.   The  user must
!        therefore use another sort routine which will sort X  and  Y
!        according  to X.
!             To summarize the scatterplot, YS,  the  fitted  values,
!        should  be  plotted  against X.   No  graphics  routines are
!        available in the package and must be supplied by the user.
! 
!             The FORTRAN code for the routines LOWESS and LOWEST has
!        been   generated   from   higher   level   RATFOR   programs
!        (B. W. Kernighan, ``RATFOR:  A Preprocessor for  a  Rational
!        Fortran,''  Software Practice and Experience, Vol. 5 (1975),
!        which are also included.
! 
!             The following are data and output from LOWESS that  can
!        be  used  to check your implementation of the routines.  The
!        notation (10)v means 10 values of v.
! 
! 
! 
! 
!        X values:
!          1  2  3  4  5  (10)6  8  10  12  14  50
! 
!        Y values:
!           18  2  15  6  10  4  16  11  7  3  14  17  20  12  9  13  1  8  5  19
! 
! 
!        YS values with F = .25, NSTEPS = 0, DELTA = 0.0
!         13.659  11.145  8.701  9.722  10.000  (10)11.300  13.000  6.440  5.596
!           5.456  18.998
! 
!        YS values with F = .25, NSTEPS = 0 ,  DELTA = 3.0
!          13.659  12.347  11.034  9.722  10.511  (10)11.300  13.000  6.440  5.596
!            5.456  18.998
! 
!        YS values with F = .25, NSTEPS = 2, DELTA = 0.0
!          14.811  12.115  8.984  9.676  10.000  (10)11.346  13.000  6.734  5.744
!            5.415  18.998
! 
! 
! 
! 
!                                   LOWESS
! 
! 
! 
!        Calling sequence
! 
!        CALL LOWESS(X,Y,N,F,NSTEPS,DELTA,YS,RW,RES)
! 
!        Purpose
! 
!        LOWESS computes the smooth of a scatterplot of Y  against  X
!        using  robust  locally  weighted regression.  Fitted values,
!        YS, are computed at each of the  values  of  the  horizontal
!        axis in X.
! 
!        Argument description
! 
!              X = Input; abscissas of the points on the
!                  scatterplot; the values in X must be ordered
!                  from smallest to largest.
!              Y = Input; ordinates of the points on the
!                  scatterplot.
!              N = Input; dimension of X,Y,YS,RW, and RES.
!              F = Input; specifies the amount of smoothing; F is
!                  the fraction of points used to compute each
!                  fitted value; as F increases the smoothed values
!                  become smoother; choosing F in the range .2 to
!                  .8 usually results in a good fit; if you have no
!                  idea which value to use, try F = .5.
!         NSTEPS = Input; the number of iterations in the robust
!                  fit; if NSTEPS = 0, the nonrobust fit is
!                  returned; setting NSTEPS equal to 2 should serve
!                  most purposes.
!          DELTA = input; nonnegative parameter which may be used
!                  to save computations; if N is less than 100, set
!                  DELTA equal to 0.0; if N is greater than 100 you
!                  should find out how DELTA works by reading the
!                  additional instructions section.
!             YS = Output; fitted values; YS(I) is the fitted value
!                  at X(I); to summarize the scatterplot, YS(I)
!                  should be plotted against X(I).
!             RW = Output; robustness weights; RW(I) is the weight
!                  given to the point (X(I),Y(I)); if NSTEPS = 0,
!                  RW is not used.
!            RES = Output; residuals; RES(I) = Y(I)-YS(I).
! 
! 
!        Other programs called
! 
!               LOWEST
!               SSORT
! 
!        Additional instructions
! 
!        DELTA can be used to save computations.   Very  roughly  the
!        algorithm  is  this:   on the initial fit and on each of the
!        NSTEPS iterations locally weighted regression fitted  values
!        are computed at points in X which are spaced, roughly, DELTA
!        apart; then the fitted values at the  remaining  points  are
!        computed  using  linear  interpolation.   The  first locally
!        weighted regression (l.w.r.) computation is carried  out  at
!        X(1)  and  the  last  is  carried  out at X(N).  Suppose the
!        l.w.r. computation is carried out at  X(I).   If  X(I+1)  is
!        greater  than  or  equal  to  X(I)+DELTA,  the  next  l.w.r.
!        computation is carried out at X(I+1).   If  X(I+1)  is  less
!        than X(I)+DELTA, the next l.w.r.  computation is carried out
!        at the largest X(J) which is greater than or equal  to  X(I)
!        but  is not greater than X(I)+DELTA.  Then the fitted values
!        for X(K) between X(I)  and  X(J),  if  there  are  any,  are
!        computed  by  linear  interpolation  of the fitted values at
!        X(I) and X(J).  If N is less than 100 then DELTA can be  set
!        to  0.0  since  the  computation time will not be too great.
!        For larger N it is typically not necessary to carry out  the
!        l.w.r.  computation for all points, so that much computation
!        time can be saved by taking DELTA to be  greater  than  0.0.
!        If  DELTA =  Range  (X)/k  then,  if  the  values  in X were
!        uniformly  scattered  over  the  range,  the   full   l.w.r.
!        computation  would be carried out at approximately k points.
!        Taking k to be 50 often works well.
! 
!        Method
! 
!        The fitted values are computed by using the nearest neighbor
!        routine  and  robust locally weighted regression of degree 1
!        with the tricube weight function.  A few additional features
!        have  been  added.  Suppose r is FN truncated to an integer.
!        Let  h  be  the  distance  to  the  r-th  nearest   neighbor
!        from X(I).   All  points within h of X(I) are used.  Thus if
!        the r-th nearest neighbor is exactly the  same  distance  as
!        other  points,  more  than r points can possibly be used for
!        the smooth at  X(I).   There  are  two  cases  where  robust
!        locally  weighted regression of degree 0 is actually used at
!        X(I).  One case occurs when  h  is  0.0.   The  second  case
!        occurs  when  the  weighted  standard error of the X(I) with
!        respect to the weights w(j) is  less  than  .001  times  the
!        range  of the X(I), where w(j) is the weight assigned to the
!        j-th point of X (the tricube  weight  times  the  robustness
!        weight)  divided by the sum of all of the weights.  Finally,
!        if the w(j) are all zero for the smooth at X(I), the  fitted
!        value is taken to be Y(I).
! 
! 
! 
! 
!  subroutine lowess(x,y,n,f,nsteps,delta,ys,rw,res)
!  real x(n),y(n),ys(n),rw(n),res(n)
!  logical ok
!  if (n<2){ ys(1) = y(1); return }
!  ns = max0(min0(ifix(f*float(n)),n),2)  # at least two, at most n points
!  for(iter=1; iter<=nsteps+1; iter=iter+1){      # robustness iterations
!         nleft = 1; nright = ns
!         last = 0        # index of prev estimated point
!         i = 1   # index of current point
!         repeat{
!                 while(nright<n){
!  # move nleft, nright to right if radius decreases
!                         d1 = x(i)-x(nleft)
!                         d2 = x(nright+1)-x(i)
!  # if d1<=d2 with x(nright+1)==x(nright), lowest fixes
!                         if (d1<=d2) break
!  # radius will not decrease by move right
!                         nleft = nleft+1
!                         nright = nright+1
!                         }
!                 call lowest(x,y,n,x(i),ys(i),nleft,nright,res,iter>1,rw,ok)
!  # fitted value at x(i)
!                 if (!ok) ys(i) = y(i)
!  # all weights zero - copy over value (all rw==0)
!                 if (last<i-1) { # skipped points -- interpolate
!                         denom = x(i)-x(last)    # non-zero - proof?
!                         for(j=last+1; j<i; j=j+1){
!                                 alpha = (x(j)-x(last))/denom
!                                 ys(j) = alpha*ys(i)+(1.0-alpha)*ys(last)
!                                 }
!                         }
!                 last = i        # last point actually estimated
!                 cut = x(last)+delta     # x coord of close points
!                 for(i=last+1; i<=n; i=i+1){     # find close points
!                         if (x(i)>cut) break     # i one beyond last pt within cut
!                         if(x(i)==x(last)){      # exact match in x
!                                 ys(i) = ys(last)
!                                 last = i
!                                 }
!                         }
!                 i=max0(last+1,i-1)
!  # back 1 point so interpolation within delta, but always go forward
!                 } until(last>=n)
!         do i = 1,n      # residuals
!                 res(i) = y(i)-ys(i)
!         if (iter>nsteps) break  # compute robustness weights except last time
!         do i = 1,n
!                 rw(i) = abs(res(i))
!         call sort(rw,n)
!         m1 = 1+n/2; m2 = n-m1+1
!         cmad = 3.0*(rw(m1)+rw(m2))      # 6 median abs resid
!         c9 = .999*cmad; c1 = .001*cmad
!         do i = 1,n {
!                 r = abs(res(i))
!                 if(r<=c1) rw(i)=1.      # near 0, avoid underflow
!                 else if(r>c9) rw(i)=0.  # near 1, avoid underflow
!                 else rw(i) = (1.0-(r/cmad)**2)**2
!                 }
!         }
!  return
!  end
! 
! 
! 
! 
!                                   LOWEST
! 
! 
! 
!        Calling sequence
! 
!        CALL LOWEST(X,Y,N,XS,YS,NLEFT,NRIGHT,W,USERW,RW,OK)
! 
!        Purpose
! 
!        LOWEST is a support routine for LOWESS and  ordinarily  will
!        not  be  called  by  the  user.   The  fitted  value, YS, is
!        computed  at  the  value,  XS,  of  the   horizontal   axis.
!        Robustness  weights,  RW,  can  be employed in computing the
!        fit.
! 
!        Argument description
! 
! 
!              X = Input; abscissas of the points on the
!                  scatterplot; the values in X must be ordered
!                  from smallest to largest.
!              Y = Input; ordinates of the points on the
!                  scatterplot.
!              N = Input; dimension of X,Y,W, and RW.
!             XS = Input; value of the horizontal axis at which the
!                  smooth is computed.
!             YS = Output; fitted value at XS.
!          NLEFT = Input; index of the first point which should be
!                  considered in computing the fitted value.
!         NRIGHT = Input; index of the last point which should be
!                  considered in computing the fitted value.
!              W = Output; W(I) is the weight for Y(I) used in the
!                  expression for YS, which is the sum from
!                  I = NLEFT to NRIGHT of W(I)*Y(I); W(I) is
!                  defined only at locations NLEFT to NRIGHT.
!          USERW = Input; logical variable; if USERW is .TRUE., a
!                  robust fit is carried out using the weights in
!                  RW; if USERW is .FALSE., the values in RW are
!                  not used.
!             RW = Input; robustness weights.
!             OK = Output; logical variable; if the weights for the
!                  smooth are all 0.0, the fitted value, YS, is not
!                  computed and OK is set equal to .FALSE.; if the
!                  fitted value is computed OK is set equal to
! 
! 
!        Method
! 
!        The smooth at XS is computed using (robust) locally weighted
!        regression of degree 1.  The tricube weight function is used
!        with h equal to the maximum of XS-X(NLEFT) and X(NRIGHT)-XS.
!        Two  cases  where  the  program  reverts to locally weighted
!        regression of degree 0 are described  in  the  documentation
!        for LOWESS.
! 
! 
! 
! 
!  subroutine lowest(x,y,n,xs,ys,nleft,nright,w,userw,rw,ok)
!  real x(n),y(n),w(n),rw(n)
!  logical userw,ok
!  range = x(n)-x(1)
!  h = amax1(xs-x(nleft),x(nright)-xs)
!  h9 = .999*h
!  h1 = .001*h
!  a = 0.0        # sum of weights
!  for(j=nleft; j<=n; j=j+1){     # compute weights (pick up all ties on right)
!         w(j)=0.
!         r = abs(x(j)-xs)
!         if (r<=h9) {    # small enough for non-zero weight
!                 if (r>h1) w(j) = (1.0-(r/h)**3)**3
!                 else      w(j) = 1.
!                 if (userw) w(j) = rw(j)*w(j)
!                 a = a+w(j)
!                 }
!         else if(x(j)>xs)break   # get out at first zero wt on right
!         }
!  nrt=j-1        # rightmost pt (may be greater than nright because of ties)
!  if (a<=0.0) ok = FALSE
!  else { # weighted least squares
!         ok = TRUE
!         do j = nleft,nrt
!                 w(j) = w(j)/a   # make sum of w(j) == 1
!         if (h>0.) {     # use linear fit
!                 a = 0.0
!                 do j = nleft,nrt
!                         a = a+w(j)*x(j) # weighted center of x values
!                 b = xs-a
!                 c = 0.0
!                 do j = nleft,nrt
!                         c = c+w(j)*(x(j)-a)**2
!                 if(sqrt(c)>.001*range) {
!  # points are spread out enough to compute slope
!                         b = b/c
!                         do j = nleft,nrt
!                                 w(j) = w(j)*(1.0+b*(x(j)-a))
!                         }
!                 }
!         ys = 0.0
!         do j = nleft,nrt
!                 ys = ys+w(j)*y(j)
!         }
!  return
!  end
! 
! 
! 
!  test driver for lowess
!  for expected output, see introduction
!     real x(20), y(20), ys(20), rw(20), res(20)
!     data x /1,2,3,4,5,10*6,8,10,12,14,50/
!     data y /18,2,15,6,10,4,16,11,7,3,14,17,20,12,9,13,1,8,5,19/
!     call lowess(x,y,20,.25,0,0.,ys,rw,res)
!     write(6,*) ys
!     call lowess(x,y,20,.25,0,3.,ys,rw,res)
!     write(6,*) ys
!     call lowess(x,y,20,.25,2,0.,ys,rw,res)
!     write(6,*) ys
!     end
!**************************************************************
!  Fortran output from ratfor
!
      subroutine lowess(x, y, n, f, nsteps, delta, ys, rw, res)
      integer(kind=int32) ::  n
      integer(kind=int32) ::  nsteps
      real(kind=double) ::  x(n), y(n), f, delta, ys(n), rw(n)
      real(kind=double) ::  res(n)
      integer(kind=int32) ::  nright!, min0, max0, 
      integer(kind=int32) :: i, j!, ifix
      integer(kind=int32) ::  iter, last, m1, m2, ns, nleft
!      real(kind=double) :: abs, cut, cmad, r, d1, d2
      real(kind=double) :: cut, cmad, r, d1, d2
      real(kind=double) ::  c1, c9, alpha, denom
      logical ok
      if (n .ge. 2) goto 1
         ys(1) = y(1)
         return
! at least two, at most n points
!   1  ns = max(min(ifix(f*float(n))), n), 2)
   1  ns = max(min(int(f*float(n)), n), 2)
      iter = 1
         goto  3
   2     iter = iter+1
   3     if (iter .gt. nsteps+1) goto  22
! robustness iterations
         nleft = 1
         nright = ns
! index of prev estimated point
         last = 0
! index of current point
         i = 1
   4        if (nright .ge. n) goto  5
! move nleft, nright to right if radius decreases
               d1 = x(i)-x(nleft)
! if d1<=d2 with x(nright+1)==x(nright), lowest fixes
               d2 = x(nright+1)-x(i)
               if (d1 .le. d2) goto  5
! radius will not decrease by move right
               nleft = nleft+1
               nright = nright+1
               goto  4
! fitted value at x(i)
   5        call lowest(x, y, n, x(i), ys(i), nleft, nright, res, iter &
                 .gt. 1, rw, ok)
            if (.not. ok) ys(i) = y(i)
! all weights zero - copy over value (all rw==0)
            if (last .ge. i-1) goto 9
               denom = x(i)-x(last)
! skipped points -- interpolate
! non-zero - proof?
               j = last+1
                  goto  7
   6              j = j+1
   7              if (j .ge. i) goto  8
                  alpha = (x(j)-x(last))/denom
                  ys(j) = alpha*ys(i)+(1.0-alpha)*ys(last)
                  goto  6
   8           continue
! last point actually estimated
   9        last = i
! x coord of close points
            cut = x(last)+delta
            i = last+1
               goto  11
  10           i = i+1
  11           if (i .gt. n) goto  13
! find close points
               if (x(i) .gt. cut) goto  13
! i one beyond last pt within cut
               if (x(i) .ne. x(last)) goto 12
                  ys(i) = ys(last)
! exact match in x
                  last = i
  12           continue
               goto  10
! back 1 point so interpolation within delta, but always go forward
  13        i = max(last+1, i-1)
  14        if (last .lt. n) goto  4
! residuals
         do  15 i = 1, n
            res(i) = y(i)-ys(i)
  15        continue
         if (iter .gt. nsteps) goto  22
! compute robustness weights except last time
         do  16 i = 1, n
            rw(i) = abs(res(i))
  16        continue
!         call sort(rw, n)
         call ssort(n,rw)
         m1 = n/2+1
         m2 = n-m1+1
! 6 median abs resid
         cmad = 3.0*(rw(m1)+rw(m2))
         c9 = .999*cmad
         c1 = .001*cmad
         do  21 i = 1, n
            r = abs(res(i))
            if (r .gt. c1) goto 17
               rw(i) = 1.
! near 0, avoid underflow
               goto  20
  17           if (r .le. c9) goto 18
                  rw(i) = 0.
! near 1, avoid underflow
                  goto  19
  18              rw(i) = (1.0-(r/cmad)**2)**2
  19        continue
  20        continue
  21        continue
         goto  2
  22  return
      end subroutine lowess


      subroutine lowest(x, y, n, xs, ys, nleft, nright, w, userw &
           , rw, ok)
      integer(kind=int32) ::  n
      integer(kind=int32) ::  nleft, nright
      real(kind=double) :: x(n), y(n), xs, ys, w(n), rw(n)
      logical :: userw, ok
      integer(kind=int32) ::  nrt, j
      real(kind=double) ::  a, b, c, h, r
      real(kind=double) ::  h1, h9
!      real(kind=double) ::  amax1
      real(kind=double) ::  range
      range = x(n)-x(1)
      h = max(xs-x(nleft), x(nright)-xs)
      h9 = .999*h
      h1 = .001*h
! sum of weights
      a = 0.0
      j = nleft
         goto  2
   1     j = j+1
   2     if (j .gt. n) goto  7
! compute weights (pick up all ties on right)
         w(j) = 0.
         r = abs(x(j)-xs)
         if (r .gt. h9) goto 5
            if (r .le. h1) goto 3
               w(j) = (1.0-(r/h)**3)**3
! small enough for non-zero weight
               goto  4
   3           w(j) = 1.
   4        if (userw) w(j) = rw(j)*w(j)
            a = a+w(j)
            goto  6
   5        if (x(j) .gt. xs) goto  7
! get out at first zero wt on right
   6     continue
         goto  1
! rightmost pt (may be greater than nright because of ties)
   7  nrt = j-1
      if (a .gt. 0.0) goto 8
         ok = .false.
         goto  16
   8     ok = .true.
! weighted least squares
         do  9 j = nleft, nrt
! make sum of w(j) == 1
            w(j) = w(j)/a
   9        continue
         if (h .le. 0.) goto 14
            a = 0.0
! use linear fit
            do  10 j = nleft, nrt
! weighted center of x values
               a = a+w(j)*x(j)
  10           continue
            b = xs-a
            c = 0.0
            do  11 j = nleft, nrt
               c = c+w(j)*(x(j)-a)**2
  11           continue
            if (sqrt(c) .le. .001*range) goto 13
               b = b/c
! points are spread out enough to compute slope
               do  12 j = nleft, nrt
                  w(j) = w(j)*(b*(x(j)-a)+1.0)
  12              continue
  13        continue
  14     ys = 0.0
         do  15 j = nleft, nrt
            ys = ys+w(j)*y(j)
  15        continue
  16  return
      end subroutine lowest

subroutine ssort(n,ra)
  integer(kind=int32) :: n
  real(kind=double) :: ra(*)
  integer(kind=int32) ::  i,j,l,ir
  real(kind=double) :: rra

  l=n/2+1
  ir=n

do 
  if(l.gt.1)then
     l=l-1
     rra=ra(l)
  else
     rra=ra(ir)
     ra(ir)=ra(1)
     ir=ir-1
     if(ir.le.1)then
        ra(1)=rra
        EXIT
     endif
!     ir=ir-1
  endif
  i=l
  j=l+l
  do while (j.le.ir)
     if(j.lt.ir)then
        if(ra(j).lt.ra(j+1))j=j+1
     endif
     if(rra.lt.ra(j))then
        ra(i)=ra(j)
        i=j
        j=j+j
     else
        j=ir+1
     endif
  enddo
  ra(i)=rra
enddo
end subroutine ssort

! Mrqmin from F90
subroutine mrqmin9(x,y,sig,a,maska,covar,alpha,chisq,funcs,alamda, verbFlag)

  !use nr, only : covsrt,gaussj
  implicit none
  real(kind=double), dimension(:), intent(in) :: x,y,sig
  real(kind=double), dimension(:), intent(inout) :: a
  real(kind=double), dimension(:,:), intent(out) :: covar, alpha
  real(kind=double), intent(out) :: chisq
  real(kind=double), intent(inout) :: alamda
  logical, dimension(:), intent(in) :: maska
  integer, intent(in) :: verbFlag
  interface
     subroutine funcs(x,a,yfit,dyda)
       !use nrtype
       double precision, dimension(:), intent(in) :: x
       double precision, dimension(:), intent(inout) :: a
       double precision, dimension(:), intent(out) :: yfit
       double precision, dimension(:,:), intent(out) :: dyda
     end subroutine funcs
  end interface
  integer(kind=int32) :: ma,ndata
  integer(kind=int32), save :: mfit
  !write(*,*) " L6470 ma=",ma, " mfit=", mfit
  call mrqmin9_private
contains

  subroutine mrqmin9_private
    real(kind=double), save :: ochisq
    real(kind=double), dimension(:), allocatable, save :: atry, beta
    real(kind=double), dimension(:,:), allocatable, save :: da
    ndata=assert_eq3(size(x), size(y),size(sig),'mrqmin9: ndata')
    write(*,*) " L6479 size a=", size(a), " maska=", size(maska), &
         " covar1=", size(covar,1), " covar2=", size(covar,2)
    write(*,*) " alpha1=", size(alpha,1), " alpha2=", size(alpha,2)

    ma=assert_eqn((/size(a),size(maska),size(covar,1),size(covar,2), &
         size(alpha,1),size(alpha,2)/),'mrqmin9: ma')
    write(*,*) " L6485 ma=", ma
    mfit=count(maska)
    write(*,*) " L6487 mfit=", mfit
    write(*,*) " MRQ9 L6462 verbFlag=", verbFlag
    write(*,*) " MRQ9 a=", a
    if (alamda < 0.0) then
       write (*,*) " MRQ9 L6467 initialising (alamda<0)"
       allocate(atry(ma),beta(ma),da(ma,1))
       alamda=0.001_double
       call mrqcof9(a,alpha,beta)
       write(*,*) " L6495"
       ochisq=chisq
       atry=a
    end if
    covar(1:mfit,1:mfit)=alpha(1:mfit,1:mfit)
    write(*,*) " L6500"
    call diagmult_r(covar(1:mfit,1:mfit),1.0_double+alamda)
    ! Alter linearized fitting matrix, by augmenting diagonal elements
    da(1:mfit,1)=beta(1:mfit)
    write(*,*) " L6504"
    call gaussj9(covar(1:mfit,1:mfit),da(1:mfit,1:1)) ! matrix solution
    if (alamda == 0.0) then
       write(*,*) " L6507"
       call covsrt9(covar,maska)
       call covsrt9(alpha,maska)
       deallocate(atry,beta,da)
       return
    endif
    write(*,*) " L6513"
    atry=a+unpack(da(1:mfit,1),maska,0.0_double) ! did the trial succeed?
    write(*,*) " L6515"
    call mrqcof9(atry,covar,da(1:mfit,1))
    write(*,*) " L6517"
    if (chisq < ochisq) then ! success, accept the new solution
       alamda=0.1_double*alamda
       ochisq=chisq
       alpha(1:mfit,1:mfit)=covar(1:mfit,1:mfit)
       beta(1:mfit)=da(1:mfit,1)
       a=atry
    else ! failure, increase alamda and return
       alamda=10.0_double*alamda
       chisq=ochisq
    endif
  end subroutine mrqmin9_private
  !
  subroutine mrqcof9(a,alpha,beta)
    real(kind=double), dimension(:), intent(inout) :: a
    real(kind=double), dimension(:), intent(out) :: beta
    real(kind=double), dimension(:,:), intent(out) :: alpha
    integer(kind=int32) :: j,k,l,m
    real(kind=double), dimension(size(x),size(a)) :: dyda
    real(kind=double), dimension(size(x)) :: dy,sig2i,wt,ymod
    !write(*,*) " MRQCOF9 L6506 a=", a
    !write(*,*) " L6515 dyda=", dyda(1,:)
    call funcs(x,a,ymod,dyda)
    !write(*,*) " L6526 ma=", ma 
    sig2i=1.0_double/(sig**2)
    !write(*,*) " L6528"
    dy=y-ymod
    !write(*,*) " L6530"
    j=0
    do l=1,ma
       !write (*,*) " L6540 l=", l
       if (maska(l)) then
          j=j+1
          !write(*,*) "       L6543  j=", j
          wt=dyda(:,l)*sig2i
          k=0
          do m=1,l
             !write(*,*) "        L6547 m=",m
             if (maska(m)) then
                k=k+1
                !write (*,*) "      L6550 k=", k
                alpha(j,k)=dot_product(wt,dyda(:,m))
                alpha(k,j)=alpha(j,k) ! fill in the symmetric side
             endif
          enddo
          beta(j)=dot_product(dy,wt)
          !write(*,*) " L6556 beta(j)=", beta(j)
       endif
    enddo
    chisq=dot_product(dy**2,sig2i) ! find chi squared
  end subroutine mrqcof9
end subroutine mrqmin9
!
subroutine gaussj9(a,b)
  !use nrtype; use nrutil, only :: assert_eq,nrerror,outerand,outerprod,swap
  implicit none
  real(kind=double), dimension(:,:), intent(inout) :: a,b
  ! Linear equation solution by Gauss-Jordan elimination
  integer(kind=int32), dimension(size(a,1)) :: ipiv,indxr,indxc
  logical, dimension(size(a,1)) :: lpiv
  real(kind=double) :: pivinv
  real(kind=double), dimension(size(a,1)) :: dumc
  integer(kind=int32), target :: irc(2)
  integer(kind=int32) :: i,l,n
  integer(kind=int32), pointer :: irow,icol
  write(*,*) " GAUSSJ9 L6572 size(a,1)=", size(a,1), " size(a,2)=", size(a,2), &
       " size(b,1)=", size(b,1), " size(b,2)=", size(b,2)
  n=assert_eq3(size(a,1),size(a,2),size(b,1),'gaussj9')
  write(*,*) " GAUSSJ9 L6575 n=", n
  irow => irc(1)
  icol => irc(2)
  ipiv=0
  
  do i=1,n ! main lopp over columns to be reduced
     write(*,*) " gaussj9 L6581 i=", i
     lpiv = (ipiv == 0) ! begin search for a pivot element
     irc=maxloc(abs(a),outerand(lpiv,lpiv))
     write(*,*) " L6584 irc=", irc
     ipiv(icol)=ipiv(icol)+1
     if (ipiv(icol) >1) write(*,*) "gaussj9: singular matrix (1, L6297)"
     if (irow /= icol) then
        call swap_rv(a(irow,:), a(icol,:))
        call swap_rv(b(irow,:),b(icol,:))
     endif
     indxr(i)=irow ! we are now ready to divode the pivot row by 
     indxc(i)=icol ! the pivotelement, located at irow and icol
     write(*,*) " L6593"
     if (a(icol,icol) == 0.0) &
          write(*,*) " ERROR gauss: singular matrix (2, L6305)"
     pivinv=1.0_double/a(icol,icol)
     write(*,*) " L6597"    
     a(icol,icol)=1.0
     a(icol,:)=a(icol,:)*pivinv
     b(icol,:)=b(icol,:)*pivinv
     write(*,*) " L6601"
     dumc=a(:,icol) ! next we reduce the rows, except for the 
     a(:,icol)=0.0  ! pivot one, of course
     a(icol,icol)=pivinv
     write(*,*) " L6605"
     a(1:icol-1,:)=a(1:icol-1,:)-outerprod(dumc(1:icol-1),a(icol,:))
     write(*,*) " L6607"
     b(1:icol-1,:)=b(1:icol-1,:)-outerprod(dumc(1:icol-1),b(icol,:))
     write(*,*) " L6609"
     a(icol+1:,:)=a(icol+1:,:)-outerprod(dumc(icol+1:),a(icol,:))
     write(*,*) " L6611"
     b(icol+1:,:)=b(icol+1:,:)-outerprod(dumc(icol+1:),b(icol,:))
  enddo
  write(*,*) " L6608"
  ! it only remains to unscramble the solution in view of the 
  ! column interchange. We do this by interchanging pairs of 
  ! columns in the reverse order thet the permutation was 
  ! built up. 
  do l=n,1,-1
     call swap_rv(a(:,indxr(l)),a(:,indxc(l)))
  enddo
  write(*,*) " L6616"
end subroutine gaussj9

subroutine diagmult_rv(mat,diag)
  ! Multiplies vector or scalar diag into the diagonal of matrix mat
  real(kind=double), dimension(:,:), intent(inout) :: mat
  real(kind=double), dimension(:), intent(in) :: diag
  integer(kind=int32) :: j,n
  n=assert_eq2(size(diag), min(size(mat,1),size(mat,2)),'diagmult_rv')
  do j=1,n
     mat(j,j)=mat(j,j)*diag(j)
  enddo
end subroutine diagmult_rv

subroutine diagmult_r(mat,diag)
  ! Multiplies vector or scalar diag into the diagonal of matrix mat
  real(kind=double), dimension(:,:), intent(inout) :: mat
  real(kind=double),  intent(in) :: diag
  integer(kind=int32) :: j,n
  n=min(size(mat,1),size(mat,2))
  do j=1,n
     mat(j,j)=mat(j,j)*diag
  enddo
end subroutine diagmult_r

function assert_eq2(n1,n2,string)
  ! report and die if integers not all equal (used for size checking)
  character(len=*), intent(in) :: string
  integer, intent(in) :: n1,n2
  integer :: assert_eq2
  if (n1 == n2) then
     assert_eq2=n1
  else
     write(*,*) 'nrerror: an assert_eq failed with this tag:', &
          string
     stop 'programme terminated by assert_eq2'
  endif
end function assert_eq2

subroutine covsrt9(covar,maska)
  implicit none
  real(kind=double), dimension(:,:), intent(inout) :: covar
  logical, dimension(:), intent(in) :: maska
  ! expand in storage the covariance matrix covar to take
  ! into account parameters that are being held fixed
  ! (for the latter return zero covariances)
  integer(kind=int32) :: ma, mfit,j,k
  ma=assert_eq3(size(covar,1),size(covar,2),size(maska),'covsrt')
  mfit=count(maska)
  covar(mfit+1:ma,1:ma)=0.0
  covar(1:ma,mfit+1:ma)=0.0
  k=mfit
  do j=ma,1,-1
     if (maska(j)) then
        call swap_rv(covar(1:ma,k),covar(1:ma,j))
        call swap_rv(covar(k,1:ma),covar(j,1:ma))
        k=k-1
     endif
  enddo
end subroutine covsrt9

subroutine swap_rv(a,b)
  real(kind=double), dimension(:), intent(inout) :: a,b
  real(kind=double), dimension(size(a)) :: dum
  dum=a
  a=b
  b=dum
end subroutine swap_rv

function outerand(a,b)
  logical, dimension(:), intent(in) :: a,b
  logical, dimension(size(a),size(b)) :: outerand
  outerand= spread(a,dim=2,ncopies=size(b)) .and. &
       spread(b,dim=1,ncopies=size(a))
end function outerand

function outerprod(a,b)
  real(kind=double), dimension(:), intent(in) :: a,b
  real(kind=double), dimension(size(a),size(b)) :: outerprod
  !write(*,*) " outerprod L6701"
  !write(*,*) " L6702 spread(a)=", spread(a,dim=2,ncopies=size(b))
  !write(*,*) " L6703 spread(b)=", spread(b,dim=1,ncopies=size(a))
  outerprod=spread(a,dim=2,ncopies=size(b)) * &
       spread(b,dim=1,ncopies=size(a))
end function outerprod

function assert_eq3(n1,n2,n3,string)
  character(len=*), intent(in) :: string
  integer, intent(in) :: n1,n2,n3
  integer :: assert_eq3
  if(n1 == n2 .and. n2 == n3) then
     assert_eq3=n1
  else
     write(*,*) 'nrerror: an assert-eq failed with this tag:', &
          string
     stop 'programme terminated by assert_eq3'
  endif
end function assert_eq3


function assert_eqn(nn,string)
  character(len=*), intent(in) :: string
  integer, dimension(:), intent(in) :: nn
  integer :: assert_eqn
  if (all(nn(2:) == nn(1))) then
     assert_eqn=nn(1)
  else
     write(*,*) 'nrerror: an assert_eq failed with this tag:', &
          string
     stop ' programme terminated by assert_eqn'
  endif
end function assert_eqn

!-----------------------------------------------
! Marquardt-code from Simon
!-----------------------------------------------
! Levenberg-Marquardt code (uses mrqcof, gaussj and covsrt and the 
! supplied function, funcs). Note adaptation to pass the extra 
! information (i.e. x and y pixel coordinates, not just x.
!-----------------------------------------------

    SUBROUTINE mrqminSim(x,y,sig,ndata,a,ia,ma,covar,alpha,nca,chisq,&
     &alamda)

      !INTEGER*4 ma,nca,ndata
      integer(kind=int32) ::  ma,nca,ndata
      !integer*4 ia(ma),MMAX
      integer(kind=int32) :: ia(ma),MMAX
      !REAL*8 alamda,chisq,a(ma),alpha(nca,nca),covar(nca,nca)
      real(kind=double) :: alamda,chisq,a(ma),alpha(nca,nca),covar(nca,nca)
      PARAMETER (MMAX=20)
!    USES covsrt,gaussj,mrqcof
!      INTEGER*4 j,k,l,m,mfit
      integer(kind=int32) :: j,k,l,m,mfit
      !REAL*8 ochisq,atry(5),beta(5),da(5)
      real(kind=double) :: ochisq,atry(5),beta(5),da(5)
      SAVE ochisq,atry,beta,da,mfit

      real(kind=double) :: x(ndata), y(ndata), sig(ndata)
!      real, pointer :: x(:), y(:), sig(:)

!      allocate(x(1:ndata))
!      allocate(y(1:ndata))
!      allocate(sig(1:ndata))

      if(alamda.lt.0.)then
        mfit=0
        do 11 j=1,ma
          if (ia(j).ne.0) mfit=mfit+1
11      continue
        alamda=0.001
        call mrqcofSim(x,y,sig,ndata,a,ia,ma,alpha,beta,nca,chisq)

        ochisq=chisq
        do 12 j=1,ma
          atry(j)=a(j)
12      continue
      endif
      j=0
      do 14 l=1,ma
        if(ia(l).ne.0) then
          j=j+1
          k=0
          do 13 m=1,ma
            if(ia(m).ne.0) then
              k=k+1
              covar(j,k)=alpha(j,k)
            endif
13        continue
          covar(j,j)=alpha(j,j)*(1.+alamda)
          da(j)=beta(j)
        endif
14    continue
      call gaussjSim(covar,mfit,nca,da,1,1)
      if(alamda.eq.0.)then
        call covsrtSim(covar,nca,ma,ia,mfit)
        return
      endif
      j=0
      do 15 l=1,ma
        if(ia(l).ne.0) then
          j=j+1
          atry(l)=a(l)+da(j)
        endif
15    continue
      call mrqcofSim(x,y,sig,ndata,atry,ia,ma,covar,da,nca,chisq)
      if(chisq.lt.ochisq)then
        alamda=0.1*alamda
        ochisq=chisq
        j=0
        do 17 l=1,ma
          if(ia(l).ne.0) then
            j=j+1
            k=0
            do 16 m=1,ma
              if(ia(m).ne.0) then
                k=k+1
                alpha(j,k)=covar(j,k)
              endif
16          continue
            beta(j)=da(j)
            a(l)=atry(l)
          endif
17      continue
      else
        alamda=10.*alamda
        chisq=ochisq
      endif

!      deallocate(x)
!      deallocate(y)
!      deallocate(sig)

      return
      end subroutine mrqminSim

      SUBROUTINE mrqcofSim(x,y,sig,ndata,a,ia,ma,alpha,beta,nalp,chisq)

      integer(kind=int32) :: ma,nalp,ndata
      integer(kind=int32) :: ia(ma)
      real(kind=double) :: chisq,a(ma),alpha(nalp,nalp),beta(5), xv!, yv
      !EXTERNAL funcsSim
      integer(kind=int32) ::  mfit,i,j,k,l,m!,kk
      real(kind=double) :: dz,sig2i, wt, zmod, dzda(5)!, fp(5), fm(5), delt(5)
!      real*8 x(1000), y(1000), sig(1000)
      real(kind=double) :: x(ndata), y(ndata), sig(ndata)

!      real, pointer :: x(:), y(:), sig(:)

      mfit=0

!      allocate(x(1:ndata))
!      allocate(y(1:ndata))
!      allocate(sig(1:ndata))

      do 11 j=1,ma
        if (ia(j).ne.0) mfit=mfit+1
11    continue
      do 13 j=1,mfit
        do 12 k=1,j
          alpha(j,k)=0.
12      continue
        beta(j)=0.
13    continue
      chisq=0.

     do 16 i=1,ndata

        xv=x(i)
        !call funcsSim(xv,ma,a,zmod,dzda)
        call fgausslSim(xv,ma,a,zmod,dzda)

        sig2i=1./(sig(i)*sig(i))
        dz=y(i)-zmod

        j=0
        do 15 l=1,ma
          if(ia(l).ne.0) then
            j=j+1
            wt=dzda(l)*sig2i
            k=0
            do 14 m=1,l
              if(ia(m).ne.0) then
                k=k+1
                alpha(j,k)=alpha(j,k)+wt*dzda(m)
              endif
14          continue
            beta(j)=beta(j)+dz*wt
          endif
15      continue
        chisq=chisq+dz*dz*sig2i
16    continue

      do 18 j=2,mfit
        do 17 k=1,j-1
          alpha(k,j)=alpha(j,k)
17      continue
18    continue
          
!          deallocate(x)
!          deallocate(y)
!          deallocate(sig)

      return
      end SUBROUTINE mrqcofSim


      SUBROUTINE covsrtSim(covar,npc,ma,ia,mfit)

      integer(kind=int32) :: ma,mfit,npc,ia(ma)
      real(kind=double) :: covar(npc,npc)
      integer(kind=int32) :: i,j,k
      real(kind=double) :: swap
      do 12 i=mfit+1,ma
        do 11 j=1,i
          covar(i,j)=0.
          covar(j,i)=0.
11      continue
12    continue
      k=mfit
      do 15 j=ma,1,-1
        if(ia(j).ne.0)then
          do 13 i=1,ma
            swap=covar(i,k)
            covar(i,k)=covar(i,j)
            covar(i,j)=swap
13        continue
          do 14 i=1,ma
            swap=covar(k,i)
            covar(k,i)=covar(j,i)
            covar(j,i)=swap
14        continue
          k=k-1
        endif
15    continue
      return
      end SUBROUTINE covsrtSim


      SUBROUTINE gaussjSim(a,n,np,b,m,mp)
      integer(kind=int32) :: m,mp,n,np,NMAX
      real(kind=double) :: a(np,np),b(np,mp)
      PARAMETER (NMAX=50)
      integer(kind=int32) :: i,icol,irow,j,k,l,ll,indxc(NMAX),indxr(NMAX),ipiv(NMAX)
      real(kind=double) :: big,dum,pivinv
      do 11 j=1,n
        ipiv(j)=0
11    continue
      do 22 i=1,n
        big=0.
        do 13 j=1,n
          if(ipiv(j).ne.1)then
            do 12 k=1,n
              if (ipiv(k).eq.0) then
                if (abs(a(j,k)).ge.big)then
                  big=abs(a(j,k))
                  irow=j
                  icol=k
                endif
              else if (ipiv(k).gt.1) then
                write(*,*) 'L8682 singular matrix in gaussj '
              endif
12          continue
          endif
13      continue
        ipiv(icol)=ipiv(icol)+1
        if (irow.ne.icol) then
          do 14 l=1,n
            dum=a(irow,l)
            a(irow,l)=a(icol,l)
            a(icol,l)=dum
14        continue
          do 15 l=1,m
            dum=b(irow,l)
            b(irow,l)=b(icol,l)
            b(icol,l)=dum
15        continue
        endif
        indxr(i)=irow
        indxc(i)=icol
        if (a(icol,icol).eq.0.) write(*,*) 'L8702 singular matrix in gaussj a=0'
        pivinv=1./a(icol,icol)
        a(icol,icol)=1.
        do 16 l=1,n
          a(icol,l)=a(icol,l)*pivinv
16      continue
        do 17 l=1,m
          b(icol,l)=b(icol,l)*pivinv
17      continue
        do 21 ll=1,n
          if(ll.ne.icol)then
            dum=a(ll,icol)
            a(ll,icol)=0.
            do 18 l=1,n
              a(ll,l)=a(ll,l)-a(icol,l)*dum
18          continue
            do 19 l=1,m
              b(ll,l)=b(ll,l)-b(icol,l)*dum
19          continue
          endif
21      continue
22    continue
      do 24 l=n,1,-1
        if(indxr(l).ne.indxc(l))then
          do 23 k=1,n
            dum=a(k,indxr(l))
            a(k,indxr(l))=a(k,indxc(l))
            a(k,indxc(l))=dum
23        continue
        endif
24    continue
      return
      END SUBROUTINE gaussjSim

! subroutine to compute the model function. Needs derivatives. Since we 
! are integrating function over arbitrary sized pixels, must get these 
! numerically using finite differences.
 
      subroutine funcsSim(x,ma,a,zfit,dzda)

      integer(kind=int32) :: ma
      real(kind=double) :: x, a(ma), zfit, dzda(4)!, sigma, astore(ma)
      real(kind=double) :: off, amp, t0, p!, r, vpix, gauss
      real(kind=double) :: arg !, ys, err, pixv, delta_a, alpha, xs
      real(kind=double) :: pi !,ang, den, delt(4), fp(4), fm(4) 

      !integer(kind=int32) ::  k, l, nsub, ix, jy, i, j, nsubpix,
      !integer(kind=int32) ::  jj, ii


      pi=3.141592654d0

! set up parameters

      off=a(1)
      amp=a(2)
      t0=a(3)
      p=a(4)


! For the given data pixel, integrate function over it by dividing the pixel
! into nsub sub pixels, summing the function over these and then getting
! the average value. 

      arg=2.d0*pi*(x-t0)/p

      zfit = off + amp*sin(arg)


! and determine gradient

      dzda(1)=1.d0
      dzda(2)=sin(arg)
      dzda(3)=(-2.d0*pi*amp/p) * cos(arg)
      dzda(4)=(-2.d0*pi*(x-t0)*amp/(p*p)) * cos(arg)


      return
    end subroutine funcsSim


! Superposition of the Gaussian and a linear function
subroutine fgausslSim(x, np, p, y, dydp)
! The function is: y=p(1) + p(5)*(x-p(3)) + p(2)*exp(-((x-p(3))/p(4))**2)
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac

  y=0.0
  dydp=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  y=p(1)+p(2)*ex +p(5)*x
  dydp(1)=1.0
  dydp(2)=ex

  dydp(3)=fac/p(4)
  dydp(4)=fac*arg/p(4)
  dydp(5)=x-p(3)

end subroutine fgausslSim


! Gaussian function
subroutine fgaussupSim(x, np, p, y, dydp)

  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac
  integer(kind=int32) :: i

  y=0.0
  dydp=0.0
  do i=1,np-1,3
     if (p(i+2) .eq. 0.0) p(i+2)=1.0
     arg=(x-p(i+1))/p(i+2)
     ex=exp(-arg*arg)
     fac=p(i)*ex*2.0*arg
     y=y+p(i)*ex
     dydp(i)=ex
     dydp(i+1)=fac/p(i+2)
     dydp(i+2)=fac*arg/p(i+2)
  enddo
  return
end subroutine fgaussupSim


end module omgextraction_mod








