!+
! Module: omgrism_mod
!
! Authors:  Vladimir Yershov - MSSL
!          
! Description: Constructs an OM OSW FITS source spectra from the grism 
! image and source list  files produced by using omdetect
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!
! History:
! 
!    2003.10.27 VNY -  First version
!    2003.10.31, 2003.11.07 VNY - V0.2 Parameters for spectra regions 
!    are passed from omgchain; parameters for image smoothing removed
!    2003.11.10 V0.3 two lines exceeding 132 characters are shorten
!    2003.11.11 V0.4 a keyword (COMMENT) with the spectrum filename is added
!                    to the source-list table (of the input source list file) 
!    12.11.2003 VNY Checking the headerss of the output spectrum file 
!               Applying the input parameters for the spectrum and background 
!               extraction regions
!    17.11.2003 VNY making extraction with a separated routine
!    28.11.2003 VNY set EFLAG to 2 if it corresponds to a genuine spectrum
!    02.12.2003 VNY the parameters src2spectrum and spectrumlength are removed as misleading
!    04.12.2003 VNY extraction of the undetected spectra is included
!    09.12.2003 VNY The spectra labels in the source-list file are changed: 
!                   now they correspod to the zero-order spectra 
!    07.01.2004 VNY Checks for optimal extraction 
!    19.01.2004 VNY Fixing the mistaken coordinates of zero- and first-order 
!               images in the headers of the spectra tables (SSC-SPR-3220)
!    26.01.2004 VNY extractionmode parameter is activated for the values 0 and 2
!    28.01.2004 - 01.02.2004 VNY adding a new output file - SPECLI - 
!                 list of the extracted spectra 
!    02.02.2004 VNY removing the column ORIGIN_ID from the table SPECLI
!    04.02.2004 VNY Checking the possibility of flagging those sources 
!    that fall inside of the spectra areas
!    06.02.2004 VNY Creating quality flags for the spectra based on 
!    solely zero-orders
!    12.02.2004 VNY Selectrion criteria for spectra are improved 
!    19.02.2004 VNY Protected from crushing when called by omgsource
!    23.02.2004 VNY a small bug removed (the task was always in
!                   the regime for omgsource)
!    03.03.2004 VNY Rejection of spurious spectra improved
!    29.03.2004 VNY Check if the zero order found for a not-related
!                   first-order belongs to another first-order
!    01.04.2004 VNY removing lateral spectra
!    20.04.2004 VNY computing zero-order coordinates with different methods
!    04.05.2004 VNY Temporarily restoring the wrong algorithm of computing
!                   the binning factors (to make it compatible with the imaging tasks
!    10.05.2004 VNY removing the keywords INNBCKAN and OUTBCKAN from the header of 
!                   the source list file
!    02.06.2004 VNY Use OAL for getting the prime_instrument value
!    20.07.2004 VNY Changing the names to the columns that are added to 
!                   the (input) SRCLST file: 
!                   SPECTR_ID column is renamed by SPECTR_NO
!                   ORIGIN column is renamed by SPECTR_ID
!    21.07.2004 VNY For the time being the procedure for recomputing zero-orders
!                   is switched off because the calibration was made with the 
!                   coordinates given by omdetect 
!    21.07.2004 VNY Splitting the output files into the product (for the 
!                   target) and intermediate files (for the rest of the 
!                   spectra); the intermediate files will start with "g"
!    26.07.2004 VNY Introducing new parameters to cover the names of all output
!                   files (not just to generate them automatically):
!                   1) outspectralistfile
!                   2) extractfieldspectra
!                   3) regionfile
!
!    02.09.2004 VNY CONTENT header keyword  "OM GRISM TARGET SPECTRUM" is replaced with
!                   "OM GRISM SPECTRA" in order to match the check made in omgrismplot-1.0.4
!                   Set the zero-order spectrum region vertical size to the value of its 
!                   horizontal size
!   06.09.2004 VNY The use of the summary file is modified. Now in order to get the primary 
!                  instrument information the task is searchin both SUM.SAS and SUM.ASC files
!                  (for the case the SUM.SAS file is missing in the ODF directory)
!        
! 08.09.2004 VNY  Interpolation of the wavelengths and fluxes at the end of the flux 
! calibration array (omgextraction_mod.f90) is improved in order to avoid 
! missing points at the end of the output flux table
!       
! 13.09.2004 VNY Checking for the possible shift of the zero-order
! 15.09.2004 VNY check of the image binning in GetTargetPosition is improved
! 01.10.2004 VNY the elimination of zero-orders that overlap the extracted 
!            spectra is not implemented anymore since it leads to 
!            information losses (SSC-SPR-3396)
! 14.10.2004 VNY (Vilspa) Correctin the position of the first-order region (in many cases the first-order
! spectra are shifted to the left with respect to the zero-orders)
! 18.10.2004 VNY (Vilspa) The bug related to the missed target on the full-frame
! low-resolution images is fixed
! 25.01.2005 VNY The limits of the background interpolation region across the spectrum are 
! corrected (SSC-SPR-3433), module omgextraction_mod.f90 
! 25.04.2005 VNY The number of warning messages about the contaminating zero-orders is reduced
! 06.09.2005 VNY Changing the reference point in the zero-order from maximum to centroid
!  (SPR-3515)
! 18.10.1005 VNY protected agains substript in the case of very small windows;
!            a warning message is issued in the case the spectrum falls out of the 
!            window (SPR-3526)
! 25.10.2005 (VNY) a bug introduced on the previous upload is fixed (it resulted in
!            zero coordinate for the X-position of the zero-order spectrum) 
! 01.12.2005 (VNY) Labelling of target/default extraction is slightly changed:
!            now the source is defined as "default extraction" only when the 
!            signal-to-noise ration is small for both the first and zero orders
!            (before it was checking only the first-order)
! 03.07.2006 VNY A slightly more complicated analysis of the target position and 
!            brightness is introduced in order to prevent missclassification of obvious  
!            bright targets (they sometimes are classified as "default extraction" although
!            being correctly extracted) - SSC-SPR-3617
! 09.11.2006 VNY Adjusted to the changes with respect to the more accurate grism source 
!            positions (SSC-SPR-3633)
! 27.11.2006 VNY QFLAG changed from I8 to I16 type (after omdetect). NB: this is 
!            not uploaded
! 16.01.2007 VNY Creating a warning message for the case of crowded field (nSources>1000)
!            to respond to SPR-3643
! 22.01.2007 Checking the failue of extracting multiple spectra 
! 30.01.2007 The above failure was actually a bug in omatt (corrected and uploaded)
!            There is another failure of recovering a very bright source (target)
!            in the obs. 0462_0125910201, exp003. Checking ... 
! 06.03.2007 Introduced the checking of the QFLAG type (int8 or int16) in the source
!            list table (to avoid problems with the older versions of omdetect)
! 12.03.2007 Closing SPR-3643 (crowded field) and SPR-3653 (QFLAG type)
! 15.04.2007 Improved the field spectra extraction (SSC-SPR-3644) by allowing fainter
!            spectra to be extracted
! 15.05.2007 More complicated analysis is introduced for close spectra (fixing SPR-3660)
! 08.07.2007 Fixing a bug, which results in a mistaken window sizes and non-extraction of the 
!            target spectra in the case of the UV exposures for the calibration source
!            HZ2 (observation 0503_0125910901). SSC-SPR-3694
! 20.10.2007 Fixing a bug that lead to allocation of zero-size array to the spectrum 
!            cross-section when processing a crowded field (SSC-SPR-3700). The int8-type 
!            for the detected source list index is changed to int32-type (this would 
!            allow processing fields with more than 255 detected sources)  
! 09.11.2007 Fixing a bug mentioned in SSC-SPR-3702: the spectrum extraction 
!            region was changing when omgrism is called from the interactive tool 
!            omgsource. 

module omgrism_mod

  use oal
  use caloaldefs
  use dal
  use cal
  use param  
  use errorHandling
  use frameTime_mod
  use datafilter
  use sasdal
!  use minim_mod
!  use mssllib
!  use omgnumerical_mod
  use omgextraction_mod

  ! VNY 04.06.2005
  use OmSourceParam_mod ! Source definitions
  use findsourceslist_mod


  implicit none 
  
  private 

  public omgrism

contains 
  
  !---------------------------------------------------------------------
  subroutine omgrism
  !---------------------------------------------------------------------

   ! variables
   type(ArrayT) :: inArrayH                                          ! Input array handle.
   integer(kind=int32) :: dim(2),nx, ny !+
   integer(kind=int32) :: i,j! ,k !+
   real(kind=single), pointer :: oswImage(:, :)
   real(kind=single)          :: exposureTime         ! The exposure time for an input image
   integer :: ysize, xsize
!   real(kind=single)          :: omgXmin, omgYmin ! the coordinates of the lower left corner of the image

!   character(len=1024) :: expression !+
   character(len=1024) :: sourceListFile !+
   real(kind=single) :: sourceArea !+
!   type(DataSetT) :: sourceRegSetH !+
   type(DataSetT) :: grismSetH !+ 
   ! Input table handle
   type(TableT) :: inTableH !+
   integer, parameter :: noColumns = 20 !+
   type(ColumnT), dimension(noColumns) :: inColumnH !+
   integer(kind=int8), dimension(noColumns) :: inColumnFlag !+
   ! Input dataset handle
   type(DataSetT) :: sourceListSetH !+
   real(kind=double)              :: detectionLimitCountRate !+         ! The detection limit count rate (counts/second)
!   real(kind=double)              :: detectionLimitMagnitude         ! The detection limit magnitude
!   real(kind=double) :: error = 0.0 
!   real(kind=single) :: calApertureRadius = 12.0 !+         ! Radius used for coincidence losses   
!   real(kind=single) :: photometryApertureRadius=36.0 !+          ! Radius used for converting counts to magnitudes
 
   character(len=8) :: dataMode !+                                      ! Input file data mode.
   character(len=9) :: filter !+                                        ! Keywords.
   real(kind=single) :: exposure !+
   real(kind=double) :: frameTime !+                                    ! OM CCD frame time.
   real(kind=double) :: deadFrac  !+                                   ! Dead time in units of CCD frame time.
   integer(kind=int32) ccdMode !+                                       ! OM CCD data mode.
   integer(kind=int32) :: filterIdInt !+
!
   integer(kind=int32) :: tmexpId, frmTime
   character(len=80) :: dateObs, dateEnd, dataType, obsId, expId

   logical :: vignApp !+p
   logical :: backApp !+
   logical :: deadApp !+
   logical :: psfApp !+
   real(kind=single), dimension(:), pointer :: rate !+                  ! Rates pointer.
   real(kind=single), dimension(:), pointer :: rateErr !+               ! Rates error pointer.
   real(kind=double), dimension(:), pointer :: sourceRA !+              ! Source Right Ascension
   real(kind=double), dimension(:), pointer :: sourceDEC !+             ! Source Declination
   real(kind=double), dimension(:), pointer :: sourceLII !+             ! Source Galactic Longitude
   real(kind=double), dimension(:), pointer :: sourceBII !+             ! Source Galactic Lattitude

   ! 06.03.2007 for two possible types of column we need two different arrays
   integer(kind=int8), dimension(:), pointer :: sourceQF8 !+             ! Source quality flag
   integer(kind=int16), dimension(:), pointer :: sourceQF16 !+             ! Source quality flag
   integer(kind=int16) :: columnDataTypeFlag

   integer(kind=int8), dimension(:), pointer :: sourceCF !+             ! Source confusion flag
   real(kind=single), dimension(:), pointer :: sourcePA !+              ! Source positional angle
   real(kind=single), dimension(:), pointer :: sourcePA0 !+              ! Source positional angle
   real(kind=single), dimension(:), pointer :: sourceFWHMaj !+              ! Source major semiaxis
   real(kind=single), dimension(:), pointer :: sourceFWHMin !+              ! Source minor semiaxis

   real(kind=single), dimension(:), pointer :: sourceFWHMaj2 !+              ! Source major semiaxis
   real(kind=single), dimension(:), pointer :: sourceFWHMin2 !+              ! Source minor semiaxis

   integer(kind=int32) :: iSource, eSource
   real(kind=single) :: sourceWx, sourceWy

   real(kind=single), dimension(:), pointer :: limitLeftMax
   real(kind=single), dimension(:), pointer :: limitRightMax

   integer(kind=int8), dimension(:), pointer :: sourceShape  ! Source shape flag (point/spectrum)
   real(kind=single), dimension(:), pointer :: sourceSize,sourceRawCountRate
   integer(kind=int32), dimension(:), pointer :: sourceRelation ! Source relation flag
   integer(kind=int32), dimension(:), pointer :: sourceRelationCheck ! Source relation flag
   integer(kind=int32), dimension(:), pointer :: sourceRelationBright ! Source relation flag

   integer(kind=int8), dimension(:), pointer :: targetFlag ! target flag (1-target, 2-dubious target, -1:field)
   integer(kind=int32), dimension(:), pointer :: procFlag ! Processing flag (to label already processed sources)
   integer(kind=int32), dimension(:), pointer :: sourceProximity ! would contain the number of a near-by source !   

   real(kind=single), dimension(:), pointer :: proximityX
   real(kind=single), dimension(:), pointer :: proximityXcheck
   real(kind=single), dimension(:), pointer :: proximityXbright
   real(kind=single) :: proximityX1, proximityX1check, proximityX1bright
   real(kind=single), dimension(:), pointer :: proximityCounts 
   real(kind=single), dimension(:), pointer :: proximityCountsCheck 
   real(kind=single), dimension(:), pointer :: proximityCountsBright
   integer(kind=int32), dimension(:), pointer :: sourceProximityCheck ! would contain the number of a near-by source !   
   integer(kind=int32), dimension(:), pointer :: sourceProximityBright ! would contain the number of a near-by source !

   real(kind = double), parameter :: PI = 3.141592653589793_double  
   real(kind = double), parameter :: degreesToRadians = PI / 180.0_double
   real(kind = double), parameter :: radiansToDegrees = 180.0_double / PI
   real(kind = double), parameter :: radiansToArcSecs = radiansToDegrees * 3600.0_double 

   real(kind=single), dimension(:), pointer :: sourceXCoord0 !+          ! Source x-coordinate.
   real(kind=single), dimension(:), pointer :: sourceYCoord0 !+         ! Source y-coordinate. 

   real(kind=single), dimension(:), pointer :: sourceXCoord !+          ! Source x-coordinate.
   real(kind=single), dimension(:), pointer :: sourceYCoord !+         ! Source y-coordinate. 
!---
   real(kind=single), dimension(:), pointer :: sourceApertureRadius !+  ! Source aperture radius
   real(kind=single), dimension(:), pointer :: backGroundRate !+        ! Source background rate pointer.
   real(kind=single), dimension(:), pointer :: mag !+                   ! Magnitude pointer.
   real(kind=single), dimension(:), pointer :: magErr !+                ! Magnitude error pointer.
   integer(kind=int8), dimension(:), pointer :: extended !+
   integer(kind=int8), dimension(:), pointer :: extended0 !+

   real(kind=double) :: xZero, yZero ! XY-coordinates of zero-order image
   real(kind=double) :: xZeroA, yZeroA ! XY-coordinates of zero-order image
   real(kind=double) :: xFirst, yFirst ! XY-coordinates of the first-order image
!   real(kind=double) :: yFeature

   real(kind=double) :: zeroMaxValue
   real(kind=double) :: spectrumHalfLength
   integer :: smoothWindowHalfWidth

   real(kind=single), dimension(:), pointer :: x0aux, x1aux, y0aux, y1aux
   integer(kind=int8), dimension(:), pointer :: specTypeAux
   real(kind=single), dimension(:), pointer :: snRatio
   real(kind=single), dimension(:), pointer :: snRatioCheck
   real(kind=single), dimension(:), pointer :: snRatioBright

   logical :: extend
!   integer :: iLoop !+
   integer(kind=int32) :: nSources, nSpectraSources !+  ! Number of sources.
   integer(kind=int32) :: nBins !+
   integer(kind=int32) :: nSourcesMax
   integer(kind=int32) :: nSources2
   ! output variables
!   type(DataSetT) :: outFieldSet0H ! output file including all the spectra
   type(DataSetT) :: outTargetSet0H ! output file including all the spectra
   !type(TableT) :: outTableH !+
!   type(TableT) :: outTableLH !+
!   type(TableT) :: outTableRH !+
!   type(TableT) :: outFieldTable0H
   type(TableT) :: outTargetTable0H
! VNY 28.01.2004
   type(DataSetT) :: outSpectraListSetH
   type(TableT) :: outSpectraTableH
!
   character(len=1024) :: outFile !+
   character(len=1024) :: grismFile !+
   character(len=1024) :: outFile2, outFile3 !+
   character(len=7) :: blockName                ! name for the current 
!   character(len=8) :: blockNameL, blockNameR   ! block in the output file with tables 
                                                ! corresponding to the spectra
   character(len=1024) :: outFile4 !+
   !character(len=1024) :: outFileL, outFileR !+
   character(len=1024) :: outFileShortenName
   character(len=1024) :: outDirectory, summaryShortFileName
   character(len=80) :: spectrumFileComment
   character(len=3) :: sourceIndex
   integer(kind=int16) :: posPoint

   ! rates file columns
   type(ColumnT) :: rateSourceOutCol0 !+
   type(ColumnT) :: rateSourceErrOutCol0 !+
   type(ColumnT) :: rateBackOutCol0 !+
   type(ColumnT) :: rateBackErrOutCol0 !+
!
!   type(ColumnT) :: rateSourceOutCol1 !+
!   type(ColumnT) :: rateSourceErrOutCol1 !+
!   type(ColumnT) :: rateBackOutCol1 !+
!   type(ColumnT) :: rateBackErrOutCol1 !+
!
!   type(ColumnT) :: rateSourceOutColL !+
!   type(ColumnT) :: rateSourceErrOutColL !+
!   type(ColumnT) :: rateBackOutColL !+
!   type(ColumnT) :: rateBackErrOutColL !+
!   type(ColumnT) :: rateSourceOutColR !+
!   type(ColumnT) :: rateSourceErrOutColR !+
!   type(ColumnT) :: rateBackOutColR !+
!   type(ColumnT) :: rateBackErrOutColR !+

   type(ColumnT) :: fluxOutCol0 !+
   type(ColumnT) :: fluxErrOutCol0 !+
!
!   type(ColumnT) :: fluxOutCol1 !+
!   type(ColumnT) :: fluxErrOutCol1 !+
!
!   type(ColumnT) :: fluxOutColL !+
!   type(ColumnT) :: fluxErrOutColL !+
!   type(ColumnT) :: fluxOutColR !+
!   type(ColumnT) :: fluxErrOutColR !+

!   type(ColumnT) :: waveSourceOutColL
!   type(ColumnT) :: waveSourceOutColR

   type(ColumnT) :: spectrumNoCol
   type(ColumnT) :: spectrumOriginCol
   integer(kind=int32), dimension(:), pointer :: spectrumNoArray0
   integer(kind=int32), dimension(:), pointer :: spectrumNoArray
   integer(kind=int32), dimension(:), pointer :: spectrumOriginArray
   integer(kind=int32), dimension(:), pointer :: spectrumNoArray2
   integer(kind=int32), dimension(:), pointer :: spectrumOriginArray2
   integer(kind=int32), dimension(:), pointer :: srcNoArray2
   integer(kind=int32) :: spectrumNoCount
! VNY 28.01.2004
   type(ColumnT) :: specNoListCol
   type(ColumnT) :: specX0Col
   type(ColumnT) :: specY0Col
   type(ColumnT) :: specX1Col
   type(ColumnT) :: specY1Col
!   type(ColumnT) :: specIdCol
! VNY 13.07.2004
   type(ColumnT) :: specTypeCol
!   character(len=18), dimension(:), pointer :: specTypeArray
   !character(len=18), dimension(:), pointer :: specTypeCol

   !type(ColumnT) :: specOriginListCol
   integer(kind=int32), dimension(:), pointer :: specNoListArray
   real(kind=single), dimension(:), pointer :: specX0Array
   real(kind=single), dimension(:), pointer :: specY0Array
   real(kind=single), dimension(:), pointer :: specX1Array
   real(kind=single), dimension(:), pointer :: specY1Array
!   integer(kind=int32), dimension(:), pointer :: specIdArray
   !integer(kind=int32), dimension(:), pointer :: specOriginListArray
!

   type(ColumnT) :: waveSourceOutCol0
   real(kind=Single), pointer :: waveSourceOutArray0(:)
!   type(ColumnT) :: waveSourceOutCol1
!   real(kind=Single), pointer :: waveSourceOutArray1(:)
!--- 
!   real(kind=Single), pointer :: waveSourceOutArrayL(:) 
!   real(kind=Single), pointer :: waveSourceOutArrayR(:) 

   ! rates file pointers
   real(kind=Single), pointer :: rateSourceOutArray0(:) !+
   real(kind=Single), pointer :: rateSourceErrOutArray0(:) !+
   real(kind=Single), pointer :: rateBackOutArray0(:) !+
   real(kind=Single), pointer :: rateBackErrOutArray0(:) !+
!
!   real(kind=Single), pointer :: rateSourceOutArray1(:) !+
!   real(kind=Single), pointer :: rateSourceErrOutArray1(:) !+
!   real(kind=Single), pointer :: rateBackOutArray1(:) !+
!   real(kind=Single), pointer :: rateBackErrOutArray1(:) !+
!
   real(kind=Single), pointer :: fluxOutArray0(:) !+
   real(kind=Single), pointer :: fluxErrOutArray0(:) !+
!
!   real(kind=Single), pointer :: fluxOutArray1(:) !+
!   real(kind=Single), pointer :: fluxErrOutArray1(:) !+
!
!   real(kind=Single), pointer :: fluxOutArrayL(:) !+
!   real(kind=Single), pointer :: fluxErrOutArrayL(:) !+
!   real(kind=Single), pointer :: fluxOutArrayR(:) !+
!   real(kind=Single), pointer :: fluxErrOutArrayR(:) !+
!

!   real(kind=Single), pointer :: rateSourceOutArrayL(:) !+
!   real(kind=Single), pointer :: rateSourceErrOutArrayL(:) !+
!   real(kind=Single), pointer :: rateBackOutArrayL(:) !+
!   real(kind=Single), pointer :: rateBackErrOutArrayL(:) !+
!   real(kind=Single), pointer :: rateSourceOutArrayR(:) !+
!   real(kind=Single), pointer :: rateSourceErrOutArrayR(:) !+
!   real(kind=Single), pointer :: rateBackOutArrayR(:) !+
!   real(kind=Single), pointer :: rateBackErrOutArrayR(:) !+
!
   real(kind=double) :: timeExposure !+
   integer(kind=int32) :: srcNo !+
   integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1 !+   ! Coordinates of the OSW origin
   integer(kind=int32) :: windowDX, windowDY !+
   logical :: binBpe !+
   real(kind=single),  dimension(:), pointer :: psf1Correction !+          ! psf1 correction
   real(kind=single),  dimension(:), pointer :: psf2Correction !+         ! psf2 correction
   real(kind=single) :: nullValue = -999.999  !+
   integer :: columnDim(1) =(/1/) !+ 
   integer :: stringColumnDim(1) =(/18/) !+ 
!   real(kind=double) :: lambda
   real(kind=double), dimension(:), pointer :: calLambda
   integer(kind=int32) :: lambdaSize
   real(kind=double) :: plateScale
   integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
   integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
   integer(kind=int32) :: binX, xbin                                       ! Bin x.
   integer(kind=int32) :: binY, ybin                                       ! Bin y.
   !integer(kind=int32) :: srcNumber                                ! Number of source in source list.
   real(kind=double) :: xs0,x1,x2,ys0,y1,y2
   real(kind=double) :: mx1,mx2,mx3,px1,px2,px3
!    real(kind=single) :: averAngle, averAngle2, sumAngle2
    real(kind=single) :: averImage
    real(kind=single) :: sum, cosPA, sinPA, black, white !, xMax, yMax
    integer(kind=int32) :: jup, jdown, i1, ii1, j1, limitUp, limitDown,limitLeft,limitRight, sourceDeformation
    integer(kind=int32) :: jRight, jLeft
    real(kind=single) :: level
!    real(kind=single) :: aL0,aR0,aL1,aR1
!    integer(kind=int32) :: iMax
!    real(kind=single) :: sMax, dist, sinMax, cosMax , angleMax
    real(kind=single) :: averLeft, averRight, numLeft, numRight

!    real(kind=single) :: averBkgL, averBkgR, numBkgL, numBkgR 
!    real(kind=single) :: averSpec, numSpec, averStrip, numStrip

!    integer :: jcross, jbkg1, jbkg2, jspc1, jspc2, jyspc1, jyspc2 

!    real :: nLevel, wSpec, dSpec

   real(kind=double), dimension(4) :: param

!   real(kind=double)   :: gres
   integer(kind=int32) :: numparam, numpoints
!   real  :: varparam

!   integer(kind=int32) :: sizeOfData
!   real, dimension (:), pointer :: xinput
!   real, dimension (:), pointer :: yinput
!---
!   real  :: xtest , ytest ! this must be uncomm. for opt. extr. 
!   real, dimension(4) ::dydptest ! this must be uncomm. for opt. extr. 
!   real :: sigm ! this must be uncomm. for opt. extr. 
!---
!   real :: sum1, sum2, sumTest, sumGauss, sumWeight
   real(kind=double), dimension(:), pointer :: bkgVect, bkgVect2, bkgVect3
   real(kind=double), dimension(:), pointer :: bkgVectR, bkgVectL, specVect, smoothBkgVectR, smoothBkgVectL

   ! VNY 17.11.2003
   real(kind=double), dimension(:), pointer :: wavelengthArray, spectrumArray, bkgArray
   real(kind=double), dimension(:), pointer :: spectrumErrArray, bkgErrArray
   real(kind=double), dimension(:), pointer :: fluxArray, fluxErrArray
   integer(kind=int32), dimension(:), pointer :: maskArray
   !real(kind=double), dimension(:), pointer :: spectrumMaskArray
   integer :: sizeMask

   real(kind=double), dimension(:), pointer :: countsScaleFactor
   real(kind=double) :: deltaX, deltaY, deltaCriterium, predictedSpectrumPosition, blueEdge, redEdge
   real(kind=double) :: deltaXY
   real(kind=double) :: brightness1, brightness2, brightness3, meanBr3, brightnessRatio
   real(kind=double) :: brightness4, brightness5
   integer :: ii,jj,jx, iy, iy1,iy2
   integer :: iblueEdge, iredEdge
   real(kind=double) :: sbRatio

   real(kind=double) :: axialRatioCriterium, axialRatio
   real(kind=double) :: shapeCriterium1, shapeCriterium2 ! criteria for assigning a specific shape to the source 
   real(kind=double) :: shapeCriterium3
   real(kind=double) :: separation1, separation2 ! separation between the point source and its spectrum (in pixels)
   real(kind=double) :: source2spectrum, spectrumLength, checkWidth, checkLength 

   real(kind=double) :: nbr1, nbr2, nbr3
   integer :: jxR
   real(kind=double) :: mn1a,mn2a,mn3a,mn1b,mn2b,mn3b, mnab, critMin, critMax
   real(kind=double) :: mnx1a,mnx2a,mnx3a,mnx1b,mnx2b,mnx3b, mnxab, mxaver
   integer :: imn
   logical :: lateralSpectrumL, lateralSpectrumR
   real(kind=double) :: critMin0, critMax0
   real(kind=double) :: latShift, latWidth ! Parameters for the lateral spectra extraction in cases of saturated images
   integer :: blkNum
!---
   real(kind=double) :: bkgOffsetLeft, bkgWidthLeft, bkgOffsetRight, bkgWidthRight
   real(kind=double) :: spectrumHalfWidth !, imageSmoothBoxWidth
   real(kind=double) :: bkgWidth, spectrumWidth
   integer :: spectrumSmoothLength
   real(kind=double) :: spectrumLengthParU, spectrumLengthParV
   real(kind=double) :: spectrumToleranceU, spectrumToleranceV, spectrumTolerance
   real(kind=double) :: src2spectrumU, src2spectrumV
   real(kind=double) :: binningInterval
!   integer :: nPixSpec, nPixBkgL, nPixBkgR
   integer :: extractionMode, evaluationMode
   logical :: subtractBkg
   integer :: iyUp, iyDown, iyUp2
   real(kind=double) :: availableSpectrumFraction, availableSpectrumLength
   real(kind=double) :: xFirstOrder, yFirstOrder, xZeroOrder, yZeroOrder
   real(kind=double) :: xZeroRef, yZeroref
! VNY 28.01.2004
   character(len=1024) :: spectraListFile, outSpectraListFile !+
   character(len=1024) :: fieldSpectraFile
!   integer(kind=int32) :: nAux
   real(kind=double) :: regionU1, regionU2, regionV1, regionV2, region1, region2
   character(len=1024) :: tempFileName
   logical :: omgsourceRegime
   integer :: verbFlag
   real(kind=double) :: maxValue, maxValue2, meanValue2, yZero2
!   real(kind=double) :: searchLength
   logical :: checkFlag
   integer :: nFeatures
   ! VNY 21.04.2004
   real(kind=double) :: xZeroIn, yZeroIn, xZeroOut, yZeroOut
   real(kind=double) :: averCountIn, averCountOut, averCountLimit
   real(kind=double) :: spectrumHalfHeight ! for recomputation of the zero-order
   integer :: computingMode
! VNY 14.10.2004
   real(kind=double) :: xFirstIn, yFirstIn, xFirstOut, yFirstOut
! VNY 02.06.2004
    character(4)   :: exposureIdentifier
!    integer        :: instrumentMode = IMAGING
!    character(len=32)  :: primeInstrString 
    integer :: nExps, count, primPos, pos1, pos2
    character(len=1024)  :: summaryFileName
    character(len=1024) :: modifiedSummaryFileName
    logical :: existFile, eofReached, primInstrFound, existModifiedFile
    logical :: existWithinOutDirectory
    character(len=1024) :: summaryString!, regionString
    character(len=10) :: primInstrValue
    integer(kind=int32) :: xTarget, yTarget
    real(kind=single) :: distance, proximity, distanceX, distanceY
    integer :: iTarget, iDubiousTarget, nTargets, targetIndex
    logical(bool) :: extractFieldSpectra, produceSpectraList
    logical(bool) :: addedRegionFile
    character(len=1024)  :: regionFileName, spectraRegionFileName
    character(len=1024)  :: addedRegionFileName
    logical(bool) :: givenRegionFile 
    logical :: givenSpectraRegionFile
    ! VNY 23.07.2004 Test with OAL
    type(ProposalInfoType) :: propInfo      
    character(20) :: object  
    character(80) :: observer
    real(kind=double) :: proposedRA, proposedDEC
!    integer :: countMax, icount
    real(kind=double) ::  box0left,box0right,box0down,box0up
    real(kind=double) ::  box1left,box1right,box1down,box1up
    logical :: firstPass, nonTargetExtraction, putContaminationComment
    character(80) :: contaminationComment
    logical :: foundRealTarget, foundDubiousTarget, foundNoTarget
!------------
    ! vny 04.06.2005
    type(sourceParameters), pointer :: sourceData(:) => null()
    type(sourceParameters), pointer :: sourceDataBright(:) => null()
    type(SourceLinkedList) :: sourcesListCheck 
    integer(int32) :: nSourcesListCheck

    type(SourceLinkedList) :: sourcesListBright 
    integer(int32) :: nSourcesListBright, nSourcesBright

    integer(int32) :: xMin, xMax, yMin, yMax, ismoothWindow, nSourcesCheck
    real(single) :: nSigma
    real(single), allocatable, dimension(:, :) :: checkOSW 
    real(single), allocatable, dimension(:, :) :: brightOSW 
    real(single), allocatable, dimension(:, :) :: patternOSW 
    real(single), allocatable, dimension(:, :) :: safeOSW 
    integer(int32) :: allocErr   

    integer(int32), allocatable, dimension(:) :: indexCheckSources 
    real(single), allocatable, dimension(:) :: distCheckSources

    integer(int32), allocatable, dimension(:) :: indexBrightSources 
    real(single), allocatable, dimension(:) :: distBrightSources

    real(single) :: xyDist, distCrit, distCritY, xDist, yDist
    real(single) :: targetDistance, targetDistance2
    integer(int32) :: nContaminatedSources

    integer        :: istat
    !--------------------------------------------------------
    !! Following are the variables for saving testimage99.fit
    character(len=255) :: outSetCheck
    integer(int32) :: outDimCheck(2) 
    type(DataSetT) :: outDataSetCheckH 
    type(ArrayT) :: outArrayCheckH 
    real(single), dimension(:, :), pointer :: outOSWcheckP  
    !--------------------------------------------------------
    real(kind=double) :: zeroOrderCriterium
    integer(int32) :: iCheck
    integer(int32) :: iMin, iMax, jMin, jMax
    real(kind=single) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 
    logical :: createSpectrumFile, createSpectraListFile
    real(kind=double) :: s2n, signal, noise,s2n0, snCriterium, snCriterium2
    real(single), allocatable, dimension(:) :: sProfile, dProfile 
    integer(int32) :: spiMin, spiMax, spjMin, spjMax, spiMin2, spiMax2, iL, iR
    real(kind=double) :: dmx1, dmn1, dpx1, dpn1
    real(kind=double) :: aux1, aux2,aux3
    real(kind=double) :: crossShape
    real(kind=double) :: minDistance, minDistance2
    integer(kind=int32) :: iMinDistance, crowdedField, crowdedFieldBright
    real(kind=single) :: zero2firstRatio
    integer :: targetListFlag ! 0: main list, 1: bright list, 2: check list, -1: no target
    integer :: foundZeroOrder
    integer :: relationIndex
    real(kind=single) :: proximityCounts1, proximityCounts2
    integer(int32) :: nSp
    real(kind=single), dimension(1:2) :: peacks, leftBorder, rightBorder
    logical :: profileChecked
    real(kind=single) :: spiX2
    real(kind=double) :: grismAngle
    character(len=255) :: detsecString
    character(len=1) :: ch1, ch2, ch3, ch4, ch5
    real(kind=single) :: abratio


    crowdedField=0 ! not crowded
    zero2firstRatio=10.0 ! this is for the visible grism

    createSpectrumFile=.false.
    createSpectraListFile=.false.
!***************************************************
   omgsourceRegime=.false.
!   omgsourceRegime=.true.
!****************************************************
   firstPass=.true.

   param(1)=2.0 ! The original values of the parameters
   param(2)=6.0
   param(3)=10.0
   param(4)=5.0
   numparam=4
   numpoints=21

   critMin0=0.5
   critMax0=0.97
   ! get parameters
   grismFile = stringParameter("set")
   sourceListFile = stringParameter("sourcelistset")
   !srcNumber = intParameter("srcnumber")
   outFile = stringParameter("outset")
   posPoint=index(trim(outFile), 'SPECTR', .true.)
   spectraListFile=outFile
   fieldSpectraFile=outFile
   if (posPoint .gt. 0) then
      spectraListFile(posPoint:posPoint+5)="SPECLI"
   else
      spectraListFile="gSPECLI.fit"
      fieldSpectraFile="gSPECTR.fit"
   endif
   !write(*,*) "L612 spectraListFile=", trim(spectraListFile)
   !write(*,*) "L613 fieldSpectraFile=", trim(fieldSpectraFile)
    extractFieldSpectra = booleanParameter("extractfieldspectra")
    addedRegionFile=booleanParameter("addedregionfile")
    outSpectraListFile=""
    produceSpectraList=.false.


    ! Get the name of the spectra list file
    outSpectraListFile=stringParameter("outspectralistset")
    if (.not. trim(outSpectraListFile) .eq. "") then
       produceSpectraList=.true.
    else
       if (extractFieldSpectra) then
          outSpectraListFile=spectraListFile
          produceSpectraList=.true.
       endif
    endif
    !write(*,*) "L629 outSpectraListFile=", trim(outSpectraListFile)
    regionFileName=""
    givenRegionFile=.false.
    regionFileName=stringParameter("regionfile")
    if (.not. trim(regionFileName) .eq. "") then
       givenRegionFile=.true.
    endif

    spectraRegionFileName=""
    givenSpectraRegionFile=.false.
    spectraRegionFileName=stringParameter("spectraregionfile")
    if (.not. trim(spectraRegionFileName) .eq. "") then
       givenSpectraRegionFile=.true.
    endif

!.................................................
   posPoint=index(trim(outFile), '/p', .true.)
   outFileShortenName=outFile(posPoint+1:1024)
   write(spectrumFileComment, '(a, a)') "spectrumFile=", &
        trim(outFileShortenName)
   !write(*,*) "L649 outFileShortenName=", trim(outFileShortenName)
   bkgOffsetLeft = realParameter("bkgoffsetleft")
   bkgWidthLeft = realParameter("bkgwidthleft")
   bkgOffsetRight = realParameter("bkgoffsetright")
   bkgWidthRight = realParameter("bkgwidthright")
   extractionMode= intParameter("extractionmode")

   evaluationMode=2 ! the spectrum is extracted for evaluation with the extraction mode=2
                    ! (Gaussian fit)
   spectrumLengthParU = 1100.0
   spectrumLengthParV = 600.0

   src2spectrumU = 900.0 
   src2spectrumV = 785.0

   spectrumToleranceU= 600.0
   spectrumToleranceV=400.0

   spectrumHalfWidth = realParameter("spectrumhalfwidth")
   spectrumSmoothLength = intParameter("spectrumsmoothlength")

   bkgOffsetLeft=abs(bkgOffsetLeft)
   bkgWidthLeft=abs(bkgWidthLeft)
   bkgOffsetRight=abs(bkgOffsetRight)
   bkgWidthRight=abs(bkgWidthRight)
   spectrumHalfWidth=abs(spectrumHalfWidth)

   grismSetH = dataSet(TRIM(grismFile), read)
   sourceListSetH=dataSet(TRIM(sourceListFile), modify)

   !---Protection against the absence of the SRCLIST block
   if (hasBlock(sourceListSetH, "SRCLIST")) then
      if (hasAttribute(sourceListSetH, "INNBCKAN")) then
         call deleteAttribute(sourceListSetH, "INNBCKAN")
      endif
      if (hasAttribute(sourceListSetH, "OUTBCKAN")) then
         call deleteAttribute(sourceListSetH, "OUTBCKAN")
      endif

      !**************************************************************
      ! Read in the detection limit count rate from the PRIMARY header
      !**************************************************************
      detectionLimitCountRate = real(real32Attribute(sourceListSetH, "CRLIMIT"))

      ! Get a handle on the source list binary table extension.
      inTableH = table(sourceListSetH, "SRCLIST")

      tempFileName = " "
      if (hasAttribute(inTableH, "FILENAME")) then
         tempFileName= stringAttribute(inTableH, "FILENAME")
         posPoint=index(trim(tempFileName), 'OMGSOURCE')
         if (posPoint .gt. 0) omgsourceRegime=.true.
      else
         call setAttribute(inTableH, "FILENAME", trim(outFileShortenName), "Name of the spectrum file")
      endif

      !====================================================================
      ! VNY 02.09.2004 Test of the omgsource regime
      !omgsourceRegime=.true.
      !
      !====================================================================
      if (omgsourceRegime) then
         write(errstr,*) "Working in the omgsource task regime"
         call message(AppMsg, 2)
      endif

      ! Read the source list table extension keywords.
      dataMode = stringAttribute(sourceListSetH, "DATAMODE")
      filter = stringAttribute(sourceListSetH, "FILTER")
      exposure = real32Attribute(sourceListSetH, "EXPOSURE") 
      binBpe= booleanAttribute(sourceListSetH, "BINBPE")
      frameTime=real64Attribute(sourceListSetH, "FRAMTIME")
      frameTime=frameTime/1000.0 ! convert frametaime into seconds
      deadFrac=real64Attribute(sourceListSetH, "DEADFRAC")

      binAx1 = int32Attribute(sourceListSetH, "BINAX1")
      binAx2 = int32Attribute(sourceListSetH, "BINAX2")
!
      tmExpId = int32Attribute(grismSetH, "TMEXP_ID")
      frmTime = int32Attribute(grismSetH, "FRMTIME")
      dataType = stringAttribute(grismSetH, "DATATYPE")
      obsId = stringAttribute(grismSetH, "OBS_ID")
      expId = stringAttribute(grismSetH, "EXP_ID")
      dateObs = stringAttribute(grismSetH, "DATE-OBS")
      dateEnd = stringAttribute(grismSetH, "DATE-END")

      ! Calculate image binning.          
      ! VNY 04.05.2004 Temporarily restoring the wrong algorithm of computing
      ! the binning factors (to make it compatible with the imaging tasks
      !----
      !binX = binAx1
      !if (binX .lt. 1.0) binX = 1.0
      !binY = binAx2
      !if (binY .lt. 1.0) binY = 1.0
      ! --- The above 4 lines is the right algorithm, but it is teporarily switched off
      
      binX = 2 ** binAx1
      binY = 2 ** binAx2

      if (binBPE) then
         binX = 2.0
         binY = 2.0
      end if
 
      xbin=binX
      ybin=binY

      !Criteria for extraction
      axialRatioCriterium=3.5
      snCriterium=5.0
      snCriterium2=0.35
      if (filter .eq. "Grism2  " .or. filter .eq. "GRISM2  ")  then ! (visual grism)
         axialRatioCriterium=2.8         
         snCriterium=7.0
         snCriterium2=0.5
         ! VNY 02.03.2004
         ! shape criteria for assigning the shape to the source
         shapeCriterium1=50.0
         shapeCriterium2=30.0 
         shapeCriterium3=20.0

         separation1=290.0
         separation2=450.0
         checkWidth=10.0
         checkLength=100.0

         ! Checking the specrum regions for both filters
         source2spectrum=src2spectrumV ! 800.0
         spectrumLength=spectrumLengthParV !700.0 !400.0
         spectrumTolerance=spectrumToleranceV
         if (binY .gt. 1.0) then 
            source2spectrum=source2spectrum/binY
            spectrumLength=spectrumLength/binY
            spectrumTolerance=spectrumTolerance/binY
            shapeCriterium1=shapeCriterium1/binY
            shapeCriterium2=shapeCriterium2/binY
            shapeCriterium3=shapeCriterium3/binY
            checkLength=checkLength/binY
         endif
         !if (binX .gt. 1.0) then
         !   checkWidth=checkWidth/binX
         !endif
      endif
!---
      if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
         shapeCriterium1=200.0 ! shape criteria for assigning the shape to the source
         shapeCriterium2=100.0 
         shapeCriterium3=30.0 ! shape criterium for searchin the overlapping sources
         separation1=290.0
         separation2=450.0

         !checkWidth=20.0 ! criterium for the x-separation of the zero and first orders
         checkWidth=13.0 ! criterium for the x-separation of the zero and first orders
         checkLength=200.0
         ! Checking the same specrum regions for both filters
         source2spectrum=src2spectrumU ! 800.0 ! 1100.0 
         spectrumLength=spectrumLengthParU ! 700.0 !1000.0
         spectrumTolerance=spectrumToleranceU
         if (binY .gt. 1.0) then 
            source2spectrum=source2spectrum/binY
            spectrumLength=spectrumLength/binY
            shapeCriterium1=shapeCriterium1/binY
            shapeCriterium2=shapeCriterium2/binY
            shapeCriterium3=shapeCriterium3/binY
            spectrumTolerance=spectrumTolerance/binY
            checkLength=checkLength/binY
         endif
         !if (binX .gt. 1.0) then
         !   checkWidth=checkWidth/binX
         !endif
         zero2firstRatio=50.0 ! (for the visible it is 10)
      endif
      spectrumHalfHeight=checkLength ! the length of the zero-order
      smoothWindowHalfWidth=shapeCriterium1

! Checking different shifts (these shifts are for the lateral wings spectra)
      latWidth=10.0
      latShift=20.0
      if (binX .gt. 1.0) then
         latShift=latShift/binX
         latWidth=latWidth/binX
         checkWidth=checkWidth/binX
         spectrumHalfWidth=spectrumHalfWidth/binX
         bkgWidthRight=bkgWidthRight/binX
         bkgWidthLeft=bkgWidthLeft/binX
      endif
!---
      !********************
      ! Get the ccd mode.
      !********************
      if (trim(dataMode) .eq. "IMAGING" .or. trim(dataMode) .eq. "GRISM") then
         ccdMode = IMAGE
      else
         ccdMode = FAST
      end if

      ! using OAL
      !--------------------------------------------------------
      exposureIdentifier = GetExposureIdentifier(trim(grismFile))
      if (.not. omgsourceRegime) then
 
         call OAL_proposalInfo(propInfo)
         observer = propInfo%GoName
         object = propInfo%targetName
         proposedRA = propInfo%targetRa
         proposedDEC = propInfo%targetDec
 
      	nExps = OAL_selectScope(XMM, SUMMARY_INFORMATION)        
      	!primeInstrString  = OAL_getIPPVstring("PRIME_INSTR")
      	call OAL_selectFile(0, summaryFileName);
        ! Find the shorten name of the summary file
        posPoint=0
        posPoint=index(trim(summaryFileName),'/',.true.) ! .true. means backward
        summaryShortFileName=summaryFileName
        if (posPoint .gt. 3) then
           summaryShortFileName(1:posPoint)=' '
           summaryShortFileName=adjustl(summaryShortFileName)
        endif
        ! Find the name of the output directory
        posPoint=0        
        outDirectory=outFile
        posPoint=index(trim(outDirectory),'/',.true.) ! .true. means backward
        if (posPoint .ge. 1) then
           outDirectory(posPoint:1024)=' '
           ! combine the name of the summary file and the output directory
           outDirectory(posPoint:posPoint+1)='/ '
           outDirectory(posPoint+1:1024)=summaryShortFileName(1:1024-posPoint)
        else
           outDirectory='./'
           outDirectory(3:1024)=summaryShortFileName(1:1021)
        end if
        existFile=.false.
        existModifiedFile=.false.
        existWithinOutDirectory=.false.
        primPos=0
        primPos=index(trim(summaryFileName),'SUM.')
        ! Find summary file in the usual directory (ODF)
      	inquire (FILE=trim(summaryFileName), EXIST=existFile)
        if (.not. existFile) then
           ! Search for the original (ascii) summary file 
           modifiedSummaryFileName=trim(summaryFileName)
           if (modifiedSummaryFileName(primPos+4:primPos+6) .eq. 'SAS') then
              modifiedSummaryFileName(primPos+4:primPos+6)='ASC'
           else
              modifiedSummaryFileName(primPos+4:primPos+6)='SAS'
           endif
           inquire(FILE=trim(modifiedSummaryFileName), EXIST=existModifiedFile)
           if (.not. existModifiedFile) then
              inquire(FILE=trim(outDirectory),EXIST=existWithinOutDirectory)
           endif
        endif

      	if (existFile .or. existModifiedFile .or. existWithinOutDirectory) then
           ! Open the summary file
           if (existFile) open(2,file=trim(summaryFileName))
           if (existModifiedFile) open(2,file=trim(modifiedSummaryFileName))
           if (existWithinOutDirectory) open(2,file=trim(outDirectory))
           eofReached=.false.
           primInstrFound=.false.
           count=0
           primPos=0
           primInstrValue=''
           do
              read(2, '(a)', end=110) summaryString
              count=count+1
              primPos=index(summaryString,'PRIME_INSTR')
              if (primPos .gt. 0) then
                 primInstrFound=.true.
                 exit
              endif
           enddo
110        continue
120        close(2)         
           if (primInstrFound) then
              pos1=index(summaryString, '=')
              pos2=index(summaryString, '/')
              primInstrValue=summaryString(pos1+2:pos2-1)
           endif
        endif
     else
       primInstrValue=''
       call warning("SummaryFile", "No summary file found")
     endif ! .not. omgsourceRegime

     !-------------------------------------------------------
     !**************************************
     ! Get the filter Id integer value
     !**************************************
     filterIdInt = filterIdToInt(filter)
     !*******************
     ! Set the CAL state
     !********************
     call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)
     !************************************* 
     ! Get the Plate scale (arcsecs/pixel)
     !*************************************
     ! VNY 12.11.2003 - removing plateScale, since the accuracy doesn't matter
     ! call CAL_omGetPlateScale(lambda, plateScale)
      plateScale=0.476513

    allocate(calLambda(0:2048))
    calLambda=0.0
    call CAL_omGetGrismWavelength(calLambda)
    lambdaSize=size(calLambda)
    regionU1=-1.0
    regionU2=-1.0
    regionV1=-1.0
    regionV2=-1.0
    if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
       do iSource=1, lambdaSize-1
          if (calLambda(iSource) .ge. 1800.0 .and. calLambda(iSource-1) .lt. 1800.0) then
             regionU1=iSource
          endif
          if (calLambda(iSource) .ge. 3600.0 .and. calLambda(iSource-1) .lt. 3600.0) then
             regionU2=iSource
          endif
       enddo
       if (binY .gt. 1.0) then 
          regionU1=regionU1/binY
          regionU2=regionU2/binY
       endif
    endif

    if (filter .eq. "Grism2  " .or. filter .eq. "GRISM2  ")  then ! (Vis grism)
       do iSource=1, lambdaSize-1
          if (calLambda(iSource) .ge. 3000.0 .and. calLambda(iSource-1) .lt. 3000.0) then
             regionV1=iSource
          endif
          if (calLambda(iSource) .ge. 6000.0 .and. calLambda(iSource-1) .lt. 6000.0) then
             regionV2=iSource
          endif
       enddo
       if (binY .gt. 1.0) then 
          regionV1=regionV1/binY
          regionV2=regionV2/binY
       endif
    endif

    if (filter .eq. "Grism2  " .or. filter .eq. "GRISM2  ")  then ! (Vis grism)
       grismAngle=-5.6
       region1=regionV1
       region2=regionV2
    endif
    if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
       grismAngle=6.3
       region1=regionU1
       region2=regionU2
    endif
    
      !******************************************************************
      ! Get the number of rows in the source list binary table extension.
      !******************************************************************
      nSources = numberOfRows(inTableH)
      nSpectraSources=nSources
      !write(*,*) "L1058 nSources=", nSources
      nSources2=nSources

      inArrayH = array(grismSetH, "PRIMARY", read)
      exposureTime = real32Attribute(grismSetH, "EXPOSURE") ! Image exposure time
      ! VNY 20.05.2007 read the coordinates of the lower left corner of the image
!         omgXmin=real32Attribute(grismSetH, "OMGXMIN")
!         omgYmin=real32Attribute(grismSetH, "OMGYMIN")
         !write(*,*) "L1220 omgXmin=", omgXmin, " omgYmin=", omgYmin, " <=================="
         !write(*,*) "L1221 grismAngle=", grismAngle

      !********************************************************************
      ! Read in the DETSEC values (to reconstruct the unrotated image coordinates)
      !********************************************************************
      if (hasAttribute(grismSetH,"DETSEC")) then
         detsecString = stringAttribute(grismSetH, "DETSEC")
         read (detsecString, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
              ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
      endif
      !write(*,*) "L1080 wX0=", windowX0, " wX1=", windowX1, " wY0=", windowY0, &
      !     " wY1=", windowY1
      windowX0=windowX0+10/xBin
      windowX1=windowX1-10/xBin
      windowY0=windowY0+20/yBin
      windowY1=windowY1-20/yBin
      !write(*,*) "L1240 wX0=", windowX0, " wX1=", windowX1, " wY0=", windowY0, &
      !     " wY1=", windowY1   

      oswImage => real32array2data(inArrayH)
      dim = shape(oswImage)
      nx = dim(1)
      ny = dim(2) 

      xMin=0
      xMax=nx-1
      yMin=0
      yMax=ny-1

      allocate (checkOSW(xMin: xMax, yMin: yMax), stat = allocErr)
      call CheckAllocation(allocErr, "checkOSW")
      allocate (patternOSW(xMin: xMax, yMin: yMax), stat = allocErr)
      call CheckAllocation(allocErr, "patternOSW")
      allocate (safeOSW(xMin: xMax, yMin: yMax), stat = allocErr)
      call CheckAllocation(allocErr, "safeOSW")
      allocate (brightOSW(xMin: xMax, yMin: yMax), stat = allocErr)
      call CheckAllocation(allocErr, "brightOSW")
      brightOSW=oswImage
      checkOSW=oswImage
      safeOSW=oswImage
      ismoothWindow=8/xBin
      nSigma=1
!*****************************************************************************
! 08.07.2007 Creating the list of very bright sources
!******************************************************************************

      if (.not. omgsourceRegime) then

         !crowdedFieldBright=15
         crowdedFieldBright=10
         call DeleteSourceList(sourcesListBright)
         call FindSourcesList(brightOSW, xMin, xMax, yMin, yMax, ismoothWindow, &
              nSourcesBright, nSigma,nSourcesListBright, SourcesListBright, &
              crowdedFieldBright, windowX0, windowX1, windowY0, windowY1)
         !write(*,*) "L1211 nSourcesBright=", nSourcesBright, " nSourcesListBright=", nSourcesListBright
         ! Printing the bright sources list
         sourceDataBright => SourceDataArray(sourcesListBright, nSourcesListBright)
         !write(*,*) "******************************************* Bright source list:"
         !do i=1, nSourcesListBright
         !   write(*,*) " L1126 source=",i, " x=", sourceDataBright(i)%xPos, &
         !        " y=", sourceDataBright(i)%yPos, &
         !        " rate=", sourceDataBright(i)%rawCountRate, " nPix=", sourceDataBright(i)%noSourcePixels
         !enddo
         !write(*,*) "*******************************************"        
         ! VNY 25.10.07 save the bright sources image
         !----------------------
         !write(*,*) "L1301 creating testimage98bright.fit"
         !outSetCheck="testimage98bright.fit"
         !outDimCheck(1)=xMax-xMin+1
         !outDimCheck(2)=yMax-yMin+1
         !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
         !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
         !     real32, outDimCheck, "Counts", "IMAGE")!             
         !outOSWcheckP => real32Array2Data(outArrayCheckH)
         !outOSWcheckP = brightOSW 
         !! Release handles on data.
         !call release(outArrayCheckH)        
         !call release(outDataSetCheckH) 
         !--------------------------------


         !***************
         !if (nSources .gt. 300) then
         !   nSigma=2
         !   crowdedField=15
         !   ismoothWindow=ismoothWindow/4
         !endif
         if (nSources .gt. 300) then
            nSigma=1
            crowdedField=4
            ismoothWindow=ismoothWindow/4
         endif
         if (nSources .gt. 550) then
            nSigma=2
            crowdedField=15
            ismoothWindow=ismoothWindow/4
         endif
         !******************
         !write(*,*) "L1208 nSigma=", nSigma, " crowdedField=", crowdedField
         ! VNY 04.06.2005 Check the missing zero-order sources sources 
         !by using a simplified detection on oswImage
         !******************************************************************************
         ! Create the list of the Check sources (not only bright)
         !******************************************************************************
         call DeleteSourceList(sourcesListCheck)
         call FindSourcesList(checkOSW, xMin, xMax, yMin, yMax, ismoothWindow, &
              nSourcesCheck, nSigma,nSourcesListCheck, SourcesListCheck, &
              crowdedField, windowX0, windowX1, windowY0, windowY1)

         !patternOSW=checkOSW ! here checkOSW is already a mask
         patternOSW=brightOSW !
         ! VNY 03.06.05 save the check image
         !----------------------
         !write(*,*) "L1173 creating testimage99.fit"
         !outSetCheck="testimage99.fit"
         !outDimCheck(1)=xMax-xMin+1
         !outDimCheck(2)=yMax-yMin+1
         !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
         !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
         !     real32, outDimCheck, "Counts", "IMAGE")!             
         !outOSWcheckP => real32Array2Data(outArrayCheckH)
         !outOSWcheckP = checkOSW 
         !! Release handles on data.
         !call release(outArrayCheckH)        
         !call release(outDataSetCheckH) 
         !write(*,*) "L1197 nsourcesList=", nSourcesListCheck
         !write(*,*) "L1198 Test image 99 saved:", trim(outSetCheck)!
         !--------------------------------

         if(allocated(checkOSW)) then
            deallocate(checkOSW,stat=istat)
            call CheckDeallocation(allocErr, "checkOSW")
         end if

         if(allocated(brightOSW)) then
            deallocate(brightOSW,stat=istat)
            call CheckDeallocation(allocErr, "brightOSW")
         end if

         !--------------------
         ! print the newly obtained source list
         sourceData => SourceDataArray(sourcesListCheck, nSourcesListCheck)
         !write(*,*) "******************************************* Check source list:"
         !do i=1, nSourcesListCheck
         !   write(*,*) " L1201 sourceCheck=",i, " x=", sourceData(i)%xPos, &
         !        " y=", sourceData(i)%yPos, &
         !        " rate=", sourceData(i)%rawCountRate, " nPix=", sourceData(i)%noSourcePixels
         !enddo
         !write(*,*) "*******************************************"

      endif ! .not. omgsourceRegime

         !


      !--- Setting a limit to the input rows to 1000 sources
      if (nSources .gt. 0 .and. nSources .lt. 1000) then
         !write(*,*) "L1214 nSourcesListCheck=", nSourcesListCheck, &
         !     " nSourcesListBright=", nsourcesListBright
         if (.not. omgsourceRegime) then
            nSourcesMax=nSources+nSourcesListCheck+nSourcesListBright+1
         else
            nSourcesMax=nSources
         endif
         ! "+1" is added to make room for the possible target source
         !write(*,*) "L1234 nSourcesMax=", nSourcesMax
         !allocate(limitLeftMax(0:nSources-1))
         !allocate(limitRightMax(0:nSources-1))
         allocate(limitLeftMax(0:nSourcesMax))
         allocate(limitRightMax(0:nSourcesMax))
         limitLeftMax=0.0
         limitRightMax=0.0

         !  Allocate space for the source relation and shape flags
         allocate(sourceShape(0:nSourcesMax))
         allocate(sourceSize(0:nSourcesMax))
         allocate(sourceRawCountRate(0:nSourcesMax))
         allocate(sourceRelation(0:nSourcesMax))
         allocate(targetFlag(0:nSourcesMax))
         allocate(procFlag(0:nSourcesMax))
         allocate(sourceProximity(0:nSourcesMax))
         allocate(proximityX(0:nSourcesMax))
         allocate(proximityCounts(0:nSourcesMax))
!

         allocate(x0aux(0:nSourcesMax))
         allocate(x1aux(0:nSourcesMax))
         allocate(y0aux(0:nSourcesMax))
         allocate(y1aux(0:nSourcesMax))
         allocate(specTypeAux(0:nSourcesMax))
!
         allocate(spectrumNoArray2(0:nSourcesMax))
         allocate(spectrumNoArray(0:nSourcesMax))
         allocate(spectrumOriginArray2(0:nSourcesMax))
         allocate(srcNoArray2(0:nSourcesMax))
         allocate(snRatio(0:nSourcesMax))
         spectrumNoArray2=99999
         spectrumNoArray=99999
         spectrumOriginArray2=99999
         srcNoArray2=99999
         spectrumNoCount=-1
         snRatio=-1.0
!
         sourceShape=-1
         sourceSize=1.0E9
         sourceRawCountRate=-1.0
         !sourceRelation=-1
         sourceRelation=-999999
         targetFlag=-1
         procFlag=-1
         sourceProximity=-1
         proximityX=0.0
         proximityX1=0.0
         x0aux=-1.0
         x1aux=-1.0
         y1aux=-1.0
         y0aux=-1.0
         specTypeAux=-1
         xZeroRef=0.0
         yZeroRef=0.0
!
         ! Get handles and pointers on the source list binary table extension colums.
         inColumnFlag=1

         if (hasColumn(inTableH, "RATE")) then
            inColumnH(1) = column(inTableH, "RATE", modify)
            rate => real32Data(inColumnH(1))
         else
            inColumnFlag(1)=0
            allocate(rate(0:nSources-1))
         endif

         if (hasColumn(inTableH, "RATE_ERR")) then
            inColumnH(2) = column(inTableH, "RATE_ERR", read)
            rateErr => real32Data(inColumnH(2))
         else
            inColumnFlag(2)=0
            allocate(rateErr(0:nSources-1))
         endif

         if (hasColumn(inTableH, "MAG")) then
            inColumnH(3) = column(inTableH, "MAG", modify)
            mag => real32Data(inColumnH(3))
         else
            inColumnFlag(3)=0
            allocate(mag(0:nSources-1))
         endif

         if (hasColumn(inTableH, "MAGERR")) then
            inColumnH(4) = column(inTableH, "MAGERR", modify)
            magErr => real32Data(inColumnH(4))
         else
            inColumnFlag(4)=0
            allocate(magErr(0:nSources-1))
         endif

         if (hasColumn(inTableH, "CORR_BACKGROUND_RATE")) then
            inColumnH(17)=column(inTableH, "CORR_BACKGROUND_RATE", modify)
         else
            inColumnFlag(17)=0
         endif

         if (hasColumn(inTableH, "EFLAG")) then
            inColumnH(5) = column(inTableH, "EFLAG", modify)
            extended0 => int8Data(inColumnH(5))
         else
            inColumnFlag(5)=0
            allocate(extended0(0:nSources-1))
            extended0=0
         endif
         allocate(extended(0:nSourcesMax))
         extended=-1
         do i=0,nSources-1
            extended(i)=extended0(i)
         enddo

         if (hasColumn(inTableH, "BACKGROUND_RATE")) then
            inColumnH(6) = column(inTableH, "BACKGROUND_RATE", read)
            backGroundRate => real32Data(inColumnH(6))
         else
            inColumnFlag(6)=0
            allocate(backGroundRate(0:nSources-1))
         endif

         inColumnH(7) = column(inTableH, "XPOS", read)
         sourceXCoord0 => real32Data(inColumnH(7))

         ! VNY 04.06.05 add one element more in order to store 
         ! (if necessary) the target coordinates 
!         allocate(sourceXCoord(0:nSources-1))
!         sourceXCoord=sourceXCoord0
         allocate(sourceXCoord(0:nSourcesMax))
         sourceXCoord=-1.0
         do i=0,nSources-1
            sourceXCoord(i)=sourceXCoord0(i)
         enddo
         !sourceXCoord(nSources)=-1.0

         inColumnH(8) = column(inTableH, "YPOS", read)
         sourceYCoord0 => real32Data(inColumnH(8))

!         allocate(sourceYCoord(0:nSources-1))
!         sourceYCoord=sourceYCoord0
         allocate(sourceYCoord(0:nSourcesMax))
         sourceYCoord=-1.0
         do i=0,nSources-1
            sourceYCoord(i)=sourceYCoord0(i)
         enddo
         !sourceYCoord(nSources)=-1.0

         if (hasColumn(inTableH, "APERTURE")) then
            inColumnH(9) = column(inTableH, "APERTURE", modify)
            sourceApertureRadius => real32Data(inColumnH(9))
         else
            inColumnFlag(9)=0
           allocate(sourceApertureRadius(0:nSources-1))
         endif

         inColumnH(18) = column(inTableH, "FWHM_MAJ", read)
         sourceFWHMaj => real32Data(inColumnH(18))

!         allocate(sourceFWHMaj2(0:nSources-1))
!         sourceFWHMaj2=sourceFWHMaj
         allocate(sourceFWHMaj2(0:nSourcesMax))
         sourceFWHMaj2=-1.0
         do i=0,nSources-1
            sourceFWHMaj2(i)=sourceFWHMaj(i)
         enddo
         !sourceFWHMaj2(nSources)=-1.0

         inColumnH(19) = column(inTableH, "FWHM_MIN", read)
         sourceFWHMin => real32Data(inColumnH(19))

!         allocate(sourceFWHMin2(0:nSources-1))
!         sourceFWHMin2=sourceFWHMin
         allocate(sourceFWHMin2(0:nSourcesMax))
         sourceFWHMin2=-1.0
         do i=0,nSources-1
            sourceFWHMin2(i)=sourceFWHMin(i)
         enddo
         !sourceFWHMin2(nSources)=-1.0

         inColumnH(20) = column(inTableH, "PA", read)
         sourcePA0 => real32Data(inColumnH(20))
         allocate(sourcePA(0:nSourcesMax))
         sourcePA=-1.0
         do i=0,nSources-1
            sourcePA(i)=sourcePA0(i)
         enddo
         !write(*,*) "L1414 ***************************** Initial Source list:"
         !do iSource=0, nSourcesMax
         !!do iSource=0, nSources
         !   write(*,'(i3,a,f6.1,a,f6.1,a,f6.1,a, f6.1,a,f6.1)') &
         !        iSource, " X=", sourceXCoord(iSource), & 
         !        " Y=", sourceYCoord(iSource), & 
         !        " Maj=", sourceFWHMaj2(iSource), " Min=", sourceFWHMin2(iSource), &
         !        " PA=", sourcePA(iSource)
         !enddo
         !write(*,*) "L1426******************************************************************"
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


         allocate(sourceProximityCheck(1:nSourcesListCheck))
         allocate(proximityXcheck(1:nSourcesListCheck))
         allocate(proximityCountsCheck(1:nSourcesListCheck))
         allocate(snRatioCheck(1:nSourcesListCheck))
         allocate(sourceRelationCheck(1:nSourcesListCheck))
         sourceProximityCheck=-1
         proximityXcheck=0.0
         proximityX1check=0.0
         snRatioCheck=-1.0
         sourceRelationCheck=-1

         if (.not. omgsourceRegime) then
            !----------- Check if some newly detected BRIGHT sources are not in the main source list
            allocate (indexBrightSources(1:nSourcesListBright), stat = allocErr)
            call CheckAllocation(allocErr, "indexBrightSources")
            allocate (distBrightSources(1:nSourcesListBright), stat = allocErr)
            call CheckAllocation(allocErr, "distBrightSources")
            indexBrightSources=-1
            distBrightSources=999999.0
            distCrit=6.0/xBin
            distCritY=200.0/xBin
            !write(*,*) "L1440 nSources=", nSources, " nSourcesBright=", nSourcesBright, &
            !     " nSourcesListCheck=", nSourcesListCheck
            !write(*,*) "-----------------------"

            allocate(sourceProximityBright(1:nSourcesListBright))
            allocate(proximityXbright(1:nSourcesListBright))
            allocate(proximityCountsBright(1:nSourcesListBright))
            allocate(snRatioBright(1:nSourcesListBright))
            allocate(sourceRelationBright(1:nSourcesListBright))
            sourceProximityBright=-1
            proximityXbright=0.0
            proximityX1bright=0.0
            snRatioBright=-1.0
            sourceRelationBright=-1



            !--Check if some of the check list sources are not in the main source list
            allocate (indexCheckSources(1:nSourcesListCheck), stat = allocErr)
            call CheckAllocation(allocErr, "indexCheckSources")
            allocate (distCheckSources(1:nSourcesListCheck), stat = allocErr)
            call CheckAllocation(allocErr, "distCheckSources")
            indexCheckSources=-1
            distCheckSources=999999.0
            distCrit=6.0/xBin
            !write(*,*) "L1466 nSources=", nSources, " nSourcesCheck=", nSourcesCheck
            !write(*,*) "-----------------------"
            do i=1, nSourcesCheck
               do iSource=0, nSources-1
                  xDist=sourceData(i)%xPos-sourceXCoord(iSource)
                  yDist=sourceData(i)%yPos-sourceYCoord(iSource)
                  xyDist=sqrt(xDist*xDist+yDist*yDist)
                  if (xyDist .lt. distCrit) then
                     if (xyDist .lt. distCheckSources(i)) then
                        indexCheckSources(i)=iSource
                        distCheckSources(i)=xyDist
                        !write(*,*) " L1418 Check i=", i, " iSource=", &
                        !     iSource, " x1=", sourceData(i)%xPos, " y1=", sourceData(i)%yPos, &
                        !     " x2=", sourceXCoord(iSource), " y2=", sourceYCoord(iSource), &
                        !     " xyDist=", xyDist
                        !write(*,*) "     xdist=", xDist, " yDist=", yDist, " distCheckSources=", &
                        !     distCheckSources(i)
                     endif
                  endif
               enddo
            enddo
            !write(*,*) "L1487---------------------- binX=", binX
            do i=1,nSourcesCheck
               !write(*,*) " L1498 i=", i, " indexCheckSources=", indexCheckSources(i), &
               !     " x=",sourceData(i)%xPos, " y=",sourceData(i)%yPos, "A, B=", &
               !     sourceData(i)%semiA, sourceData(i)%semiB, " semiA/semiB=", &
               !     sourceData(i)%semiA/sourceData(i)%semiB
               abratio=0
               sourceData(i)%extensionFlag=0
               if (sourceData(i)%semiB .gt. 0.0) then 
                  abratio=sourceData(i)%semiA/sourceData(i)%semiB
               endif
               if (abratio .gt. 6.0 .and. sourceData(i)%semiA .gt. 12.0/binX) then
                  sourceData(i)%extensionFlag=1
               endif
               !write(*,*) " L1489 i=", i, " indexCheckSources=", indexCheckSources(i), &
               !     " x=",sourceData(i)%xPos, " y=",sourceData(i)%yPos, " A=", &
               !     sourceData(i)%semiA, " B=", sourceData(i)%semiB, " abratio=", &
               !     abratio, " ext=", sourceData(i)%extensionFlag       
               !write(*,*) " L1517 i=", i, " indexCheckSources=", indexCheckSources(i), &
               !     " x=",sourceData(i)%xPos, " y=",sourceData(i)%yPos, &
               !     " noPix=", sourceData(i)%noSourcePixels             
            enddo
         endif ! .no. omgsourceRegime

            !write(*,*) "L1352 shapeCriterium1=", shapeCriterium1, " axialRatioCriterium=", &
            !     axialRatioCriterium
            ! Assign flags for the extended and point-like sources in the main source list
         do iSource=0, nSources-1
            sourceShape(iSource)=0
            sourceSize(iSource)=0.0
            sourceRawCountRate(iSource)=rate(iSource)
            axialRatio=sourceFWHMaj2(iSource)/sourceFWHMin2(iSource)
            if (sourceFWHMaj2(iSource) .gt. shapeCriterium1 .and. axialRatio .gt. &
                 axialRatioCriterium .and. abs(sourcePA(iSource)) .gt. 82.0) then  
               sourceShape(iSource)=1
               sourceSize(iSource)=1.0E9
            endif

            !write(*,*) "L1541 i=",iSource, " FWHM=", sourceFWHMaj2(iSource), &
            !     " axRat=", axialRatio, " PA=", sourcePA(iSource), " shape=", &
            !     sourceShape(iSource)
         enddo
         
         if (.not. omgsourceRegime) then
            !==== Check whether the sources belong to the main source list
            !write(*,*) "L1531 distCrit=", distCrit
            do i=1, nSourcesCheck
               do iSource=0, nSources-1
                  xDist=sourceData(i)%xPos-sourceXCoord(iSource)
                  yDist=sourceData(i)%yPos-sourceYCoord(iSource)
                  xyDist=sqrt(xDist*xDist+yDist*yDist)
                  if (abs(xDist) .lt. distCrit .and. abs(yDist) .lt. distCritY) then
                     if (sourceShape(iSource) .eq. 1) then
                        !write(*,*) "L1539 **** iCheck=", i, " iSource=", iSource, &
                        !     " xDist=", xDist, " yDist=", yDist
                     
                        ! Ensure that the y-coordinates of the two sources are
                        ! not overlapping
                        iy1=sourceYCoord(iSource)
                        iy2=sourceData(i)%yPos
                        if (abs(iy1-iy2) .le. 5) then
                           ! adjust the y-coordinate of the first source
                           if ((iy1-iy2) .lt. 0) iy1=iy1-5
                           if ((iy1-iy2) .ge. 0) iy1=iy1+5
                        endif
                        ! compare brightnesses of these sources within 3x3 pixels
                        brightness1=0.0 ! Brightness of the first source
                        do ii=-2,2
                           iy=iy1+ii
                           if (iy .ge. 0 .and. iy .lt. ny) then
                              do jj=-2,2
                                 jx=sourceXCoord(iSource)+jj
                                 if (jx .ge. 0 .and. jx .lt. nx) then
                                    brightness1=brightness1+oswImage(jx,iy)
                                 endif
                              enddo
                           endif
                        enddo ! ii=-2,2
                      
                        brightness2=0.0 ! Brightness of the second source
                        do ii=-2,2
                           iy=iy2+ii
                           if (iy .ge. 0 .and. iy .lt. ny) then
                              do jj=-2,2
                                 jx=sourceData(i)%xPos+jj
                                 if (jx .ge. 0 .and. jx .lt. nx) then
                                    brightness2=brightness2+oswImage(jx,iy)
                                 endif
                              enddo
                           endif
                        enddo ! ii=-2,2
                        ! compute the ratio between two brightnesses
                        brightnessRatio=brightness1/brightness2
                        !write(*,*) "L1505 brightness1=", brightness1, " brightness2=", brightness2
                        !write(*,*) "      x1=", sourceXCoord(iSource), " y1=", sourceYCoord(iSource)
                        !write(*,*) "      x2=",sourceDataBright(i)%xPos, " y2=", sourceDataBright(i)%yPos  
                        !write(*,*) "      brightnessRatio=", brightnessRatio
                        if (brightnessRatio .gt. 0.6) then
                           indexCheckSources(i)=iSource
                           distCheckSources(i)=xyDist
                        endif
                     endif
                  endif

                  if (indexCheckSources(i) .lt. 0 .and. xyDist .lt. distCrit) then
                     if (xyDist .lt. distCheckSources(i)) then
                        indexCheckSources(i)=iSource
                        distCheckSources(i)=xyDist
                        !write(*,*) " L1594 Check i=", i, " main N=", &
                        !     iSource, " x1=", sourceData(i)%xPos, &
                        !     " y1=", sourceData(i)%yPos, &
                        !     " x2=", sourceXCoord(iSource), " y2=", sourceYCoord(iSource), &
                        !     " xyDist=", xyDist
                        !write(*,*) "     xdist=", xDist, " yDist=", yDist, " distCheckSources=", &
                        !     distCheckSources(i)
                     endif
                  endif
               enddo
            enddo

            !write(*,*) "L1606---------------------------------"
            !do i=1,nSourcesCheck
            !   if (indexCheckSources(i) .lt. 0) then
            !      write(*,*) " L1608 iCheck=", i, " x=", sourceData(i)%xPos, &
            !           " y=", sourceData(i)%yPos, " indexCheck=", indexCheckSources(i), &
            !           " dist=", distCheckSources(i)
            !   endif
            !enddo        
            !write(*,*) "L1600---------------------------------"
            !====
            nSp=0
            ! include the newly detected sources into the main source list
            do i=1,nSourcesCheck
               if (indexCheckSources(i) .lt. 0 .and. sourceData(i)%extensionFlag .eq. 0) then
                  ! this source was not in the list, include it
                  sourceXCoord(nSources2)=sourceData(i)%xPos
                  sourceYCoord(nSources2)=sourceData(i)%yPos
                  sourceFWHMaj2(nSources2)=sourceData(i)%semiA/2.0
                  sourceFWHMin2(nSources2)=sourceData(i)%semiB/2.0
                  sourcePA(nSources2)=sourceData(i)%theta
                  if (sourcePA(nSources2) .le. 82.0)sourcePA(nSources2)=83.0
                  sourceShape(nSources2)=0
                  sourceSize(nSources2)=sourceData(i)%noSourcePixels
                  sourceRawCountRate(nSources2)=sourceData(i)%rawCountRate
                  nSources2=nSources2+1
                  nSp=nSp+1
               endif
            enddo


            !=== Check for the bright sources in the main source list

            !write(*,*) "L1635 distCrit=", distCrit, " Number of included Check sources=", &
            !     nSp
            !write(*,*) "++++++++++++++++++++++++++++++++++++++++++++++++++"
            do i=1, nSourcesBright
               !write(*,*) " iBright=", i, " x=", sourceDataBright(i)%xPos, &
               !     " y=", sourceDataBright(i)%yPos, " nPix=", &
               !     sourceDataBright(i)%noSourcePixels
               do iSource=0, nSources-1
                  xDist=sourceDataBright(i)%xPos-sourceXCoord(iSource)
                  yDist=sourceDataBright(i)%yPos-sourceYCoord(iSource)
                  xyDist=sqrt(xDist*xDist+yDist*yDist)
                  
                  if (abs(xDist) .lt. distCrit .and. abs(yDist) .lt. distCritY) then
                     if (sourceShape(iSource) .eq. 1) then
                        !write(*,*) "L1644 **** iBright=", i, " iSource=", iSource, &
                        !     " xDist=", xDist, " yDist=", yDist
                        
                        ! Ensure that the y-coordinates of the two sources are
                        ! not overlapping
                        iy1=sourceYCoord(iSource)
                        iy2=sourceDataBright(i)%yPos
                        if (abs(iy1-iy2) .le. 5) then
                           ! adjust the y-coordinate of the first source
                           if ((iy1-iy2) .lt. 0) iy1=iy1-5
                           if ((iy1-iy2) .ge. 0) iy1=iy1+5
                        endif
                        ! compare brightnesses of these sources within 3x3 pixels
                        brightness1=0.0 ! Brightness of the first source
                        do ii=-2,2
                           iy=iy1+ii
                           if (iy .ge. 0 .and. iy .lt. ny) then
                              do jj=-2,2
                                 jx=sourceXCoord(iSource)+jj
                                 if (jx .ge. 0 .and. jx .lt. nx) then
                                    brightness1=brightness1+oswImage(jx,iy)
                                 endif
                              enddo
                           endif
                        enddo ! ii=-2,2
                      
                        brightness2=0.0 ! Brightness of the second source
                        do ii=-2,2
                           iy=iy2+ii
                           if (iy .ge. 0 .and. iy .lt. ny) then
                              do jj=-2,2
                                 jx=sourceDataBright(i)%xPos+jj
                                 if (jx .ge. 0 .and. jx .lt. nx) then
                                    brightness2=brightness2+oswImage(jx,iy)
                                 endif
                              enddo
                           endif
                        enddo ! ii=-2,2
                        ! compute the ratio between two brightnesses
                        brightnessRatio=brightness1/brightness2
                        !write(*,*) "L1505 brightness1=", brightness1, " brightness2=", brightness2
                        !write(*,*) "      x1=", sourceXCoord(iSource), " y1=", sourceYCoord(iSource)
                        !write(*,*) "      x2=",sourceDataBright(i)%xPos, " y2=", sourceDataBright(i)%yPos  
                        !write(*,*) "      brightnessRatio=", brightnessRatio
                        if (brightnessRatio .gt. 0.6) then
                           indexBrightSources(i)=iSource
                           distBrightSources(i)=xyDist
                        endif
                     endif
                  endif

                  if (indexBrightSources(i) .lt. 0 .and. xyDist .lt. distCrit) then
                     if (xyDist .lt. distBrightSources(i)) then
                        indexBrightSources(i)=iSource
                        distBrightSources(i)=xyDist
                        !write(*,*) " L1482 Bright i=", i, " main N=", &
                        !     iSource, " x1=", sourceDataBright(i)%xPos, &
                        !     " y1=", sourceDataBright(i)%yPos, &
                        !     " x2=", sourceXCoord(iSource), " y2=", sourceYCoord(iSource), &
                        !     " xyDist=", xyDist
                        !write(*,*) "     xdist=", xDist, " yDist=", yDist, " distBrightSources=", &
                        !     distBrightSources(i)
                     endif
                  endif
               enddo
            enddo
            !write(*,*) "L1710----------------------------"


            ! check for the bright sources in the check list
            do i=1, nSourcesBright
               do iSource=1, nSourcesCheck
                  xDist=sourceDataBright(i)%xPos-sourceData(iSource)%xPos
                  yDist=sourceDataBright(i)%yPos-sourceData(iSource)%yPos
                  xyDist=sqrt(xDist*xDist+yDist*yDist)

                  if (indexBrightSources(i) .lt. 0 .and. xyDist .lt. distCrit) then
                     if (xyDist .lt. distBrightSources(i)) then
                        indexBrightSources(i)=iSource
                        distBrightSources(i)=xyDist
                        !write(*,*) " L1776 Bright i=", i, " Check N=", &
                        !     iSource, " x1=", sourceDataBright(i)%xPos, &
                        !     " y1=", sourceDataBright(i)%yPos, &
                        !     " x2=", sourceData(iSource)%xPos, " y2=", sourceData(iSource)%yPos, &
                        !     " xyDist=", xyDist
                        !write(*,*) "     xdist=", xDist, " yDist=", yDist, " distBrightSources=", &
                        !     distBrightSources(i)
                     endif
                  endif
               enddo
            enddo
            !write(*,*) "L1787----------------------------"

            !=====
            !do i=1,nSourcesBright
            !   write(*,*) " L1791 i=", i, " indexBrightSources=", indexBrightSources(i)
            !enddo
            !        write(*,*) "L1593---------------------------------"
            !        do i=1,nSourcesBright
            !           write(*,*) " L1595 i=", i, " x=", sourceDataBright(i)%xPos, &
            !                " y=", sourceDataBright(i)%yPos, " iMin=", sourceDataBright(i)%iMin, &
            !                " iMax=", sourceDataBright(i)%iMax, " jMin=", sourceDataBright(i)%jMin, &
            !                " jMax=", sourceDataBright(i)%jMax 
            !        enddo        
            !        write(*,*) "L1600---------------------------------"
            !do i=1,nSourcesBright
            !   write(*,*) " L1595 i=", i, " x=", sourceDataBright(i)%xPos, &
            !        " y=", sourceDataBright(i)%yPos, " semiA=", sourceDataBright(i)%semiA, &
            !        " semiB=", sourceDataBright(i)%semiB, " theta=", sourceDataBright(i)%theta
            !enddo        
            ! include the new bright sources into the main list
            do i=1,nSourcesBright
               if (indexBrightSources(i) .lt. 0) then
                  sourceXCoord(nSources2)=sourceDataBright(i)%xPos
                  sourceYCoord(nSources2)=sourceDataBright(i)%yPos
                  sourceFWHMaj2(nSources2)=sourceDataBright(i)%semiA/2.0
                  sourceFWHMin2(nSources2)=sourceDataBright(i)%semiB/2.0
                  sourcePA(nSources2)=sourceDataBright(i)%theta
                  if (sourcePA(nSources2) .le. 82.0) sourcePA(nSources2)=83.0
                  sourceShape(nSources2)=0
                  sourceSize(nSources2)=sourceDataBright(i)%noSourcePixels
                  sourceRawCountrate(nSources2)=sourceDataBright(i)%rawCountRate
                  nSources2=nSources2+1
               endif
            enddo
            !write(*,*) "L1766 nSources2=", nSources2
            !write(*,*) "********************************************* Modified source list:"
         

!***********************************************************************
! Create a region file corresponding to the newly detected sources
!***********************************************************************
            !write(*,*) "L1793 regionFileName=", trim(regionFileName)
            i=index(trim(regionFileName),".ASC")
            !write(*,*) "L1796 index=", i
            if (i .gt. 9) then
               addedRegionFileName=regionFileName
               addedRegionFileName(i-10:i-5)="REGADD"
               if (i .gt. 26) then
                  addedRegionFileName(i-27:i-27)="g"
               endif
            else
               addedRegionFileName="gREGADD.ASC"
            endif
            !write(*,*) "L1814-------------------------------------"
            !do iSource=0, nSourcesMax
            !   write(*,'(i3,a,f6.1,a,f6.1,a,f6.1,a, f6.1,a,i3,a,f6.1, a, f6.1)') &
            !        iSource, " X=", sourceXCoord(iSource), & 
            !        " Y=", sourceYCoord(iSource), & 
            !        " Maj=", sourceFWHMaj2(iSource), " Min=", sourceFWHMin2(iSource), &
            !        " sourceShape=", sourceShape(iSource), &
            !        " PA=", sourcePA(iSource), " size=", sourceSize(iSource)
            !enddo

            !write(*,*) "L1784 addedRegionFileName=", trim(addedRegionFileName)
            !write(*,*) "L1785 addedRegionFile=", addedRegionFile, " nSources2-nSources=", &
            !     nSources2-nSources
            if (addedRegionFile) then
               !write(*,*) "L1787 Producing an additional region file"
               outSetCheck=addedRegionFileName
               open(8,file=trim(outSetCheck), status='replace')
               write(8,*) " global select=0 edit=0 move=0 delete=1"
               do iSource=nSources, nSources2-1
                  if (sourceFWHMaj2(iSource) .lt. 0.1) sourceFWHMaj2(iSource)=1.0
                  if (sourceFWHMin2(iSource) .lt. 0.1) sourceFWHMin2(iSource)=1.0
                  write(8,'(a,f6.1,a,f6.1,a,f6.1,a, f6.1,a,f6.1,a,i3,a)') &
                       "ellipse(", sourceXCoord(iSource), ",", sourceYCoord(iSource), & 
                       ",", sourceFWHMaj2(iSource)*5.0, ",", sourceFWHMin2(iSource)*5.0, &
                       ",", sourcePA(iSource), ") # color=red text={",iSource,"}"
               enddo
               close(8)
               !write(*,*) "L1812******************************************************************"
            endif
         else
            nSources2=nSources
         endif ! .not. omgsourceRegime

         !   Search for the contaminating sources (point-like sources at the vicinity of a primary source).
         deltaCriterium=24.0/xBin!12.0 ! 12-pixel minimal distance between the sources 
         zeroOrderCriterium=1000.0/xBin ! criterium for the size of zero-orders
         !write(*,*) "L1847-------------------------------------------"
!***********************************
!         do iSource=0, nSources-1
         do iSource=0, nSources2-1
            ! Do it only for the point-like sources
            if (sourceShape(iSource) .eq. 0) then
!*********************************
               !               do i=0, nSources-1
               do i=0, nSources2-1
                  if (i .ne. iSource) then ! calculate distance between the sources 
                     if (sourceSize(i) .lt. zeroOrderCriterium) then
                        deltaX=abs(sourceXCoord(i)-sourceXCoord(iSource))
                        deltaY=abs(sourceYCoord(i)-sourceYCoord(iSource))
                        deltaXY=sqrt(deltaX*deltaX+deltaY*deltaY)
                        if (deltaXY .lt. deltaCriterium .and. .not. omgsourceRegime) then
                           !write(*,*) "L1862 iSource=", iSource, " i=", i, &
                           !     " dX=", deltaX, " dY=", deltaY, " dXY=", deltaXY
                           !*******************************************************
                           ! assign -1 to the source shape 
                           ! compare brightnesses of these sources within 3x3 pixels
                           !*******************************************************
                           brightness1=0.0 ! Brightness of the first source
                           do ii=-2,2
                              iy=sourceYCoord(iSource)+ii
                              if (iy .ge. 0 .and. iy .lt. ny) then
                                 do jj=-2,2
                                    jx=sourceXCoord(iSource)+jj
                                    if (jx .ge. 0 .and. jx .lt. nx) then
                                       brightness1=brightness1+oswImage(jx,iy)
                                    endif
                                 enddo
                              endif
                           enddo ! ii=-2,2
                        

                           brightness2=0.0 ! Brightness of the second source
                           do ii=-2,2
                              iy=sourceYCoord(i)+ii
                              if (iy .ge. 0 .and. iy .lt. ny) then
                                 do jj=-2,2
                                    jx=sourceXCoord(i)+jj
                                    if (jx .ge. 0 .and. jx .lt. nx) then
                                       brightness2=brightness2+oswImage(jx,iy)
                                    endif
                                 enddo
                              endif
                           enddo ! ii=-2,2
                           
                           if (brightness1 .ge. brightness2) then ! eliminating the second source
                              sourceShape(i)=-1
                           else ! eliminate the first source
                              sourceShape(iSource)=-1
                           endif
                        endif ! deltaX .lt. ...
                        ! VNY 04.02.04 Check if the source fall inside of a spectrum
                        if (sourceShape(i) .eq. 1) then
                           if (deltaX .le. sourceFWHMin2(i) .and. deltaY .le. sourceFWHMaj2(i) &
                                .and. .not. omgsourceRegime) then
                              sourceShape(iSource)=-2
                           endif ! (deltaX .le. sourceFWHMin
                        endif
                     endif ! sourceSize .lt. zeroOrderCriterium
                  endif ! i .ne. iSource
               enddo
!               !
            endif ! sourceShape
            !****************************************************************
            ! 06.11.2007
            ! Check whether the supposedely point=like source is a spectrum
            !****************************************************************
            if (.not. omgsourceRegime) then
               !if (iSource .eq. 441) then
               if (sourceShape(iSource) .eq. 0) then
                  ! Compute the brightness of the source itself
                  brightness1=0.0 ! Brightness of the first source
                  do ii=-2,2
                     iy=sourceYCoord(iSource)+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-2,2
                           jx=sourceXCoord(iSource)+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness1=brightness1+oswImage(jx,iy)
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  brightness1=brightness1/25.0

                  brightness2=0.0 ! Brightness of the upper region
                  aux1=sourceYCoord(iSource)+sourceFWHMaj2(iSource)+2.0
                  do ii=-2,2
                     iy=aux1+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-2,2
                           jx=sourceXCoord(iSource)+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness2=brightness2+oswImage(jx,iy)
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  brightness2=brightness2/25.0

                  brightness3=0.0 ! Brightness of the lower region
                  aux1=sourceYCoord(iSource)-sourceFWHMaj2(iSource)-2.0
                  do ii=-2,2
                     iy=aux1+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-2,2
                           jx=sourceXCoord(iSource)+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness3=brightness3+oswImage(jx,iy)
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  brightness3=brightness3/25.0
                  
                  brightness4=0.0 ! Brightness of the left region
                  aux1=sourceXCoord(iSource)-sourceFWHMin2(iSource)-2.0
                  do ii=-2,2
                     iy=sourceYCoord(iSource)+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-2,2
                           jx=aux1+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness4=brightness4+oswImage(jx,iy)
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  brightness4=brightness4/25.0
                  
                  brightness5=0.0 ! Brightness of the right region
                  aux1=sourceXCoord(iSource)+sourceFWHMin2(iSource)+2.0
                  do ii=-2,2
                     iy=sourceYCoord(iSource)+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-2,2
                           jx=aux1+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness5=brightness5+oswImage(jx,iy)
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  brightness5=brightness5/25.0


                  !write(*,*) " L1941 iSource=", iSource, " br1=", brightness1, &
                  !     " br2=", brightness2, " br3=", brightness3
                  brightness2=(brightness2+brightness3)/2.0
                  aux1=1000.0
                  brightness3=sqrt(brightness1+brightness2)
                  if (brightness3 .gt. 0.0) then
                     aux1=brightness1-brightness2
                     if (aux1 .lt.0.0) aux1=0.0
                     aux1=aux1/brightness3
                  endif

                  brightness4=(brightness4+brightness5)/2.0
                  aux2=1000.0
                  brightness5=sqrt(brightness1+brightness4)
                  if (brightness5 .gt. 0.0) then
                     aux2=brightness1-brightness4
                     if (aux2 .lt.0.0) aux2=0.0
                     aux2=aux2/brightness5
                  endif
                  aux3=99999.0
                  if (aux1 .gt. 0.0) aux3=aux2/aux1
                  ! Here "aux1" is s/n in the vertical direction, "aux2" - in the horizontal
                  if (aux1 .lt. 4.0 .or. aux2 .lt. 4.0) then
                     if (aux2 .gt. 0.2 .and. aux3 .gt. 1.5) then
                        if (aux1 .lt. aux2) sourceShape(iSource)=1
                     else
                        if (aux2 .lt. 0.5) then
                           sourceShape(iSource)=-1
                        endif
                     endif
                  endif
                  !write(*,'(i3,a,f6.1,a,f6.1,a,f10.1,a, f10.1,a,f8.3,a,f8.3,a,f8.3,a,i3)') iSource, &
                  !     " x=", sourceXCoord(iSource), &
                  !     " y=", sourceYCoord(iSource), " br1=", brightness1, &
                  !     " br2=", brightness2, " s/nV=", aux1 , " s/nH=", aux2, " s2/s1=", aux3, &
                  !     " shape=", sourceShape(iSource)
                  !endif ! iSource .eq. 441
               endif ! sourceShape .eq. 0
            endif ! .not. omgsourceRegime

            !****************************************************************
         enddo ! iSource-0, nSources-1
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^66666
         targetListFlag=-1 ! no target detected yet
         
         
         ! Look for the target (among the point-like sources with sourceShape=0)
         call GetTargetPosition(primInstrValue,filter,binBpe, xbin, ybin, xTarget, yTarget)
         nonTargetExtraction=.true.
         !write(*,*) "L1643 xTarget=", xTarget, " yTarget=", yTarget
         !write(*,*) "L1644 binBpe=", binBpe, " nonTargetExtraction=", nonTargetExtraction 
         nTargets=0
         iTarget=-9999
         iDubiousTarget=-9999
         targetDistance=999999.0
         !***************************************** 03.11.2007
         !         do iSource=0,nSources-1
         do iSource=0,nSources2-1
            if (sourceShape(iSource) .eq. 0) then
               ! Checking whether it is the target
               !*************************************
               if (.not. omgsourceRegime) then 
                  !*********************************
                  distance=(sourceXCoord(iSource)-xTarget)*(sourceXCoord(iSource)-xTarget)+ &
                       (sourceYCoord(iSource)-yTarget)*(sourceYCoord(iSource)-yTarget)
                  distance=sqrt(distance)
                  if (distance .lt. 20.0/binY) then
                     if (distance .lt. targetDistance) then
                        if (nTargets .eq.0 .or. targetDistance .gt. 20.0/binY) then
                           if (nTargets .eq. 0) then
                              nTargets=nTargets+1
                           endif
                           iTarget=iSource
                           targetFlag(iSource)=1 ! real target found
                           
                           targetDistance=distance
!----------------------------------------------------------
                           if (iSource .lt. nSources) then
                              targetListFlag=0 ! main source list
                           else
                              targetListFlag=1 ! additional source list
                           endif
!----------------------------------------------------------
                        endif
                        !write(*,*) "L1663 target=", nTargets, " iSource=", iSource, &
                        !     " x=", sourceXcoord(iSource), " y=", sourceYCoord(iSource)
                     endif
                  else
                     ! Rather it is a dubious target
                     if (distance .lt. 40.0/binY) then
                        if (distance .lt. targetDistance) then
                           targetFlag(iSource)=2 ! dubious target found
                           if (nTargets .eq. 0 .or. targetDistance .gt. 40.0/binY) then
                              if (nTargets .eq. 0) then
                                 nTargets=nTargets+1
                              endif
                              iDubiousTarget=iSource
                              if (iSource .lt. nSources) then
                                 targetListFlag=0 ! main source list
                              else
                                 targetListFlag=1
                              endif
                              !write(*,*) "L1674 dubipus target=", nTargets, " iSource=", iSource, &
                              !     " x=", sourceXcoord(iSource), " y=", sourceYCoord(iSource)
                              targetDistance=distance
                           endif
                        endif
                     endif
                  endif
               else ! omgsourceRegime
                  ! 
                  ! Force the target extraction in the case of omgsourceRegime
                  targetFlag(iSource)=1
                  nTargets=nTargets+1
                  iTarget=iSource
                  targetListFlag=0
               endif ! .not. omgsourceregime
            endif ! (sourceShape(iSource) .eq. 0)
         enddo ! iSource=0,nSources2-1
         !write(*,*) "L1690 nonTargetExtraction=", nonTargetExtraction
         !write(*,*) "L1691 targetListFlag=", targetListFlag
         ! Check also whether the target is present among the newly detected sources


         !write(*,*) "L2113 nTargets=", nTargets, " nonTargetExtraction=", nonTargetExtraction
         !write(*,*) "L2114 targetListFlag=", targetListFlag
         if (nTargets .gt. 0) then
            nonTargetExtraction=.false.
            !write(*,*) " L2226 iTarget=", iTarget, " iDubiousTarget=", iDubiousTarget, &
            !     " targetDistance=", &
            !     targetDistance, " targetDistance2=", targetDistance2
            !***************
            ! VNY 26.03.2007 here is a small bug: the target index could be 
            ! either in the variable iTarget or iDubiousTarget
            ! Therefore, first checking whether the target index in negative
            
            targetIndex=-9999
            if (iTarget .ge. 0) then
               targetIndex=iTarget   
            else
               if (iDubiousTarget .ge. 0) then                 
                  targetIndex=iDubiousTarget
               endif
            endif
            !write(*,*) "L1820 targetIndex=", targetIndex
            if (targetIndex .ge. 0) then ! we can pick up the coordinates from the source list
               if (targetListFlag .eq. 0) then ! target is found in the main list
                  
                  xZeroRef=sourceXCoord(targetIndex)
                  yZeroRef=sourceYCoord(targetIndex)
!-----------------------------
               else                        
                  if (xTarget .gt. 0 .and. xTarget .lt. nx .and. yTarget .gt. 0 &
                       .and. yTarget .lt. ny/2) then
                     !write(*,*) "L1864 yZeroRef=", yZeroRef
                     xZeroRef=xTarget
                     yZeroRef=yTarget
                     nonTargetExtraction=.true.
                  endif
!                     endif ! targetListFlag .eq. 1
                  !  endif ! targetListFlag .eq. 2
               endif ! iTarget .eq. 0
               !******************
            else ! there were no targets               
               if (xTarget .gt. 0 .and. xTarget .lt. nx .and. yTarget .gt. 0 &
                    .and. yTarget .lt. ny/2) then
                  
                  nonTargetExtraction =.true.
                  xZeroRef=xTarget
                  yZeroRef=yTarget
               endif
            endif ! targetIndex .ge.0
         else ! there were no targets detected
            
            if (xTarget .gt. 0 .and. xTarget .lt. nx .and. yTarget .gt. 0 &
                 .and. yTarget .lt. ny/2) then
               
               nonTargetExtraction =.true.
               xZeroRef=xTarget
               yZeroRef=yTarget
            endif
         endif ! nTargets .gt. 0

         !write(*,*) " L1894 nonTargetExtraction=", nonTargetExtraction
         !write(*,*) " L1895 xZeroRef=", xZeroRef, " yZeroRef=", yZeroRef, " <--------"
         !write (*,*) "L1896 xTarget=", xTarget, " yTarget=", yTarget, &
         !     " nx=", nx, " ny/2=", ny/2
!********************************************* 03.11.2007
         !--- Analysis of the source interconnections
!         do iSource=0, nSources-1
         do iSource=0, nSources2-1
            !write(*,*) "L1909 iSource=", iSource, " sourceRelation=", sourceRelation(iSource), &
            !     " sourceShape=", sourceShape(iSource)
            if (sourceRelation(iSource) .lt. 0) then 
               !  should analyse distances only for point-like sources
               if (sourceShape(iSource) .eq. 0) then
                  !
                  !if (iSource .eq. 73) then
                  !   write(*,*) "L1886 iSource=", iSource, " sourceShape=", sourceShape(iSource)
                  !endif
!                  do i=0, nSources-1
                  do i=0, nSources2-1
                     if (sourceShape(i) .eq. 1) then 
                        if (i .ne. iSource .and. sourceRelation(i) .lt. 0) then
                           deltaX=abs(sourceXCoord(i)-sourceXCoord(iSource))
                           if (omgsourceRegime) deltaX=0.0
                           deltaY=(sourceYCoord(i)-sourceYCoord(iSource))
                           predictedSpectrumPosition=sourceYCoord(iSource)+region2
                           deltaX=abs(sourceXCoord(i)-sourceXCoord(iSource))
                           !if (i .eq. 73) then
                           !   write(*,*) "L1897 i=", i, " deltaX=", deltaX, " deltaY=", deltaY, &
                           !        " checkWidth=", checkWidth
                           !endif
                           blueEdge=sourceYCoord(i)-spectrumTolerance/2.0
                           redEdge=sourceYCoord(i)+1.8*spectrumTolerance/2.0
                           if (omgsourceRegime) predictedSpectrumPosition=(blueEdge+redEdge)/2.0
                            !if (iSource .eq. 73 .and. i .eq. 29) then
                            !   write(*,*) "L1904 blueEdge=", blueEdge, " redEdge=", redEdge, &
                            !        " predictedSpectrumPosition=", predictedSpectrumPosition, &
                            !        " checkWidth=", checkWidth
                            !endif
                           if (deltaX .lt. checkWidth &
                                .and. predictedSpectrumPosition .gt. blueEdge &
                                .and. predictedSpectrumPosition .lt. redEdge) then
                              if (sourceRelation(iSource) .lt. 0) then
                                 sourceRelation(i)=iSource
                                 sourceRelation(iSource)=i
                              endif
                           endif
                        endif
                     endif
                  enddo 
               endif ! sourceShape .eq. 0
            endif ! sourceRelation .lt. 0
         enddo ! iSource=0, nSources2-1
         !
         !write(*,*) "-----------------------------------------------"
         !do iSource=0,nSources2-1
         !   write(*,*) "L2231 iSource=", iSource, " x=", sourceXCoord(iSource), &
         !        " y=", sourceYCoord(iSource), " sourceRelation=", sourceRelation(iSource), &
         !        " sourceShape=", sourceShape(iSource)
         !enddo
         !write(*,*) "L2234------------------"
         !do i=1,nSourcesListCheck
         !   write(*,*) "L1951 iCheck=",i, " indexCheckSources=", indexCheckSources(i)
         !enddo
         !write(*,*) "L1953------------------"
!--
!  Checking if there are some undetected extended sources   
!                related to some of the rest of the point-like sources 
!                by using source2spectrum and spectrumLength parameters
         !write(*,*) "L1462 nSources=", nSources
!         do iSource=0, nSources-1
         do iSource=0, nSources2-1
            !write(*,*) "L1463 iSource=", iSource
            if (sourceShape(iSource) .eq. 0 .and. sourceRelation(iSource) .lt. 0) then 
               ! Check only the point-like sources
               ! calculate flux within the expected area of    
               brightness1=0.0
               nbr1=0.0
               brightness2=0.0
               nbr2=0.0
               brightness3=0.0 ! the maximal value of the strip
               meanBr3=0.0
               nbr3=0.0
               iy=sourceYCoord(iSource)+source2spectrum
               if (iy .ge. 0 .and. iy .lt. ny) then                                    
                  do ii=iy-int(spectrumLength/2), iy+int(spectrumLength/2)
                     if (ii .ge. 0 .and. ii .lt. ny) then
                        brightness3=0.0
                        do jj=-int(checkWidth),int(checkWidth)
                           jx=sourceXCoord(iSource)+jj
                           jxR=jx+checkWidth*3+1
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness1=brightness1+oswImage(jx,ii)
                              nbr1=nbr1+1.0
                              ! compute the maximal value on the strip
                              if (oswImage(jx,ii) .gt. brightness3) brightness3=oswImage(jx,ii)
                           endif
                           if (jxR .ge. 0 .and. jxR .lt. nx) then
                              brightness2=brightness2+oswImage(jxR,ii)
                              nbr2=nbr2+1.0
                           endif
                        enddo
                        meanBr3=meanBr3+brightness3
                        nbr3=nbr3+1.0
                     endif
                  enddo
                  if (nbr1 .gt. 0.0) brightness1=brightness1/nbr1
                  if (nbr2 .gt. 0.0) brightness2=brightness2/nbr2
                  if (nbr3 .gt. 0.0) meanBr3=meanBr3/nbr3
               endif
            endif ! sourceShape .eq. 0
         enddo ! iSource=0, nSource-1

         if (hasColumn(inTableH, "RA")) then
            inColumnH(10) = column(inTableH, "RA", modify)
            sourceRA => real64Data(inColumnH(10))
         else
            inColumnFlag(10)=0
            allocate(sourceRA(0:nSources-1))
         endif

         if (hasColumn(inTableH, "DEC")) then
            inColumnH(11) = column(inTableH, "DEC", modify)
            sourceDEC => real64Data(inColumnH(11))
         else
            inColumnFlag(11)=0
            allocate(sourceDEC(0:nSources-1))
         endif

         if (hasColumn(inTableH, "LII")) then
            inColumnH(12) = column(inTableH, "LII", modify)
            sourceLII => real64Data(inColumnH(12))
         else
            inColumnFlag(12)=0
            allocate(sourceLII(0:nSources-1))
         endif

         if (hasColumn(inTableH, "BII")) then
            inColumnH(13) = column(inTableH, "BII", modify)
            sourceBII => real64Data(inColumnH(13))
         else
            inColumnFlag(13)=0
            allocate(sourceBII(0:nSources-1))
         endif
!--------------------------------------------- NB changes 27.11.2006
         columnDataTypeFlag=0
         if (hasColumn(inTableH, "QFLAG")) then
            inColumnH(14) = column(inTableH, "QFLAG", read)

            columnDataTypeFlag=columnDataType(inColumnH(14))
            !write(*,*) "L1617 columnDataType=", columnDataTypeFlag
            !write(*,*) "****************************************************"
            if (columnDataTypeFlag .eq. 2) then
               sourceQF8 => int8Data(inColumnH(14))
            else ! columnDataTypeFlag=4
               sourceQF16 => int16Data(inColumnH(14))
            endif
         else
            inColumnFlag(14)=0
            allocate(sourceQF16(0:nSources-1))
         endif
!---------------------------------------------
         if (hasColumn(inTableH, "CFLAG")) then
            inColumnH(15) = column(inTableH, "CFLAG", read)
            sourceCF => int8Data(inColumnH(15))
         else
            inColumnFlag(15)=0
            allocate(sourceCF(0:nSources-1))
         endif
         if (hasColumn(inTableH, "PSF1_CORR")) then
            inColumnH(16) = column(inTableH, "PSF1_CORR", modify)
            psf1Correction => real32Data(inColumnH(16))
         else
            inColumnFlag(16)=0
            allocate(psf1Correction(0:nSources-1))
            psf1Correction = nullValue
         endif
         allocate(psf2Correction(0:nSources-1))    
         psf2Correction = nullValue
         ! Add columns with the spectrum identification number to the (input) source list table 
         if (.not. hasColumn(inTableH, "SPECTR_ID")) then
            spectrumNoCol = addColumn(inTableH, "SPECTR_ID", integer32, "", columnDim, "spectrum identifier")
         else
            spectrumNoCol = column(inTableH, "SPECTR_ID", modify)
         endif
         spectrumNoArray0 => int32data(spectrumNoCol)
         do iSource=0, nSources-1
            spectrumNoArray(iSource)=spectrumNoArray0(iSource)
         enddo
         call setNullValue(spectrumNoCol, 999999 )
         do iSource=0, nSources-1
            call setToNull(spectrumNoCol, iSource)
         enddo
         ! The column with the spectrum origin (the number of the corresponding point-like source)
         if (.not. hasColumn(inTableH, "REL2SRCNUM")) then
            spectrumOriginCol = addColumn(inTableH, "REL2SRCNUM", integer32, "", columnDim, "number of the related source")
         else
            spectrumOriginCol = column(inTableH, "REL2SRCNUM", modify)
         endif
         spectrumOriginArray => int32data(spectrumOriginCol)
         call setNullValue(spectrumOriginCol, 999999 )
         do iSource=0, nSources-1
            call setToNull(spectrumOriginCol, iSource)
         enddo

         !-(origin of the OSW)
         windowX0 = int16Attribute(sourceListSetH, "WINDOWX0")
         windowY0 = int16Attribute(sourceListSetH, "WINDOWY0")
         windowDX = int16Attribute(sourceListSetH, "WINDOWDX")
         windowDY = int16Attribute(sourceListSetH, "WINDOWDY")
         windowX1=windowX0+windowDX-1
         windowY1=windowY0+windowDY-1
         !write(*,*) "L2036 windowX0=", windowX0, " windowY0=", windowY0, &
         !     " windowDX=", windowDX, " windowDY=", windowDY, " <================="
         allocate(bkgVect(0:ny-1))
         allocate(bkgVect2(0:ny-1))
         allocate(bkgVect3(0:ny-1))
         allocate(bkgVectR(0:ny-1))
         allocate(bkgVectL(0:ny-1))
         allocate(specVect(0:ny-1))
         allocate(smoothBkgVectR(0:ny-1))
         allocate(smoothBkgVectL(0:ny-1))
         allocate(countsScaleFactor(0:ny-1))

         ! VNY 17.11.2003
         ysize=ny
         xsize=nx
         allocate(wavelengthArray(0:ny-1))
         allocate(spectrumArray(0:ny-1))
         allocate(maskArray(0:ny-1))
         !allocate(spectrumMaskArray(0:ny-1))
         allocate(spectrumErrArray(0:ny-1))
         allocate(bkgArray(0:ny-1))
         allocate(bkgErrArray(0:ny-1))
         allocate(fluxArray(0:ny-1))
         allocate(fluxErrArray(0:ny-1))

         !***********************************************************
         ! Compute the average value of the image
         !***********************************************************
         black = minval(oswImage)
         white = maxval(oswImage)
         averImage=0.0
         sum=0.0
         do i=0,nx-1
            do j=0,ny-1
               averImage=averImage+oswImage(i,j)
               sum=sum+1.0
            enddo
         enddo

         if (sum .gt. 0.0) then
            averImage=averImage/sum
         endif

         !write(*,*) "L2126------------------"
         !do i=1,nSourcesListCheck
         !   write(*,*) "L2128 iCheck=",i, " indexCheckSources=", indexCheckSources(i)
         !enddo
         !write(*,*) "L2130"
!***********************************************************************
!***********************************************************************
!***********************************************************************
! Target spectrum
!***********************************************************************
!***********************************************************************
!***********************************************************************
         nContaminatedSources=0
         !write(*,*) "L2139 nTargets=", nTargets, " nontargetExtraction=", nonTargetExtraction
         if (nTargets .gt. 0 .or. nonTargetExtraction) then 
            xZeroOrder=0.0
            yZeroOrder=0.0
            xFirstOrder=0.0
            yFirstOrder=0.0
!            iSource=-1
            iSource=-9999
            !write(*,*) "L2453 iSource=", iSource, " nTargets=", nTargets, " nonTargetExtraction=", &
            !     nonTargetExtraction
            if (iDubiousTarget .gt. -1) iSource=iDubiousTarget
            !write(*,*) "L2456 iSource=", iSource, " iDubiousTarge=", iDubiousTarget
            !write(*,*) "L2457 iTarget=", iTarget, " xZeroref=", xZeroRef, " yZeroref=", &
            !     yZeroref
            ! VNY 04.06.05
            !if (iTarget .gt. -1) iSource=iTarget
            if (iTarget .gt. -9000) iSource=iTarget
            !write(*,*) "L2462 iSource=", iSource, " iTarget=", iTarget
            if (iSource .gt. -1) then
               !write(*,*) "L2157 iSource=", iSource, " iTarget=", iTarget
               !write(*,*) "L2158 xZeroRef=", xZeroref, " sourceXCoord=",sourceXCoord(iSource)
               !write(*,*) "L2159 sourceRelation=", sourceRelation(iSource)
               !write(*,*) "L2160 sourceYCoord=", sourceYCoord(iSource)
               !i1=int(sourceXCoord(iSource)+0.5)
               i1=int(xZeroRef+0.5)

               !xZero=sourceXCoord(iSource)
               xZero=xZeroRef
               xZeroA=xZero
               !
               !yZero=sourceYCoord(iSource)
               yZero=yZeroRef
               !
               xFirst=-1
               yFirst=-1
               ! The position of the corresponding zero-order is taken as reference
               !xZeroA=sourceXCoord(iSource)

               ! VNY 13.07.2004 Rounding the Y-position in order not to have problems with interpolation 
               ! of the wavelengths
               !!yZeroA=int(sourceYCoord(iSource)+0.5)
               !yZeroA=sourceYCoord(iSource)
               yZeroA=yZeroRef
               if (i1 .ge. nx-1) then
                  i1=nx-2
                  xZero=i1
                  xZeroA=i1
               endif
               !
               xZeroOrder=xZeroA
               yZeroOrder=yZeroA
               !write(*,*) "L2496 xZeroOrder=", xZeroOrder, " yZeroOrder=", yZeroOrder
               procFlag(iSource)=1
               if (sourceRelation(iSource) .ge. 0) then
                  i=sourceRelation(iSource)
                  procFlag(i)=1
                  if (i .lt. nSources) then
                     spectrumOriginArray(i)=iSource+1
                  endif
                  spectrumOriginArray2(i)=iSource+1
                  !
                  lateralSpectrumL = .false.
                  lateralSpectrumR = .false.
                  extend = .false.
                  if (sourceShape(i) .eq. 1) then
                     extended(i)=2
                  endif
                  xFirst=sourceXCoord(i)
                  yFirst=sourceYCoord(i)
                  xFirstOrder=xFirst
                  yFirstOrder=yFirst
                  !write(*,*) "L2660 xFirst=", xFirst, " yFirst=", yFirst
               else
                  ! VNY 08.07.2007 replacing sourceYCoord with yZeroRef 
                  !yFirst=sourceYCoord(iSource)+source2spectrum ! position of the first order
                  !xFirstOrder=xZeroOrder
                  yFirst=yZeroRef+source2spectrum ! position of the first order
                  xFirstOrder=xZeroref
                  yFirstOrder=yFirst
                  !write(*,*) "L2668 yFirst=", yFirst
               endif ! sourceRelation(iSource .ge. 0
            else
               !*********************************************************************
               ! VNy 03.11.2007 The "else" condition is probably not needed here anymore
               !*********************************************************************
               !write(*,*) "L1821 iSource =", iSource
               if (iSource .gt. -900) then
                  !write(*,*) "L2532 iSource=", iSource, " xZeroRef=", xZeroRef
                  !write(*,*) "---------------------------------"
                  ii1=-iSource
                  !write(*,*) "L1825 sourceData%x=", sourceData(ii1)%xPos, " sourceData%y=", sourceData(ii1)%yPos
                  !xZero=sourceData(ii1)%xPos
                  !yZero=sourceData(ii1)%yPos
                  xZero=xZeroRef
                  yZero=yZeroRef
                  xZeroA=xZero
                  yZeroA=yZero
                  !
                  i1=int(xZero+0.5)
                  xFirst=-1
                  yFirst=-1
                  !xZeroA=sourceData(ii1)%xPos
                  !!yZeroA=int(sourceData(ii1)%yPos+0.5)
                  !
                  !yZeroA=sourceData(ii1)%yPos
                  yZeroA=yZeroRef
                  if (i1 .ge. nx-1) then
                     i1=nx-2
                     xZero=i1
                     xZeroA=i1
                  endif

                  xZeroOrder=xZeroA
                  yZeroOrder=yZeroA
                  yFirst=sourceData(ii1)%yPos+source2spectrum
                  xFirstOrder=xZeroOrder
                  yFirstOrder=yFirst
                  !write(*,*) "L2242 yFirst=", yFirst
               else
                  !write(*,*) "L2564 iSource=", iSource, " xtarget=", xTarget
                  i1=int(xTarget+0.5)
                  xZero=xTarget
                  !yZero=int(yTarget+0.5)
                  yZero=yTarget
                  xFirst=-1
                  yFirst=-1
                  xZeroA=xTarget
                  if (i1 .ge. nx-1) then
                     i1=nx-2
                     xZero=i1
                     xZeroA=i1
                  endif
                  !yZeroA=int(yTarget+0.5)
                  yZeroA=yTarget
                  xZeroOrder=xZeroA
                  yZeroOrder=yZeroA
                  xFirst=xTarget
                  yFirst=yTarget+source2spectrum
                  xFirstOrder=xFirst
                  yFirstOrder=yFirst
                  !write(*,*) "L2264 yFirst=", yFirst
               endif ! iSource .gt. -900
            endif ! iSource .gt. -1
            !write(*,*) "L2588 xZero=", xZero, " yZero=", yZero 
            !write(*,*) "L2589 yFirst=", yFirst, " spectrumLength=", spectrumLength
            iyUp=int(yFirst)
            iyDown=int(yFirst-spectrumLength/2.0)
            !write(*,*) "L2270 iyUp=", iyUp, " iyDown=", iyDown

            if (iyUp .ge. ny-2) iyUp=ny-2
            if (iyDown .ge. iyUp) iyDown=iyUp-1
            !write(*,*) "L2274 iyUp=", iyUp, " iyDown=", iyDown

            availableSpectrumLength=0.0
            availableSpectrumFraction=0.0
            !write(*,*) " L2278 yZeroA=", yZeroA, " ny=", ny, " i=", i
            !write(*,*) "L2279 iyDown=", iyDown, " yZZeroA=", int(yZeroA), &
            !     " ny-1=", ny-1, " iyUp=", iyUp, " i1=", i1
            if (iyDown .gt. int(yZeroA) .and. iyDown .lt. ny-1 .and. iyUp .lt. ny-1) then
               !write(*,*) "L2282 i1=", i1, " oswImage(i1,iyDown)=", oswImage(i1,iyDown), &
               !     " oswImage(i1,iyUp)=", oswImage(i1, iyUp), " i=", i
               !write(*,*) "L1896"
               if (oswImage(i1,iyDown) .gt. -0.9 .and. oswImage(i1,iyUp) .gt. -0.9) then
                  ! determine the upper available limit
                  availableSpectrumLength=0.0
                  checkUpA: do iyUp2=iyDown+1,iyDown+int(spectrumLength)
                     if (iyUp2 .lt. ny) then
                        if (oswImage(i1,iyUp2) .gt. -0.9) then
                           availableSpectrumLength=availableSpectrumLength+1.0
                        else
                           exit checkUpA
                        endif
                     endif
                  enddo checkUpA
                  !write(*,*) "L2297 availableSpectrumLength=", availableSpectrumLength
               endif
            else
               !*******************************************************
               ! NB here is the problem revealed by SPR-3522 22.09.2005
               !*******************************************************
               !write(*,*) "L2303 spectrumLength=", spectrumLength
               availableSpectrumLength=spectrumLength
               yFirst=yTarget+source2spectrum
               yFirstOrder=yFirst
               !sourceYCoord(i)=yFirst
               xFirst=xZeroOrder
               xFirstOrder=xFirst
               
            endif
            !write(*,*) "L2634 nTargets=", nTargets, " nonTargetExtraction=", nonTargetExtraction
            !write(*,*) "L2635 availableSpectrumLength=", availableSpectrumlength
            !write(*,*) "L2636 spectrumLength=", spectrumLength, " xZeroA=", xZeroA, " yZeroA=", yZeroA
            availableSpectrumFraction=availableSpectrumLength/spectrumLength
            ! Extract the spectrum only if more than 50 percent of the spectrum is available
               !write(*,*) "L2317 availableSpectrumFraction=", availableSpectrumFraction
               if (availableSpectrumFraction .gt. 0.5) then !extract
!**********************************************************************
         !  Create a single file containing spectra for all the sources            
!**********************************************************************
                  !write(*,*) "L1874 before creation of the file: ", trim(outFile)
                  outFile4=outFile
                  outTargetSet0H = dataSet(outFile4, create) ! This file will contain 
                  ! only the target spectrum
                  !write(*,*) "L1878 after creation of the file: ", trim(outFile4)
                  createSpectrumFile=.true.
                  createSpectraListFile=.true.
                  call setAttribute(outTargetSet0H, "CONTENT", "OM GRISM SPECTRA", &
                       "File contains OM grsim spectrum/spectra")
                  call setAttribute(outTargetSet0H, "HDUCLASS", "OGIP    ", &
                       "Format conforms to OGIP/GSFC conventions")
                  call setAttribute(outTargetSet0H, "HDUCLAS1", "SPECTRUM", &
                       "File contains spectral data")
                  call setAttribute(outTargetSet0H, "HDUVERS", "1.2.0    ", &
                       "The version number of the format")
                  call setAttribute(outTargetSet0H, "TELESCOP", "XMM     ", &
                       "XMM mission")
                  call setAttribute(outTargetSet0H, "INSTRUME", "OM      ", &
                       "Instrument name")
                  call setAttribute(outTargetSet0H, "DATAMODE", dataMode, &
                       "Data mode")
                  call setAttribute(outTargetSet0H, "FILTER", filter, &
                       "Filter used for obtaining the data")
                  call setAttribute(outTargetSet0H, "FRAMTIME", frameTime, &
                       "Frame time")
                  call setAttribute(outTargetSet0H, "DEADFRAC", deadFrac, &
                       "Frame dead fraction")
                  call setAttribute(outTargetSet0h, "TMEXP_ID", tmExpId, &
                       "Exposure Id from telemetry")
                  call setAttribute(outTargetSet0H, "FRMTIME", frmTime, &
                       "Frametime from telemetry")
                  call setAttribute(outTargetSet0H, "DATATYPE", dataType, &
                       "Data type")
                  call setAttribute(outTargetSet0H, "OBS_ID", obsId, &
                       "Observation identifier")
                  call setAttribute(outTargetSet0H, "EXP_ID", expId, &
                       "Exposure identifier")
                  call setAttribute(outTargetSet0H, "DATE-OBS", dateObs, &
                       "Observation start time")
                  call setAttribute(outTargetSet0H, "DATE-END", dateEnd, &
                       "Observation end time")
!*****************************************************************
                  sourceArea=0.0
                  srcNo=iSource
                  timeExposure = exposure
                  nBins=ny
                  xFirst=xFirstOrder
                  yFirst=yFirstOrder
                  !extractionMode=1
                  subtractBkg=.true.
                  verbFlag=0
                  ! extract spectrum with the evaluation purposes (mode=2)
                  !----------------------------------------------------------------------
                  xZeroIn=xZeroA
                  yZeroIn=yZeroA
                  verbFlag=0
                  !write(*,*) "OMGRISM L1532 omgsourceRegime=", omgsourceRegime

                  !*******************************************
                  ! TEST OMGSOURCE REGIME
                  !*******************************************
                  !omgsourceRegime=.true.
                  !*******************************************
                  if (.not. omgsourceRegime) then
                     !verbFlag=1
                     verbFlag=0
                     call CheckZeroOrder(grismSetH, xZeroIn, yZeroIn, &
                          spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                          verbFlag, xZeroOut, yZeroOut, averCountIn, averCountOut)
                     !write(*,*) "L1993 averCountIn=", averCountIn, " averCountOut=", averCountOut
                     ! Check if the new position is different from the initial one
                     distance=(xZeroOut-xZeroIn)*(xZeroOut-xZeroIn)+(yZeroOut-yZeroIn)*(yZeroOut-yZeroIn)
                     distance= sqrt(distance)
                     averCountLimit=averCountIn+3.0*sqrt(averCountIn)
                     !write(*,*) "L2713 distance=", distance, " averCountLimit=", averCountLimit
                     !write(*,*) "L2714 xZeroIn=", xZeroIn, " yZeroIn=", yZeroIn, &
                     !     " xZeroOut=", xZeroOut, " yZeroOut=", yZeroOut
                     !********************************************************************************
                     ! NB (VNY) this is commented 06.09.2005 in order to make uniform 
                     ! the position of the zero order 
                     ! 
                     ! Correcting the coordinates if the maximum is too far from the 
                     ! previously measured position of the source
                     ! if (distance .gt. 0.7*spectrumHalfWidth .and. averCountOut .gt. 
                     ! averCountLimit) then
                     !   call warning("ObjectPosition", "Coordinates of the 
                     ! zero-order spectrum seem to be wrong (corrected)")
                     !   xZeroA=xZeroOut
                     !   yZeroA=yZeroOut
                     !   xFirst=xZeroA
                     !   xZeroOrder=xZeroA
                     !   yZeroOrder=yZeroA
                     !   xFirstOrder=xZeroA
                     !endif
                     !***************************
                     ! Check if the target position is improved
                     foundRealTarget=.false.
                     foundDubiousTarget=.false.
                     foundNoTarget=.false.
                     !write(*,*) "L2738 distance=", distance, " averCountOut=", averCountOut, &
                     !     " averCountLimit=", averCountLimit
                     !write(*,*) "L2024 20/binY=", 20/binY, " binY=", binY, " averCountIn=", averCountIn
! VNY 03.07.2006 A little bit more sophysticated check for very bright sources 
!                     if (distance .lt. 20/binY .and. averCountOut .gt. averCountLimit) then
                     if (distance .lt. 8/binY .and. averCountOut .gt. averCountIn .or. &
                          distance .lt. 20/binY .and. averCountOut .gt. averCountLimit) then
                        foundRealTarget=.true.
                        call warning("ObjectPosition", "Default target coordinates corrected")
                        !write(*,*) "L2027 found real target"
                        !write(*,*) "L2748 xZeroOut=", xZeroOut, " yZeroOut=", yZeroOut
                        ! here the centroid must be re-computed (6.09.05)
                        iMin=int(xZeroOut)-10
                        if (iMin .lt. xMin) iMin=xMin 
                        iMax=int(xZeroOut)+10
                        if (iMax .gt. xMax) iMax=xMax
                        jMin=int(yZeroOut)-10
                        if (jMin .lt. yMin) jMin=yMin
                        jMax=int(yZeroOut)+10
                        if (jMax .gt. yMax) jMax=yMax
                        !write(*,*) "L2758 xMin=", xMin, " xmax=", xMax, " yMin=", yMin, " yMax=", yMax
                        call AnalyseZeroOrder(safeOSW, patternOSW, xMin, xMax, yMin, yMax, &
                             iMin,iMax,jMin,jMax,xPos,yPos,xPosErr,yPosErr,&
                             aValue,bValue,theta )
                        !write(*,*) "L2762 xPos=", xPos, " yPos=", yPos
                        xZeroA=xPos!xZeroOut
                        yZeroA=yPos!yZeroOut
                        xFirst=xZeroA
                        xZeroOrder=xZeroA
                        yZeroOrder=yZeroA
                        xFirstOrder=xZeroA
                        !write(*,*) "L2769 xZeroOrder=", xZeroorder, " yZeroOrder=", yZeroOrder
                     else
                        !if (distance .lt. 40/binY .and. averCountOut .gt. averCountLimit) then
                        if (distance .lt. 30/binY .and. averCountOut .gt. averCountIn .or. &
                             distance .lt. 40/binY .and. averCountOut .gt. averCountLimit) then
                           foundDubiousTarget=.true.
                           !write(*,*) " L2774 found DUBIOUS target"
                           call warning("ObjectPosition", "Default target coordinates corrected")

                           ! here the centroid must be re-computed (6.09.05)

                           iMin=int(xZeroOut)-10
                           if (iMin .lt. xMin) iMin=xMin 
                           iMax=int(xZeroOut)+10
                           if (iMax .gt. xMax) iMax=xMax
                           jMin=int(yZeroOut)-10
                           if (jMin .lt. yMin) jMin=yMin
                           jMax=int(yZeroOut)+10
                           if (jMax .gt. yMax) jMax=yMax

                           call AnalyseZeroOrder(safeOSW, patternOSW, xMin, xMax, yMin, yMax, &
                                iMin,iMax,jMin,jMax,xPos,yPos,xPosErr,yPosErr,&
                                aValue,bValue,theta )
                           xZeroA=xPos!xZeroOut
                           yZeroA=yPos!yZeroOut
                           xFirst=xZeroA
                           xZeroOrder=xZeroA
                           yZeroOrder=yZeroA
                           xFirstOrder=xZeroA
                           !write(*,*) "L2797 xZeroOrder=", xZeroorder, " yZeroOrder=", yZeroOrder
                        endif
                     endif
                     !write(*,*) "L2801 xZeroOrder=", xZeroOrder, " yzeroOrder=", yZeroOrder
                     if(allocated(patternOSW)) then
                        deallocate(patternOSW,stat=istat)
                        call CheckDeallocation(allocErr, "patternOSW")
                     end if
                     if(allocated(safeOSW)) then
                        deallocate(safeOSW,stat=istat)
                        call CheckDeallocation(allocErr, "safeOSW")
                     end if

                     !----------------------------------------------------------------------
                     ! VNY 14.10.2004 Check the x-position of the first order
                     xFirstIn=xFirst
                     yFirstIn=yFirst
                     verbFlag=0
                     call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                          spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                          verbFlag, xFirstOut, yFirstOut, averCountIn, &
                          averCountOut)
                     !write(*,*) "L2820 OMGRISM L1815 xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn, &
                     !     " xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                     verbFlag=0
                     !write(*,*) " abs(dif)=", abs(xFirstIn-xFirstOut), "spectrumHalfWidth=", &
                     !     spectrumHalfWidth
!                     if (abs(xFirstIn-xFirstOut) .lt. spectrumHalfWidth) then ! correct the position
                     if (abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then ! correct the position
                        xFirst=xFirstOut
                        xFirstOrder=xFirst
                     endif
                  else
                     xZeroOrder=xZeroA
                     yZeroOrder=yZeroA
                     xFirstOrder=xFirst
                     yFirstOrder=yFirst
                  endif ! not omgsourceRegime
                  
                  ! ----- 11.01.05
                  verbFlag=0
                  !write(*,*) "L2834 xZeroA=", xZeroA, " yZeroA=", yZeroA, " xFirst=", &
                  !     xFirst, " yFirst=", yFirst
                  !write(*,*) "L2475 proximityX1=", proximityX1
                  call ExtractSpectrum(grismSetH, xZeroA, yZeroA, xFirst, yFirst, &
                       spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                       bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                       spectrumArray, spectrumErrArray, bkgArray, &
                       bkgErrArray, fluxArray, fluxErrArray, ysize, &
                       evaluationMode, subtractBkg, verbFlag,s2n,s2n0, &
                       crossShape, leftBorder, rightBorder, proximityX1, nSp)

                  !write(*,*) "L2079 Target spectrum s2n=", s2n, " s2n0=", s2n0, " crossShape=", crossShape, " <-----"
                  if (s2n0 .lt. s2n*10) then
                     foundNoTarget=.true.
                  endif
                  blueEdge=yZeroA+region1
                  redEdge=yZeroA+region2
                  iblueEdge=int(blueEdge)
                  iredEdge=int(redEdge)
                  verbFlag=0
                  yFirst=(blueEdge+redEdge)/2.0
                  spectrumHalfLength=(redEdge-blueEdge)/2.0
                  sizeMask=ny
                  !verbFlag=1

!-----------------------
                  call FindPossibleFeatures2 (spectrumArray, smoothWindowHalfWidth, &
                       maxValue, maxValue2, meanValue2, maskArray, &
                       ny, iblueEdge, iredEdge, verbFlag, nFeatures) 
!----------------------
                  contaminationComment=" "
                  putContaminationComment=.false.
                  if (nFeatures .gt. 0) then ! issue a warning
                     !!call warning("Overlapping", "The spectrum is likely to be contaminted by zero-order features")
                     !!write(errstr,*) "The number of likely contaminating zero-order features:  ", nFeatures
                     if (iSource .gt. -1) then
                        write(errstr,*) "Source: ", iSource, " is likely to be contaminated; zero-order features:  ", nFeatures
                     else
                        write(errstr,*) "Source (default target) is likely to be contaminated; zero-order features:  ", nFeatures
                     endif
                     nContaminatedSources=nContaminatedSources+1
                     call message(AppMsg, 3)
                     write(contaminationComment, '(a, i3, a)') &
                          "The spectrum is likely to be contaminated by (",  &
                          nFeatures, ") zero-order features"
                     putContaminationComment=.true.
                  endif

                  verbFlag=0
                  call EvaluateSpectrum(spectrumArray, bkgArray, &
                       ysize, iblueEdge, iredEdge, bkgWidth, &
                       spectrumWidth, sbRatio, maskArray, verbFlag)
                  !write(*,*) "L2136 iSource=", iSource, " sbRatio=", sbRatio, " s2n=", s2n, " s2n0=", s2n0
                  !if (iSource .gt. -1 .and. sbRatio .lt. 0.8) then
                  if (iSource .gt. -1 .and. sbRatio .lt. 0.8 .and. s2n0 .lt. 6.0) then
                     !write(*,*) "L2138 iSource=", iSource
                     procFlag(iSource)=3 ! flag 3 means Default extraction 
                  endif ! sbRatio .lt. 0.8 
!!
                  !write(*,*) "L1869 extractionMode=", extractionMode, &
                  !     " evaluationMode=", evaluationMode
                  if (extractionMode .ne. evaluationMode) then
                     if (.not. omgsourceRegime) then
                        ! VNY 14.10.2004 Check the x-position of the first order
                        xFirstIn=xFirst
                        yFirstIn=yFirst
                        verbFlag=0
                        call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                             spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                             verbFlag, xFirstOut, yFirstOut, averCountIn, &
                             averCountOut)
                        !write(*,*) " OMGRISM L1890 xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn, &
                        !     " xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                        verbFlag=0
                        !write(*,*) "L1893 difference=", abs(xFirstIn-xFirstOut), " spectrumHalfWidth=", &
                        !     spectrumHalfWidth, " iSource=", iSource
!                        if (abs(xFirstIn-xFirstOut) .lt. spectrumHalfWidth) then ! correct the position
                        if (abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then ! correct the position
                           !write(*,*) "L1894 difference=", abs(xFirstIn-xFirstOut)
                           xFirst=xFirstOut
                           xFirstOrder=xFirst
                        endif
                     endif
                     !--- 11.01.05
                     verbFlag=0
                     !----
                     !write(*,*) "L2155 xZeroA=", xZeroA, " yZeroA=", yZeroA, " xFirst=", &
                     !     xFirst, " yFirst=", yFirst
                     !if (iSource .eq. 17) then
                     !   verbFlag=1
                     !endif
                     !write(*,*) "L2563 proximityX1=", proximityX1
                     call ExtractSpectrum(grismSetH, xZeroA, yZeroA, xFirst, yFirst, &
                          spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                          bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                          spectrumArray, spectrumErrArray, bkgArray, &
                          bkgErrArray, fluxArray, fluxErrArray, ysize, &
                          extractionMode, subtractBkg, verbFlag,s2n,s2n0, &
                          crossShape, leftBorder, rightBorder, proximityX1, nSp)
                     !write(*,*) "L2156 Target spectrum s2n=", s2n, " s2n0=", s2n0, " crossShape=", crossShape, " <-----"
                     if (s2n0 .lt. s2n*10) then
                        foundNoTarget=.true.
                     endif
                  endif
                  ! First check if there was some first-order within the detected spectrum
                  ! and mark this first-order as already identified
                  if (.not. nonTargetExtraction) then
! VNY 03.11.2007
                     !do i=0, nSources-1
                     do i=0, nSources2-1
                        if (sourceShape(i) .eq. 1) then 
                           if (i .ne. iSource .and. sourceRelation(i) .lt. 0) then
                              if (sourceXcoord(i) .gt. xZeroA-shapeCriterium2 &
                                   .and. sourceXcoord(i) .lt. xZeroA+shapeCriterium2) then
                                 if (sourceYcoord(i) .gt. blueEdge .and. &
                                      sourceYcoord(i) .lt. redEdge) then
                                    ! Label the source
                                    sourceRelation(i)=iSource
                                    !write(*,*) "L1912 i=", i, " sourceRelation(i)=", &
                                    !     sourceRelation(i)
                                 endif
                              endif
                           endif ! i .ne. iSource
                        endif ! sourceShape(i) .eq. 1
                     enddo ! i=0,nSources-1
                  endif ! .not. nonTargetExtraction

                  verbFlag=0
                  !verbFlag=1
                  !call RemoveZeroOrders(spectrumArray, bkgArray, fluxArray, &
                  !     maskArray, ySize, iblueEdge, iredEdge, verbFlag)
                  nFeatures=0

!                     spectrumNoArray(iSource)=iSource+1

! VNY 20.07.2004                  
!                  spectrumOriginArray(iSource)=iSource+1
                  spectrumNoCount=spectrumNoCount+1
!                  if (nonTargetExtraction) then
                  if (nonTargetExtraction .or. iSource .lt. 0) then
                     spectrumNoArray2(spectrumNoCount)=0
                     spectrumOriginArray2(spectrumNoCount)=-1
                     srcNoArray2(spectrumNoCount)=0
                  else
                     if (iSource .lt. nSources) then
                        spectrumNoArray0(iSource)=spectrumNoCount+1
                     endif
                     spectrumNoArray(iSource)=spectrumNoCount+1
                     spectrumNoArray2(spectrumNoCount)=iSource+1
                  
                     if (sourceRelation(iSource) .ge. 0) then
                        jj=sourceRelation(iSource)
                        if (iSource .lt. nSources) then
                           spectrumOriginArray(iSource)=jj+1
                        endif
                        spectrumOriginArray2(iSource)=jj+1                     
                        sourceRelation(jj)=iSource+1
                        if (jj .lt. nSources) then
                           spectrumNoArray0(jj)=spectrumNoCount+1
                        endif
                        spectrumNoArray(jj)=spectrumNoCount+1
                        spectrumNoArray2(spectrumNoCount)=iSource+1                    
                     else
                        if (iSource .lt. nSources) then
                           spectrumOriginArray(iSource)=-1
                        endif
                        spectrumOriginArray2(iSource)=-1
                     endif
                     if (iSource .lt. nSources) then
                        spectrumOriginArray2(spectrumNoCount)=spectrumOriginArray(iSource)
                     endif
                     !
                     srcNoArray2(spectrumNoCount)=iSource
                  endif
                  x0aux(spectrumNoCount)=xZeroOrder
                  y0aux(spectrumNoCount)=yZeroOrder
                  x1aux(spectrumNoCount)=xFirstOrder
                  y1aux(spectrumNoCount)=yFirstOrder
                  !write(*,*) "L2248 iSource =", iSource, " xZeroOrder=", xZeroOrder, &
                  !     " yZeroOrder=", yZeroOrder
                  !write(*,*) "L2249^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
                  !write(*,*) "L2149 iTarget=", iTarget
                  !if (iTarget .gt. -1) then
                  if (iTarget .gt. -900) then
                     !write(*,*) "L1963 iTarget=", iTarget
                     specTypeAux(spectrumNoCount)=1 ! Target
                     if (iTarget .lt. 0) specTypeAux(spectrumNoCount)=5 ! Re-detected target
                     !write(*,*) "L1966 spectrumNoCount=", spectrumNoCount, " specTypeAux=", &
                     !     specTypeAux(spectrumNoCount)
                  else
                     specTypeAux(spectrumNoCount)=2 ! Dubious target
                  endif

                  !if (nonTargetExtraction) then
                  !write(*,*) "L2236 nonTargetExtraction=", nonTargetExtraction, " iSource=", iSource, &
                  !     " foundNoTarget=", foundNoTarget
!                  if (nonTargetExtraction .or. iSource .lt. 0 ) then
                  if (nonTargetExtraction .or. iSource .lt. -900  .or. foundNoTarget) then
                     !write(*,*) "L2265"
                     specTypeAux(spectrumNoCount)=4 ! default extraction with no zero order
                     if (foundRealTarget) specTypeAux(spectrumNoCount)=1 
                     ! real target found during the second check
                     if (foundDubiousTarget) specTypeAux(spectrumNoCount)=2 
                     ! found dubious target during the second check
                  else
                     !write(*,*) "L2272"
                     if (iSource .gt. 0) then
                        !write(*,*) "L2274"
                        if (procFlag(iSource) .eq. 3) specTypeAux(spectrumNoCount)=3 ! default extraction
                     endif
                  endif
                  !write(*,*) "L2278 iSource=", iSource
                  write(sourceIndex, '(i3.3)') spectrumNoCount+1
                  !
                  write(blockName, '(a,a)') "RATE", sourceIndex
                  outTargetTable0H=addTable(outTargetSet0H, blockName, nBins, &
                       "Table with the target source spectrum")
                  !outFieldTable0H=addTable(outFieldSet0H, blockName, nBins, &
                  !     "Table with the target source spectrum")
                  !  Adding the spectrum number to the source list file
                  blkNum=blockNumber(outTargetSet0H, blockName)
                  !spectrumNoArray(iSource)=blkNum+1
                  !spectrumOriginArray(iSource)=sourceRelation(iSource)+1
                  ! 
                  ! Computing the binning interval
                  if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                     binningInterval=2.40774*binY ! in Angstroms
                  else
                     binningInterval=5.626*binY
                  endif
                  !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                  call setAttribute(outTargetTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                  call setAttribute(outTargetTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                  call setAttribute(outTargetTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                  call setAttribute(outTargetTable0H, "HDUCLAS2", "NET", "Background subtracted")
                  call setAttribute(outTargetTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                  call setAttribute(outTargetTable0H, "DLAMBDA", binningInterval, "Angstrom", "binning interval")
                  call setAttribute(outTargetTable0H, "EXTRMODE", extractionMode, "Spectra extraction mode")
                  
                  !--Adding columns to the output table
                  waveSourceOutCol0 = addColumn(outTargetTable0H, "WAVELENGTH", real32, &
                       "ANGSTROM", columnDim, "wavelengths ")
                  waveSourceOutArray0 => real32data(waveSourceOutCol0)
                  !
                  !------------------
                  rateSourceOutCol0 = addColumn(outTargetTable0H,"RATE", real32, &
                       "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                  rateSourceOutArray0 => real32data(rateSourceOutCol0)
                  !
                  !--------------
                  rateSourceErrOutCol0 = addColumn(outTargetTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                       columnDim, "source counts rate error")
                  rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                  !
                  !------------------
                  rateBackOutCol0 = addColumn(outTargetTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                       columnDim, " background counts rate")
                  rateBackOutArray0 => real32data(rateBackOutCol0)
                  !
                  !----------------
                  rateBackErrOutCol0 = addColumn(outTargetTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                       columnDim, " background counts rate error")
                  rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                  !
                  !------------------
                  fluxOutCol0 = addColumn(outTargetTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                       columnDim, " background counts rate")
                  fluxOutArray0 => real32data(fluxOutCol0)
                  !-------------
                  fluxErrOutCol0 = addColumn(outTargetTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                       columnDim, " background counts rate error")
                  fluxErrOutArray0 => real32data(fluxErrOutCol0)
                  !
                  !
                  
                  !
                  waveSourceOutArray0=wavelengthArray
                  rateSourceOutArray0=spectrumArray
                  rateSourceErrOutArray0=spectrumErrArray
                  rateBackOutArray0=bkgArray
                  rateBackErrOutArray0=bkgErrArray
                  fluxOutArray0=fluxArray
                  fluxErrOutArray0=fluxErrArray

                  !
                  vignApp=.false.
                  deadApp=.false.
                  backApp=.true.
                  psfApp=.false.
                  call setAttribute(outTargetTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                  call setAttribute(outTargetTable0H, "NPIXSOU", sourceArea, "pix", "Source area")
                  call setAttribute(outTargetTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                  call setAttribute(outTargetTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                  call setAttribute(outTargetTable0H, "BACKAPP", backApp,  "Background is subtracted")
                  call setAttribute(outTargetTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            
         
                  !-------------
                  sourceRA(0)=proposedRA
                  sourceDEC(0)=proposedDEC
         
                  call setAttribute(outTargetTable0H, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                  call setAttribute(outTargetTable0H, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                  !call setAttribute(outTargetTable0H, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                  !call setAttribute(outTargetTable0H, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                  if (columnDataTypeFlag .eq. 2) then
                     call setAttribute(outTargetTable0H, "QFLAG", sourceQF8(0), " ",  "Quality flag")
                  else
                     call setAttribute(outTargetTable0H, "QFLAG", sourceQF16(0), " ",  "Quality flag")
                  endif
                  call setAttribute(outTargetTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                  !write(*,*) "L2413"
                  !-------------
                  !write(*,*) "L2375 iSource=", iSource
                  !if (iSource .ge. 0) then
                  !   write(*,*) "L2376 procFlag=", procFlag(iSource), " specTypeAux=", specTypeAux(iSource)
                  !endif
                  !write(*,*) "L2379  foundNoTarget=", foundNoTarget, " iTarget=", iTarget
                  if (iTarget .gt. -1) then
                     if (procFlag(iSource) .eq. 1) then
                        call setAttribute(outTargetTable0H, "OBJ_TYPE", "TARGET", &
                             "Target flag     ")
                     else
                        if (procFlag(iSource) .eq. 3) then
                           if (foundNoTarget) then
                              call setAttribute(outTargetTable0H, "OBJ_TYPE", "DEFAULT EXTRACTION (NO TARGET)", &
                                   "Target flag  ")
                           else
                              call setAttribute(outTargetTable0H, "OBJ_TYPE", "DEFAULT EXTRACTION", &
                                   "Target flag  ")
                           endif
                        else
                           call setAttribute(outTargetTable0H, "OBJ_TYPE", "DUBIOUS TARGET", &
                                "Target flag  ")
                        endif
                     endif
                  else
                     !write(*,*) "L2283 iTarget=", iTarget, " nonTargetExtraction=", &
                     !     nonTargetExtraction, " foundRealTarget=", foundRealTarget
                     if (nonTargetExtraction) then
                        if (foundRealTarget) then 
                           call setAttribute(outTargetTable0H, "OBJ_TYPE", "TARGET (RECOVERED FROM ZERO-ORDER)")
                        else
                           if (foundDubiousTarget) then
                              call setAttribute(outTargetTable0H, "OBJ_TYPE", "DUBIOUS TARGET (RECOVERED FROM ZERO-ORDER)")
                           else
                              call setAttribute(outTargetTable0H, "OBJ_TYPE", "DEFAULT (NO TARGET)", &
                                   "Target flag  ")
                           endif
                        endif
                     else
                        if (iTarget .gt. -900) then
                           call setAttribute(outTargetTable0H, "OBJ_TYPE", "TARGET (RE-DETECTED ZERO-ORDER)", &
                                "Target flag  ")
                        endif
                     endif
                  endif

                  call setAttribute(outTargetTable0H, "XFIRST", xFirstOrder, &
                       "pix", "First-order (spectrum) image X-coordiate")
                  call setAttribute(outTargetTable0H, "YFIRST", yFirstOrder, &
                       "pix", "First-order (spectrum) image Y-coordiate")
                  !----
                  !
                  ! VNY 25.04.2005 Recompute zero-order coordinates
                  !computingMode=0
                  !verbFlag=1
                  !xZeroIn=xZeroOrder
                  !yZeroIn=yZeroOrder
                  !call RecomputeZeroOrder(grismSetH, xZeroIn, yZeroIn, &
                  !     spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                  !     verbFlag, xZeroOut, yZeroOut)
                  !write(*,*) "L1912 xZeroIn=", xZeroIn, " xZeroOut=", xZeroOut, " yZeroIn=", &
                  !     yZeroIn, " yZeroOut=", yZeroOut

                  !xZeroOrder=xZeroOut
                  !yZeroOrder=yZeroOut
                  !
                  call setAttribute(outTargetTable0H, "XZERO", xZeroOrder, &
                       "pix", "Zero-order (object) image X-coordiate")
                  call setAttribute(outTargetTable0H, "YZERO", yZeroOrder, &
                       "pix", "Zero-order (object) image Y-coordiate")
                  call setAttribute(outTargetTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                       "pix", "Width of the spectrum extraction region")
                  call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                       "pix", "Backgr.extraction region width (left)")
                  call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthleft), &
                       "pix", "Backgr.extraction region width (right)")
                  call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                       "pix", "Spectrum to backgr.reg.separation (left)")
                  call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                       "pix", "Spectrum to backgr.reg.separation (right)")
                  !-----
                  if (putContaminationComment) then
                     call addComment(outTargetTable0H, contaminationComment)
                  endif

                  ! release the dal handles
                  !---------
                  call release(rateBackErrOutCol0)
                  call release(rateBackOutCol0)
                  call release(rateSourceErrOutCol0)
                  call release(rateSourceOutCol0)
                  call release(fluxOutCol0)
                  call release(fluxErrOutCol0)
                  call release(waveSourceOutCol0)
                  call release(outTargetTable0H)
                  !---                  
                  !box0left=int(xZeroOrder)-int(spectrumHalfWidth)
                  !box0right=int(xZeroOrder)+int(spectrumHalfWidth)
                  !box1left=int(xFirstOrder)-int(spectrumHalfWidth)
                  !box1right=int(xFirstOrder)+int(spectrumHalfWidth)
                  !box0down=int(yZeroOrder)-int(spectrumHalfWidth)
                  !box0up=int(yZeroOrder)+int(spectrumHalfWidth)
                  !---------------------------- VNY 14.10.2004
                  box0left=xZeroOrder - spectrumHalfWidth +0.5
                  box0right=xZeroOrder+spectrumHalfWidth +0.5
                  box1left=xFirstOrder-spectrumHalfWidth +0.5
                  box1right=xFirstOrder+spectrumHalfWidth+0.5
                  box0down=yZeroOrder-spectrumHalfWidth+0.5
                  box0up=yZeroOrder+spectrumHalfWidth+0.5
                  !box1down=int(yFirstOrder-spectrumHalfHeight+0.5)
                  !box1up=int(yFirstOrder+spectrumHalfHeight+0.5)
                  !write(*,*) "L2835 box1left=", box1left, " box1right=", box1right
                  !write(*,*) "L2836 nSp=", nSp, " rightBorder=", rightBorder,  &
                  !     " leftBorder=", leftBorder 
                  !-------------
                  if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                     box1down=yZeroOrder+regionU1+0.5
                     box1up=yZeroOrder+regionU2+0.5
                  else
                     box1down=yZeroOrder+regionV1+0.5
                     box1up=yZeroOrder+regionV2+0.5
                  endif
                  if (box0left .lt. 0.0) box0left=0.0
                  if (box0right .gt. nx-1) box0right=nx-1
                  if (box0down .lt. 0.0) box0down=0.0
                  if (box1left .lt. 0.0) box1left=0.0
                  if (box1right .gt. nx-1) box1right=nx-1
                  if (box1up .gt. ny-1) box1up=ny-1

                  !firstPass=.true.
                  if (givenSpectraRegionFile) then 
                     call writeSpectrumExtractionRegion(spectraRegionFileName, &
                          box0left,box0right,box0down,box0up, &
                          box1left,box1right,box1down,box1up, &
                          spectrumNoCount, firstPass, givenSpectraRegionFile)
                     firstPass=.false.
                  else
                     if (givenRegionFile) then
                        call writeSpectrumExtractionRegion(regionFileName, &
                             box0left,box0right,box0down,box0up, &
                             box1left,box1right,box1down,box1up, &
                             spectrumNoCount, firstPass, givenSpectraRegionFile)
                        firstPass=.false.
                     endif
                  endif
               else
                  call warning("OMGRISM", "The window is too small: spectrum not extracted")
               endif ! (availableSpectrumFraction .gt. 0.5
            else
               call warning("OMGRISM", "Problem with spectrum extraction: skipped a cycle")
            endif ! (nTargets .gt. 0)
!*************************************************************************
!***********************************************************************
!***********************************************************************
         !*********************************************************
         ! Processing source by source the extended (field) spectra
         !*********************************************************
!***********************************************************************
!***********************************************************************
!***********************************************************************
            !write(*,*) "*******************************************************"
            !write(*,*) " L2358 PROCESSING FIRST_ORDERS"
            !write(*,*) "*******************************************************"
            !write(*,*) "L2487 Analysis of extended sources related to zero-orders"
            !write(*,*) "---------------------------------------------------------"
            !write(*,*) "L2573 nSources=", nSources

         if (extractFieldSpectra) then
         !write(*,*) "L2971------------------"
         !do i=1,nSourcesListCheck
         !   write(*,*) "L2973 iCheck=",i, " indexCheckSources=", indexCheckSources(i)
         !enddo
         !write(*,*) "L2975"
            !*****************************************************************
            ! first we have to check whether some of the zero-orders have
            ! very close neighbours
            !*****************************************************************

            ! compute the S/N ratio for all of the zero-order sources
            !do iSource=0, nSources-1
            do iSource=0, nSources2-1
               if (sourceShape(iSource) .eq. 0) then
                  !if (iSource .eq. 213) then
                     !write(*,*) "----------------------------------"
                     !write(*,*) "L3018 iSource=", iSource, " sourceProximity=", &
                     !     sourceProximity(iSource), proximityX(iSource), " snratio=", &
                     !     snRatio(iSource), " counts=", proximityCounts(iSource), &
                     !     " sourceRelation=", sourceRelation(iSource)
                     !write(*,*) "-----------------------------"
                  !endif
                  
                  brightness1=0.0 ! compute the brightness of the source
                  brightness2=0.0
                  do ii=-2,2
                     iy=sourceYCoord(iSource)+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-2,2
                           jx=sourceXCoord(iSource)+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness1=brightness1+oswImage(jx,iy)
                              brightness2=brightness2+1.0
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  proximityCounts(iSource)=brightness1
                  if (brightness2 .gt. 1.0) then
                     proximityCounts(iSource)=proximityCounts(iSource)/brightness2
                  endif
               
                  brightness1=0.0 ! compute the background around the source
                  brightness2=0.0
                  do ii=-3,3,6
                     iy=sourceYCoord(iSource)+ii
                     if (iy .ge. 0 .and. iy .lt. ny) then
                        do jj=-3,3,6
                           jx=sourceXCoord(iSource)+jj
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness1=brightness1+oswImage(jx,iy)
                              brightness2=brightness2+1.0
                           endif
                        enddo
                     endif
                  enddo ! ii=-2,2
                  if (brightness2 .gt. 1.0) brightness1=brightness1/brightness2
                  brightness2=sqrt(proximityCounts(iSource))
                  proximityCounts(iSource)=proximityCounts(iSource)-brightness1              
                  snRatio(iSource)=proximityCounts(iSource)/brightness2
               endif
            enddo

            
            ! Compute the distances between the sources 
            !do iSource=0, nSources-1
            do iSource=0, nSources2-1
               if (sourceShape(iSource) .eq. 0) then
                  !do i=0,nSources-1
                  do i=0,nSources2-1
                     if (i .ne. iSource .and. sourceShape(i) .eq. 0) then
                        distance=(sourceXCoord(iSource)-sourceXCoord(i)) * &
                             (sourceXCoord(iSource)-sourceXCoord(i)) + &
                             (sourceYCoord(iSource)-sourceYCoord(i)) * &
                             (sourceYCoord(iSource)-sourceYCoord(i))
                        distance=sqrt(distance)
                        distanceX=abs(sourceXCoord(iSource)-sourceXCoord(i))
                        distanceY=abs(sourceYCoord(iSource)-sourceYCoord(i))
                        !if (iSource .eq. 453 .and. distance .lt.100.0) then
                        !   write(*,*) "L3072 iSource=", iSource, " i=", i, " distance=", &
                        !        distance, " dX=", distanceX, &
                        !        " dY=", distanceY
                        !   !write(*,*) "L3068 sourceXCoord(i)=", sourceXCoord(i), &
                        !   !     " sourceYCoord(i)=", sourceYCoord(i)
                        !endif
                        !if (distance .lt. 100.0/binY) then
                        if (distanceX .lt. 25.0/binX .and. distanceY .lt. 140.0/binY) then
                           proximity=sourceXCoord(i)-sourceXCoord(iSource)
                           proximityX(iSource)=proximity
                           proximityX(i)=-proximity
                           proximity=abs(proximity)
                           if (proximity .lt. 25.0/binX) then
                              sourceProximity(iSource)=i
                              sourceProximity(i)=iSource
                              !if (iSource .eq. 453) then
                              !   write(*,*) "L3089 iSource=", iSource, " i=", i, " sourceProximity=", &
                              !        sourceProximity(iSource), proximityX(iSource), " distanceX=", &
                              !        distanceX, " distanceY=", distanceY
                              !endif
                           endif
                        endif
                     endif
                  end do ! i=0, nSources-1
                  !if (iSource .eq. 213) then
                  !   write(*,*) "----------------------------------"
                  !   write(*,*) "L3099 iSource=", iSource, " sourceProximity=", &
                  !        sourceProximity(iSource), proximityX(iSource), " snratio=", &
                  !        snRatio(iSource), " counts=", proximityCounts(iSource)
                  !   write(*,*) "-----------------------------"
                  !endif
                  !**************************************************************************
               endif
            enddo
            !write(*,*) "--------------------------------------------"


!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

!*******************************************
! Start processing the normal source list
!********************************************            
         !do iSource=0, nSources-1
         do iSource=0, nSources2-1
            lateralSpectrumL = .false.
            lateralSpectrumR = .false.
            if (extended(iSource) .eq. 0) then
               extend = .false.
            else
               extend = .true.
            end if
            !write(*,*) "L3731 iSource=", iSource, " shape=", sourceShape(iSource), & 
            !     " sourceRelation=", sourceRelation(iSource), " procFlag=", &
            !     procFlag(iSource)
            xZeroOrder=0.0
            yZeroOrder=0.0
            xFirstOrder=0.0
            yFirstOrder=0.0

            ! Set the EXTENDED parameter to 2 in it is a genuine spectrum
            if (sourceShape(iSource) .eq. 1) then
               extended(iSource)=2
            endif
            !---------------------------------------------------------------------
            !  analysis of the extended sources related to their zero-orders
            !---------------------------------------------------------------------
            if (sourceShape(iSource) .eq. 1 .and. sourceRelation(iSource) .ge. 0 &
                 .and. procFlag(iSource) .lt. 0) then
              
               relationIndex=sourceRelation(iSource)
               !if (iSource .eq. 43) then
               !   write(*,*) "L3462 iSource=", iSource, " relation=", &
               !        sourceRelation(iSource), " procFlag=", procFlag(iSource), &
               !        " backRelation=", sourceRelation(relationIndex)
               !endif
               proximityX1=0.0
               if (sourceProximity(relationIndex) .gt. -1) then
                  proximityX1=proximityX(relationIndex)
                  !if (iSource .eq. 43) then
                  !   write(*,*) " L3157 the source ", iSource, " --->", relationIndex, &
                  !        " will be checked proximityX=", proximityX(relationIndex)
                  !   write(*,*) " L3159 sourceProximity=", sourceProximity(relationIndex)
                  !   if (relationIndex .eq. 106) then
                  !      write(*,*) "+++++++++++++++++++++++++++++++++++++++++++++"
                  !   endif
                  !endif
               endif
               
               !  Calculation of the spectrum edge coordinates
               sourceWx=sourceFWHMin2(iSource)
               sourceWy=sourceFWHMaj2(iSource)

               ! VNY 20.07.2004
               jj=sourceRelation(iSource)
               if (iSource .lt. nSources) then
                  spectrumOriginArray(iSource)=jj+1
               endif
               if (jj .lt. nSources) then
                  spectrumOriginArray(jj)=iSource+1
               endif
               !
               
               xs0=sourceXCoord(iSource)

               x1=xs0
               x2=x1
               xFirstOrder=xs0
               !
               ys0=int(sourceYCoord(iSource)+0.5)
               !
               y1=ys0
               y2=y1
               yFirstOrder=ys0              
               if (sourcePA(iSource) .ne. 0.) then
                  x1=x1+0.5*sourceWy/tan(-sourcePA(iSource)*PI/180.0_double)
                  x2=x2-0.5*sourceWy/tan(-sourcePA(iSource)*PI/180.0_double)
               else
                  x1=x1+0.5*sourceWy
                  x2=x2-0.5*sourceWy
               endif

               y1=y1-0.25*spectrumLength
               y2=y2+0.25*spectrumLength

               iblueEdge=int(y1)
               iredEdge=int(y2)
               !if (iSource .eq. 34) then
               !   write(*,*) "L3200 x1=", x1, " x2=", x2, " y1=", y1, " y2=", y2, &
               !        " iblueEdge=", iblueEdge, " iredEdge=", iredEdge
               !endif

               ! correcting the width for profile extraction
               if (sourceProximity(relationIndex) .gt. -1) then
                  x1=min(sourceXcoord(iSource),sourceXcoord(relationIndex))
                  x2=max(sourceXcoord(iSource),sourceXcoord(relationIndex))
                  x1=x1-abs(proximityX1)
                  x2=x2+abs(proximityX1)
                  !write(*,*) "L3210 x1=", x1, " x2=", x2, " <-------"                  
               endif
               !+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
               ! VNY 9.11.2005 extracting the averaged one-dimensional profile 
               ! of the spectrum
               !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++=

               spiMin=(x1+x2)/2.0-spectrumHalfWidth*1.5
               spiMax=(x1+x2)/2.0+spectrumHalfWidth*1.5

               spjMin=iblueEdge
               spjMax=iredEdge
        
               if (spiMin .lt. 0) spiMin=0
               if (spiMax .gt. nx-1) spiMax=nx-1
               if (spjMin .lt. 0) spjMin=0
               if (spjMax .gt. ny-1) spjMax=ny-1
               !if (iSource .eq. 108) then
               !   write(*,*) "L3829 spiMin=", spiMin, " spiMax=", spiMax, &
               !        " spjMin=", spjMin, " spjMAx=", spjMax
               !endif

               allocate (sProfile(spiMin: spiMax), stat = allocErr)
               call CheckAllocation(allocErr, "sProfile")
               allocate (dProfile(spiMin: spiMax), stat = allocErr)
               call CheckAllocation(allocErr, "dProfile")
               
               do i=spiMin, spiMax
                  mx1=0.0
                  px1=0.0
                  do j=spjMin,spjMax
                     mx1=mx1+oswImage(i,j)
                     px1=px1+1.0
                  enddo ! j=spjMin
                  if (px1 .gt. 1.0) then
                     mx1=mx1/px1
                  endif
                  sProfile(i)=mx1
               enddo ! i=spiMin

               !if (iSource .eq. 34) then
               !   write(*,*) "L3244 Profile:"
               !   write(*,*) "--------------"
               !endif
               do i=spiMin,spiMax
                  mx1=0.0
                  if (i-spiMin .gt. 0) then
                     mx1=sProfile(i)-sProfile(i-1)
                  endif
                  dProfile(i)=mx1
                  !if (iSource .eq. 34) then
                  !   write(*,*) " i=", i, " sProfile=", sProfile(i), " dProfile=", dProfile(i)
                  !endif
               enddo

               profileChecked=.false.
               nSp=1
               profileChecked=.false.
               if (sourceProximity(relationIndex) .gt. -1) then
                  verbFlag=0
                  !if (iSource .eq. 34) then
                  !   verbFlag=1
                  !endif
                  !write(*,*) "L3321 ---------------"
                  call EvaluateProfile(sProfile, dProfile, spiMin, spiMax, nSp, &
                       peacks, leftBorder, rightBorder, verbFlag)
                  !if (iSource .eq. 108) then                
                  !   write(*,*) "L3878 nSp=", nSp, " spectrumHalfWidth=", spectrumHalfWidth
                  !   do i=1,nSp
                  !      write(*,*) "L3880 i=", i, " peack=", peacks(i), "left=",leftBorder(i), &
                  !           " right=", rightBorder(i)
                  !   enddo
                  !endif
                  if (nSp .eq. 1) then
                     spiMin2=leftBorder(1)
                     spiMax2=rightBorder(1)
                     spiX2=peacks(1)
                  else
                     if (proximityX(relationIndex) .lt. 0) then 
                        ! Picking up the right peak of the profile
                        spiMin=leftBorder(2)
                        spiMin2=leftBorder(2)
                        spiMax2=rightBorder(2)
                        spiX2=peacks(2)
                        xs0=peacks(2)
                        !write(*,*) "L3287 picking up the RIGHT maximum"
                     else ! picking up the left peak of the profile
                        spiMax=rightBorder(1)
                        spiMin2=leftBorder(1)
                        spiMax2=rightBorder(1)
                        spiX2=peacks(1)
                        xs0=peacks(1)
                        !write(*,*) "L3294 picking up the LEFT maximum"
                     endif
                     profileChecked=.true.
                  endif
                  !if (iSource .eq. 108) then            
                  !   write(*,*) "L3299 spiMin=", spiMin, " spiMax=", spiMax
                  !   write(*,*) "L3300 spiMin2=", spiMin2, " spiMax2=", spiMax2
                  !endif
               else

                  ! search for the maxima and minima
                  mx1=0.0
                  px1=(x1+x2)/2.0
                  dmx1=0.0
                  dmn1=999999.0
                  dpx1=px1
                  dpn1=px1
                  aux1=0.0
                  aux2=0.0
                  do i=spiMin, spiMax ! compute the average derivative of the profile
                     if (i .ne. spiMin) then
                        aux1=aux1+ abs(dProfile(i))
                        aux2=aux2+1.0
                     endif
                     if (sProfile(i) .gt. mx1) then ! maximum opf the profile
                        mx1=sProfile(i)
                        px1=i
                     endif
                     if (dProfile(i) .gt. dmx1) then ! maximum of the derivative of the profile
                        dmx1=dProfile(i)
                        dpx1=i
                     endif
                     if (dProfile(i) .lt. dmn1) then !minimum of the derivative of the profile
                        dmn1=dProfile(i)
                        dpn1=i
                     endif
                  enddo
                  if (aux2 .gt. 1.0) then
                     aux1=aux1/aux2
                  endif
                  ! Check the left and right limits of the spectrum
                  spiMin2=0 ! left limit of the profile
                  spiMax2=0 ! right limit of the profile
                  aux1=aux1/2.0
                  do i=spiMin,spiMax
                     if (abs(dProfile(i)) .gt. aux1) then
                        spiMin2=i
                        exit
                     endif
                  enddo
                  !write(*,*) "L3400 ---------------"
                  do i=spiMax,spiMin+1,-1
                     if (abs(dProfile(i)) .gt. aux1) then
                        spiMax2=i
                        exit
                     endif
                  enddo
               
                  !if (iSource .eq. 108) then
                  !   write(*,*) "--------------------"
                  !   write(*,*) " L3963 dmx1=", dmx1, " dpx1=", dpx1, &
                  !        " dmn1=", dmn1, " dpn1=", dpn1
                  !   write(*,*) " L3965 aux1=", aux1, " aux2=", aux2, &
                  !        " spiMin2=", spiMin2, " spiMax2=", spiMax2
                  !endif
               endif ! sourceProximity .gt. -1

               ! Compute signal-to-noise ratio
               !write(*,*) "L3971 iSource=",iSource
               noise =0.0
               aux2=0.0
               iR=spiMin2-1
               iL=spiMin2
               iL=iL-spiMax2+spiMin2-1
               if (iR .lt. spiMin+1) iR=spiMin+1
               if (iL .lt. spiMin) iL=spiMin
               !if (iSource .eq. 34) then
               !   write(*,*) " L3369 iL=", iL, " iR=", iR
               !endif
               do i=iR,iL,-1
                  noise=noise+sProfile(i)
                  aux2=aux2+1.0
               enddo
               iL=spiMax2+1
               iR=spiMax2
               iR=iR+spiMax2-spiMin2+1
               if (iL .gt. spiMax-1) iL=spiMax-1
               if (iL .lt. spiMin) iL=spiMin
               if (iR .gt. spiMax) iR=spiMax
               do i=iL,iR
                  noise=noise+sProfile(i)
                  aux2=aux2+1.0
               enddo
               if (aux2 .gt. 1.0) noise=noise/aux2

               signal=0.0
               aux2=0.0
               iL=spiMin2
               iR=spiMax2
               if (iL .lt. spiMin) iL=spiMin
               if (iR .gt. spiMax) iR=spiMax
               do i=iL,iR
                  signal=signal+sProfile(i)
                  aux2=aux2+1.0
               enddo
               if (aux2 .gt. 1.0) signal=signal/aux2!

               snRatio(iSource)=0.0
               if (noise .gt. 0.0) then
                  snRatio(iSource)=(signal-noise)/sqrt(noise)
               endif
               !if (iSource .eq. 43) then
               !   write(*,*) " L3732 signal=", signal, " noise=", noise, &
               !       " s-n=", signal-noise, " snRatio=", snRatio(iSource), " iSource=", iSource
               !   write(*,*) "------------------------------------- "
               !endif

               if(allocated(sProfile)) then
                  deallocate(sProfile,stat=istat)
                  call CheckDeallocation(allocErr, "sProfile")
               end if
               if(allocated(dProfile)) then
                  deallocate(dProfile,stat=istat)
                  call CheckDeallocation(allocErr, "dProfile")
               end if
               !+++++++++++++++ 09.11.2005 +++++++++++++++++++++++++++++++++++++++

               !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
               !  Check the spectrum profile for three strips
               !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
               j=int(y2)
               mx1=0.0
               px1=xs0

               if (j .ge. 0 .and. j .lt. ny) then
                  do i=-20,20
                     i1=int(xs0)+i
                     if (i1 .ge. 0 .and. i1 .lt. nx) then
                        if (oswImage(i1,j) .gt. mx1) then
                           mx1=oswImage(i1,j)
                           px1=i1
                        endif
                     endif
                  enddo

                  !...................1
                  ! Test the left part of the strip for the presence of a minimum
                  mn1a=mx1
                  imn=int(px1)-1
                  do i=int(xs0)-20,int(px1)-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn1a) then
                           mn1a=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo!

                  ! Test the left part of the strip for the second maximum
                  mnx1a=mn1a
                  do i=int(xs0)-20,imn-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx1a) then!
                           mnx1a=oswImage(i,j)
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the minimum
                  mn1b=mx1
                  imn=int(px1)+1
                  do i=int(px1)+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn1b) then
                           mn1b=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the second maximum
                  mnx1b=mn1b
                  do i=imn+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx1b) then
                           mnx1b=oswImage(i,j)
                        endif
                     endif
                  enddo

                  ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
                  !.......................1
                  ! VNY test 07.01.2004 !

               endif ! If j .ge. 0
               j=int(ys0)
               mx2=0.0
               px2=xs0!

               if (j .ge. 0 .and. j .lt. ny) then
                  do i=-20,20
                     i1=int(xs0)+i
                     if (i1 .ge. 0 .and. i1 .lt. nx) then
                        if (oswImage(i1,j) .gt. mx2) then
                           mx2=oswImage(i1,j)
                           px2=i1
                        endif
                     endif
                  enddo

                  !...................2
                  ! Test the left part of the strip for the minimum
                  mn2a=mx2
                  imn=int(px2)-1
                  do i=int(xs0)-20,int(px2)-1
                     if (i .ge. 0 .and. i .lt. nx) then!
                        if (oswImage(i,j) .lt. mn2a) then
                           mn2a=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo

                  ! Test the left part of the strip for the second maximum
                  mnx2a=mn2a
                  do i=int(xs0)-20,imn-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx2a) then
                           mnx2a=oswImage(i,j)
                        endif
                     endif
                  enddo
!
                  ! Test the right part of the strip for the minimum!
                  mn2b=mx2
                  imn=int(px2)+1
                  do i=int(px2)+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn2b) then
                           mn2b=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the second maximum!
                  mnx2b=mn2b
                  do i=imn+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx2b) then!
                           mnx2b=oswImage(i,j)
                        endif
                     endif
                  enddo!
                  ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
                  !.......................2
               endif

               j=int(y1)

               mx3=0.0
               px3=xs0
               if (j .ge. 0 .and. j .lt. ny) then 
                  do i=-20,20
                     i1=int(xs0)+i
                     if (i1 .ge. 0 .and. i1 .lt. nx) then!

                        if (oswImage(i1,j) .gt. mx3) then
                           mx3=oswImage(i1,j)
                           px3=i1
                        endif
                     endif
                  enddo

                  !...................3
                  ! Test the left part of the strip for the minimum
                  mn3a=mx3
                  imn=int(px3)-1
                  do i=int(xs0)-20,int(px3)-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn3a) then
                           mn3a=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo

                  ! Test the left part of the strip for the second maximum
                  mnx3a=mn3a
                  do i=int(xs0)-20,imn-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx3a) then
                           mnx3a=oswImage(i,j)
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the minimum
                  mn3b=mx3
                  imn=int(px3)+1
                  do i=int(px3)+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn3b) then
                           mn3b=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the second maximum
                  mnx3b=mn3b
                  do i=imn+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx3b) then
                           mnx3b=oswImage(i,j)
                        endif
                     endif
                  enddo
                  ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
                  !.......................3
               endif

               ! Averaging secondary minima and maxima
               mnab=(mn1a+mn1b+mn2a+mn2b+mn3a+mn3b)/6.0
               mnxab=(mnx1a+mnx1b+mnx2a+mnx2b+mnx3a+mnx3b)/6.0

               ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
               !.......................1
               

               mxaver=(mx1+mx2+mx3)/3.0

               critMax=(mxaver-mnab)
               critMin=(mnxab-mnab)
               if (mxaver .gt. 0.) critMax=critMax/mxaver
               if (mnxab .gt. 0.)  critMin=critMin/mnxab

               xs0=(px1+px2+px3)/3.0
!

               ! 09.11.2005                
               !sourceXCoord(iSource)=xs0
               !if (iSource .eq. 34) then
               !   write(*,*) "L3635 sourceXCoord=", sourceXCoord(iSource)
               !endif
               if (profileChecked) then
                  sourceXCoord(iSource)=spiX2
               else
                  sourceXCoord(iSource)=real(spiMax2+spiMin2)/2.0
               endif
               !if (iSource .eq. 108) then
               !   write(*,*) "L4254 sourceXCoord=", sourceXCoord(iSource)
               !endif
               if (critMin .gt. critMin0 .and. critMax .gt. critMax0) then
                  write(errstr,*) "The spectrum No. ", iSource, &
                       " is probably corrupted by coincidence losses/Mod-8 noise"
                  call message(AppMsg, 2)
               endif

               !***********************************************************
               ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
               !***********************************************************
               !call readModes(grismSetH, x0, y0, xsize, ysize)  

               i=int(sourceXCoord(iSource)+0.5)
               j=int(sourceYCoord(iSource)+0.5)
               cosPA=cos(sourcePA(iSource)*degreesToRadians)
               sinPA=sin(sourcePA(iSource)*degreesToRadians)

               !if (iSource .eq. 25) then
               !   write(*,*) "L3375 i(xcoord)=", i, " j(ycoord)=",j
               !endif
               ! Find limits of the current source upwards and downwards
               sourceDeformation=0.0
               limitUp=0
               limitDown=0  
               level=averImage!*1.05

               do jup=0,500
                  j1=j+jup
                  i1=int(sourceXCoord(iSource)+0.5)
                  if (j1 .lt. 0 .or. j1 .ge. ny .or. i1 .ge. nx .or. i1 .lt. 0) exit
                  if (oswImage(i1,j1) .lt. level) exit
               enddo

               limitUp=jup
               do jdown=0,500
                  j1=j-jdown
                  i1=int(sourceXCoord(iSource)+0.5)
                  if (j1 .lt. 0 .or. j1 .ge. ny .or. i1 .ge. nx .or. i1 .lt. 0) exit
                  if (oswImage(i1,j1) .lt. level) exit
               enddo
               limitDown=jdown

               ! Calculate limits left and right
               limitLeft=0.0
               limitRight=0.0
               if (cosPA .ne. 0.0) then
                  !            if (cosPA .ge. 0.5) then
                  !
                  do jRight=0,100
                     i1=i+jRight
                     j1=int(sourceYCoord(iSource)+0.5)
                     if (i1 .lt. 0 .or. i1 .ge. nx .or. j1 .ge. ny .or. j1 .lt. 0) exit
                     if (oswImage(i1,j1) .lt. level) exit
                  enddo
                  limitRight=jRight 
                  do jLeft=0,100
                     i1=i-jLeft
                     j1=int(sourceYCoord(iSource)+0.5)
                     if (i1 .lt. 0 .or. i1 .ge. nx .or. j1 .ge. ny .or. j1 .lt. 0) exit
                     if (oswImage(i1,j1) .lt. level) exit
                  enddo
                  limitLeft=jLeft
               endif

               if (limitLeft .gt. limitLeftMax(iSource)) limitLeftMax(iSource)=limitLeft
               if (limitRight .gt. limitRightMax(iSource)) limitRightMax(iSource)=limitRight
               
               if ((limitUp+limitDown+1.0) .gt. (limitRight+limitLeft+1.0)) then
                  sourceDeformation=(limitUp+limitDown+1.0)/(limitRight+limitLeft+1.0)
               endif

               sourceArea=0.0
               srcNo=iSource
               timeExposure = exposure
               !--- Create output only for the sources with the proper shape and relation
               !           
               ! Create the output file only for the extended sources
               !  Name the output block using the source number (iSource)
               posPoint=scan(trim(outFile), '.', .true.)
               i=sourceRelation(iSource)
               
               if (i .gt. 0) then
                  if (procFlag(i) .lt. 0) then 
                     write(sourceIndex, '(i3.3)') i+1
                     outFile2=outFile
                     if (posPoint .ne. 0) then
                        outFile2(posPoint:1024)=' '
                        write(outFile3, '(a,a,a)') trim(outFile2), sourceIndex, ".FIT"
                     else
                        write(outFile3, '(a,a,a)') trim(outFile2), sourceIndex, ".FIT"
                     endif

                     write(blockName, '(a,a)') "RATE", sourceIndex
                     !---
                     ! The extracted region will have the full length of the y-column
                     nBins=ny

                     i1=int(sourceXCoord(iSource)+0.5)

                     averLeft=real(i1)
                     averRight=averLeft
                     numLeft=0.0
                     numRight=0.0

                     deltaX=0.0
                     ! The position of the corresponding zero-order is taken as reference
                     xZero=-1.0
                     yZero=-1.0
                     xFirst=sourceXCoord(iSource)
                     yFirst=sourceYCoord(iSource)
                     !if (iSource .eq. 107) then
                     !   write(*,*) "L4366 xFirst=", xFirst, " yFirst=", yFirst
                     !endif
                     if (sourceRelation(iSource) .gt. -1) then
                        i=sourceRelation(iSource)
                        deltaX=sourceYCoord(i)
                        ! XY-coordinates of the zero-order image
                        xZero=sourceXCoord(i)
                        !
                        yZero=int(sourceYCoord(i)+0.5)
                        !
                        xZeroOrder=xZero
                        yZeroOrder=yZero
                        !if (iSource .eq. 43) then
                        !   write(*,*) "L4096 xZero=", xZero, " yZero=", yZero
                        !endif
                        !---VNY 10.06.2005 Check if there is a bright zero order nearby
                        !---------------------------------------------------------------------------
                        zeroOrderCriterium=90.0/xBin ! criterium for checking the erroneous zero-orders
                        !do iCheck=1,nSourcesListCheck
                        do iCheck=nSources,nSources2-1
                           !deltaX=abs(sourceData(iCheck)%xPos-xZeroOrder)
                           !deltaY=abs(sourceData(iCheck)%yPos-yZeroOrder)
                           deltaX=abs(sourceXCoord(iCheck)-xZeroOrder)
                           deltaY=abs(sourceYCoord(iCheck)-yZeroOrder)
                           deltaXY=sqrt(deltaX*deltaX+deltaY*deltaY)
                           if (deltaXY .lt. zeroOrderCriterium .and. .not. omgsourceRegime) then 
                              !write(*,*) "L3610 iCheck=", iCheck, " deltaXY=", deltaXY, " index=", &
                              !     indexCheckSources(iCheck), " x=", sourceData(iCheck)%xPos, &
                              !     " y=", sourceData(iCheck)%yPos, " x0=", xZeroOrder, " y0=", yZeroOrder

                              !write(*,*) "L3611 iSource=", i, " iCheck=", iCheck, " deltaXY=", deltaXy, &
                              !     " noPix(iCheck)=", sourceData(iCheck)%noSourcePixels, &
                              !     " rate(iCheck)=", sourceData(iCheck)%rawCountRate
                                 
                              !Replace the coordinates only for very big and bright sources (zero-orders)
                              ! and only if the spectrum is not double (two close spectra)
                              ! The logical variable profileChecked is set .true. if there 
                              ! are two peaks in the spectrum
                              !write(*,*) "L3624 profileChecked=", profileChecked
                              if (sourceSize(iCheck) .gt. 2000/xBin .and. &
                                   !sourceData(iCheck)%rawCountRate .gt. 3000.0 .and. .not. profileChecked) then
                                   sourceRawCountRate(iCheck) .gt. 3000.0 .and. .not. profileChecked) then
                                 !write(*,*) "          L3617 xold=", xZeroOrder, " xnew=", sourceData(iCheck)%xPos
                                 !write(*,*) "          L3618 yold=", yZeroOrder, " ynew=", sourceData(iCheck)%yPos
                                 !xZero=sourceData(iCheck)%xPos-1
                                 !yZero=sourceData(iCheck)%yPos-1
                                 xZero=sourceXCoord(iCheck)-1
                                 yZero=sourceYCoord(iCheck)-1
                                 xZeroOrder=xZero
                                 yZeroOrder=yZero 
                              endif
                              
                           endif
                        enddo ! iCheck=nSources, nSources2-1
                        !----------------------------------------------------------------------------

                     endif ! sourceRelation .gt. -1
                     subtractBkg=.true.
                     verbFlag=0
                     !if (iSource .eq. 43) then
                     !   write(*,*) "L4426 iSource=", iSource, " xZero=", xZero, " yZero=", yZero
                     !endif

                     if (.not. omgsourceRegime) then
                        ! VNY 14.10.2004 Check the x-position of the first order
                        xFirstIn=xFirst
                        yFirstIn=yFirst
                        verbFlag=0
                        !if (iSource .eq. 2) verbFlag=1
                        call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                             spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                             verbFlag, xFirstOut, yFirstOut, averCountIn, &
                             averCountOut)
                        !if (iSource .eq. 43) then
                        !   write(*,*) "L4157  xFirstIn=", xFirstIn, " yFirstIn=", &
                        !        yFirstIn, " xFirstOut=", xFirstOut, " yFirstOut=", &
                        !        yFirstOut
                        !   write(*,*) "L4160 profileChecked=", profileChecked 
                        !endif
                        verbFlag=0
                        !write(*,*) " L2629 iSource=", iSource, " difference=", abs(xFirstIn-xFirstOut), &
                        !     " spectrumHalfWidth=", spectrumHalfWidth
!                        if (abs(xFirstIn-xFirstOut) .lt. spectrumHalfWidth) then ! correct the position
                        if (.not. profileChecked .and. abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then 
                           ! correct the position
                           xFirst=xFirstOut
                           xFirstOrder=xFirst
                        endif
                     endif
                     !if (iSource .eq. 34) then
                     !   write(*,*) "L3839 xFirst=", xFirst
                     !endif
                     !--- 11.01.05
                     verbFlag=0
                     ! First extract for the evaluation purporses with mode=2 (evaluationMode)
                     !if (iSource .eq. 86) then
                     !   verbFlag=1
                     !   write(*,*) "L3679 ****************************** source=", iSource
                     !   write(*,*) "L3680 xZer=", xZero, " yZero=", yZero, " xFirst=", xFirst, " yFirst=", yFirst
                     !   write(*,*) "L3681 relationIndex=", relationIndex, " proximityX=", &
                     !        proximityX(relationIndex), " proximityX1=", proximityX1
                     !endif

                     call ExtractSpectrum(grismSetH, xZero, yZero, xFirst, yFirst, &
                          spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                          bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                          spectrumArray, spectrumErrArray, bkgArray, &
                          bkgErrArray, fluxArray, fluxErrArray, ysize, &
                          evaluationMode, subtractBkg, verbFlag,s2n,s2n0, &
                          crossShape, leftBorder, rightBorder, proximityX1, nSp)

                     !if (iSource .eq. 86) then
                     !   write(*,*) "************************************** source=", iSource
                     !endif
                     !
                     maskArray=1
                     verbFlag=0
                     spectrumWidth=spectrumHalfWidth*2.0
                     bkgWidth=(bkgWidthRight+bkgWidthLeft)/2.0
                     !if (iSource .eq. 16) verbFlag=1
                     call EvaluateSpectrum(spectrumArray, bkgArray, &
                          ysize, iblueEdge, iredEdge, bkgWidth, &
                          spectrumWidth, sbRatio, maskArray, verbFlag)
! VNY 25.10.2007 relaxing the criterium for 2n to s2n0 ratio
                     if (iSource .eq. 43) then
                        !if (s2n0 .gt. s2n*7.0 .and. s2n .gt. 0.5 .and. crossShape .gt. 0.0 &
                        if (s2n0 .gt. s2n*snCriterium .and. s2n .gt. snCriterium2 &
                             .and. crossShape .gt. 0.0 .and. sbRatio .gt. 0.1) then
                           write(*,'(a,i5,7(a,f11.3))') "L4210 ISOURCE=", iSource, " s2n=", s2n, &
                                " s2n0=", s2n0, " x0=", xZero, &
                                " y0=", yZero, " sbratio=", sbRatio, " crossShape=", crossShape
                        else
                           write(*,'(a,i5,7(a,f11.3))') "L4213 iSource=", iSource, " s2n=", &
                                s2n, " s2n0=", s2n0, " x0=", xZero, &
                                " y0=", yZero, " sbratio=", sbRatio, " crossShape=", crossShape
                        endif
                     endif
                     ! VNY 06.02.2004 Add table only if the spectrum significance is larger than 2.0
!                     if (s2n .gt. 0.5 .and. crossShape .gt. 0.0 .and. &
                     if (s2n .gt. snCriterium2 .and. crossShape .gt. 0.0 .and. &
                          sbRatio .gt. 0.1 .or. omgsourceRegime) then
! VNY 25.10.2007 relaxing the criterium for 2n to s2n0 ratio
                        !if (s2n0 .lt. s2n*10.0) then 
!                        if (s2n0 .lt. s2n*7.0) then 
                        if (s2n0 .lt. s2n*snCriterium) then 
                           ! check whether the zero-order is correct
                           !write(*,*) "L4225 the zero-order is possibly wrong iSource=", iSource
                           continue
                        else
                           ! extract with the requested mode
                           ! VNY 14.10.2004 Check the x-position of the first order
                           if (.not. omgsourceRegime) then
                              xFirstIn=xFirst
                              yFirstIn=yFirst
                              verbFlag=0
                              call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                                   spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                                   verbFlag, xFirstOut, yFirstOut, averCountIn, &
                                   averCountOut)
                              !if (iSource .eq. 43) then 
                              !   write(*,*) "L4239 xFirstIn=", xFirstIn, &
                              !        " yFirstIn=", yFirstIn, &
                              !        " xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                              !   write(*,*) "L4242  iSource=", iSource, " difference+", &
                              !        abs(xFirstIn-xFirstOut), &
                              !        " spectrumHalfWidth=", spectrumHalfWidth
                              !   write(*,*) "L4245 profileChecked=", profileChecked
                              !endif
                              verbFlag=0
                              if (.not. profileChecked .and. &
                                   abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then 
                                 ! correct the position
                                 xFirst=xFirstOut
                                 xFirstOrder=xFirst
                              else 
                                 xFirst=xFirstIn
                                 xFirstOrder=xFirst
                              endif
                           endif
                           !--- 11.01.05
                           verbFlag=0
                           ! 25.04.2005
                           !if (iSource .eq. 43) then 
                           !   !verbFlag=1
                           !   write(*,*) "L4263 xFirstOrder=", xFirstOrder, " yFirstOrder=", yFirstOrder
                           !endif
                           !write(*,*) "L3782 proximityX1=", proximityX1
                           call ExtractSpectrum(grismSetH, xZero, yZero, xFirst, yFirst, &
                                spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                                bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                                spectrumArray, spectrumErrArray, bkgArray, &
                                bkgErrArray, fluxArray, fluxErrArray, ysize, &
                                extractionMode, subtractBkg, verbFlag,s2n,s2n0,&
                                crossShape, leftBorder, rightBorder, proximityX1, nSp)

                           !if (iSource .eq. 86) then 
                           !   write(*,*) "L4024 xFirstOrder=", xFirstOrder, " leftBorder(2)=", &
                           !     leftBorder(2), " rightBorder(2)=", rightBorder(2)
                           !endif
!*********************************************************************** 30.10.2005
                           if (.not. createSpectrumFile) then
                              outFile4=outFile
                              outTargetSet0H = dataSet(outFile4, create) ! This file will contain 
                              createSpectrumFile=.true.
                              createSpectraListFile=.true.
                              call setAttribute(outTargetSet0H, "CONTENT", "OM GRISM SPECTRA", &
                                   "File contains OM grsim spectrum/spectra")
                              call setAttribute(outTargetSet0H, "HDUCLASS", "OGIP    ", &
                                   "Format conforms to OGIP/GSFC conventions")
                              call setAttribute(outTargetSet0H, "HDUCLAS1", "SPECTRUM", &
                                   "File contains spectral data")
                              call setAttribute(outTargetSet0H, "HDUVERS", "1.2.0    ", &
                                   "The version number of the format")
                              call setAttribute(outTargetSet0H, "TELESCOP", "XMM     ", &
                                   "XMM mission")
                              call setAttribute(outTargetSet0H, "INSTRUME", "OM      ", &
                                   "Instrument name")
                              call setAttribute(outTargetSet0H, "DATAMODE", dataMode, &
                                   "Data mode")
                              call setAttribute(outTargetSet0H, "FILTER", filter, &
                                   "Filter used for obtaining the data")
                              call setAttribute(outTargetSet0H, "FRAMTIME", frameTime, &
                                   "Frame time")
                              call setAttribute(outTargetSet0H, "DEADFRAC", deadFrac, &
                                   "Frame dead fraction")
                              call setAttribute(outTargetSet0h, "TMEXP_ID", tmExpId, &
                                   "Exposure Id from telemetry")
                              call setAttribute(outTargetSet0H, "FRMTIME", frmTime, &
                                   "Frametime from telemetry")
                              call setAttribute(outTargetSet0H, "DATATYPE", dataType, &
                                   "Data type")
                              call setAttribute(outTargetSet0H, "OBS_ID", obsId, &
                                   "Observation identifier")
                              call setAttribute(outTargetSet0H, "EXP_ID", expId, &
                                   "Exposure identifier")
                              call setAttribute(outTargetSet0H, "DATE-OBS", dateObs, &
                                   "Observation start time")
                              call setAttribute(outTargetSet0H, "DATE-END", dateEnd, &
                                   "Observation end time")
                           endif
!*********************************************************************** 30.10.2005
                           spectrumNoCount=spectrumNoCount+1
                           write(sourceIndex, '(i3.3)') spectrumNoCount+1
                           !
                           write(blockName, '(a,a)') "RATE", sourceIndex
                           outTargetTable0H=addTable(outTargetSet0H, blockName, nBins, &
                                "Table with the field source spectrum")

                           blkNum=blockNumber(outTargetSet0H, blockName)
                           
                           !spectrumNoArray(iSource)=blkNum+1
                           jj = sourceRelation(iSource)
                           
                           !if (spectrumNoCount .eq. 71) then
                           !   write(*,*) "L4332 spNo=", spectrumNoCount, " iSource=", &
                           !        iSource, " relation=", sourceRelation(iSource)
                           !endif
                           if (jj .lt. nSources) then
                              spectrumOriginArray(jj)=iSource+1
                           endif
                           !
                           if (iSource .lt. nSources) then
                              spectrumOriginArray(iSource)=jj+1
                              spectrumNoArray0(iSource)=spectrumNoCount+1
                           endif
                           spectrumNoArray(iSource)=spectrumNoCount+1
                           spectrumNoArray2(spectrumNoCount)=jj+1
                           spectrumOriginArray2(spectrumNoCount)=jj+1
                           srcNoArray2(spectrumNoCount)=iSource
                           x0aux(spectrumNoCount)=xZeroOrder
                           y0aux(spectrumNoCount)=yZeroOrder
                           x1aux(spectrumNoCount)=sourceXCoord(iSource)
                           y1aux(spectrumNoCount)=sourceYCoord(iSource)
                           !write(*,*) "L4637 iSource =", iSource, " xZeroOrder=", xZeroOrder, " yZeroOrder=", yzeroOrder
                           !write(*,*) "L4638^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
                           !
                           ! 
                           ! Computing the binning interval
                           if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                              binningInterval=2.40774*binY ! in Angstroms
                           else
                              binningInterval=5.626*binY
                           endif

                           call setAttribute(outTargetTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                           call setAttribute(outTargetTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                           call setAttribute(outTargetTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                           call setAttribute(outTargetTable0H, "HDUCLAS2", "NET", "Background subtracted")
                           call setAttribute(outTargetTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                           call setAttribute(outTargetTable0H, "DLAMBDA", binningInterval, "Angstrom", "binning interval")
                           ! VNY 12.05.04 attribute extractionMode
                           call setAttribute(outTargetTable0H, "EXTRMODE", extractionMode, "Spectra extraction mode")
                           !--Adding columns to the output table
                           waveSourceOutCol0 = addColumn(outTargetTable0H, "WAVELENGTH", real32, &
                                "ANGSTROM", columnDim, "wavelengths ")
                           waveSourceOutArray0 => real32data(waveSourceOutCol0)
                           !
                           rateSourceOutCol0 = addColumn(outTargetTable0H,"RATE", real32, &
                                "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                           rateSourceOutArray0 => real32data(rateSourceOutCol0)
                           !
                           rateSourceErrOutCol0 = addColumn(outTargetTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                                columnDim, "source counts rate error")
                           rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                           !
                           rateBackOutCol0 = addColumn(outTargetTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                                columnDim, " background counts rate")
                           rateBackOutArray0 => real32data(rateBackOutCol0)
                           !
                           rateBackErrOutCol0 = addColumn(outTargetTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                                columnDim, " background counts rate error")
                           rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                           !
                           fluxOutCol0 = addColumn(outTargetTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                                columnDim, " background counts rate")
                           fluxOutArray0 => real32data(fluxOutCol0)
                           !
                           fluxErrOutCol0 = addColumn(outTargetTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                                columnDim, " background counts rate error")
                           fluxErrOutArray0 => real32data(fluxErrOutCol0)
                           !
                     
                           waveSourceOutArray0=wavelengthArray
                           rateSourceOutArray0=spectrumArray
                           rateSourceErrOutArray0=spectrumErrArray
                           rateBackOutArray0=bkgArray
                           rateBackErrOutArray0=bkgErrArray
                           fluxOutArray0=fluxArray
                           fluxErrOutArray0=fluxErrArray
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                     !
                           vignApp=.false.
                           deadApp=.false.
                           backApp=.true.
                           psfApp=.false.
                           call setAttribute(outTargetTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                           call setAttribute(outTargetTable0H, "NPIXSOU", sourceArea, "pix", "Source area")
                           call setAttribute(outTargetTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                           call setAttribute(outTargetTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                           call setAttribute(outTargetTable0H, "BACKAPP", backApp,  "Background is subtracted")
                           call setAttribute(outTargetTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            
                           !write(*,*) "L3416"
                           if (columnDataTypeFlag .eq. 2) then
                              call setAttribute(outTargetTable0H, "QFLAG", sourceQF8(0), " ",  "Quality flag")
                           else
                              call setAttribute(outTargetTable0H, "QFLAG", sourceQF16(0), " ",  "Quality flag")
                           endif
                           call setAttribute(outTargetTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                           call setAttribute(outTargetTable0H, "OBJ_TYPE", "FIELD OBJECT", &
                                "Target flag                                 ")
                           call setAttribute(outTargetTable0H, "XFIRST", xFirstOrder, &
                                "pix", "First-order (spectrum) image X-coordiate")
                           call setAttribute(outTargetTable0H, "YFIRST", yFirstOrder, &
                                "pix", "First-order (spectrum) image Y-coordiate")

                           !if (iSource .eq. 25) then
                           !   write(*,*) "L3918 xFirstOrder=", xFirstOrder, " yFirstOrder=", yFirstOrder
                           !endif
                        !


                        !---------------------------- VNY 14.10.2004
                           box0left=xZeroOrder - spectrumHalfWidth +0.5
                           box0right=xZeroOrder+spectrumHalfWidth +0.5
                           box1left=xFirstOrder-spectrumHalfWidth +0.5
                           box1right=xFirstOrder+spectrumHalfWidth-0.5
                           box0down=yZeroOrder-spectrumHalfWidth+0.5
                           box0up=yZeroOrder+spectrumHalfWidth+0.5
                           box1down=yFirstOrder-spectrumHalfHeight+0.5
                           box1up=yFirstOrder+spectrumHalfHeight+0.5
                           !-------------
                           !if (iSource .eq. 25) then
                           !   write(*,*) "L3936 box1left=", box1left, " box1right=", box1right
                           !   write(*,*) "L3937 nSp=", nSp, " leftBorder=", leftBorder,  &
                           !        " rightBorder=", rightBorder, " proximityX1=", proximityX1
                           !endif
                           ! Change the outline of the box for the double (close) spectra
                           if (nSp .eq. 2) then
                              if (proximityX1 .lt. 0) then ! using the right spectrum limits
                                 box1left=leftBorder(2)
                                 box1right=rightBorder(2)
                              endif
                              if (proximityX1 .gt. 0) then ! using the right spectrum limits
                                 box1left=leftBorder(1)
                                 box1right=rightBorder(1)
                              endif
                              !if (iSource .eq. 25) then
                              !   write(*,*) "L3949 the left/right limits of the extraction box are changed"
                              !   write(*,*) "L3950 box1left=", box1left, " box1right=", box1right
                              !endif
                           endif ! nSp .eq. 2

                           call setAttribute(outTargetTable0H, "XZERO", xZeroOrder, &
                                "pix", "Zero-order (object) image X-coordiate")
                           call setAttribute(outTargetTable0H, "YZERO", yZeroOrder, &
                                "pix", "Zero-order (object) image Y-coordiate")

                           if (nSp .eq. 2) then
                              call setAttribute(outTargetTable0H, "SPECREGW", &
                                   box1right-box1left+1, &
                                   "pix", "Width of the spectrum extraction region")
                              !if (iSource .eq. 25) then
                              !   write(*,*) "L3982 SPECREGW=", box1right-box1left+1
                              !endif
                              if (proximityX1 .lt. 0) then ! no left bkg region
                                 call setAttribute(outTargetTable0H, "BKGREGWL", 0.0, &
                                      "pix", "Backgr.extraction region width (left)")
                                 call setAttribute(outTargetTable0H, "BKGSEPL", 0.0, &
                                      "pix", "Spectrum to backgr.reg.separation (left)")
                              else
                                 call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                                      "pix", "Backgr.extraction region width (left)")
                                 call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                                      "pix", "Spectrum to backgr.reg.separation (left)")
                              endif

                              if (proximityX1 .gt. 0) then ! no right bkg region
                                 call setAttribute(outTargetTable0H, "BKGREGWR", 0.0, &
                                      "pix", "Backgr.extraction region width (right)")
                                 call setAttribute(outTargetTable0H, "BKGSEPR", 0.0, &
                                      "pix", "Spectrum to backgr.reg.separation (right)")
                              else
                                 call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthright), &
                                      "pix", "Backgr.extraction region width (right)")
                                 call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                                      "pix", "Spectrum to backgr.reg.separation (right)")
                              endif
                           else
                              call setAttribute(outTargetTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                                   "pix", "Width of the spectrum extraction region")
                              !write(*,*) "L3962 SPECREGW=", 2.0*abs(spectrumHalfWidth)
                              call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                                   "pix", "Backgr.extraction region width (left)")
                              call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                                   "pix", "Spectrum to backgr.reg.separation (left)")
                              call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthleft), &
                                   "pix", "Backgr.extraction region width (right)")
                              call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                                   "pix", "Spectrum to backgr.reg.separation (right)")
                           endif

                           if (putContaminationComment) then
                              call addComment(outTargetTable0H, contaminationComment)
                           endif

                           ! release the dal handles
                           !---VNY 30.08.03
                           call release(rateBackErrOutCol0)
                           call release(rateBackOutCol0)
                           call release(rateSourceErrOutCol0)
                           call release(rateSourceOutCol0)
                           call release(fluxOutCol0)
                           call release(fluxErrOutCol0)
                           call release(waveSourceOutCol0)
                           call release(outTargetTable0H)


                           if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                              box1down=yZeroOrder+regionU1+0.5
                              box1up=yZeroOrder+regionU2+0.5
                           else
                              box1down=yZeroOrder+regionV1+0.5
                              box1up=yZeroOrder+regionV2+0.5
                           endif
                           if (box0left .lt. 0.0) box0left=0.0
                           if (box0right .gt. nx-1) box0right=nx-1
                           if (box0down .lt. 0.0) box0down=0.0
                           if (box1left .lt. 0.0) box1left=0.0
                           if (box1right .gt. nx-1) box1right=nx-1
                           if (box1up .gt. ny-1) box1up=ny-1
                           !write(*,*) "L4834"
                           !firstPass=.false.
                           if (givenSpectraRegionFile) then 
                              call writeSpectrumExtractionRegion(spectraRegionFileName, &
                                   box0left,box0right,box0down,box0up, &
                                   box1left,box1right,box1down,box1up, &
                                   spectrumNoCount, firstPass, givenSpectraRegionFile)
                              firstPass=.false.
                           else
                              if (givenRegionFile) then
                                 call writeSpectrumExtractionRegion(regionFileName, &
                                      box0left,box0right,box0down,box0up, &
                                      box1left,box1right,box1down,box1up, &
                                      spectrumNoCount, firstPass, givenSpectraRegionFile)
                                 firstPass=.false.
                              endif
                           endif
                        endif ! s2n0 .lt. s2n*10.0
                     endif ! sbRatio .gt. 2.0
                  endif ! if (procFlag(i) .lt. 0
               endif ! if (i .gt. 0)
            endif ! sourceDeformation .gt. 10.0
         enddo ! iSource=0, nSource-1
         !write(*,*) "---------------------------------------------------- checkLength=", checkLength
         !***********************************************************************
         !***********************************************************************
         !***********************************************************************
         ! Processing source by source for those point source that were not related
         ! to their spectra
         !***********************************************************************
         !***********************************************************************
         !***********************************************************************
         !write(*,*) "L4866 Processing standing-alone  zero-orders"
         !write(*,*) "L4867 nSources=", nSources
         !write(*,*) "---------------------------------------------------------"

         !do iSource=0, nSources-1
         do iSource=0, nSources2-1
            lateralSpectrumL = .false.
            lateralSpectrumR = .false.
            xZeroOrder=0.0
            yZeroOrder=0.0
            xFirstOrder=0.0
            yFirstOrder=0.0
            !if (iSource .eq. 107) then
            !   write(*,*) "L4879 iSource=", iSource, " relation=", sourceRelation(iSource), &
            !        " FWHM=", sourceFWHMaj2(iSource), " PA=", sourcePA(iSource), " procFlag=", procFlag(iSource)
            !   write(*,*) "    x=", sourceXCoord(iSource), " y=", sourceYCoord(iSource)
            !   write(*,*) "    sourceShape=", sourceShape(iSource)
            !endif

            if (sourceShape(iSource) .eq. 0 .and. sourceRelation(iSource) .lt. 0 &
                 .and. sourceFWHMaj2(iSource) .lt. checkLength &
                 .and. abs(sourcePA(iSource)) .gt. 82.0 &
                 .and. procFlag(iSource) .lt. 0 .and. snRatio(iSource) .gt. 4.0) then 
               ! Check only the point-like sources
               i1=int(sourceXCoord(iSource)+0.5)
               ! The position of the corresponding zero-order is taken as reference
               xZeroA=sourceXCoord(iSource)
               !
               !yZeroA=int(sourceYCoord(iSource)+0.5)
               yZeroA=sourceYCoord(iSource)
               !
               !write(*,*) "L4113------------------------------------------"
               !if (iSource .eq. 70) then
               !   write(*,*) "      L4277 iSource=", iSource, " xZeroA=", xZeroA, &
               !        " yZeroA=", yZeroA, " sourceRelation=", sourceRelation(iSource)
               !   
               !   write(*,*) "L4280 iSource=", iSource, " sourceProximity=", &
               !        sourceProximity(iSource), proximityX(iSource), " snratio=", &
               !        snRatio(iSource), " counts=", proximityCounts(iSource)
               !endif
               xZeroOrder=xZeroA
               yZeroOrder=yZeroA
               yFirst=sourceYCoord(iSource)+source2spectrum ! position of the first order
               xFirstOrder=xZeroOrder
               yFirstOrder=yFirst
               iyUp=int(yFirst)
               iyDown=int(yFirst-spectrumLength/2.0)
               availableSpectrumLength=0.0
               availableSpectrumFraction=0.0
               if (iyDown .gt. int(yZeroA) .and. iyDown .lt. ny-1 .and. iyUp .lt. ny-1) then 
                  if (oswImage(i1,iyDown) .gt. -0.9 .and. oswImage(i1,iyUp) .gt. -0.9) then
                     ! determine the upper available limit
                     availableSpectrumLength=0.0
                     checkUp: do iyUp2=iyDown+1,iyDown+int(spectrumLength)
                        if (iyUp2 .lt. ny) then
                           if (oswImage(i1,iyUp2) .gt. -0.9) then
                              availableSpectrumLength=availableSpectrumLength+1.0
                           else
                              exit checkUp
                           endif
                        endif
                     enddo checkUp
                  endif
               endif
               availableSpectrumFraction=availableSpectrumLength/spectrumLength
               !if (iSource .eq. 453) then
               !   write(*,*) " L4361 iSource=", iSource, " availableSpectrumFraction=", &
               !        availableSpectrumFraction
               !   write(*,*) " L4363 xFirstOrder=", xFirstOrder, " yFirstOrder=", yFirstOrder
               !endif
               ! Extract the spectrum only if more than 80 percent of the spectrum is available
               if (availableSpectrumFraction .gt. 0.6) then !extract
                  !if (iSource .eq. 453) then
                  !   write(*,*) "L4363 iSource=", iSource, " availableSpectrumFraction=", &
                  !        availableSpectrumFraction
                  !endif
                  sourceArea=0.0
                  srcNo=iSource
                  timeExposure = exposure

                  nBins=ny
                  xFirst=xFirstOrder
                  yFirst=yFirstOrder
                  subtractBkg=.true.
                  verbFlag=0

                  ! check whether the source (zero-order) is close to another source
                  proximityX1=0.0
                  if (sourceProximity(iSource) .gt. -1) then
                     proximityX1=proximityX(iSource)
                     relationIndex=sourceRelation(iSource)
                     !if (iSource .eq. 82) then
                     !   write(*,*) " L4335 the source ", iSource, &
                     !        " will be checked proximityX=", proximityX(iSource)
                     !   write(*,*) " L4337 sourceProximity=", sourceProximity(iSource)
                     !   write(*,*) "+++++++++++++++++++++++++++++++++++++++++++++"
                     !endif
                  endif


                  ! VNY 14.10.2004 Check the x-position of the first order
                  if (.not. omgsourceRegime) then
                     xFirstIn=xFirst
                     yFirstIn=yFirst
                     verbFlag=0
                     !if (iSource .eq. 49) then
                     !   verbFlag=1
                     !endif
                     !if (iSource .eq. 70) then
                     !   if (sourceProximity(iSource) .gt. -1) then
                     !      write(*,*) "L4353 the source ", iSource, " will be checked for proximity "
                     !   endif
                     !   write(*,*) "L4355 xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn
                     !endif
                     call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                          spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                          verbFlag, xFirstOut, yFirstOut, averCountIn, averCountOut)
                     !if (iSource .eq. 70) then
                     !   write(*,*) " L4361 xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn, &
                     !        " xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                     !   write(*,*) "L4363  isource=", iSource, " difference=", &
                     !        abs(xFirstIn-xFirstOut), &
                     !        " spectrumHalfWidth=", spectrumHalfWidth
                     !endif
                     verbFlag=0
!                    
                     if (abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then ! correct the position
                        xFirst=xFirstOut
                        xFirstOrder=xFirst
                        !if (iSource .eq. 70) then
                        !   write(*,*) " L4373 xFirst=", xFirst, " (corrected) "
                        !endif
                        ! VNY 04.11.2005 Check if there are detected extended sources nearby
                     else
                        if (abs(xFirstIn-xFirstOut) .lt. 2.5*spectrumHalfWidth) then ! correct the position
                           !if (iSource .eq. 66) then
                           !   write(*,*) " L4217 analysing the source X-position"
                           !endif
                           !do eSource=0,nSources-1
                           do eSource=0,nSources2-1
                              if (eSource .ne. iSource .and. sourceShape(eSource) .eq. 1 .and. sourceRelation(eSource) .lt. 0) then
                                 distance=(xFirstOut-sourceXCoord(eSource))*(xFirstOut-sourceXCoord(eSource))+&
                                      (yFirstOut-sourceYCoord(eSource))*(yFirstOut-sourceYCoord(eSource))
                                 distance= sqrt(distance)
                                 xDist=abs(xFirstOut-sourceXCoord(eSource))
                                 yDist=abs(yFirstOut-sourceYCoord(eSource))
                                 if (distance .lt. 30.0*spectrumHalfWidth .and. xDist .lt. 1.5*spectrumHalfWidth) then
                                    
                                    !write(*,*) "L3308 eSource=", eSource, " distance=", distance, &
                                    !     " xDist=", xDist, " yDist=", yDist
                                    ! Changing the coordinate of the first-order
                                    xFirst=sourceXCoord(eSource)
                                    xFirstOrder=xFirst
                                    sourceRelation(eSource)=iSource
                                    sourceRelation(iSource)=eSource
                                 endif
                              endif ! eSource .ne. iSource
                           enddo
                        endif ! abs(xFirstIn-xFirstOut) .lt. 2.5
                     endif ! abs(xFirstIn-xFirstOut) .lt. 1.5
                  endif !if .not omgsourceRegime

                  ! Calculate the spectrum edge coordinates

                  xs0=xFirstOrder
                  x1=xs0
                  x2=xs0
                  ys0=yFirst
                  y1=ys0
                  y2=y1
                  yFirstOrder=ys0

                  y1=y1-0.25*spectrumLength
                  y2=y2+0.25*spectrumLength

                  iblueEdge=int(y1)
                  iredEdge=int(y2)

                  if (sourceProximity(iSource) .gt. -1) then
                     relationIndex=sourceProximity(iSource)
                     x1=min(sourceXcoord(iSource),sourceXcoord(relationIndex))
                     x2=max(sourceXcoord(iSource),sourceXcoord(relationIndex))
                     !if (iSource .eq. 453) then
                     !   write(*,*) "L4777 relationIndex=", relationIndex, " x1=", x1, &
                     !        " x2=", x2
                     !endif
                     x1=x1-abs(proximityX1)
                     x2=x2+abs(proximityX1)
                     !if (iSource .eq. 453) then
                     !   write(*,*) "L4783  x1=", x1, " x2=", x2
                     !endif
                  endif
                  !+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                  ! VNY 17.05.2007 extracting the averaged one-dimensional profile 
                  ! of the spectrum
                  !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++=
                  spiMin=(x1+x2)/2.0-spectrumHalfWidth*1.5
                  spiMax=(x1+x2)/2.0+spectrumHalfWidth*1.5
                  spjMin=iblueEdge
                  spjMax=iredEdge
        
                  if (spiMin .lt. 0) spiMin=0
                  if (spiMax .gt. nx-1) spiMax=nx-1
                  if (spjMin .lt. 0) spjMin=0
                  if (spjMax .gt. ny-1) spjMax=ny-1
                  !if (iSource .eq. 453) then
                  !   write(*,*) "L4493 spiMin=", spiMin, " spiMax=", spiMax, &
                  !        " spjMin=", spjMin, " spjMAx=", spjMax
                  !endif

                  allocate (sProfile(spiMin: spiMax), stat = allocErr)
                  call CheckAllocation(allocErr, "sProfile")
                  allocate (dProfile(spiMin: spiMax), stat = allocErr)
                  call CheckAllocation(allocErr, "dProfile")
               
                  do i=spiMin, spiMax
                     mx1=0.0
                     px1=0.0
                     do j=spjMin,spjMax
                        mx1=mx1+oswImage(i,j)
                        px1=px1+1.0
                     enddo ! j=spjMin
                     if (px1 .gt. 1.0) then
                        mx1=mx1/px1
                     endif
                     sProfile(i)=mx1
                  enddo ! i=spiMin

                  !if (iSource .eq. 453) then
                  !   write(*,*) "L4516 Profile:"
                  !   write(*,*) "--------------"
                  !endif
                  do i=spiMin,spiMax
                     mx1=0.0
                     if (i-spiMin .gt. 0) then
                        mx1=sProfile(i)-sProfile(i-1)
                     endif
                     dProfile(i)=mx1
                     !if (iSource .eq. 453) then
                     !   write(*,*) " i=", i, " sProfile=", sProfile(i), " dProfile=", dProfile(i)
                     !endif
                  enddo

                  profileChecked=.false.
                  nSp=1
                  
                  if (sourceProximity(iSource) .gt. -1) then
                     verbFlag=0
                     !if (iSource .eq. 238) then
                     !   verbFlag=1
                     !endif
                     !write(*,*) "L4537 --------------- iSource =", iSource
                     call EvaluateProfile(sProfile, dProfile, spiMin, spiMax, nSp, &
                          peacks, leftBorder, rightBorder, verbFlag)
                     !if (iSource .eq. 453) then
                     !   write(*,*) "L4542 nSp=", nSp, " spectrumHalfWidth=", spectrumHalfWidth
                     !   do i=1,nSp
                     !      write(*,*) "L4544 i=", i, " peack=", peacks(i), "left=",leftBorder(i), &
                     !           " right=", rightBorder(i)
                     !   enddo
                     !endif
                     if (nSp .eq. 1) then
                        spiMin2=leftBorder(1)
                        spiMax2=rightBorder(1)
                        spiX2=peacks(1)
                     else
                        if (proximityX(iSource) .lt. 0) then 
                           ! Picking up the right peak of the profile
                           spiMin=leftBorder(2)
                           spiMin2=leftBorder(2)
                           spiMax2=rightBorder(2)
                           spiX2=peacks(2)
                           xs0=peacks(2)
                           !write(*,*) "L4507 picking up the RIGHT maximum"
                        else ! picking up the left peak of the profile
                           spiMax=rightBorder(1)
                           spiMin2=leftBorder(1)
                           spiMax2=rightBorder(1)
                           spiX2=peacks(1)
                           xs0=peacks(1)
                           !write(*,*) "L4514 picking up the LEFT maximum"
                        endif
                        profileChecked=.true.
                     endif
                     !if (iSource .eq. 453) then
                     !   write(*,*) "L4572 spiMin=", spiMin, " spiMax=", spiMax
                     !   write(*,*) "L4573 spiMin2=", spiMin2, " spiMax2=", spiMax2
                     !endif
                  else
                     ! search for the maxima and minima
                     mx1=0.0
                     px1=(x1+x2)/2.0
                     dmx1=0.0
                     dmn1=999999.0
                     dpx1=px1
                     dpn1=px1
                     aux1=0.0
                     aux2=0.0
                     do i=spiMin, spiMax ! compute the average derivative of the profile
                        if (i .ne. spiMin) then
                           aux1=aux1+ abs(dProfile(i))
                           aux2=aux2+1.0
                        endif
                        if (sProfile(i) .gt. mx1) then ! maximum opf the profile
                           mx1=sProfile(i)
                           px1=i
                        endif
                        if (dProfile(i) .gt. dmx1) then ! maximum of the derivative of the profile
                           dmx1=dProfile(i)
                           dpx1=i
                        endif
                        if (dProfile(i) .lt. dmn1) then !minimum of the derivative of the profile
                           dmn1=dProfile(i)
                           dpn1=i
                        endif
                     enddo
                     if (aux2 .gt. 1.0) then
                        aux1=aux1/aux2
                     endif
                     ! Check the left and right limits of the spectrum
                     spiMin2=0 ! left limit of the profile
                     spiMax2=0 ! right limit of the profile
                     aux1=aux1/2.0
                     do i=spiMin,spiMax
                        if (abs(dProfile(i)) .gt. aux1) then
                           spiMin2=i
                           exit
                        endif
                     enddo
                     do i=spiMax,spiMin+1,-1
                        if (abs(dProfile(i)) .gt. aux1) then
                           spiMax2=i
                           exit
                        endif
                     enddo
               
                  endif ! sourceProximity .gt. -1

                  !if (iSource .eq. 453) then
                  !   write(*,*) "--------------------"
                  !   write(*,*) " L4627 dmx1=", dmx1, " dpx1=", dpx1, &
                  !        " dmn1=", dmn1, " dpn1=", dpn1
                  !   write(*,*) " L4629 aux1=", aux1, " aux2=", aux2, &
                  !        " spiMin2=", spiMin2, " spiMax2=", spiMax2
                  !endif

                  ! Compute signal-to-noise ratio
                  noise =0.0
                  aux2=0.0
                  iR=spiMin2-1
                  iL=spiMin2
                  iL=iL-spiMax2+spiMin2-1
                  if (iR .lt. spiMin+1) iR=spiMin+1
                  if (iL .lt. spiMin) iL=spiMin
                  !if (iSource .eq. 453) then
                  !   write(*,*) "L4642 iL=", iL, " iR=", iR
                  !endif
                  do i=iR,iL,-1
                     noise=noise+sProfile(i)
                     aux2=aux2+1.0
                  enddo
                  iL=spiMax2+1
                  iR=spiMax2
                  iR=iR+spiMax2-spiMin2+1
                  if (iL .gt. spiMax-1) iL=spiMax-1
                  if (iL .lt. spiMin) iL=spiMin
                  if (iR .gt. spiMax) iR=spiMax
                  do i=iL,iR
                     noise=noise+sProfile(i)
                     aux2=aux2+1.0
                  enddo
                  if (aux2 .gt. 1.0) noise=noise/aux2

                  signal=0.0
                  aux2=0.0
                  iL=spiMin2
                  iR=spiMax2
                  if (iL .lt. spiMin) iL=spiMin
                  if (iR .gt. spiMax) iR=spiMax
                  do i=iL,iR
                     signal=signal+sProfile(i)
                     aux2=aux2+1.0
                  enddo
                  if (aux2 .gt. 1.0) signal=signal/aux2!

                  snRatio(iSource)=0.0
                  if (noise .gt. 0.0) then
                     snRatio(iSource)=(signal-noise)/sqrt(noise)
                  endif
                  !if (iSource .eq. 453) then
                  !   write(*,*) " L4676 signal=", signal, " noise=", noise, &
                  !        " s-n=", signal-noise, " snRatio=", snRatio(iSource), " iSource=", iSource
                  !   write(*,*) "------------------------------------- xFirst=",xFirst
                  !endif

                  if(allocated(sProfile)) then
                     deallocate(sProfile,stat=istat)
                     call CheckDeallocation(allocErr, "sProfile")
                  end if
                  if(allocated(dProfile)) then
                     deallocate(dProfile,stat=istat)
                     call CheckDeallocation(allocErr, "dProfile")
                  end if

!--------------------------------------------------------------------
                  !if (iSource .eq. 453) then
                  !   write(*,*) " L4692 iSource=", iSource, " relation=", sourceRelation(iSource)
                  !endif
                  ! extract spectrum with the evaluation purposes with mode=2 (evaluationMode)
                  !-- 11.01.05
                  verbFlag=0
                  !if (iSource .eq. 453) then
                  !   verbFlag=1
                  !   write(*,*) "L4699 xZeroA=", xZeroA, " yZeroA=", yzeroA, " xFirst=", xFirst, &
                  !        " yFirst=", yFirst ! xFirst=1003
                  !   write(*,*) "L4701 proximityX1=", proximityX1
                  !endif
                  call ExtractSpectrum(grismSetH, xZeroA, yZeroA, xFirst, yFirst, &
                       spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                       bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                       spectrumArray, spectrumErrArray, bkgArray, &
                       bkgErrArray, fluxArray, fluxErrArray, ysize, &
                       evaluationMode, subtractBkg, verbFlag,s2n,s2n0, &
                       crossShape, leftBorder, rightBorder, proximityX1, nSp)

                
                  blueEdge=yZeroA+region1
                  redEdge=yZeroA+region2
                  iblueEdge=int(blueEdge)
                  iredEdge=int(redEdge)
                  !if (iSource .eq. 453) then
                  !   write(*,*) "L4717 iblueEdge=", iblueEdge, " iredEdge=", iredEdge, &
                  !        " subtractBkg=", subtractBkg
                  !   !do ii=iblueEdge,iredEdge
                  !   !   write(*,*) "ii=", ii, "   sp=", spectrumArray(ii), "  bkg=", bkgArray(ii), &
                  !   !        " ratio=", spectrumArray(ii)/(bkgarray(ii)+0.00001)
                  !   !enddo
                  !endif
                  verbFlag=0
                  yFirst=(blueEdge+redEdge)/2.0
                  spectrumHalfLength=(redEdge-blueEdge)/2.0

                  sizeMask=ny

!-----------------------
                  call FindPossibleFeatures2 (spectrumArray, smoothWindowHalfWidth, &
                       maxValue, maxValue2, meanValue2, maskArray, &
                       ny, iblueEdge, iredEdge, verbFlag, nFeatures) 
!----------------------
                  !if (iSource .eq. 66) then
                  !   write(*,*) "L4512 maxValue=", maxValue, " maxValue2=", maxValue2, &
                  !        " meanValue2=", meanValue2, " nFeatures=", nFeatures
                  !endif

                  contaminationComment= " "
                  putContaminationComment=.false.
                  if (nFeatures .gt. 0) then ! issue a warning
                     !call warning("Overlapping", "The spectrum is likely to be contaminated by zero-order features")
                     !write(errstr,*) "The number of contaminating sources:  ", nFeatures
                     if (iSource .gt. -1) then
                        write(errstr,*) "Source: ", iSource, &
                             " is likely to be contaminated; zero-order features:  ", nFeatures
                     else
                        write(errstr,*) "Source is likely to be contaminated; zero-order features:  ", nFeatures
                     endif
                     nContaminatedSources=nContaminatedSources+1
                     call message(AppMsg, 3)
                     write(contaminationComment, '(a, i3, a)') &
                          "The spectrum is likely to be contaminated by (",  &
                          nFeatures, ") zero-order features"
                     putContaminationComment=.true.
                  endif
                  verbFlag=0
                  !if (iSource .eq. 16) then
                  !   verbFlag=1
                  !endif
                  call EvaluateSpectrum(spectrumArray, bkgArray, &
                       ysize, iblueEdge, iredEdge, bkgWidth, &
                       spectrumWidth, sbRatio, maskArray, verbFlag)
                  !if (iSource .eq. 70) then
                  !   write(*,*) "L4714 iSourc=", iSource, " sbRatio=", sbRatio
                  !   write(*,*) " s2n=", s2n, " crossShape=", crossShape
                 ! 
                 !    if (s2n0 .gt. s2n*10.0 .and. s2n .gt. 0.5 .and. crossShape .gt. 0.0 &
                 !         .and. sbRatio .gt. 0.1) then
                 !       write(*,'(a,i5,6(a,f11.3))') "L4719 ISOURCE=", iSource, &
                 !            " s2n=", s2n, " s2n0=", s2n0, " x0=", xZeroA, &
                 !            " y0==", yZeroA, " sbRatio=", sbRatio, " crossShape=", crossShape
                 !    else
                 !       write(*,'(a,i5,6(a,f11.3))') "L4723 isource=", iSource, " s2n=", &
                 !            s2n, " s2n0=", s2n0, " x0=", xZeroA, &
                 !            " y0==", yZeroA, " sbRatio=", sbRatio, " crossShape=", crossShape
                 !    endif
                 ! endif
!                  02.04.2007
                  if (s2n .gt. 0.2 .and. crossShape .gt. 0.0 .and. sbRatio .gt. 0.05) then
                     if (s2n0 .lt. s2n*zero2firstRatio) then ! check the position of zero-order
                        !write(*,*) "L4731 the position of zero-order is possibly wrong, iSource=", &
                        !     iSource
                        !write(*,*) "l3684 sourceRelation=", sourceRelation(iSource)
                        if (sourceRelation(iSource) .gt. -1) then
                           iCheck=sourceRelation(iSource)
                           !write(*,*) "L3656 iCheck=", iCheck, " sourceRelation(iCheck)=", &
                           !     sourceRelation(iCheck)
                           sourceRelation(iSource)=-1
                           sourceRelation(iCheck)=-1
                        endif
                     else
                        if (extractionMode .ne. evaluationMode) then
                           ! VNY 14.10.2004 Check the x-position of the first order
                           if (.not. omgsourceRegime) then
                              xFirstIn=xFirst
                              yFirstIn=yFirst
                              verbFlag=0
                              !if (iSource .eq. 70) verbFlag=1

                              call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                                   spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                                   verbFlag, xFirstOut, yFirstOut, averCountIn, &
                                   averCountOut)
                              !if (iSource .eq. 70) then
                              !   write(*,*) " L4584 xqFirstIn=", xFirstIn, " yFirstIn=", yFirstIn, &
                              !        " xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                              !   write(*,*) " L3699 iSource=", iSource, " diff=", abs(xFirstIn-xFirstOut), &
                              !        " spectrumHalfWidth=", spectrumHalfWidth
                              !endif
                              !if (iSource .eq. 70) then 
                              !   write(*,*) "L4760 xFirstIn=", xFirstIn, &
                              !        " yFirstIn=", yFirstIn, &
                              !        " xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                              !   write(*,*) "L4764  iSource=", iSource, " difference+", &
                              !        abs(xFirstIn-xFirstOut), &
                              !        " spectrumHalfWidth=", spectrumHalfWidth
                              !   write(*,*) "L4767 profileChecked=", profileChecked
                              !endif
                              verbFlag=0
                              !if (abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then ! correct the position
                              !   xFirst=xFirstOut
                              !   xFirstOrder=xFirst
                              !endif
                              if (.not. profileChecked .and. &
                                   abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then 
                                 ! correct the position
                                 xFirst=xFirstOut
                                 xFirstOrder=xFirst
                              else 
                                 xFirst=xFirstIn
                                 xFirstOrder=xFirst
                              endif
                           endif
                           ! --- 11.01.05
                           verbFlag=0
                           !if (iSource .eq. 70) then 
                           !   !verbFlag=1
                           !   write(*,*) "L4788 xFirstOrder=", xFirstOrder
                           !   write(*,*) "L4789 proximityX1=", proximityX1
                           !endif

                           call ExtractSpectrum(grismSetH, xZeroA, yZeroA, xFirst, yFirst, &
                                spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                                bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                                spectrumArray, spectrumErrArray, bkgArray, &
                                bkgErrArray, fluxArray, fluxErrArray, ysize, &
                                extractionMode, subtractBkg, verbFlag,s2n,s2n0, &
                                crossShape, leftBorder, rightBorder, proximityX1, nSp)
                        endif
                        !if (iSource .eq. 107) then
                        !   write(*,*) "L5433 xFirstOrder=", xFirstOrder, " leftBorder(1)=", &
                        !        leftBorder(1), " rightBorder(1)=", rightBorder(1)
                        !endif
                        ! First check if there was some first-order within the detected spectrum
                        ! and mark this first-order as already identified
!                        do i=0, nSources-1
                        do i=0, nSources2-1
                           if (sourceShape(i) .eq. 1) then 
                              if (i .ne. iSource .and. sourceRelation(i) .lt. 0) then
                                 if (sourceXcoord(i) .gt. xZeroA-shapeCriterium2 &
                                      .and. sourceXcoord(i) .lt. xZeroA+shapeCriterium2) then
                                    if (sourceYcoord(i) .gt. blueEdge .and. &
                                         sourceYcoord(i) .lt. redEdge) then
                                       ! Label the source
                                       sourceRelation(i)=iSource
                                       if (i .lt. nSources) then
                                          spectrumOriginArray(i)=iSource+1
                                       endif
                                       if (iSource .lt. nSources) then
                                          spectrumOriginArray(iSource)=i+1
                                       endif
                                    endif
                                 endif
                              endif ! i .ne. iSource
                           endif ! sourceShape(i) .eq. 1
                        enddo ! i=0,nSources-1

                        verbFlag=0
                        nFeatures=0

                        spectrumNoCount=spectrumNoCount+1
                        !if (spectrumNoCount .eq. 71) then
                        !   write(*,*) "L5179 spNo=", spectrumNoCount, " iSource=", iSource
                        !endif
                        if (iSource .lt. nSources) then
                           spectrumNoArray0(iSource)=spectrumNoCount+1
                        endif
                        spectrumNoArray(iSource)=spectrumNoCount+1
                        spectrumNoArray2(spectrumNoCount)=iSource+1
                        spectrumOriginArray2(spectrumNoCount)=iSource+1
                        srcNoArray2(spectrumNoCount)=iSource
                        x0aux(spectrumNoCount)=xZeroOrder
                        y0aux(spectrumNoCount)=yZeroOrder
                        x1aux(spectrumNoCount)=xFirstOrder
                        y1aux(spectrumNoCount)=yFirstOrder
                        !write(*,*) "L5474 iSource =", iSource, " xZeroOrder=", xZeroOrder, " yZeroOrder=", yZeroOrder
                        !write(*,*) "L5475^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"

!*********************************************************************************** 30.10.2005
                        if (.not. createSpectrumFile) then
                           outFile4=outFile
                           outTargetSet0H = dataSet(outFile4, create) ! This file will contain 
                           ! only the target spectrum
                           createSpectrumFile=.true.
                           createSpectraListFile=.true.
                           call setAttribute(outTargetSet0H, "CONTENT", "OM GRISM SPECTRA", &
                                "File contains OM grsim spectrum/spectra")
                           call setAttribute(outTargetSet0H, "HDUCLASS", "OGIP    ", &
                                "Format conforms to OGIP/GSFC conventions")
                           call setAttribute(outTargetSet0H, "HDUCLAS1", "SPECTRUM", &
                                "File contains spectral data")
                           call setAttribute(outTargetSet0H, "HDUVERS", "1.2.0    ", &
                                "The version number of the format")
                           call setAttribute(outTargetSet0H, "TELESCOP", "XMM     ", &
                                "XMM mission")
                           call setAttribute(outTargetSet0H, "INSTRUME", "OM      ", &
                                "Instrument name")
                           call setAttribute(outTargetSet0H, "DATAMODE", dataMode, &
                                "Data mode")
                           call setAttribute(outTargetSet0H, "FILTER", filter, &
                                "Filter used for obtaining the data")
                           call setAttribute(outTargetSet0H, "FRAMTIME", frameTime, &
                                "Frame time")
                           call setAttribute(outTargetSet0H, "DEADFRAC", deadFrac, &
                                "Frame dead fraction")
                           call setAttribute(outTargetSet0h, "TMEXP_ID", tmExpId, &
                                "Exposure Id from telemetry")
                           call setAttribute(outTargetSet0H, "FRMTIME", frmTime, &
                                "Frametime from telemetry")
                           call setAttribute(outTargetSet0H, "DATATYPE", dataType, &
                                "Data type")
                           call setAttribute(outTargetSet0H, "OBS_ID", obsId, &
                                "Observation identifier")
                           call setAttribute(outTargetSet0H, "EXP_ID", expId, &
                                "Exposure identifier")
                           call setAttribute(outTargetSet0H, "DATE-OBS", dateObs, &
                                "Observation start time")
                           call setAttribute(outTargetSet0H, "DATE-END", dateEnd, &
                                "Observation end time")
                        endif
!*********************************************************************************** 30.10.2005

                        write(sourceIndex, '(i3.3)') spectrumNoCount+1
                        !
                        !if (spectrumNoCount .eq. 71) then
                        !   write(*,*) "L5239 spNo=", spectrumNoCount, " iSource=", iSource
                        !endif
                        write(blockName, '(a,a)') "RATE", sourceIndex
                        outTargetTable0H=addTable(outTargetSet0H, blockName, nBins, &
                             "Table with the field source spectrum")
                        !  Adding the spectrum number to the source list file
                        blkNum=blockNumber(outTargetSet0H, blockName)
                        ! 
                        ! Computing the binning interval
                        if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                           binningInterval=2.40774*binY ! in Angstroms
                        else
                           binningInterval=5.626*binY
                        endif
                        !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                        call setAttribute(outTargetTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                        call setAttribute(outTargetTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                        call setAttribute(outTargetTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                        call setAttribute(outTargetTable0H, "HDUCLAS2", "NET", "Background subtracted")
                        call setAttribute(outTargetTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                        call setAttribute(outTargetTable0H, "DLAMBDA", binningInterval, "Angstrom", "binning interval")
                        call setAttribute(outTargetTable0H, "EXTRMODE", extractionMode, "Spectra extraction mode")
                        !--Adding columns to the output table
                        waveSourceOutCol0 = addColumn(outTargetTable0H, "WAVELENGTH", real32, &
                             "ANGSTROM", columnDim, "wavelengths ")
                        waveSourceOutArray0 => real32data(waveSourceOutCol0)
                        !
                        rateSourceOutCol0 = addColumn(outTargetTable0H,"RATE", real32, &
                             "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                        rateSourceOutArray0 => real32data(rateSourceOutCol0)
                        !
                        rateSourceErrOutCol0 = addColumn(outTargetTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                             columnDim, "source counts rate error")
                        rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                        !
                        rateBackOutCol0 = addColumn(outTargetTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate")
                        rateBackOutArray0 => real32data(rateBackOutCol0)
                        !
                        rateBackErrOutCol0 = addColumn(outTargetTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate error")
                        rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                        !
                        fluxOutCol0 = addColumn(outTargetTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate")
                        fluxOutArray0 => real32data(fluxOutCol0)
                        !
                        fluxErrOutCol0 = addColumn(outTargetTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate error")
                        fluxErrOutArray0 => real32data(fluxErrOutCol0)
                        !
                        !
                        waveSourceOutArray0=wavelengthArray
                        rateSourceOutArray0=spectrumArray
                        rateSourceErrOutArray0=spectrumErrArray
                        rateBackOutArray0=bkgArray
                        rateBackErrOutArray0=bkgErrArray
                        fluxOutArray0=fluxArray
                        fluxErrOutArray0=fluxErrArray

                        !
                        vignApp=.false.
                        deadApp=.false.
                        backApp=.true.
                        psfApp=.false.
                        call setAttribute(outTargetTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                        call setAttribute(outTargetTable0H, "NPIXSOU", sourceArea, "pix", "Source area")
                        call setAttribute(outTargetTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                        call setAttribute(outTargetTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                        call setAttribute(outTargetTable0H, "BACKAPP", backApp,  "Background is subtracted")
                        call setAttribute(outTargetTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            

                        !call setAttribute(outTargetTable0H, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                        !call setAttribute(outTargetTable0H, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                        !call setAttribute(outTargetTable0H, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                        !call setAttribute(outTargetTable0H, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                        if (columnDataTypeFlag .eq. 2) then
                           call setAttribute(outTargetTable0H, "QFLAG", sourceQF8(0), " ",  "Quality flag")
                        else
                           call setAttribute(outTargetTable0H, "QFLAG", sourceQF16(0), " ",  "Quality flag")
                        endif
                        !write(*,*) "L3946"
                        call setAttribute(outTargetTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                        call setAttribute(outTargetTable0H, "OBJ_TYPE", "FIELD OBJECT", &
                             "Target flag                                 ")
                        call setAttribute(outTargetTable0H, "XFIRST", xFirstOrder, &
                             "pix", "First-order (spectrum) image X-coordiate")
                        call setAttribute(outTargetTable0H, "YFIRST", yFirstOrder, &
                             "pix", "First-order (spectrum) image Y-coordiate")
                        !
                        ! VNY 21.07.2004 Recompute zero-order coordinates
                        !computingMode=0
                        !verbFlag=0
                        !xZeroIn=xZeroOrder
                        !yZeroIn=yZeroOrder
                        !call RecomputeZeroOrder(grismSetH, xZeroIn, yZeroIn, &
                        !     spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                        !     verbFlag, xZeroOut, yZeroOut)
                        !xZeroOrder=xZeroOut
                        !yZeroOrder=yZeroOut
                        !

                        !---------------------------- VNY 14.10.2004
                        box0left=xZeroOrder - spectrumHalfWidth +0.5
                        box0right=xZeroOrder+spectrumHalfWidth +0.5
                        box1left=xFirstOrder-spectrumHalfWidth +0.5
                        box1right=xFirstOrder+spectrumHalfWidth-0.5
                        box0down=yZeroOrder-spectrumHalfWidth+0.5
                        box0up=yZeroOrder+spectrumHalfWidth+0.5
                        box1down=yFirstOrder-spectrumHalfHeight+0.5
                        box1up=yFirstOrder+spectrumHalfHeight+0.5
                        !-------------
                        !if (iSource .eq. 66) then
                        !   write(*,*) "L4824 box1left=", box1left, " box1right=", box1right
                        !   write(*,*) "L4825 nSp=", nSp, " leftBorder=", leftBorder,  &
                        !        " rightBorder=", rightBorder, " proximityX1=", proximityX1 
                        !endif

                        if (nSp .eq. 2) then
                           if (proximityX1 .lt. 0) then ! using the right spectrum limits
                              box1left=leftBorder(2)
                              box1right=rightBorder(2)
                           endif
                           if (proximityX1 .gt. 0) then ! using the right spectrum limits
                              box1left=leftBorder(1)
                              box1right=rightBorder(1)
                           endif

                        !if (iSource .eq. 66) then                           
                        !   write(*,*) "L4840 the left/right limits of the extraction box are changed"
                        !   write(*,*) "L4841 box1left=", box1left, " box1right=", box1right
                        !endif

                        endif ! nSp .eq. 2

                        call setAttribute(outTargetTable0H, "XZERO", xZeroOrder, &
                             "pix", "Zero-order (object) image X-coordiate")
                        call setAttribute(outTargetTable0H, "YZERO", yZeroOrder, &
                             "pix", "Zero-order (object) image Y-coordiate")

                        if (nSp .eq. 2) then
                           call setAttribute(outTargetTable0H, "SPECREGW", &
                                box1right-box1left+1, &
                                "pix", "Width of the spectrum extraction region")

                           !if (iSource .eq. 66) then
                           !   write(*,*) "L4856 SPECREGW=", box1right-box1left+1
                           !endif

                           if (proximityX1 .lt. 0) then ! no left bkg region
                              call setAttribute(outTargetTable0H, "BKGREGWL", 0.0, &
                                   "pix", "Backgr.extraction region width (left)")
                              call setAttribute(outTargetTable0H, "BKGSEPL", 0.0, &
                                   "pix", "Spectrum to backgr.reg.separation (left)")
                           else
                              call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                                   "pix", "Backgr.extraction region width (left)")
                              call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                                   "pix", "Spectrum to backgr.reg.separation (left)")
                           endif

                           if (proximityX1 .gt. 0) then ! no right bkg region
                              call setAttribute(outTargetTable0H, "BKGREGWR", 0.0, &
                                   "pix", "Backgr.extraction region width (right)")
                              call setAttribute(outTargetTable0H, "BKGSEPR", 0.0, &
                                   "pix", "Spectrum to backgr.reg.separation (right)")
                           else
                              call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthright), &
                                   "pix", "Backgr.extraction region width (right)")
                              call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                                   "pix", "Spectrum to backgr.reg.separation (right)")
                           endif
                        else
                           call setAttribute(outTargetTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                                "pix", "Width of the spectrum extraction region")
                           !write(*,*) "L4934 SPECREGW=", 2.0*abs(spectrumHalfWidth)
                           call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                                "pix", "Backgr.extraction region width (left)")
                           call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                                "pix", "Spectrum to backgr.reg.separation (left)")
                           call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthleft), &
                                "pix", "Backgr.extraction region width (right)")
                           call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                                "pix", "Spectrum to backgr.reg.separation (right)")
                        endif

                        !call setAttribute(outTargetTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                        !     "pix", "Width of the spectrum extraction region")
                        !call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                        !     "pix", "Backgr.extraction region width (left)")
                        !call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthleft), &
                        !     "pix", "Backgr.extraction region width (right)")
                        !call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                        !     "pix", "Spectrum to backgr.reg.separation (left)")
                        !call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                        !     "pix", "Spectrum to backgr.reg.separation (right)")

                        !write(*,*) "L3180 nContaminatedSources=", nContaminatedSources
                        if (putContaminationComment) then
                           call addComment(outTargetTable0H, contaminationComment)
                        endif
                        ! release the dal handles
                        !---------
                        !---VNY 30.08.03
                        call release(rateBackErrOutCol0)
                        call release(rateBackOutCol0)
                        call release(rateSourceErrOutCol0)
                        call release(rateSourceOutCol0)
                        call release(fluxOutCol0)
                        call release(fluxErrOutCol0)
                        call release(waveSourceOutCol0)
                        call release(outTargetTable0H)
                        !---


                        if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                           box1down=yZeroOrder+regionU1+0.5
                           box1up=yZeroOrder+regionU2+0.5
                        else
                           box1down=yZeroOrder+regionV1+0.5
                           box1up=yZeroOrder+regionV2+0.5
                        endif
                        if (box0left .lt. 0.0) box0left=0.0
                        if (box0right .gt. nx-1) box0right=nx-1
                        if (box0down .lt. 0.0) box0down=0.0
                        if (box1left .lt. 0.0) box1left=0.0
                        if (box1right .gt. nx-1) box1right=nx-1
                        if (box1up .gt. ny-1) box1up=ny-1

                        !firstPass=.false.
                        if (givenSpectraRegionFile) then 
                           call writeSpectrumExtractionRegion(spectraRegionFileName, &
                                box0left,box0right,box0down,box0up, &
                                box1left,box1right,box1down,box1up, &
                                spectrumNoCount, firstPass, givenSpectraRegionFile)
                           firstPass=.false.
                        else
                           if (givenRegionFile) then
                              call writeSpectrumExtractionRegion(regionFileName, &
                                   box0left,box0right,box0down,box0up, &
                                   box1left,box1right,box1down,box1up, &
                                   spectrumNoCount, firstPass, givenSpectraRegionFile)
                              firstPass=.false.
                           endif
                        endif
                     endif !  if (s2n0 .lt. s2n*10.0) 
                  endif ! s2n .gt. 2.0
!----------------------------------------------------------------
               endif ! if availableSpectrumFraction .gt. 0.6
            endif ! sourceShape .gt. 0
         enddo ! iSource=0, nSources-1
         if (nContaminatedSources .gt. 0) then
            write(errstr,*) "Number of likely contaminated (by zero-order features) sources:", nContaminatedSources
            call message(AppMsg, 2)
         endif



         !write(*,*) "======================================================"
         !write(*,*) "L3940 Processing extended sources without zeros"
         !write(*,*) "======================================================"
         !*********************************************************
         !*********************************************************         
         !*********************************************************
         ! Processing source by source those extended sources (spectra) 
         ! that were not related to their zero-orders
         !*********************************************************
         !*********************************************************
         !*********************************************************

         
!         do iSource=0, nSources-1
         do iSource=0, nSources2-1

            !write(*,*) "L3950 iSource=", iSource, " shape=", sourceShape(iSource), &
            !     " relation=", sourceRelation(iSource), " procFlag=", procFlag(iSource)
            if (sourceShape(iSource) .eq. 1 .and. sourceRelation(iSource) .lt. 0 &
                 .and. procFlag(iSource) .lt. 0) then
               spectrumHalfLength=sourceFWHMaj2(iSource)
               !if (iSource .eq. 383) then
               !   write(*,*) "L5522 iSource=", iSource, " shape=", sourceShape(iSource), &
               !        " relation=", sourceRelation(iSource), " procFlag=", procFlag(iSource), &
               !        " x=", sourceXCoord(iSource), " y=", sourceYCoord(iSource)
               !   write(*,*) "L5525 spectrumHalfLength=", spectrumHalfLength, &
               !        " sourceYcoord=", sourceYCoord(iSource), &
               !        " region1=", region1
               !   write(*,*) "L5528 criterium=",sourceYCoord(iSource)-spectrumHalfLength-region1*0.8  
               !endif
               if(sourceYCoord(iSource)-spectrumHalfLength-region1*0.8 .gt. 0) then 
                  ! Trying to find a zero-order
                  xZero=sourceXCoord(iSource)
                  yZero=sourceYCoord(iSource)-sourceFWHMaj2(iSource)-region1
                  xFirst=sourceXCoord(iSource)
                  yFirst=sourceYCoord(iSource)
                  if (yZero .lt. 0.0) then
                     yZero=1.0
                  endif
                  !if (iSource .eq. 383) then
                  !   write(*,*) "L5541 yZero=", yZero, " distance=", &
                  !        sourceYCoord(iSource)-spectrumHalfLength-region1 
                  !endif
                  verbFlag=0

                  !   write(*,*) "L3975 =============== iSource=10"
                     !---VNY 18.11.2005 Check if there is a bright zero order nearby
                     !---------------------------------------------------------------------------
                  !zeroOrderCriterium=180.0/xBin ! criterium for checking the erroneous zero-orders
                  zeroOrderCriterium=220.0/xBin ! criterium for checking the erroneous zero-orders
                  !if (iSource .eq. 9) then
                  !   write(*,*) "L4681 nSourcesListCheck=", nSourcesListCheck, &
                  !        " nSourcesListBright=", nSourcesListBright
                  !   write(*,*) "L4683 zeroOrderCriterium=", zeroOrderCriterium, " xBin=", xBin
                  !endif
                  ! scaninng the list of bright sources
                  foundZeroOrder=0 ! initially none is found (1: bright list, 2: check list)
                  !do iCheck=1,nSourcesListBright
                  do iCheck=nSources,nSources2-1
                     !deltaX=abs(sourceDataBright(iCheck)%xPos-xFirst)
                     !deltaY=abs(sourceDataBright(iCheck)%yPos-yZero)
                     deltaX=abs(sourceXCoord(iCheck)-xFirst)
                     deltaY=abs(sourceYCoord(iCheck)-yZero)
                     deltaXY=sqrt(deltaX*deltaX+deltaY*deltaY) 
                     !if (iSource .eq. 449) then
                     !   write(*,*) "L5566 iCheck=", iCheck, " deltaXY=", deltaXY, &
                     !        " zeroOrderCriterioum=", zeroOrderCriterium
                     !   write(*,*) "L5568 indexCheckSources(10)=", indexCheckSources(10) 
                     !endif
                     if (deltaXY .lt. zeroOrderCriterium .and. .not. omgsourceRegime) then 
                        !if (indexBrightSources(iCheck) .lt. 0) then
                           !if (iSource .eq. 383) then 
                           !   write(*,*) "L5573 iSource=", iSource, " iCheck=", iCheck, &
                           !        " deltaXY=", deltaXy, &
                           !        " sourceSize(i)=", sourceSize(iCheck), &
                           !        " rate(i)=", sourceRawCountRate(iCheck)
                           !   write(*,*) "L5571 x=",sourceXCoord(iCheck), &
                           !        " y=", sourceYCoord(iCheck)
                           !   
                           !endif
                           !Replace the coordinates only for very big and bright sources (zero-orders)
                           !if (sourceDataBright(iCheck)%noSourcePixels .gt. 400/xBin .and. &
                           !     sourceDataBright(iCheck)%noSourcePixels .lt. 3000/xBin .and. &
                           !     sourceDataBright(iCheck)%rawCountRate .gt. 500.0) then
!                           if (sourceDataBright(iCheck)%noSourcePixels .gt. 4/xBin .and. &
!                                sourceDataBright(iCheck)%noSourcePixels .lt. 200/xBin .and. &
!                                sourceDataBright(iCheck)%rawCountRate .gt. 400.0) then

                           if (sourceSize(iCheck) .gt. 4/xBin .and. &
                                sourceSize(iCheck) .lt. 200/xBin .and. &
                                sourceRawCountRate(iCheck) .gt. 400.0) then
                              foundZeroOrder=1 ! found in the bright list
                              !if (iSource .eq. 383) then
                              !   write(*,*) "          L5594 xold=", xFirst, &
                              !        " xnew=", sourceXCoord(iCheck)
                              !   write(*,*) "          L5596 yold=", yZero, &
                              !        " ynew=", sourceYCoord(iCheck)
                              !endif
                              xZero=sourceXCoord(iCheck)-1
                              yZero=sourceYCoord(iCheck)-1
                              !sourceRelation(iSource)=-iCheck
                              sourceRelation(iSource)=iCheck
                              !indexBrightSources(iCheck)=iSource
                              sourceRelation(iCheck)=iSource
                           endif
                        !endif !indexBrightSources .lt. 0
                     endif ! deltaXY
                  enddo ! iCheck=1, nSourcesBright
                  !if (iSource .eq. 383) then
                  !   write(*,*) "L5610 =============== iSource=", iSource
                  !   write(*,*) "L5611 foundZeroOrder=", foundZeroOrder, " yZero=", yZero 
                  !   verbFlag=1
                  !endif
                  call FindZeroOrder(grismSetH, xFirst, yFirst, &
                       spectrumHalfLength, yZero, zeroMaxValue, &
                       spectrumHalfWidth, shapeCriterium1, verbFlag, &
                       maxValue2, meanValue2, yZero2, spectrumHalfHeight)

                  !if (iSource .eq. 383) then
                  !   write(*,*) "L5620 ============="
                  !   write(*,*) "L5621 maxValue2=", maxValue2, &
                  !        " meanValue2=", meanValue2, " yZero=", yZero, &
                  !        " yZero2=", yZero2
                  !   write(*,*) "L5624 foundZeroOrder=", foundZeroOrder
                  !   write(*,*) "L5625 xZero=", xZero 
                  !endif


                  !**************************
                  !************************** 
                  ! yZero2 is the estimated  y-coordinate of the zero-order
                  if (yZero .gt. 0 .and. abs(maxValue2/meanValue2) .gt. 2.0) then
                     !if (iSource .eq. 10) then
                     !   write(*,*) " L5336 Trying to extract spectrum iSource=", iSource
                     !endif
                     ! extract the spectrum                     
                     subtractBkg=.true.
                     xZeroA=xFirst
                     !
                     !yZeroA=int(yZero+0.5)
                     yZeroA=yZero
                     !
                     ! VNY 29.03.2004 Check if this found zero-order belongs to some other
                     ! first order. If so, it must not be considered
                     checkFlag = .false.
                     !if (iSource .eq. 10) then
                     !   write(*,*) " L4574 checkWidth=", checkWidth
                     !endif
                     iminDistance=-999999
                     minDistance=999999.0
!                     do i=0,nSources-1
                     do i=0,nSources2-1
                        if (sourceShape(i) .eq. 0) then
                           deltaX=abs(sourceXCoord(i)-xZeroA)
                           deltaY=abs(sourceYCoord(i)-yZeroA)
                           distance=sqrt(deltaX*deltaX+deltaY*deltaY)
                           if (distance .lt. minDistance) then
                              iminDistance=i
                              minDistance=distance
                           endif
                           !if (iSource .eq. 383) then
                           !   write(*,*) "L5662 deltaX=", deltaX, " deltaY=", deltaY
                           !endif

                           !if (deltaX .lt. checkWidth .and. deltaY .lt. checkWidth) then
                           !   if (iSource .eq. 383) then
                           !      write(*,*) "L5667 iSource=", iSource, " deltaX=", deltaX, &
                           !           " deltaY=", deltaY, " i=",i, " sourceRelation=", &
                           !           sourceRelation(i)
                           !      write(*,*) "L5670 xZeroA=", xZeroA, " yZeroA=", yZeroA, &
                           !           " xi=", sourceXCoord(i), " yi=", sourceYCoord(i)
                           !   endif
                           !endif

                           if (deltaX .lt. checkWidth .and. deltaY .lt. checkWidth) then
                              if (sourceRelation(i) .lt. 0) then
                                 checkFlag=.true. ! found a source in the main list
                              endif
                           endif
                        endif
                     enddo ! i=0,nSources-1
                     !if (iSource .eq. 383) then
                     !   write(*,*) "L5683 iminDistance=", iminDistance, " minDistance=", minDistance, &
                     !        " sourceRelation=", sourceRelation(iSource), " checkFlag=", checkFlag
                     !endif
                     ! if there are no occupied zero-orders, extract the spectrum
                     !if (iSource .eq. 10) then
                     !   write(*,*) "L4612 checkFlag=", checkFlag, " iSource=", iSource, &
                     !        " iminDistance=", iMinDistance, " minDistance=", minDistance
                     !endif
                     !if (checkFlag) then
                     !   !write(*,*) "L4137 iSource=", iSource, " sourceRelation=", &
                     !   !     sourceRelation(iSource)
                     !   ! check whether the source from the additional list is closer
                     !   if (sourceRelation(iSource) .gt. -99999) then
                     !      iCheck=-sourceRelation(iSource) ! the index from either bright or check list
                     !      if (foundZeroOrder .eq. 1) then ! tacking the data from the bright list
                     !         deltaX=abs(sourceDataBright(iCheck)%xPos-xZeroA)
                     !         deltaY=abs(sourceDataBright(iCheck)%yPos-yZeroA)
                     !      else ! taking from the check list
                     !         deltaX=abs(sourceData(iCheck)%xPos-xZeroA)
                     !         deltaY=abs(sourceData(iCheck)%yPos-yZeroA)
                     !      endif
                     !      minDistance2=sqrt(deltaX*deltaX+deltaY*deltaY)
                     !      !write(*,*) "--------------------------"
                     !      !write(*,*) "L4134 iSource=", iSource, " sourceRelation=", &
                     !      !     sourceRelation(iSource), &
                     !      !     " x=", sourceData(iCheck)%xPos, " y=", sourceData(iCheck)%yPos, &
                     !      !     " minDistance2=", minDistance2
                     !      if (minDistance2 .lt. checkWidth .and. minDistance2 .lt. minDistance) then 
                     !         ! use the source from the additional list
                     !         iCheck=-sourceRelation(iSource)
                     !         if (foundZeroOrder .eq. 1) then ! taking from the bright list
                     !            xZeroA=sourceDataBright(iCheck)%xPos
                     !            yZeroA=sourceDataBright(iCheck)%yPos
                     !            xZeroOrder=xZeroA
                     !            yZeroOrder=yZeroA
                     !         else ! taking from the check list
                     !            xZeroA=sourceData(iCheck)%xPos
                     !            yZeroA=sourceData(iCheck)%yPos
                     !            xZeroOrder=xZeroA
                     !            yZeroOrder=yZeroA
                     !         endif
                     !      else ! use the source from the main list
                     !         xZeroA=sourceXCoord(iminDistance)
                     !         yZeroA=sourceYCoord(iminDistance)
                     !         xZeroOrder=xZeroA
                     !         yZeroOrder=yZeroA
                     !      endif
                     !   else
                     !      ! 09.12.2005
                     !      checkFlag=.false.
                     !   endif
                     !else
                     !   if (sourceRelation(iSource) .gt. -99999) then
                     !      iCheck=-sourceRelation(iSource)
                     !      if (foundZeroOrder .eq. 1) then ! using the bright list
                     !         deltaX=abs(sourceDataBright(iCheck)%xPos-xZeroA)
                     !         deltaY=abs(sourceDataBright(iCheck)%yPos-yZeroA)
                     !      else ! isung the check list
                     !         deltaX=abs(sourceData(iCheck)%xPos-xZeroA)
                     !         deltaY=abs(sourceData(iCheck)%yPos-yZeroA)
                     !      endif
                     !      minDistance2=sqrt(deltaX*deltaX+deltaY*deltaY)
                     !      !write(*,*) "L4158 iSource=", iSource, " sourceRelation=", sourceRelation(iSource), &
                     !      !     " x=", sourceData(iCheck)%xPos, " y=", sourceData(iCheck)%yPos, &
                     !      !     " minDistance2=", minDistance2
                     !      if (minDistance2 .lt. minDistance) then ! use the source from the additional list
                     !         checkFlag=.true.
                     !         if (foundZeroOrder .eq. 1) then ! using the bright list
                     !            xZeroA=sourceDataBright(iCheck)%xPos
                     !            yZeroA=sourceDataBright(iCheck)%yPos
                     !         else ! using the check list
                     !            xZeroA=sourceData(iCheck)%xPos
                     !            yZeroA=sourceData(iCheck)%yPos
                     !         endif
                     !         xZeroOrder=xZeroA
                     !         yZeroOrder=yZeroA
                     !      endif
                     !   endif
                     !endif
                     !write(*,*) "L4170 checkFlag=", checkFlag
                     if (checkFlag) then ! exctract this spectrum as the zero-order was found 
                        !write(*,*) "L4172 extracting spectrum"
                        if (.not. omgsourceRegime) then
                           ! VNY 14.10.2004 Check the x-position of the first order
                           xFirstIn=xFirst
                           yFirstIn=yFirst
                           verbFlag=0
                           !if (iSource .eq. 383) then
                           !   verbFlag=1
                           !endif
                           call CheckFirstOrder(grismSetH, xFirstIn, yFirstIn, &
                                spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                                verbFlag, xFirstOut, yFirstOut, averCountIn, &
                                averCountOut)
                           verbFlag=0
                           !if (iSource .eq. 383) then
                           !   write(*,*) "L5775 xFirstIn=", xFirstIn, " yFirstIn=", yFirstIn
                           !   write(*,*) "L5776 xFirstOut=", xFirstOut, " yFirstOut=", yFirstOut
                           !   write(*,*) " isource=", iSource, " difference=", abs(xFirstIn-xFirstOut), &
                           !        " spectrumHalfWidth=", spectrumHalfWidth
                           !endif
                           if (abs(xFirstIn-xFirstOut) .lt. 1.5*spectrumHalfWidth) then ! correct the position
                              xFirst=xFirstOut
                              xFirstOrder=xFirst
                           endif
                        endif
                        !--- 11.01.05
                           verbFlag=0
                           !write(*,*) "L4995 proximityX1=", proximityX1
                           !if (iSource .eq. 383) then
                           !   write(*,*) "L5792 xFirst=", xFirst, " yFirst=", yFirst
                           !endif
                        call ExtractSpectrum(grismSetH, xZeroA, yZeroA, xFirst, yFirst, &
                             spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                             bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                             spectrumArray, spectrumErrArray, bkgArray, &
                             bkgErrArray, fluxArray, fluxErrArray, ysize, &
                             extractionMode, subtractBkg, verbFlag,s2n,s2n0, &
                             crossShape, leftBorder, rightBorder, proximityX1, nSp)
                        !if (iSource .eq. 383) then
                        !   write(*,*) "L5796 the spectrum is extracted xZeroA=", xZeroA, &
                        !        " yZeroA=",yZeroA, " xFirst=", xFirst, " yFirst=", yFirst 
                        !endif
                        ! Include the spectrum into the output file
!
!********************************************************************* 30.10.2005
                        if (.not. createSpectrumFile) then
                           outFile4=outFile
                           outTargetSet0H = dataSet(outFile4, create) ! This file will contain 
                           createSpectrumFile=.true.
                           createSpectraListFile=.true.
                           call setAttribute(outTargetSet0H, "CONTENT", "OM GRISM SPECTRA", &
                                "File contains OM grsim spectrum/spectra")
                           call setAttribute(outTargetSet0H, "HDUCLASS", "OGIP    ", &
                                "Format conforms to OGIP/GSFC conventions")
                           call setAttribute(outTargetSet0H, "HDUCLAS1", "SPECTRUM", &
                                "File contains spectral data")
                           call setAttribute(outTargetSet0H, "HDUVERS", "1.2.0    ", &
                                "The version number of the format")
                           call setAttribute(outTargetSet0H, "TELESCOP", "XMM     ", &
                                "XMM mission")
                           call setAttribute(outTargetSet0H, "INSTRUME", "OM      ", &
                                "Instrument name")
                           call setAttribute(outTargetSet0H, "DATAMODE", dataMode, &
                                "Data mode")
                           call setAttribute(outTargetSet0H, "FILTER", filter, &
                                "Filter used for obtaining the data")
                           call setAttribute(outTargetSet0H, "FRAMTIME", frameTime, &
                                "Frame time")
                           call setAttribute(outTargetSet0H, "DEADFRAC", deadFrac, &
                                "Frame dead fraction")
                           call setAttribute(outTargetSet0h, "TMEXP_ID", tmExpId, &
                                "Exposure Id from telemetry")
                           call setAttribute(outTargetSet0H, "FRMTIME", frmTime, &
                                "Frametime from telemetry")
                           call setAttribute(outTargetSet0H, "DATATYPE", dataType, &
                                "Data type")
                           call setAttribute(outTargetSet0H, "OBS_ID", obsId, &
                                "Observation identifier")
                           call setAttribute(outTargetSet0H, "EXP_ID", expId, &
                                "Exposure identifier")
                           call setAttribute(outTargetSet0H, "DATE-OBS", dateObs, &
                                "Observation start time")
                           call setAttribute(outTargetSet0H, "DATE-END", dateEnd, &
                                "Observation end time")
                        endif
!********************************************************************* 30.10.2005
                        spectrumNoCount=spectrumNoCount+1
                        if (iSource .lt. nSources) then
                           spectrumNoArray0(iSource)=spectrumNoCount+1
                        endif
                        !if (spectrumNoCount .eq. 71) then
                        !   write(*,*) "L5940 spNo=", spectrumNoCount, " iSource=", iSource
                        !endif
                        spectrumNoArray(iSource)=spectrumNoCount+1
                        spectrumNoArray2(spectrumNoCount)=iSource+1
                        srcNoArray2(spectrumNoCount)=iSource
                        x0aux(spectrumNoCount)=xZeroA
                        y0aux(spectrumNoCount)=yZeroA
                        x1aux(spectrumNoCount)=xFirst
                        y1aux(spectrumNoCount)=yFirst
                        xFirstOrder=xFirst
                        yFirstOrder=yFirst
                        xZeroOrder=xZeroA
                        yZeroOrder=yZeroA
                        !if (iSource .eq. 383) then
                        !   write(*,*) "L5858 iSource =", iSource, " xZeroA=", xZeroA, " yZeroA=", yzeroA
                        !   write(*,*) "L5865 xFirst=", xFirst, " yFirst=", yFirst
                        !   write(*,*) "L5866^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ spectrumNoCount=", &
                        !        spectrumNoCount
                        !endif
                        write(sourceIndex, '(i3.3)') spectrumNoCount+1
                        !
                        write(blockName, '(a,a)') "RATE", sourceIndex
                        outTargetTable0H=addTable(outTargetSet0H, blockName, nBins, &
                             "Table with the field source spectrum")
                        !  Adding the spectrum number to the source list file
                        blkNum=blockNumber(outTargetSet0H, blockName)
                        if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                           binningInterval=2.40774*binY ! in Angstroms
                        else
                           binningInterval=5.626*binY
                        endif
                        !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                        call setAttribute(outTargetTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                        call setAttribute(outTargetTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                        call setAttribute(outTargetTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                        call setAttribute(outTargetTable0H, "HDUCLAS2", "NET", "Background subtracted")
                        call setAttribute(outTargetTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                        call setAttribute(outTargetTable0H, "DLAMBDA", binningInterval, "Angstrom", "binning interval")
                        call setAttribute(outTargetTable0H, "EXTRMODE", extractionMode, "Spectra extraction mode")
                        !--Adding columns to the output table
                        waveSourceOutCol0 = addColumn(outTargetTable0H, "WAVELENGTH", real32, &
                             "ANGSTROM", columnDim, "wavelengths ")
                        waveSourceOutArray0 => real32data(waveSourceOutCol0)
                        !
                        rateSourceOutCol0 = addColumn(outTargetTable0H,"RATE", real32, &
                             "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                        rateSourceOutArray0 => real32data(rateSourceOutCol0)
                        !
                        rateSourceErrOutCol0 = addColumn(outTargetTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                             columnDim, "source counts rate error")
                        rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                        !
                        rateBackOutCol0 = addColumn(outTargetTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate")
                        rateBackOutArray0 => real32data(rateBackOutCol0)
                        !
                        rateBackErrOutCol0 = addColumn(outTargetTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate error")
                        rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                        !
                        fluxOutCol0 = addColumn(outTargetTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate")
                        fluxOutArray0 => real32data(fluxOutCol0)
                        !
                        fluxErrOutCol0 = addColumn(outTargetTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate error")
                        fluxErrOutArray0 => real32data(fluxErrOutCol0)
                        !
                        !
                        waveSourceOutArray0=wavelengthArray
                        rateSourceOutArray0=spectrumArray
                        rateSourceErrOutArray0=spectrumErrArray
                        rateBackOutArray0=bkgArray
                        rateBackErrOutArray0=bkgErrArray
                        fluxOutArray0=fluxArray
                        fluxErrOutArray0=fluxErrArray

                        !
                        vignApp=.false.
                        deadApp=.false.
                        backApp=.true.
                        psfApp=.false.
                        call setAttribute(outTargetTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                        call setAttribute(outTargetTable0H, "NPIXSOU", sourceArea, "pix", "Source area")

                        call setAttribute(outTargetTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                        call setAttribute(outTargetTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                        call setAttribute(outTargetTable0H, "BACKAPP", backApp,  "Background is subtracted")
                        call setAttribute(outTargetTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            
                     
                        !call setAttribute(outTargetTable0H, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                        !call setAttribute(outTargetTable0H, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                        !call setAttribute(outTargetTable0H, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                        !call setAttribute(outTargetTable0H, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                        if (columnDataTypeFlag .eq. 2) then
                           call setAttribute(outTargetTable0H, "QFLAG", sourceQF8(0), " ",  "Quality flag")
                        else
                           call setAttribute(outTargetTable0H, "QFLAG", sourceQF16(0), " ",  "Quality flag")
                        endif
                        !write(*,*) "L4399"
                        call setAttribute(outTargetTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                        call setAttribute(outTargetTable0H, "OBJ_TYPE", "FIELD OBJECT", &
                             "Target flag                                 ")
                        call setAttribute(outTargetTable0H, "XFIRST", xFirstOrder, &
                             "pix", "First-order (spectrum) image X-coordiate")
                        call setAttribute(outTargetTable0H, "YFIRST", yFirstOrder, &
                             "pix", "First-order (spectrum) image Y-coordiate")
                        !
                        ! VNY 21.07.2004 Recompute zero-order coordinates
                        !computingMode=0
                        !verbFlag=0
                        !xZeroIn=xZeroOrder
                        !yZeroIn=yZeroOrder
                        !call RecomputeZeroOrder(grismSetH, xZeroIn, yZeroIn, &
                        !     spectrumHalfWidth, spectrumHalfHeight, computingMode, &
                        !     verbFlag, xZeroOut, yZeroOut)
                        !xZeroOrder=xZeroOut
                        !yZeroOrder=yZeroOut
                        !
                        call setAttribute(outTargetTable0H, "XZERO", xZeroOrder, &
                             "pix", "Zero-order (object) image X-coordiate")
                        call setAttribute(outTargetTable0H, "YZERO", yZeroOrder, &
                             "pix", "Zero-order (object) image Y-coordiate")
                        call setAttribute(outTargetTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                             "pix", "Width of the spectrum extraction region")
                        call setAttribute(outTargetTable0H, "BKGREGWL", abs(bkgwidthleft), &
                             "pix", "Backgr.extraction region width (left)")
                        call setAttribute(outTargetTable0H, "BKGREGWR", abs(bkgwidthleft), &
                             "pix", "Backgr.extraction region width (right)")
                        call setAttribute(outTargetTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                             "pix", "Spectrum to backgr.reg.separation (left)")
                        call setAttribute(outTargetTable0H, "BKGSEPR", abs(bkgoffsetright), &
                             "pix", "Spectrum to backgr.reg.separation (right)")
                     
                        if (putContaminationComment) then
                           call addComment(outTargetTable0H, contaminationComment)
                        endif
                        ! release the dal handles
                        !---------
                        !---VNY 30.08.03
                        call release(rateBackErrOutCol0)
                        call release(rateBackOutCol0)
                        call release(rateSourceErrOutCol0)
                        call release(rateSourceOutCol0)
                        call release(fluxOutCol0)
                        call release(fluxErrOutCol0)
                        call release(waveSourceOutCol0)
                        call release(outTargetTable0H)
                        !---
                        ! VNY 26.07.2004
                        !box0left=int(xZeroOrder)-int(spectrumHalfWidth)
                        !box0right=int(xZeroOrder)+int(spectrumHalfWidth)
                        !box1left=int(xFirstOrder)-int(spectrumHalfWidth)
                        !box1right=int(xFirstOrder)+int(spectrumHalfWidth)
                        !! VNY 02.09.2004
                        !box0down=int(yZeroOrder)-int(spectrumHalfWidth)!50
                        !box0up=int(yZeroOrder)+int(spectrumHalfWidth)!50
                        !!
                        !box1down=int(yFirstOrder)-int(spectrumHalfHeight)
                        !box1up=int(yFirstOrder)+int(spectrumHalfHeight)
                        !---------------------------- VNY 14.10.2004
                        box0left=xZeroOrder - spectrumHalfWidth +0.5
                        box0right=xZeroOrder+spectrumHalfWidth +0.5
                        box1left=xFirstOrder-spectrumHalfWidth +0.5
                        box1right=xFirstOrder+spectrumHalfWidth+0.5
                        box0down=yZeroOrder-spectrumHalfWidth+0.5
                        box0up=yZeroOrder+spectrumHalfWidth+0.5
                        box1down=yFirstOrder-spectrumHalfHeight+0.5
                        box1up=yFirstOrder+spectrumHalfHeight+0.5
                        !-------------

                       !if (iSource .eq. 383) then
                       !   write(*,*) "L6015 yFirstOrder=", yFirstOrder, " box1down=", &
                       !        box1down, " box1up=", box1up
                       !endif
                        if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                           box1down=yZeroOrder+regionU1+0.5
                           box1up=yZeroOrder+regionU2+0.5
                        else
                           box1down=yZeroOrder+regionV1+0.5
                           box1up=yZeroOrder+regionV2+0.5
                        endif
                        if (box0left .lt. 0.0) box0left=0.0
                        if (box0right .gt. nx-1) box0right=nx-1
                        if (box0down .lt. 0.0) box0down=0.0
                        if (box1left .lt. 0.0) box1left=0.0
                        if (box1right .gt. nx-1) box1right=nx-1
                        if (box1up .gt. ny-1) box1up=ny-1
                        !if (iSource .eq. 383) then
                        !   write(*,*) "L6032 box1down=", &
                        !        box1down, " box1up=", box1up
                        !endif
                        !firstPass=.false.
                        if (givenSpectraRegionFile) then 
                           call writeSpectrumExtractionRegion(spectraRegionFileName, &
                                box0left,box0right,box0down,box0up, &
                                box1left,box1right,box1down,box1up, &
                                spectrumNoCount, firstPass, givenSpectraRegionFile)
                           firstPass=.false.
                        else
                           if (givenRegionFile) then
                              call writeSpectrumExtractionRegion(regionFileName, &
                                   box0left,box0right,box0down,box0up, &
                                   box1left,box1right,box1down,box1up, &
                                   spectrumNoCount, firstPass, givenSpectraRegionFile)
                              firstPass=.false.
                           endif
                        endif
                     endif ! if checkFlag
                  endif ! if yZero .gt. 0
               endif
            endif
         enddo ! iSource=0, nSources-1


         !****************************************************************
         !****************************************************************
         !****************************************************************
      endif ! extractFieldSpectra
      ! Check the source list

      nSpectraSources=0
!      do iSource=0, nSources-1
      do iSource=0, nSources2-1
         if (spectrumNoArray(iSource) .lt. 99999 .and. spectrumNoArray(iSource) .gt. 0) then
            nSpectraSources=nSpectraSources+1
         endif
      enddo ! iSource=0, nSources-1

      if (spectrumNoCount .ge. 0) then
         do i=0,spectrumNoCount
            if (spectrumOriginArray2(i) .ge. 99999) then
               spectrumOriginArray2(i) =spectrumNoArray2(i)
            endif
         enddo
      endif

!**********************************************************************
         !  Create a file containing list of the extracted spectra for             
!**********************************************************************   
!         if (produceSpectraList) then !30.10.2005
         if (produceSpectraList .and. createSpectraListFile) then
            outSpectraListSetH = dataSet(spectraListFile, create)
            call setAttribute(outSpectraListSetH, "CONTENT", "OM GRISM SPECTRA LIST", &
                 "File contains the list of the extracted OM grsim spectra")
            call setAttribute(outSpectraListSetH, "HDUCLASS", "OGIP    ", &
                 "Format conforms to OGIP/GSFC conventions")
            call setAttribute(outSpectraListSetH, "HDUCLAS1", "SRCLIST", &
                 "File contains a source/spectra list")
            call setAttribute(outSpectraListSetH, "HDUVERS", "1.2.0    ", &
                 "The version number of the format")
            call setAttribute(outSpectraListSetH, "TELESCOP", "XMM     ", &
                 "XMM mission")
            call setAttribute(outSpectraListSetH, "INSTRUME", "OM      ", &
                 "Instrument name")
            call setAttribute(outSpectraListSetH, "DATAMODE", dataMode, &
                 "Data mode")
            call setAttribute(outSpectraListSetH, "FILTER", filter, &
                 "Filter used for obtaining the data")
            call setAttribute(outSpectraListSetH, "FRAMTIME", frameTime, &
                 "Frame time")
            call setAttribute(outSpectraListSetH, "DEADFRAC", deadFrac, &
                 "Frame dead fraction")
            call setAttribute(outSpectraListSetH, "TMEXP_ID", tmExpId, &
                 "Exposure Id from telemetry")
            call setAttribute(outSpectraListSetH, "FRMTIME", frmTime, &
                 "Frametime from telemetry")
            call setAttribute(outSpectraListSetH, "DATATYPE", dataType, &
                 "Data type")
            call setAttribute(outSpectraListSetH, "OBS_ID", obsId, &
                 "Observation identifier")
            call setAttribute(outSpectraListSetH, "EXP_ID", expId, &
                 "Exposure identifier")
            call setAttribute(outSpectraListSetH, "DATE-OBS", dateObs, &
                 "Observation start time")
            call setAttribute(outSpectraListSetH, "DATE-END", dateEnd, &
                 "Observation end time")

            ! Add columns with spectra identification
            nBins=spectrumNoCount+1
            write(blockName, '(a)') "SRCLIST"
            outSpectraTableH=addTable(outSpectraListSetH, blockName, nBins, &
                 "Table with the extracted spectra list")

            ! Add columns with the spectrum identification number to the (input) source list table 
            if (.not. hasColumn(outSpectraTableH, "SPECTR_ID")) then
               specNoListCol = addColumn(outSpectraTableH, "SPECTR_ID", integer32, "", columnDim, "spectrum identification ")
            else
               specNoListCol = column(outSpectraTableH, "SPECTR_ID", modify)
            endif
            specNoListArray => int32data(specNoListCol)
            call setNullValue(specNoListCol, 999999 )
 
            !do iSource=0, nSpectraSources-1
            do iSource=0, spectrumNoCount
               call setToNull(specNoListCol, iSource)
            enddo

            ! X coordinates of the zero-order
            if (.not. hasColumn(outSpectraTableH, "X0")) then
               specX0Col = addColumn(outSpectraTableH, "X0", real32, "", columnDim, "Zero-order X-coord.")
            else
               specX0Col = column(outSpectraTableH, "X0", modify)
            endif
            specX0Array => real32data(specX0Col)
            specX0Array=0.0

            ! Y coordinates of the zero-order
            if (.not. hasColumn(outSpectraTableH, "Y0")) then
               specY0Col = addColumn(outSpectraTableH, "Y0", real32, "", columnDim, "Zero-order Y-coord.")
            else
               specY0Col = column(outSpectraTableH, "Y0", modify)
            endif
            specY0Array => real32data(specY0Col)
            specY0Array=0.0

            ! X coordinates of the first-order
            if (.not. hasColumn(outSpectraTableH, "X1")) then
               specX1Col = addColumn(outSpectraTableH, "X1", real32, "", columnDim, "First-order X-coord.")
            else
               specX1Col = column(outSpectraTableH, "X1", modify)
            endif
            specX1Array => real32data(specX1Col)
            specX1Array=0.0

            ! Y coordinates of the first-order
            if (.not. hasColumn(outSpectraTableH, "Y1")) then
               specY1Col = addColumn(outSpectraTableH, "Y1", real32, "", columnDim, "First-order Y-coord.")
            else
               specY1Col = column(outSpectraTableH, "Y1", modify)
            endif
            specY1Array => real32data(specY1Col)
            specY1Array=0.0

            ! VNY 20.07.2004 removing spectrum ID column (instead, SPECTR_NO column 
            ! will be renamed by SPECTR_ID)
            ! Add columns with the spectrum identification number to the (input) source list table 
            !if (.not. hasColumn(outSpectraTableH, "SPECTR_ID")) then
            !   specIdCol = addColumn(outSpectraTableH, "SPECTR_ID", integer32, "", &
            !        columnDim, "spectrum identifier, as in the SWSRLI-file")
            !else
            !   specIdCol = column(outSpectraTableH, "SPECTR_ID", modify)
            !endif
            !specIdArray => int32data(specIdCol)
            !call setNullValue(specIdCol, 999999 )
            
            !do iSource=0, nSpectraSources-1
            !   call setToNull(specIdCol, iSource)
            !enddo
            !------------------------------------------------------------
            ! VNY 13.07.2004 Adding a column for flagging the target type
            if (.not. hasColumn(outSpectraTableH, "SPEC_TYPE")) then
               specTypeCol = addColumn(outSpectraTableH, "OBJ_TYPE", String, "", stringColumnDim, "Target flag")
            else
               specTypeCol = column(outSpectraTableH, "OBJ_TYPE", modify)
            endif
            do iSource=0, spectrumNoCount
               !write(*,*) " L4517 iSource=", iSource, " specTypeAux(iSource)=", specTypeAux(iSource)
               call SetStringCell(specTypeCol,iSource,"                  ")
               if (specTypeAux(iSource) .eq. 1) then 
                  call SetStringCell(specTypeCol,iSource,"TARGET            ")
               else
                  if (specTypeAux(iSource) .eq. 2) then
                     call SetStringCell(specTypeCol,iSource,"DUBIOUS TARGET    ")
                  else
                     if (specTypeAux(iSource) .eq. 3) then
                        call SetstringCell(specTypeCol,iSource, "DEFAULT EXTRACTION")
                     else
                        if (specTypeAux(iSource) .eq. 4) then
                           call SetStringCell(specTypeCol,iSource, "DEFAULT(NO TARGET)")
                        else
                           if (specTypeAux(iSource) .eq. 5) then
                              call SetStringCell(specTypeCol, iSource, "RE-DETECTED TARGET")
                           else
                              call SetStringCell(specTypeCol,iSource,"FIELD OBJECT      ")
                           endif
                        endif
                     endif
                  endif
               endif
            enddo
            !------------------------------------------------------------

            do iSource=0, spectrumNoCount
               specNoListArray(iSource)=iSource+1
            enddo
            
            do i=0, spectrumNoCount
               specX0Array(i)=x0aux(i)
               specX1Array(i)=x1aux(i)
               specY0Array(i)=y0aux(i)
               specY1Array(i)=y1aux(i)
            enddo ! i=0, spectrumNoCount
         endif ! produceSpectraList
!***********************************************************************
         !*************************
         ! Release handles on data
         !*************************
         if (createSpectrumFile) then
            call release(outTargetSet0H)
         endif
         call release(grismSetH)
         do i=1, noColumns
            if (inColumnFlag(i) .eq. 1) then
               call release(inColumnH(i))
            endif
         end do
         if(produceSpectraList .and. createSpectraListFile) then
            call release(spectrumNoCol)
            call release(spectrumOriginCol)
            call release(specNoListCol)
            call release(specX0Col)
            call release(specX1Col)
            call release(specY0Col)
            call release(specY1Col)
            !         call release(specIdCol)
            call release(specTypeCol)

            !call release(specOriginListCol)
            call release(outSpectraTableH)
            call release(outSpectraListSetH)
         endif
         !
         deallocate(indexCheckSources)
         deallocate(distCheckSources)
         !
         deallocate(indexBrightSources)
         deallocate(distBrightSources)
         !
         deallocate(sourceFWHMaj2)
         deallocate(sourceFWHMin2)
         deallocate(sourceXCoord)
         deallocate(sourceYCoord)
         deallocate(extended)
         deallocate(spectrumNoArray)
         !
         deallocate(limitLeftMax)
         deallocate(limitRightMax)
         deallocate(sourceShape)
         deallocate(sourceSize)
         deallocate(sourceRawCountRate)
         deallocate(sourceRelation)
         deallocate(targetFlag)
         deallocate(procFlag)
         deallocate(sourceProximity)
         deallocate(proximityX)
         deallocate(proximityCounts)
!
         deallocate(sourceProximityCheck)
         deallocate(proximityXcheck)
         deallocate(proximityCountsCheck)
         deallocate(snRatioCheck)
         deallocate(sourceRelationCheck)

         deallocate(sourceProximityBright)
         deallocate(proximityXbright)
         deallocate(proximityCountsBright)
         deallocate(snRatioBright)
         deallocate(sourceRelationBright)

         deallocate(x0aux)
         deallocate(x1aux)
         deallocate(y0aux)
         deallocate(y1aux)
         deallocate(specTypeAux)
         deallocate(snRatio)
!
         deallocate(spectrumNoArray2)
         deallocate(spectrumOriginArray2)
         deallocate(srcNoArray2)
!
         deallocate(bkgVect)
         deallocate(bkgVect2)
         deallocate(bkgVect3)
         deallocate(bkgVectR)
         deallocate(bkgVectL)
         deallocate(smoothBkgVectR)
         deallocate(smoothBkgVectL)
         deallocate(specVect)
         deallocate(countsScaleFactor)
         ! VNY 17.11.2003
         deallocate (wavelengthArray)
         deallocate (spectrumArray)
         deallocate (maskArray)
!         deallocate (spectrumMaskArray)
         deallocate (spectrumErrArray)
         deallocate (bkgArray)
         deallocate (bkgErrArray)
         deallocate (fluxArray)
         deallocate (fluxErrArray)
      else
        call warning("Crowded field", "Too many sources in the input source list")
         write(errstr,*) "Spectra will not be extracted (try using omgsource)"
         call message(AppMsg, 2)
      endif ! nSources>0

!      if(allocated(backGroundRate)) then
!         deallocate(backGroundrate,stat=istat)
!         call CheckDeallocation(allocErr, "backGroundRate")
!      end if
!
!      if(allocated(mag)) then
!         deallocate(mag,stat=istat)
!         call CheckDeallocation(allocErr, "mag")
!      end if
!
!      if(allocated(magErr)) then
!         deallocate(magErr,stat=istat)
!         call CheckDeallocation(allocErr, "magErr")
!      end if
!
!      if(allocated(sourceApertureRadius)) then
!         deallocate(sourceApertureradius,stat=istat)
!         call CheckDeallocation(allocErr, "sourceApertureRadius")
!      end if
!
!      if(allocated(sourceBII)) then
!         deallocate(sourceBII,stat=istat)
!         call CheckDeallocation(allocErr, "sourceBII")
!      end if!
!
!      if(allocated(sourceLII)) then
!         deallocate(sourceLII,stat=istat)
!         call CheckDeallocation(allocErr, "sourceLII")
!      end if
!
!      if(allocated(rateErr)) then
!         deallocate(rateErr,stat=istat)
!         call CheckDeallocation(allocErr, "rateErr")
!      end if
!
!      if(allocated(rate)) then
!         deallocate(rate,stat=istat)
!         call CheckDeallocation(allocErr, "rate")
!      end if
      ! VNY 14.10.2003 Deleting the columns in the input source list
      !************************************************************************
      if (hasColumn(inTableH, "RATE")) then
         call deleteColumn(inTableH, "RATE")
      endif
      if (hasColumn(inTableH, "RATE_ERR")) then
         call deleteColumn(inTableH, "RATE_ERR")
      endif
      if (hasColumn(inTableH, "BACKGROUND_RATE")) then
         call deleteColumn(inTableH, "BACKGROUND_RATE")
      endif
      if (hasColumn(inTableH, "SIGNIFICANCE")) then
         call deleteColumn(inTableH, "SIGNIFICANCE")
      endif
      if (hasColumn(inTableH, "CORR_RATE")) then
         call deleteColumn(inTableH, "CORR_RATE")
      endif
      if (hasColumn(inTableH, "CORR_RATE_ERR")) then
         call deleteColumn(inTableH, "CORR_RATE_ERR")
      endif
      if (hasColumn(inTableH, "CORR_BACKGROUND_RATE")) then
         call deleteColumn(inTableH, "CORR_BACKGROUND_RATE")
      endif
      if (hasColumn(inTableH, "PSF1_CORR")) then
         call deleteColumn(inTableH, "PSF1_CORR")
      endif
      if (hasColumn(inTableH, "COINCLOSS_CORR")) then
         call deleteColumn(inTableH, "COINCLOSS_CORR")
      endif
      if (hasColumn(inTableH, "APERTURE")) then
         call deleteColumn(inTableH, "APERTURE")
      endif
      !---^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      call release(inTableH)
   else ! hasBlock(sourceListSetH)
      !************************************
      ! Incompatible file - can't continue
      !************************************
      write (errstr, *) "Could not find any compatibile blocks in set ", sourceListFile
      call message(AppMsg, 2)
   endif ! hasBlock(sourceListSetH)
   call release(sourceListSetH)
   !      call release(sourceRegSetH)
!   deallocate(siginput)
 end subroutine omgrism

!*************************************************************
! Subroutines
!*************************************************************

subroutine AddReal32Column(tableSet, array, size, heading, &
        units, comment, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, units, comment
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(0 : size - 1) :: array
    real(kind=single), intent(in) :: nullValue
    integer(kind=int32), intent(in), optional :: position
    real(kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)


     if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   

    real32Col => real32Data(pcolumn)

    do i = 0 , size - 1
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i)
        else
            real32Col(i) = array(i)
        end if
    enddo
    call release(pcolumn)

end subroutine AddReal32Column

!*********************************************************
! Retrieves the exposure identifier from the given string
! ie S006, U453, X008
!*********************************************************
function GetExposureIdentifier(name) result(exposureIdentifier)
    character(*), intent(in) :: name
    character(4) :: exposureIdentifier
    integer :: i, j, length
    character(4) :: string
    character(1) :: string1

    length = len(trim(name))

loop1 : do i = 3, length - 3

        string = name(i : i + 3)
      
        if (string(1 : 1) .eq. 'S' .or. &
            string(1 : 1) .eq. 'U' .or. &
            string( 1: 1) .eq. 'X') then
            
            do j = 2, 4
                string1 = string(j : j)
                if (scan(string1, "0123456789") .eq. 0) cycle loop1
               
            end do 
            exposureIdentifier = string

            if (exposureIdentifier(1 : 1) .eq.  "X") exposureIdentifier(1 : 1) = "S" 
            return
        end if
   
    end do loop1
    call fatal("Exposure idenetifier not found")
    
end function GetExposureIdentifier

!****************************************************************
! Subroutine to get the X and Y coordinates of the target
!****************************************************************
subroutine GetTargetPosition(primInstrValue, filter, binBpe, xbin, ybin, xTarget, yTarget)
    character(len=*), intent(in) :: primInstrValue, filter
    logical, intent(in) :: binBpe
    integer(kind=int32), intent(in) :: xbin, ybin
    integer(kind=int32), intent(out) :: xTarget, yTarget
    integer(kind=int32), dimension(3,2,2) :: xCal, yCal
    integer :: i,j,k


! The calibration data for the target measured by A.Talavera 20.07.2004
    data xCal/236,252,211,242,209,205,616,618,560,602,603,545/
    data yCal/ 67, 65,124, 51,102,102,232,251,753,196,205,239/
!-------------------------------------------------------------
!  Only for test purporses 27.07.2004:
!    data xCal/236,252,211,242,209,205,616,618,560,602,150,545/
!    data yCal/ 67, 65,124, 51,102,102,232,251,753,196, 80,239/
!
!  Current table for 20.07.2004
!-------------------------------------------------------------------
!        R1            PN                OM       <-- Prim.instr
!-------------------------------------------------------------------
!     X     Y       X       Y         X      Y
!-------------------------------------------------------------------
!   236    67      252      65       211    124   <-- UV, HR
!   242    51      209     102       205    102   <--  V, HR
!
!   616   232      618     251       560    753   <-- UV, LR
!   602   196      603     205       545    239   <--  V, LR
!-------------------------------------------------------------------     
    i=2 ! primary instrument: PN or E3
    if (trim(primInstrValue) .eq. "R1") i=1
    if (trim(primInstrValue) .eq. "OM") i=3
    
    j=1 ! filter=UV (grism1)
    if (trim(filter) .eq. "GRISM2" .or. trim(filter) .eq. "Grism2") j=2

    k=1 ! resolution is High (binBpe=F)
    if (binBpe) k=2
    xTarget=xCal(i,j,k)
    yTarget=yCal(i,j,k)
    !write(*,*) " OMGRISM L3640 xbin=", xbin, " ybin=", ybin
    if (xbin .gt. 1 .and. .not. binBpe) xTarget=xTarget/xbin
    if (ybin .gt. 1 .and. .not. binBpe) yTarget=yTarget/ybin
end subroutine GetTargetPosition

!****************************************************************
! Subroutine to get the X and Y coordinates of the target
!****************************************************************
subroutine writeSpectrumExtractionRegion(regionFileName, &
     box0left,box0right,box0down,box0up, &
     box1left,box1right,box1down,box1up, spectrumNo, &
     firstPass, givenSpectraRegionFile)
!     xZeroOrder,yZeroOrder, &
!     xFirstOrder,yFirstOrder, spectrumHalfWidth, spectrumHalfHeight)
  real(kind=double), intent(in) :: box0left,box0right,box0down,box0up
  real(kind=double), intent(in) :: box1left,box1right,box1down,box1up
  character(len=1024), intent(in) :: regionFileName
  logical, intent(in) :: firstPass
  logical, intent(in) :: givenSpectraRegionFile

  integer(kind=int32) :: spectrumNo
  logical :: existFile
  integer :: count, icount, countMax
  character(len=1024) :: regionString
  real(kind=double) :: x0,y0,w0,h0,x1,y1,w1,h1

  !write(*,*) "OMGRISM L3598 box0left=", box0left, "box0right=", box0right
  !x0=(box0left+box0right+0.5)/2.0
  x0=box0left+box0right+0.5
  x0=x0/2.0
  w0=box0right-box0left+1.0
  !y0=(box0down+box0up+0.5)/2.0
  y0=box0down+box0up+0.5
  y0=y0/2.0
  h0=box0up-box0down+1.0

  !write(*,*) " L3606 x0=", x0, " y0=", y0

  !x1=(box1left+box1right+0.5)/2.0
  x1=(box1left+box1right+0.5)/2.0
  w1=box1right-box1left+1.0
!  y1=(box1down+box1up+0.5)/2.0
  y1=(box1down+box1up+0.5)/2.0
  h1=box1up-box1down+1.0

  if (.not. givenSpectraRegionFile) then 
     inquire(FILE=trim(regionFileName), EXIST=existFile)
     if (existFile) then
        ! Open region file
        open(3,file=trim(regionFileName))
        count=0
        do
           read(3, '(a)', end=105) regionString
           if (regionString(1:4) .eq. " BOX" .and. firstPass) goto 105
           count=count+1
        enddo
105     continue 
        countMax=count
        close(3)
        ! Open region file again
        open(3,file=trim(regionFileName))
        count=0
        if (countMax .gt. 0) then
           do icount=1,countMax
              read(3, '(a)', end=106) regionString
           enddo
        endif
        goto 107
106     continue 
        goto 108
107     continue
        write(3,'(a, f9.2,f9.2,f9.2,f9.2,f4.1,a)') " BOX ", &
             x0,y0,w0,h0,0.0," #COLOR=red"
        write(3,'(a, f9.2,f9.2,f9.2,f9.2,f4.1,a,i2,a)') " BOX ", &
             x1,y1,w1,h1,0.0, " #COLOR=red,TEXT={",spectrumNo+1,"}"
108     close(3)
     endif
  else
     ! Open region file
     open(3,file=trim(regionFileName))
     count=0
     do
        read(3, '(a)', end=1051) regionString
        if (regionString(1:4) .eq. " BOX" .and. firstPass) goto 1051
        count=count+1
     enddo
1051 continue 
     countMax=count
     close(3)
     ! Open region file again
     open(3,file=trim(regionFileName))
     count=0
     if (countMax .gt. 0) then
        do icount=1,countMax
           read(3, '(a)', end=1061) regionString
        enddo
     endif
     goto 1071
1061 continue 
     goto 1081
1071 continue
     write(3,'(a, f9.2,f9.2,f9.2,f9.2,f4.1,a)') " BOX ", &
          x0,y0,w0,h0,0.0," #COLOR=red"
     write(3,'(a, f9.2,f9.2,f9.2,f9.2,f4.1,a,i2,a)') " BOX ", &
          x1,y1,w1,h1,0.0, " #COLOR=red,TEXT={",spectrumNo+1,"}"
1081 close(3)
  endif
  !
end subroutine writeSpectrumExtractionRegion
!----
end module omgrism_mod











